ner_displacy_palette = {
    "CASENAME": "#da3650",
    "CITATION": "#67328b",
    "JUDGE": "#00a594",
    "COURT": "#fcd548",
    "PROVISION": "#007bac",
    "INSTRUMENT": "#6c63a5",
    "CONCEPT": "#df5a35",
}

ner_displacy_options = {
    "ents": [
        "CASENAME",
        "CITATION",
        "JUDGE",
        "COURT",
        "PROVISION",
        "INSTRUMENT",
        "CONCEPT",
    ],
    "colors": ner_displacy_palette,
}


dep_displacy_options = {
    "compact": False,
    "bg": "linear-gradient(90deg, #bdc3c7, #2c3e50)",
    "color": "white",
    "font": "Gotham",
}
