CONCEPT_PATTERNS = [
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "aarhus"}, {"LOWER": "convention"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "abandonment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "abandonment"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "to"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "abatement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "abatement"}, {"LOWER": "notice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "abduction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "abidin"}, {"LOWER": "daver"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "abortion"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "abortion"}, {"LOWER": "services"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "abridged"},
            {"LOWER": "taxation"},
            {"LOWER": "of"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "to"},
            {"LOWER": "commit"},
            {"LOWER": "acts"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "absence"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "pregnancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "absence"}, {"LOWER": "from"}, {"LOWER": "trial"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "absolute"}, {"LOWER": "privilege"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "abuse"},
            {"LOWER": "of"},
            {"LOWER": "dominant"},
            {"LOWER": "position"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "abuse"},
            {"LOWER": "of"},
            {"LOWER": "executive"},
            {"LOWER": "power"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "abuse"}, {"LOWER": "of"}, {"LOWER": "process"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "abuse"},
            {"LOWER": "or"},
            {"LOWER": "excess"},
            {"LOWER": "of"},
            {"LOWER": "power"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "academic"}, {"LOWER": "point"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "academic"},
            {"LOWER": "point"},
            {"LOWER": "of"},
            {"LOWER": "law"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "academic"}, {"LOWER": "regulations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "accelerated"}, {"LOWER": "payment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "accelerated"},
            {"LOWER": "payment"},
            {"LOWER": "notice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "acceptance"}, {"LOWER": "by"}, {"LOWER": "conduct"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acceptance"},
            {"LOWER": "of"},
            {"LOWER": "lesser"},
            {"LOWER": "sum"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acceptance"},
            {"LOWER": "of"},
            {"LOWER": "payment"},
            {"LOWER": "into"},
            {"LOWER": "court"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "accepting"}, {"LOWER": "deposits"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "access"}, {"LOWER": "request"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "access"}, {"LOWER": "to"}, {"LOWER": "church"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "access"}, {"LOWER": "to"}, {"LOWER": "documents"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "documents"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "institutions"},
            {"LOWER": "refused"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "justice"},
            {"LOWER": "in"},
            {"LOWER": "environmental"},
            {"LOWER": "matter"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "justice"},
            {"LOWER": "in"},
            {"LOWER": "environmental"},
            {"LOWER": "matters"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "social"},
            {"LOWER": "assistance"},
            {"LOWER": "and"},
            {"LOWER": "child"},
            {"LOWER": "benefit"},
            {"LOWER": "conditional"},
            {"LOWER": "on"},
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "residence"},
            {"LOWER": "in"},
            {"LOWER": "member"},
            {"LOWER": "state"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "accession"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "accident"}, {"LOWER": "insurance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "accommodation"},
            {"LOWER": "available"},
            {"LOWER": "for"},
            {"LOWER": "occupation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "accord"}, {"LOWER": "and"}, {"LOWER": "satisfaction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "account"}, {"LOWER": "of"}, {"LOWER": "costs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "account"}, {"LOWER": "of"}, {"LOWER": "profits"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "accountable"},
            {"LOWER": "person"},
            {"LOWER": "’s"},
            {"LOWER": "certificate"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "accountant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "accountants"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "accountants"}, {"LOWER": "fees"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "accounting"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "accounts"}, {"LOWER": "and"}, {"LOWER": "inquiries"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "accounts"},
            {"LOWER": "by"},
            {"LOWER": "persons"},
            {"LOWER": "liable"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "accrual"},
            {"LOWER": "of"},
            {"LOWER": "cause"},
            {"LOWER": "of"},
            {"LOWER": "action"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "accuracy"}, {"LOWER": "of"}, {"LOWER": "data"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "accused"}, {"LOWER": "person"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acknowledgement"},
            {"LOWER": "of"},
            {"LOWER": "service"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "acknowledgment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acquiring"},
            {"LOWER": "or"},
            {"LOWER": "possessing"},
            {"LOWER": "criminal"},
            {"LOWER": "property"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "acquisition"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "acquisition"}, {"LOWER": "naturalisation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "acquittal"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acquittal"},
            {"LOWER": "of"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "defendants"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "acquitted"}, {"LOWER": "defendant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "act"},
            {"LOWER": "done"},
            {"LOWER": "to"},
            {"LOWER": "another"},
            {"LOWER": "person"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "act"},
            {"LOWER": "done"},
            {"LOWER": "“"},
            {"LOWER": "to"},
            {"LOWER": "another"},
            {"LOWER": "person"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "act"},
            {"LOWER": "in"},
            {"LOWER": "furtherance"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "act"}, {"LOWER": "of"}, {"LOWER": "bankruptcy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "act"}, {"LOWER": "of"}, {"LOWER": "state"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acting"},
            {"LOWER": "in"},
            {"LOWER": "execution"},
            {"LOWER": "of"},
            {"LOWER": "office"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"ORTH": "("},
            {"LOWER": "short"},
            {"LOWER": "of"},
            {"LOWER": "dismissal"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "against"},
            {"LOWER": "several"},
            {"LOWER": "defendants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "against"},
            {"LOWER": "two"},
            {"LOWER": "defendants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "allocated"},
            {"LOWER": "to"},
            {"LOWER": "multi"},
            {"LOWER": "-"},
            {"LOWER": "track"},
            {"LOWER": "in"},
            {"LOWER": "county"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "by"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "by"},
            {"LOWER": "third"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "action"}, {"LOWER": "for"}, {"LOWER": "account"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "action"}, {"LOWER": "for"}, {"LOWER": "annulment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "action"}, {"LOWER": "for"}, {"LOWER": "damages"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "action"}, {"LOWER": "for"}, {"LOWER": "damages"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "action"}, {"LOWER": "for"}, {"LOWER": "debt"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "for"},
            {"LOWER": "fraudulent"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "trust"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "for"},
            {"LOWER": "infringement"},
            {"LOWER": "brought"},
            {"LOWER": "against"},
            {"LOWER": "undertaking"},
            {"LOWER": "with"},
            {"LOWER": "seat"},
            {"LOWER": "in"},
            {"LOWER": "third"},
            {"LOWER": "country"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "action"}, {"LOWER": "in"}, {"LOWER": "rem"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "short"},
            {"LOWER": "of"},
            {"LOWER": "dismissal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "action"},
            {"LOWER": "taken"},
            {"LOWER": "in"},
            {"LOWER": "reliance"},
            {"LOWER": "on"},
            {"LOWER": "counsel"},
            {"LOWER": "’s"},
            {"LOWER": "opinion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "action"}, {"LOWER": "to"}, {"LOWER": "recover"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "actions"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "agents"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "activities"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "sovereign"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acts"},
            {"LOWER": "contrary"},
            {"LOWER": "to"},
            {"LOWER": "united"},
            {"LOWER": "nations"},
            {"LOWER": "purposes"},
            {"LOWER": "and"},
            {"LOWER": "principles"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acts"},
            {"LOWER": "of"},
            {"LOWER": "independent"},
            {"LOWER": "contractor"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "acts"}, {"LOWER": "restricted"}, {"LOWER": "by"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "acts"},
            {"LOWER": "“"},
            {"LOWER": "more"},
            {"LOWER": "than"},
            {"LOWER": "merely"},
            {"LOWER": "preparatory"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "actual"}, {"LOWER": "undue"}, {"LOWER": "influence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "additional"},
            {"LOWER": "liabilities"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "additional"},
            {"LOWER": "or"},
            {"LOWER": "substituted"},
            {"LOWER": "charges"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "address"}, {"LOWER": "for"}, {"LOWER": "service"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ademption"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "adjoining"}, {"LOWER": "council"}, {"LOWER": "flats"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adjoining"}, {"LOWER": "properties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "adjournment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "adjournment"}, {"LOWER": "of"}, {"LOWER": "hearing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adjournment"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adjudication"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "adjudication"}, {"LOWER": "decision"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adjudication"},
            {"LOWER": "of"},
            {"LOWER": "bankruptcy"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adjudicator"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adjudicator"},
            {"LOWER": "determining"},
            {"LOWER": "sum"},
            {"LOWER": "of"},
            {"LOWER": "money"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "claimant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adjudicator"},
            {"LOWER": "determining"},
            {"LOWER": "sum"},
            {"LOWER": "of"},
            {"LOWER": "money"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "claimant"},
            {"LOWER": "subcontractor"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adjudicators"}, {"LOWER": "decision"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adjudicator"},
            {"LOWER": "’s"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "adjustment"}, {"LOWER": "of"}, {"LOWER": "awards"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adjustment"},
            {"LOWER": "of"},
            {"LOWER": "contract"},
            {"LOWER": "price"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adjustment"},
            {"LOWER": "of"},
            {"LOWER": "prior"},
            {"LOWER": "transactions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adjustment"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administering"}, {"LOWER": "of"}, {"LOWER": "cautions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "administering"},
            {"LOWER": "poison"},
            {"LOWER": "so"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "endanger"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administration"}, {"LOWER": "of"}, {"LOWER": "claims"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administration"}, {"LOWER": "of"}, {"LOWER": "estate"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administration"}, {"LOWER": "of"}, {"LOWER": "estates"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administration"}, {"LOWER": "of"}, {"LOWER": "funds"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administration"}, {"LOWER": "of"}, {"LOWER": "justice"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "administration"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "administrative"}, {"LOWER": "acts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "administrative"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administrative"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administrative"}, {"LOWER": "tribunal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appointment"},
            {"LOWER": "of"},
            {"LOWER": "administrator"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "admiralty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "admiralty"},
            {"LOWER": "and"},
            {"LOWER": "commercial"},
            {"LOWER": "court"},
            {"LOWER": "registry"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "admiralty"},
            {"LOWER": "and"},
            {"LOWER": "commercial"},
            {"LOWER": "registry"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "admiralty"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "admissibility"}, {"LOWER": "in"}, {"LOWER": "evidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "admissibility"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "admissibility"}, {"LOWER": "of"}, {"LOWER": "answers"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "admissibility"},
            {"LOWER": "of"},
            {"LOWER": "documents"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "admission"}, {"LOWER": "for"}, {"LOWER": "treatment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "admission"}, {"LOWER": "of"}, {"LOWER": "public"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "admissions"},
            {"LOWER": "made"},
            {"LOWER": "under"},
            {"LOWER": "compulsion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "adopted"}, {"LOWER": "children"}, {"LOWER": "register"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adoption"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "adoption"}, {"LOWER": "agency"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adoption"},
            {"LOWER": "and"},
            {"LOWER": "contact"},
            {"LOWER": "orders"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adoption"}, {"LOWER": "application"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adoption"},
            {"LOWER": "of"},
            {"LOWER": "statute"},
            {"LOWER": "and"},
            {"LOWER": "protocol"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adoption"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "adoption"}, {"LOWER": "orders"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "adoption"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "adult"}, {"LOWER": "dependent"}, {"LOWER": "relatives"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "adult"},
            {"LOWER": "trespasser"},
            {"LOWER": "climbing"},
            {"LOWER": "over"},
            {"LOWER": "locked"},
            {"LOWER": "gate"},
            {"LOWER": "and"},
            {"LOWER": "diving"},
            {"LOWER": "into"},
            {"LOWER": "closed"},
            {"LOWER": "swimming"},
            {"LOWER": "pool"},
            {"LOWER": "at"},
            {"LOWER": "night"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adulterated"}, {"LOWER": "drink"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "advance"}, {"LOWER": "corporation"}, {"LOWER": "tax"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "advancement"},
            {"LOWER": "of"},
            {"LOWER": "animal"},
            {"LOWER": "welfare"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "adverse"}, {"LOWER": "possession"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "advertisement"},
            {"LOWER": "containing"},
            {"LOWER": "indication"},
            {"LOWER": "of"},
            {"LOWER": "price"},
            {"LOWER": "of"},
            {"LOWER": "car"},
            {"LOWER": "excluding"},
            {"LOWER": "transfer"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "advertisement"}, {"LOWER": "on"}, {"LOWER": "bus"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "advertisements"},
            {"ORTH": ","},
            {"LOWER": "control"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "advertising"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "advertising"}, {"LOWER": "for"}, {"LOWER": "adoption"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "advertising"}, {"LOWER": "standards"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "advice"}, {"LOWER": "and"}, {"LOWER": "assistance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "advice"},
            {"LOWER": "on"},
            {"LOWER": "settlement"},
            {"LOWER": "of"},
            {"LOWER": "claim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "advice"}, {"LOWER": "to"}, {"LOWER": "client"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "aerodrome"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "affidavit"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "affidavit"}, {"LOWER": "in"}, {"LOWER": "reply"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "affidavits"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "affiliation"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "after"},
            {"LOWER": "the"},
            {"LOWER": "event"},
            {"LOWER": "insurance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "after"},
            {"LOWER": "the"},
            {"LOWER": "event"},
            {"LOWER": "insurance"},
            {"LOWER": "premiums"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "age"}, {"LOWER": "discrimination"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "age"}, {"LOWER": "assessment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "agency"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "agency"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "agency"}, {"LOWER": "worker"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "agency"}, {"LOWER": "workers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "aggravated"}, {"LOWER": "burglary"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "aggravated"}, {"LOWER": "trespass"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "aggravated"}, {"LOWER": "vehicle"}, {"LOWER": "taking"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "agreed"}, {"LOWER": "orders"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "agreement"},
            {"LOWER": "between"},
            {"LOWER": "revenue"},
            {"LOWER": "and"},
            {"LOWER": "major"},
            {"LOWER": "employer"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "nic"},
            {"LOWER": "contributions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "agreement"}, {"LOWER": "for"}, {"LOWER": "lease"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "agreement"}, {"LOWER": "for"}, {"LOWER": "sale"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "agreement"}, {"LOWER": "to"}, {"LOWER": "compromise"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "agreement"},
            {"LOWER": "to"},
            {"LOWER": "grant"},
            {"LOWER": "lease"},
            {"LOWER": "to"},
            {"LOWER": "nominee"},
            {"LOWER": "of"},
            {"LOWER": "plaintiff"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "agreement"}, {"LOWER": "to"}, {"LOWER": "negotiate"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "agreement"},
            {"LOWER": "to"},
            {"ORTH": "“"},
            {"LOWER": "like"},
            {"LOWER": "effect"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "agreement"},
            {"LOWER": "with"},
            {"LOWER": "crown"},
            {"LOWER": "dependency"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "agreement"},
            {"LOWER": "with"},
            {"LOWER": "respect"},
            {"LOWER": "to"},
            {"LOWER": "child"},
            {"LOWER": "ratified"},
            {"LOWER": "by"},
            {"LOWER": "portuguese"},
            {"LOWER": "court"},
            {"LOWER": "for"},
            {"LOWER": "parental"},
            {"LOWER": "care"},
            {"LOWER": "by"},
            {"LOWER": "two"},
            {"ORTH": "-"},
            {"LOWER": "monthly"},
            {"LOWER": "rotations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "agreements"}, {"LOWER": "registrable"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "agreements"},
            {"LOWER": "with"},
            {"LOWER": "third"},
            {"LOWER": "countries"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "decisions"},
            {"LOWER": "and"},
            {"LOWER": "concerted"},
            {"LOWER": "practice"},
            {"LOWER": "agreements"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "decisions"},
            {"LOWER": "and"},
            {"LOWER": "concerted"},
            {"LOWER": "practice"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "agricultural"}, {"LOWER": "building"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "agricultural"}, {"LOWER": "holding"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "agricultural"}, {"LOWER": "land"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "agricultural"},
            {"LOWER": "products"},
            {"LOWER": "and"},
            {"LOWER": "foodstuffs"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "agriculture"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "aid"},
            {"LOWER": "granted"},
            {"LOWER": "in"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "feu"},
            {"LOWER": "treaty"},
            {"LOWER": "article"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "aid"},
            {"LOWER": "in"},
            {"LOWER": "form"},
            {"LOWER": "of"},
            {"LOWER": "implied"},
            {"ORTH": ","},
            {"LOWER": "unlimited"},
            {"LOWER": "guarantee"},
            {"LOWER": "in"},
            {"LOWER": "favour"},
            {"LOWER": "of"},
            {"LOWER": "french"},
            {"LOWER": "public"},
            {"LOWER": "post"},
            {"LOWER": "and"},
            {"LOWER": "telecommunications"},
            {"LOWER": "service"},
            {"LOWER": "as"},
            {"LOWER": "result"},
            {"LOWER": "of"},
            {"LOWER": "its"},
            {"LOWER": "status"},
            {"LOWER": "as"},
            {"LOWER": "publicly"},
            {"ORTH": "-"},
            {"LOWER": "owned"},
            {"LOWER": "establishment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "aiding"}, {"LOWER": "and"}, {"LOWER": "abetting"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "aiding"}, {"LOWER": "unlawful"}, {"LOWER": "acts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "air"},
            {"LOWER": "accident"},
            {"LOWER": "investigation"},
            {"LOWER": "report"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "air"}, {"LOWER": "brakes"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "air"}, {"LOWER": "fares"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "air"}, {"LOWER": "quality"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "air"}, {"LOWER": "services"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "aircraft"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "airspace"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "alcohol"}, {"LOWER": "duty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "alhani"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "aliakmon"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "alien"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "allegation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "allegation"},
            {"LOWER": "that"},
            {"LOWER": "vaccine"},
            {"LOWER": "against"},
            {"LOWER": "hepatitis"},
            {"LOWER": "b"},
            {"LOWER": "defective"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "allocation"},
            {"LOWER": "between"},
            {"LOWER": "parties"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "allocation"}, {"LOWER": "of"}, {"LOWER": "cases"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "allocation"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "allocation"}, {"LOWER": "of"}, {"LOWER": "funds"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "allocation"}, {"LOWER": "scheme"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "allocation"}, {"LOWER": "to"}, {"LOWER": "judiciary"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "allocation"},
            {"LOWER": "to"},
            {"LOWER": "multi"},
            {"LOWER": "-"},
            {"LOWER": "track"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "allowances"},
            {"LOWER": "to"},
            {"LOWER": "cover"},
            {"LOWER": "costs"},
            {"LOWER": "incurred"},
            {"LOWER": "in"},
            {"LOWER": "exercise"},
            {"LOWER": "of"},
            {"LOWER": "parliamentary"},
            {"LOWER": "duties"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "almshouse"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "altar"}, {"LOWER": "frontal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "alteration"}, {"LOWER": "of"}, {"LOWER": "name"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "alteration"},
            {"LOWER": "of"},
            {"LOWER": "valuation"},
            {"LOWER": "bands"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "alternative"}, {"LOWER": "accommodation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "alternative"}, {"LOWER": "charges"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "alternative"}, {"LOWER": "counts"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "alternative"},
            {"LOWER": "dispute"},
            {"LOWER": "resolution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "alternative"},
            {"LOWER": "dispute"},
            {"LOWER": "resolution"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "alternative"}, {"LOWER": "remedy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "alternative"}, {"LOWER": "service"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "alternative"}, {"LOWER": "verdict"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "alternative"}, {"LOWER": "verdicts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ama"}, {"LOWER": "antxine"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "amendment"}, {"LOWER": "of"}, {"LOWER": "pleadings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "amendment"}, {"LOWER": "of"}, {"LOWER": "register"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "amusements"}, {"LOWER": "with"}, {"LOWER": "prizes"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "analysts"}, {"LOWER": "certificate"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ancillary"}, {"LOWER": "relief"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "ancillary"},
            {"LOWER": "relief"},
            {"LOWER": "application"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "ancillary"},
            {"LOWER": "relief"},
            {"LOWER": "proceedings"},
            {"LOWER": "costs"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "anguilla"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "animal"}, {"LOWER": "cruely"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "animal"}, {"LOWER": "feed"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "animal"}, {"LOWER": "health"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "annual"}, {"LOWER": "leave"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "annual"},
            {"LOWER": "point"},
            {"ORTH": "-"},
            {"LOWER": "to"},
            {"ORTH": "-"},
            {"LOWER": "point"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "annuity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "annulment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "anonymity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "anonymity"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "anonymous"}, {"LOWER": "witness"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "competitive"},
            {"LOWER": "practices"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "arbitration"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "competitive"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "deprivation"},
            {"LOWER": "rule"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "deprivation"},
            {"LOWER": "rule"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "anti"}, {"LOWER": "-"}, {"LOWER": "dumping"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "dumping"},
            {"LOWER": "duty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "enforcement"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "smoking"},
            {"LOWER": "measures"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "social"},
            {"LOWER": "behaviour"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "suit"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "anti"},
            {"LOWER": "-"},
            {"LOWER": "suit"},
            {"LOWER": "injunctions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "antigua"}, {"LOWER": "and"}, {"LOWER": "barbuda"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "antigua"}, {"LOWER": "and"}, {"LOWER": "barbuda"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "anton"}, {"LOWER": "piller"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "apparent"}, {"LOWER": "bias"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "against"},
            {"LOWER": "adoption"},
            {"LOWER": "order"},
            {"LOWER": "and"},
            {"LOWER": "no"},
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "direct"},
            {"LOWER": "contact"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "against"},
            {"LOWER": "decision"},
            {"LOWER": "of"},
            {"LOWER": "general"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appeal"}, {"LOWER": "against"}, {"LOWER": "dismissal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appeal"}, {"LOWER": "against"}, {"LOWER": "sentence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "as"},
            {"LOWER": "of"},
            {"LOWER": "right"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "by"},
            {"LOWER": "general"},
            {"LOWER": "medical"},
            {"LOWER": "council"},
            {"LOWER": "against"},
            {"LOWER": "failure"},
            {"LOWER": "of"},
            {"LOWER": "tribunal"},
            {"LOWER": "to"},
            {"LOWER": "impose"},
            {"LOWER": "sanction"},
            {"LOWER": "of"},
            {"LOWER": "erasure"},
            {"LOWER": "or"},
            {"LOWER": "suspension"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "by"},
            {"LOWER": "mother"},
            {"LOWER": "and"},
            {"LOWER": "child"},
            {"LOWER": "against"},
            {"LOWER": "return"},
            {"LOWER": "order"},
            {"LOWER": "dismissed"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "from"},
            {"LOWER": "disciplinary"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "from"},
            {"LOWER": "district"},
            {"LOWER": "judge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "from"},
            {"LOWER": "magistrates"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appeal"}, {"LOWER": "from"}, {"LOWER": "registrar"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "from"},
            {"LOWER": "solicitors"},
            {"LOWER": "disciplinary"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "from"},
            {"LOWER": "upper"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "on"},
            {"LOWER": "point"},
            {"LOWER": "of"},
            {"LOWER": "law"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "appeal"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appeal"}, {"LOWER": "to"}, {"LOWER": "commissioner"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "county"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "crown"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "divisional"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "high"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "house"},
            {"LOWER": "of"},
            {"LOWER": "lords"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "magistrates"},
            {"LOWER": "’"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "privy"},
            {"LOWER": "council"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "social"},
            {"LOWER": "security"},
            {"LOWER": "appeal"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "social"},
            {"LOWER": "security"},
            {"LOWER": "commissioner"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "supreme"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "the"},
            {"LOWER": "planning"},
            {"LOWER": "inspector"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appeals"}, {"LOWER": "against"}, {"LOWER": "sentence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeals"},
            {"LOWER": "against"},
            {"LOWER": "conviction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appeal"}, {"LOWER": "against"}, {"LOWER": "sentence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appeal"}, {"LOWER": "against"}, {"LOWER": "conviction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appeals"},
            {"LOWER": "from"},
            {"LOWER": "magistrates"},
            {"LOWER": "courts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appearance"}, {"LOWER": "of"}, {"LOWER": "counsel"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appellant"},
            {"LOWER": "escaping"},
            {"LOWER": "from"},
            {"LOWER": "custody"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "extradition"},
            {"LOWER": "sought"},
            {"LOWER": "by"},
            {"LOWER": "indian"},
            {"LOWER": "government"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appellate"},
            {"LOWER": "courts"},
            {"LOWER": "discretion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appellate"}, {"LOWER": "jurisdiction"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "applicable"}, {"LOWER": "law"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "applicant"}, {"LOWER": "for"}, {"LOWER": "employment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applicant"},
            {"LOWER": "making"},
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "discharge"},
            {"LOWER": "or"},
            {"LOWER": "vary"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applicant"},
            {"LOWER": "not"},
            {"LOWER": "intentionally"},
            {"LOWER": "homeless"},
            {"LOWER": "or"},
            {"LOWER": "in"},
            {"LOWER": "priority"},
            {"LOWER": "need"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applicant"},
            {"LOWER": "operating"},
            {"LOWER": "anonymous"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "wireless"},
            {"LOWER": "local"},
            {"LOWER": "area"},
            {"LOWER": "network"},
            {"LOWER": "free"},
            {"LOWER": "of"},
            {"LOWER": "charge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applicants"},
            {"LOWER": "applying"},
            {"LOWER": "to"},
            {"LOWER": "adopt"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "applicants"}, {"LOWER": "expectation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "applicants"}, {"LOWER": "interest"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applicants"},
            {"LOWER": "not"},
            {"LOWER": "informed"},
            {"LOWER": "of"},
            {"LOWER": "objection"},
            {"LOWER": "to"},
            {"LOWER": "grant"},
            {"LOWER": "of"},
            {"LOWER": "entertainments"},
            {"LOWER": "licence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applicants"},
            {"LOWER": "resident"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "against"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "against"},
            {"LOWER": "third"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "by"},
            {"LOWER": "newspaper"},
            {"LOWER": "group"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "joined"},
            {"LOWER": "as"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "by"},
            {"LOWER": "relative"},
            {"LOWER": "and"},
            {"LOWER": "wife"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "a"},
            {"LOWER": "cleud"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "accommodation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "for"}, {"LOWER": "asylum"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "lawful"},
            {"LOWER": "use"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "declaration"},
            {"LOWER": "of"},
            {"LOWER": "marital"},
            {"LOWER": "status"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "directions"},
            {"LOWER": "pending"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "for"}, {"LOWER": "discharge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "employment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "judgment"},
            {"LOWER": "summons"},
            {"LOWER": "to"},
            {"LOWER": "enforce"},
            {"LOWER": "consent"},
            {"LOWER": "order"},
            {"LOWER": "made"},
            {"LOWER": "in"},
            {"LOWER": "divorce"},
            {"LOWER": "proceedings"},
            {"LOWER": "for"},
            {"LOWER": "payment"},
            {"LOWER": "of"},
            {"LOWER": "periodical"},
            {"LOWER": "child"},
            {"LOWER": "maintenance"},
            {"LOWER": "payments"},
            {"LOWER": "and"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "leapfrog"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "for"}, {"LOWER": "leave"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "new"},
            {"LOWER": "lease"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "new"},
            {"LOWER": "tenancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "recognition"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "divorce"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "for"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "judicial"},
            {"LOWER": "review"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
            {"LOWER": "against"},
            {"LOWER": "sentence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
            {"LOWER": "against"},
            {"LOWER": "conviction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "apply"},
            {"LOWER": "for"},
            {"LOWER": "judicial"},
            {"LOWER": "review"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "port"},
            {"LOWER": "revision"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "premises"},
            {"LOWER": "licence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "registration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "for"}, {"LOWER": "rehearing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "for"}, {"LOWER": "release"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "repayment"},
            {"LOWER": "of"},
            {"LOWER": "overpaid"},
            {"LOWER": "vat"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "retrospective"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "summary"},
            {"LOWER": "judgment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "for"}, {"LOWER": "taxation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "application"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "of"}, {"LOWER": "act"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "of"},
            {"LOWER": "quota"},
            {"LOWER": "in"},
            {"LOWER": "sugar"},
            {"LOWER": "sector"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "annul"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "clarify"},
            {"LOWER": "meaning"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "discharge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "extend"},
            {"LOWER": "stay"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "grant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "lift"},
            {"LOWER": "automatic"},
            {"LOWER": "suspension"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "mental"},
            {"LOWER": "health"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "quash"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "quash"},
            {"LOWER": "acquittal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "quash"},
            {"LOWER": "verdict"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "re"},
            {"LOWER": "-"},
            {"LOWER": "open"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "reopen"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "reopen"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "reopen"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "reopen"},
            {"LOWER": "judicial"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "reopen"},
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "review"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "revoke"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
            {"LOWER": "judgment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
            {"LOWER": "or"},
            {"LOWER": "stay"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
            {"LOWER": "recognition"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "strike"},
            {"LOWER": "out"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "suspend"},
            {"LOWER": "operation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "terminate"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "to"}, {"LOWER": "vary"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "vary"},
            {"LOWER": "leave"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "application"},
            {"LOWER": "without"},
            {"LOWER": "notice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applications"},
            {"LOWER": "against"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "applications"}, {"LOWER": "by"}, {"LOWER": "children"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "applications"}, {"LOWER": "during"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applications"},
            {"LOWER": "for"},
            {"LOWER": "compensation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "applications"}, {"LOWER": "for"}, {"LOWER": "leave"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applications"},
            {"LOWER": "for"},
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "residential"},
            {"LOWER": "developments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "applications"},
            {"LOWER": "to"},
            {"LOWER": "extend"},
            {"LOWER": "stay"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appointment"}, {"LOWER": "and"}, {"LOWER": "removal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appointment"},
            {"LOWER": "of"},
            {"LOWER": "administrative"},
            {"LOWER": "receiver"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appointment"},
            {"LOWER": "of"},
            {"LOWER": "administrators"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appointment"},
            {"LOWER": "of"},
            {"LOWER": "administrators"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appointment"}, {"LOWER": "of"}, {"LOWER": "arbitrator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appointment"}, {"LOWER": "of"}, {"LOWER": "deputy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appointment"}, {"LOWER": "of"}, {"LOWER": "executors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appointment"}, {"LOWER": "of"}, {"LOWER": "receiver"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appointment"},
            {"ORTH": ","},
            {"LOWER": "conditions"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "apportionability"}, {"LOWER": "of"}, {"LOWER": "rent"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "apportionment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "apprentice"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "approach"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "taken"},
            {"LOWER": "where"},
            {"LOWER": "parties"},
            {"LOWER": "foreign"},
            {"LOWER": "nationals"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "appropriate"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "appropriate"}, {"LOWER": "list"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "appropriate"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "appropriate"}, {"LOWER": "procedure"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appropriate"},
            {"LOWER": "structure"},
            {"LOWER": "where"},
            {"LOWER": "parental"},
            {"LOWER": "order"},
            {"LOWER": "unavailable"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appropriation"}, {"LOWER": "of"}, {"LOWER": "goodwill"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appropriation"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "appropriation"},
            {"LOWER": "to"},
            {"LOWER": "beneficiary"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "approval"}, {"LOWER": "by"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "approval"}, {"LOWER": "by"}, {"LOWER": "creditors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "approval"},
            {"LOWER": "of"},
            {"LOWER": "channel"},
            {"LOWER": "islands"},
            {"LOWER": "legislation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "approval"},
            {"LOWER": "of"},
            {"LOWER": "costs"},
            {"LOWER": "budget"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "approval"},
            {"LOWER": "of"},
            {"LOWER": "costs"},
            {"LOWER": "budget"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "approval"},
            {"LOWER": "of"},
            {"LOWER": "settlement"},
            {"LOWER": "of"},
            {"LOWER": "personal"},
            {"LOWER": "injury"},
            {"LOWER": "claim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "approved"}, {"LOWER": "by"}, {"LOWER": "creditors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "approximation"}, {"LOWER": "of"}, {"LOWER": "laws"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitral"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitration"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitration"}, {"LOWER": "appeal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "arbitration"}, {"LOWER": "applications"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitration"}, {"LOWER": "award"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitration"}, {"LOWER": "claim"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitration"}, {"LOWER": "clause"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arbitrator"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "architect"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "architect"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "architects"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "area"},
            {"LOWER": "of"},
            {"LOWER": "freedom"},
            {"ORTH": ","},
            {"LOWER": "security"},
            {"LOWER": "and"},
            {"LOWER": "justice"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "armed"}, {"LOWER": "forces"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "army"}, {"LOWER": "resettlement"}, {"LOWER": "grant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "arraignment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "arrangement"},
            {"LOWER": "facilitating"},
            {"LOWER": "acquisition"},
            {"ORTH": ","},
            {"LOWER": "retention"},
            {"ORTH": ","},
            {"LOWER": "use"},
            {"LOWER": "or"},
            {"LOWER": "control"},
            {"LOWER": "of"},
            {"LOWER": "criminal"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "arrangement"}, {"LOWER": "to"}, {"LOWER": "protect"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "arrangements"}, {"LOWER": "after"}, {"LOWER": "death"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "arrangements"}, {"LOWER": "for"}, {"LOWER": "adoption"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "arranging"},
            {"LOWER": "or"},
            {"LOWER": "facilitating"},
            {"LOWER": "commission"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "sex"},
            {"LOWER": "offence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "arrears"},
            {"LOWER": "of"},
            {"LOWER": "maintenance"},
            {"LOWER": "contributions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "arrest"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "arrest"}, {"LOWER": "of"}, {"LOWER": "ship"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "arrest"}, {"LOWER": "on"}, {"LOWER": "suspicion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "arrest"}, {"LOWER": "without"}, {"LOWER": "warrant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "arrestable"}, {"LOWER": "offence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arrested"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "arson"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "article"}, {"LOWER": "with"}, {"LOWER": "blade"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "articles"},
            {"LOWER": "inconsistent"},
            {"LOWER": "with"},
            {"LOWER": "memorandum"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "articles"}, {"LOWER": "of"}, {"LOWER": "association"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "artificial"},
            {"LOWER": "insemination"},
            {"LOWER": "facilities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "artificial"}, {"LOWER": "transaction"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "artistic"}, {"LOWER": "work"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "asbestos"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "asbestos"}, {"LOWER": "dust"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "asbestos"}, {"LOWER": "exposure"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "assault"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assault"},
            {"LOWER": "by"},
            {"LOWER": "sexual"},
            {"LOWER": "penetration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assault"},
            {"LOWER": "occasioning"},
            {"LOWER": "actual"},
            {"LOWER": "bodily"},
            {"LOWER": "harm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assault"},
            {"LOWER": "on"},
            {"LOWER": "officer"},
            {"LOWER": "of"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assault"},
            {"LOWER": "with"},
            {"LOWER": "intent"},
            {"LOWER": "to"},
            {"LOWER": "resist"},
            {"LOWER": "arrest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "by"}, {"LOWER": "jury"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "of"}, {"LOWER": "bills"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "of"}, {"LOWER": "damages"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assessment"},
            {"LOWER": "of"},
            {"LOWER": "effects"},
            {"LOWER": "of"},
            {"LOWER": "certain"},
            {"LOWER": "public"},
            {"LOWER": "and"},
            {"LOWER": "private"},
            {"LOWER": "projects"},
            {"LOWER": "on"},
            {"LOWER": "environment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assessment"},
            {"LOWER": "of"},
            {"LOWER": "effects"},
            {"LOWER": "of"},
            {"LOWER": "plans"},
            {"LOWER": "and"},
            {"LOWER": "programmes"},
            {"LOWER": "on"},
            {"LOWER": "environment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "of"}, {"LOWER": "evidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assessment"},
            {"LOWER": "of"},
            {"LOWER": "implications"},
            {"LOWER": "for"},
            {"LOWER": "protected"},
            {"LOWER": "site"},
            {"LOWER": "of"},
            {"LOWER": "plan"},
            {"LOWER": "or"},
            {"LOWER": "project"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "of"}, {"LOWER": "need"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "of"}, {"LOWER": "needs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "of"}, {"LOWER": "risk"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assessment"}, {"LOWER": "of"}, {"LOWER": "validity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "appointment"}, {"LOWER": "of"}, {"LOWER": "assessors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "role"}, {"LOWER": "of"}, {"LOWER": "assessors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assets"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "retirement"},
            {"LOWER": "pension"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assignment"}, {"LOWER": "of"}, {"LOWER": "lease"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assignment"},
            {"LOWER": "of"},
            {"LOWER": "lease"},
            {"LOWER": "or"},
            {"LOWER": "underletting"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assignment"},
            {"LOWER": "of"},
            {"LOWER": "whole"},
            {"LOWER": "of"},
            {"LOWER": "property"},
            {"LOWER": "to"},
            {"LOWER": "secure"},
            {"LOWER": "existing"},
            {"LOWER": "debt"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assisting"},
            {"LOWER": "avoidance"},
            {"LOWER": "of"},
            {"LOWER": "prosecution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assisting"}, {"LOWER": "illegal"}, {"LOWER": "entrant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "assisting"}, {"LOWER": "offender"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assisting"},
            {"LOWER": "unlawful"},
            {"LOWER": "immigration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "associated"}, {"LOWER": "development"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "associated"}, {"LOWER": "employer"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "assured"}, {"LOWER": "shorthold"}, {"LOWER": "tenancy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "assured"}, {"LOWER": "tenancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "asylum"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "asylum"},
            {"LOWER": "and"},
            {"LOWER": "immigration"},
            {"LOWER": "tribunal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "asylum"}, {"LOWER": "claim"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "asylum"},
            {"LOWER": "claim"},
            {"LOWER": "refused"},
            {"LOWER": "after"},
            {"LOWER": "conclusion"},
            {"LOWER": "that"},
            {"LOWER": "claimant"},
            {"LOWER": "was"},
            {"LOWER": "not"},
            {"LOWER": "credible"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "asylum"}, {"LOWER": "seeker"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "asylum"}, {"LOWER": "seekers"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "asylum"}, {"LOWER": "support"}, {"LOWER": "benefit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "asylum"}, {"LOWER": "support"}, {"LOWER": "claims"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "attempt"},
            {"LOWER": "to"},
            {"LOWER": "cause"},
            {"LOWER": "grievous"},
            {"LOWER": "bodily"},
            {"LOWER": "harm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "attempt"},
            {"LOWER": "to"},
            {"LOWER": "pervert"},
            {"LOWER": "course"},
            {"LOWER": "of"},
            {"LOWER": "justice"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "attempted"}, {"LOWER": "murder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "attendance"}, {"LOWER": "allowance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "attendance"}, {"LOWER": "at"}, {"LOWER": "place"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "attestation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "attorney"}, {"LOWER": "general"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "attorney"},
            {"LOWER": "general"},
            {"LOWER": "’s"},
            {"LOWER": "certificate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "attorney"}, {"ORTH": "-"}, {"LOWER": "general"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "auctioneer"}, {"LOWER": "and"}, {"LOWER": "valuer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "audi"}, {"LOWER": "alteram"}, {"LOWER": "partem"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "audiovisual"},
            {"LOWER": "media"},
            {"LOWER": "services"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "audit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "audit"}, {"LOWER": "commission"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "auditor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "auditors"}, {"LOWER": "duties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "aumbry"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "authorisation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "authorisation"},
            {"LOWER": "or"},
            {"LOWER": "qualification"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "authorisation"},
            {"LOWER": "or"},
            {"LOWER": "qualification"},
            {"LOWER": "for"},
            {"LOWER": "employment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "authority"},
            {"LOWER": "requiring"},
            {"LOWER": "dismissal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "authority"}, {"LOWER": "to"}, {"LOWER": "bring"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "authority"}, {"LOWER": "to"}, {"LOWER": "proceed"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "authors"}, {"LOWER": "’"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "author"}, {"LOWER": "’s"}, {"LOWER": "rights"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "automatic"}, {"LOWER": "deportation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "automatic"},
            {"LOWER": "reference"},
            {"LOWER": "to"},
            {"LOWER": "arbitration"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "automatic"}, {"LOWER": "suspension"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "automatism"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "autrefois"}, {"LOWER": "acquit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "autrefois"}, {"LOWER": "convict"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "availability"}, {"LOWER": "of"}, {"LOWER": "remedy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "award"}, {"LOWER": "criteria"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "award"},
            {"LOWER": "for"},
            {"LOWER": "financial"},
            {"LOWER": "dependency"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "award"}, {"LOWER": "of"}, {"LOWER": "damages"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "award"},
            {"LOWER": "of"},
            {"LOWER": "equitable"},
            {"LOWER": "compensation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "award"}, {"LOWER": "on"}, {"LOWER": "damages"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "awards"}, {"LOWER": "procedure"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "back"}, {"LOWER": "duty"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bad"}, {"LOWER": "character"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bad"},
            {"LOWER": "character"},
            {"LOWER": "of"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "accused"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bad"},
            {"LOWER": "character"},
            {"LOWER": "of"},
            {"LOWER": "complainant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bad"},
            {"LOWER": "character"},
            {"LOWER": "of"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "defendant"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bahamas"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bail"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conditions"}, {"LOWER": "of"}, {"LOWER": "bail"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bail"}, {"LOWER": "breach"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bail"}, {"LOWER": "pending"}, {"LOWER": "appeal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bailiff"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bailment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "baleares"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ballot"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "banker"}, {"LOWER": "and"}, {"LOWER": "customer"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankers"}, {"LOWER": "books"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankers"}, {"LOWER": "draft"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "banking"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "banking"}, {"LOWER": "business"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankruptcy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bankruptcy"},
            {"LOWER": "act"},
            {"ORTH": ","},
            {"LOWER": "1869"},
            {"ORTH": "("},
            {"LOWER": "32"},
            {"LOWER": "&"},
            {"LOWER": "33"},
            {"LOWER": "vict"},
            {"ORTH": "."},
            {"LOWER": "c."},
            {"LOWER": "71"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "s."},
            {"LOWER": "125"},
            {"ORTH": ","},
            {"LOWER": "sub"},
            {"ORTH": "-"},
            {"LOWER": "s"},
            {"ORTH": "."},
            {"LOWER": "12"},
            {"ORTH": "."},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "notice"}, {"LOWER": "file"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "of"}, {"LOWER": "assignor"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "petition"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "petition"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bankruptcy"},
            {"LOWER": "restrictions"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bankruptcy"}, {"LOWER": "rules"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bankruptcy"},
            {"LOWER": "rules"},
            {"LOWER": "1952"},
            {"ORTH": "("},
            {"LOWER": "s.i"},
            {"ORTH": "."},
            {"LOWER": "1952"},
            {"LOWER": "no"},
            {"ORTH": "."},
            {"LOWER": "2113"},
            {"ORTH": "("},
            {"LOWER": "l."},
            {"LOWER": "14"},
            {"ORTH": ")"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "r."},
            {"LOWER": "173(1"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "banks"}, {"LOWER": "liability"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bank"}, {"LOWER": "’s"}, {"LOWER": "liability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bar"}, {"LOWER": "to"}, {"LOWER": "extradition"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "barbados"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bargaining"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "baronetcy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "barrister"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "barristers"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "basis"}, {"LOWER": "of"}, {"LOWER": "award"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "basis"}, {"LOWER": "of"}, {"LOWER": "decision"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "basis"},
            {"LOWER": "on"},
            {"LOWER": "which"},
            {"LOWER": "costs"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "awarded"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "battery"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "battlefield"}, {"LOWER": "designation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bazias"}, {"LOWER": "3"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bazias"}, {"LOWER": "4"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bearer"}, {"LOWER": "share"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bee"}, {"LOWER": "keepers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "before"},
            {"LOWER": "commencement"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "being"},
            {"LOWER": "concerned"},
            {"LOWER": "in"},
            {"LOWER": "production"},
            {"LOWER": "of"},
            {"LOWER": "controlled"},
            {"LOWER": "drug"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "being"},
            {"LOWER": "knowingly"},
            {"LOWER": "concerned"},
            {"LOWER": "in"},
            {"LOWER": "an"},
            {"LOWER": "attempt"},
            {"LOWER": "to"},
            {"LOWER": "export"},
            {"LOWER": "restricted"},
            {"LOWER": "goods"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "belize"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "belonging"},
            {"LOWER": "to"},
            {"LOWER": "proscribed"},
            {"LOWER": "organisation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "benarty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "benefice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "beneficial"}, {"LOWER": "interests"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "beneficial"}, {"LOWER": "owner"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "beneficial"},
            {"LOWER": "ownership"},
            {"LOWER": "of"},
            {"LOWER": "former"},
            {"LOWER": "matrimonial"},
            {"LOWER": "home"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "beneficial"},
            {"LOWER": "owner"},
            {"LOWER": "’s"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "possession"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "beneficiary"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "beneficiary"},
            {"LOWER": "of"},
            {"LOWER": "letter"},
            {"LOWER": "of"},
            {"LOWER": "credit"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "beneficiary"},
            {"LOWER": "’s"},
            {"LOWER": "rights"},
            {"LOWER": "against"},
            {"LOWER": "trustee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "benefit"}, {"LOWER": "and"}, {"LOWER": "burden"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "benefit"}, {"LOWER": "cap"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "benefit"}, {"LOWER": "cap"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "benefit"}, {"LOWER": "of"}, {"LOWER": "contract"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "benefits"}, {"LOWER": "claimant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "benefits"},
            {"LOWER": "granted"},
            {"LOWER": "by"},
            {"LOWER": "public"},
            {"LOWER": "airport"},
            {"LOWER": "operator"},
            {"LOWER": "to"},
            {"LOWER": "low"},
            {"ORTH": "-"},
            {"LOWER": "cost"},
            {"LOWER": "airline"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bermuda"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "best"}, {"LOWER": "interests"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "best"}, {"LOWER": "interests"}, {"LOWER": "decision"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "betting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bias"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bijela"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "costs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "exchange"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "exchange"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "indictment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "lading"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "lading"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bill"}, {"LOWER": "of"}, {"LOWER": "sale"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bills"}, {"LOWER": "of"}, {"LOWER": "exchange"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bills"}, {"LOWER": "of"}, {"LOWER": "exchange"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "binding"}, {"LOWER": "over"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bingo"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "biofuels"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "biometric"}, {"LOWER": "passport"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "biotechnical"}, {"LOWER": "invention"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "biotechnical"}, {"LOWER": "inventions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "protection"},
            {"LOWER": "areas"},
            {"LOWER": "for"},
            {"LOWER": "birds"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "birth"}, {"LOWER": "of"}, {"LOWER": "child"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "black"}, {"LOWER": "community"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "blackmail"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "blasphemous"}, {"LOWER": "libel"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "block"}, {"LOWER": "of"}, {"LOWER": "flats"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "blood"}, {"LOWER": "alcohol"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "blood"}, {"LOWER": "specimen"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "blood"}, {"LOWER": "test"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "blood"}, {"LOWER": "tests"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "board"},
            {"LOWER": "and"},
            {"LOWER": "lodging"},
            {"LOWER": "allowance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "board"}, {"LOWER": "of"}, {"LOWER": "visitors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "board"},
            {"ORTH": ","},
            {"LOWER": "provision"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bodies"},
            {"LOWER": "other"},
            {"LOWER": "than"},
            {"LOWER": "employers"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "body"}, {"LOWER": "corporate"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bone"}, {"LOWER": "marrow"}, {"LOWER": "donation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bonus"}, {"LOWER": "scheme"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bonus"},
            {"LOWER": "to"},
            {"LOWER": "bank"},
            {"LOWER": "employees"},
            {"LOWER": "structured"},
            {"LOWER": "as"},
            {"LOWER": "restricted"},
            {"LOWER": "securities"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "border"}, {"LOWER": "controls"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "both"},
            {"LOWER": "parties"},
            {"LOWER": "seeking"},
            {"LOWER": "to"},
            {"LOWER": "challenge"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "boundary"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "boundary"}, {"LOWER": "commission"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "boundary"}, {"LOWER": "fence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bow"}, {"LOWER": "spring"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bowbelle"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "brain"}, {"LOWER": "injury"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "and"},
            {"LOWER": "procuring"},
            {"LOWER": "breach"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "community"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "community"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "competition"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "conditions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "confidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contracts"},
            {"LOWER": "under"},
            {"LOWER": "seal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contractual"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "solicitors"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contractual"},
            {"LOWER": "obligations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "convention"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "court"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "court"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "covenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "duty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "duty"},
            {"LOWER": "by"},
            {"LOWER": "employer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "conduct"},
            {"LOWER": "undertaking"},
            {"LOWER": "to"},
            {"LOWER": "ensure"},
            {"LOWER": "persons"},
            {"LOWER": "not"},
            {"LOWER": "in"},
            {"LOWER": "its"},
            {"LOWER": "employment"},
            {"LOWER": "not"},
            {"LOWER": "exposed"},
            {"LOWER": "to"},
            {"LOWER": "risks"},
            {"LOWER": "to"},
            {"LOWER": "health"},
            {"LOWER": "and"},
            {"LOWER": "safety"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "fiduciary"},
            {"LOWER": "duty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "freezing"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "injunction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "natural"},
            {"LOWER": "justice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "peace"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "planning"},
            {"LOWER": "control"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "restrictive"},
            {"LOWER": "covenant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "right"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "settlement"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "statutory"},
            {"LOWER": "duty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "trust"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "undertaking"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "undertakings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "union"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "breach"}, {"LOWER": "of"}, {"LOWER": "warranty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "warranty"},
            {"LOWER": "of"},
            {"LOWER": "authority"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "break"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "breath"},
            {"LOWER": "specimen"},
            {"LOWER": "for"},
            {"LOWER": "analysis"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "breath"}, {"LOWER": "test"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bribe"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bribed"}, {"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bridleway"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "bringing"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "british"},
            {"LOWER": "broadcasting"},
            {"LOWER": "corporation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "british"}, {"LOWER": "citizen"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "british"}, {"LOWER": "citizens"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "british"}, {"LOWER": "citizenship"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "british"}, {"LOWER": "coal"}, {"LOWER": "corporation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "british"}, {"LOWER": "overseas"}, {"LOWER": "citizens"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "british"}, {"LOWER": "protected"}, {"LOWER": "person"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "broadcast"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "broadcasting"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "broadcasting"},
            {"LOWER": "complaints"},
            {"LOWER": "commission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "broadcasts"}, {"LOWER": "and"}, {"LOWER": "films"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "broker"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "brokers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "brokers"}, {"LOWER": "slip"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "brother"}, {"LOWER": "and"}, {"LOWER": "sister"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "budget"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "buggery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "builder"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "building"},
            {"LOWER": "and"},
            {"LOWER": "do"},
            {"ORTH": "-"},
            {"LOWER": "it"},
            {"ORTH": "-"},
            {"LOWER": "yourself"},
            {"LOWER": "supplier"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "building"}, {"LOWER": "contract"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "building"}, {"LOWER": "contractor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "building"},
            {"LOWER": "plan"},
            {"LOWER": "“"},
            {"LOWER": "for"},
            {"LOWER": "development"},
            {"LOWER": "within"},
            {"LOWER": "urban"},
            {"LOWER": "area"},
            {"ORTH": "”"},
            {"LOWER": "exempted"},
            {"LOWER": "from"},
            {"LOWER": "environmental"},
            {"LOWER": "assessment"},
            {"LOWER": "under"},
            {"LOWER": "national"},
            {"LOWER": "legislation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "building"}, {"LOWER": "site"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "building"}, {"LOWER": "society"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "building"}, {"LOWER": "society"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "building"}, {"LOWER": "works"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "bulk"},
            {"LOWER": "of"},
            {"LOWER": "husband"},
            {"LOWER": "’s"},
            {"LOWER": "wealth"},
            {"LOWER": "in"},
            {"LOWER": "offshore"},
            {"LOWER": "discretionary"},
            {"LOWER": "trusts"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "bumbesti"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bunkers"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "burden"}, {"LOWER": "of"}, {"LOWER": "proof"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "burden"}, {"LOWER": "of"}, {"LOWER": "proof"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "burglary"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "burial"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "burial"}, {"LOWER": "ground"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bus"}, {"LOWER": "lanes"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bus"}, {"LOWER": "passenger"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "business"}, {"LOWER": "interruption"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "business"},
            {"LOWER": "premises"},
            {"ORTH": "("},
            {"LOWER": "security"},
            {"LOWER": "of"},
            {"LOWER": "tenure"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "business"}, {"LOWER": "sale"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "business"}, {"LOWER": "services"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "business"}, {"LOWER": "tenancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "business"}, {"LOWER": "transfer"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "byelaws"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "bylaw"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "byway"},
            {"LOWER": "open"},
            {"LOWER": "to"},
            {"LOWER": "all"},
            {"LOWER": "traffic"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "c.i.f"}, {"ORTH": "."}, {"LOWER": "buyers"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "call"},
            {"LOWER": "for"},
            {"LOWER": "tenders"},
            {"LOWER": "for"},
            {"LOWER": "helicopter"},
            {"LOWER": "support"},
            {"LOWER": "for"},
            {"LOWER": "eulex"},
            {"LOWER": "mission"},
            {"LOWER": "in"},
            {"LOWER": "kosovo"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "canada"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cancellation"},
            {"LOWER": "of"},
            {"LOWER": "65"},
            {"LOWER": "%"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "tax"},
            {"LOWER": "debt"},
            {"LOWER": "in"},
            {"LOWER": "a"},
            {"LOWER": "collective"},
            {"LOWER": "bankruptcy"},
            {"LOWER": "procedure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cancellation"},
            {"LOWER": "of"},
            {"LOWER": "claimants"},
            {"LOWER": "’"},
            {"LOWER": "event"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cancellation"}, {"LOWER": "of"}, {"LOWER": "passport"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cancellation"}, {"LOWER": "of"}, {"LOWER": "passport"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "capacity"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "to"},
            {"LOWER": "borrow"},
            {"LOWER": "under"},
            {"LOWER": "private"},
            {"LOWER": "law"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "capacity"},
            {"LOWER": "to"},
            {"LOWER": "consent"},
            {"LOWER": "to"},
            {"LOWER": "adoption"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "capacity"},
            {"LOWER": "to"},
            {"LOWER": "consent"},
            {"LOWER": "to"},
            {"LOWER": "sexual"},
            {"LOWER": "relations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "capacity"}, {"LOWER": "to"}, {"LOWER": "marry"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "capacity"}, {"LOWER": "to"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cape"}, {"LOWER": "providence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "capital"}, {"LOWER": "allowance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "capital"}, {"LOWER": "allowances"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "capital"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "capital"}, {"LOWER": "gains"}, {"LOWER": "tax"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "capital"}, {"LOWER": "gains"}, {"LOWER": "tax"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "capital"}, {"LOWER": "murder"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "capital"}, {"LOWER": "transfer"}, {"LOWER": "tax"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "capitan"}, {"LOWER": "san"}, {"LOWER": "luis"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "car"},
            {"LOWER": "and"},
            {"LOWER": "driver"},
            {"LOWER": "hire"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "car"}, {"LOWER": "auction"}, {"LOWER": "business"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "car"}, {"LOWER": "hire"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "car"}, {"LOWER": "park"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "damage"}, {"LOWER": "to"}, {"LOWER": "car"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "caravan"}, {"LOWER": "site"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "care"},
            {"LOWER": "and"},
            {"LOWER": "placement"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "care"}, {"LOWER": "and"}, {"LOWER": "protection"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "care"}, {"LOWER": "component"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "care"}, {"LOWER": "home"}, {"LOWER": "fees"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "care"}, {"LOWER": "home"}, {"LOWER": "payments"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "care"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "care"}, {"LOWER": "orders"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "care"},
            {"LOWER": "orders"},
            {"LOWER": "made"},
            {"LOWER": "at"},
            {"LOWER": "conclusion"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "care"}, {"LOWER": "plan"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "care"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "care"},
            {"LOWER": "proceedings"},
            {"LOWER": "following"},
            {"LOWER": "murder"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "’s"},
            {"LOWER": "mother"},
            {"LOWER": "by"},
            {"LOWER": "father"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "care"}, {"LOWER": "standards"}, {"LOWER": "tribunal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "careless"}, {"LOWER": "driving"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "carer"}, {"LOWER": "’s"}, {"LOWER": "allowance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "damage"}, {"LOWER": "to"}, {"LOWER": "cargo"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "intermixture"}, {"LOWER": "of"}, {"LOWER": "cargo"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "loading"}, {"LOWER": "of"}, {"LOWER": "cargo"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "cargo"},
            {"ORTH": ","},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "carriage"}, {"LOWER": "by"}, {"LOWER": "air"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "carriage"},
            {"LOWER": "by"},
            {"LOWER": "air"},
            {"LOWER": "limitation"},
            {"LOWER": "of"},
            {"LOWER": "liability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "carriage"}, {"LOWER": "by"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "carriage"}, {"LOWER": "by"}, {"LOWER": "sea"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "carriage"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "carriage"}, {"LOWER": "of"}, {"LOWER": "passengers"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "carriers"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "carriers"}, {"LOWER": "liability"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "carrier"}, {"LOWER": "’s"}, {"LOWER": "liability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "carrier"},
            {"LOWER": "’s"},
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "compensate"},
            {"LOWER": "for"},
            {"LOWER": "delays"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "carrier"},
            {"LOWER": "’s"},
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "compensate"},
            {"LOWER": "passengers"},
            {"LOWER": "for"},
            {"LOWER": "cancellation"},
            {"LOWER": "of"},
            {"LOWER": "flight"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "carrier"},
            {"LOWER": "’s"},
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "compensate"},
            {"LOWER": "passengers"},
            {"LOWER": "for"},
            {"LOWER": "cancellation"},
            {"LOWER": "or"},
            {"LOWER": "delay"},
            {"LOWER": "of"},
            {"LOWER": "flight"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "carrier"},
            {"LOWER": "’s"},
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "compensate"},
            {"LOWER": "passengers"},
            {"LOWER": "for"},
            {"LOWER": "delayed"},
            {"LOWER": "flight"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "carrier"},
            {"LOWER": "’s"},
            {"LOWER": "obligation"},
            {"LOWER": "under"},
            {"LOWER": "regulation"},
            {"LOWER": "to"},
            {"LOWER": "compensate"},
            {"LOWER": "passengers"},
            {"LOWER": "for"},
            {"LOWER": "cancellation"},
            {"LOWER": "or"},
            {"LOWER": "delay"},
            {"LOWER": "of"},
            {"LOWER": "flight"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "cartel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "case"}, {"LOWER": "management"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "case"},
            {"LOWER": "management"},
            {"LOWER": "conference"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "case"}, {"LOWER": "management"}, {"LOWER": "hearing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "case"}, {"LOWER": "management"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "case"}, {"LOWER": "management"}, {"LOWER": "orders"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "case"}, {"LOWER": "management"}, {"LOWER": "powers"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "case"}, {"LOWER": "stated"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cash"},
            {"LOWER": "under"},
            {"LOWER": "control"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "categories"}, {"LOWER": "of"}, {"LOWER": "awards"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "categorisation"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "causation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cause"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cause"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "causing"}, {"LOWER": "death"}, {"LOWER": "by"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "death"},
            {"LOWER": "by"},
            {"LOWER": "dangerous"},
            {"LOWER": "driving"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "death"},
            {"LOWER": "by"},
            {"LOWER": "reckless"},
            {"LOWER": "driving"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "death"},
            {"LOWER": "by"},
            {"LOWER": "reckless"},
            {"LOWER": "driving"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "death"},
            {"LOWER": "when"},
            {"LOWER": "under"},
            {"LOWER": "influence"},
            {"LOWER": "of"},
            {"LOWER": "drink"},
            {"LOWER": "or"},
            {"LOWER": "drugs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "noxious"},
            {"LOWER": "thing"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "taken"},
            {"LOWER": "by"},
            {"LOWER": "another"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "or"},
            {"LOWER": "allowing"},
            {"LOWER": "death"},
            {"LOWER": "of"},
            {"LOWER": "child"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "or"},
            {"LOWER": "allowing"},
            {"LOWER": "the"},
            {"LOWER": "death"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "child"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "polluting"},
            {"LOWER": "matter"},
            {"LOWER": "to"},
            {"LOWER": "enter"},
            {"LOWER": "controlled"},
            {"LOWER": "waters"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "causing"},
            {"LOWER": "or"},
            {"LOWER": "allowing"},
            {"LOWER": "death"},
            {"LOWER": "of"},
            {"LOWER": "vulnerable"},
            {"LOWER": "adult"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "cayman"}, {"LOWER": "islands"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cell"}, {"LOWER": "searches"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cemetery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "central"}, {"LOWER": "funds"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "central"}, {"LOWER": "services"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "centre"},
            {"LOWER": "of"},
            {"LOWER": "main"},
            {"LOWER": "interests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "alternative"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "certificate"}, {"LOWER": "of"}, {"LOWER": "entry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "independence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "lawful"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "lawful"},
            {"LOWER": "use"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "lawfulness"},
            {"LOWER": "of"},
            {"LOWER": "proposed"},
            {"LOWER": "use"},
            {"LOWER": "or"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "lawfulness"},
            {"LOWER": "of"},
            {"LOWER": "proposed"},
            {"LOWER": "works"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "lawfulness"},
            {"LOWER": "of"},
            {"LOWER": "proposed"},
            {"LOWER": "works"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "certificate"}, {"LOWER": "of"}, {"LOWER": "title"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "certification"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certification"},
            {"LOWER": "of"},
            {"LOWER": "claim"},
            {"LOWER": "as"},
            {"LOWER": "“"},
            {"LOWER": "clearly"},
            {"LOWER": "unfounded"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certification"},
            {"LOWER": "of"},
            {"LOWER": "human"},
            {"LOWER": "rights"},
            {"LOWER": "claim"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "certification"}, {"LOWER": "officer"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "certification"},
            {"LOWER": "precluding"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "certification"}, {"LOWER": "requirement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "certiorari"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cesser"},
            {"LOWER": "of"},
            {"LOWER": "succession"},
            {"LOWER": "duty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "challenge"},
            {"LOWER": "to"},
            {"LOWER": "administrators"},
            {"LOWER": "conduct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "challenge"},
            {"LOWER": "to"},
            {"LOWER": "administrator"},
            {"LOWER": "’s"},
            {"LOWER": "conduct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "challenge"}, {"LOWER": "to"}, {"LOWER": "award"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "challenge"},
            {"LOWER": "to"},
            {"LOWER": "contract"},
            {"LOWER": "award"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "challenge"},
            {"LOWER": "to"},
            {"LOWER": "exercise"},
            {"LOWER": "of"},
            {"LOWER": "discretion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "challenge"}, {"LOWER": "to"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "challenge"},
            {"LOWER": "to"},
            {"LOWER": "lawfulness"},
            {"LOWER": "of"},
            {"LOWER": "detention"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "chambers"}, {"LOWER": "application"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "chambers"}, {"LOWER": "applications"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "champerty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "chancery"},
            {"LOWER": "and"},
            {"LOWER": "queens"},
            {"LOWER": "bench"},
            {"LOWER": "divisions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "chancery"}, {"LOWER": "chambers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "chancery"}, {"LOWER": "division"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "change"}, {"LOWER": "in"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "change"}, {"LOWER": "of"}, {"LOWER": "forename"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "change"}, {"LOWER": "of"}, {"LOWER": "parties"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "change"}, {"LOWER": "of"}, {"LOWER": "position"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "change"},
            {"LOWER": "of"},
            {"LOWER": "residence"},
            {"LOWER": "pending"},
            {"LOWER": "final"},
            {"LOWER": "disposal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "change"}, {"LOWER": "of"}, {"LOWER": "surname"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "change"}, {"LOWER": "of"}, {"LOWER": "use"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "channel"}, {"LOWER": "islands"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "channel"}, {"LOWER": "ranger"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "chapel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "characteristics"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "charge"}, {"LOWER": "on"}, {"LOWER": "property"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "charge"},
            {"LOWER": "over"},
            {"LOWER": "book"},
            {"LOWER": "debts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "chargeability"}, {"LOWER": "of"}, {"LOWER": "vat"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "charged"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charges"}, {"LOWER": "register"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charges"}, {"LOWER": "registrable"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charging"}, {"LOWER": "clause"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charging"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "charging"}, {"LOWER": "order"}, {"LOWER": "nisi"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "charging"},
            {"LOWER": "order"},
            {"LOWER": "on"},
            {"LOWER": "land"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "charitable"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charitable"}, {"LOWER": "funds"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "charitable"},
            {"LOWER": "housing"},
            {"LOWER": "association"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "charitable"}, {"LOWER": "organisation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "charitable"}, {"LOWER": "purposes"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charitable"}, {"LOWER": "trust"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charities"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charity"}, {"LOWER": "commission"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charity"}, {"LOWER": "commissioners"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "charity"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "charter"},
            {"LOWER": "of"},
            {"LOWER": "fundamental"},
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "charterparty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "chattel"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "chattel"},
            {"LOWER": "found"},
            {"LOWER": "on"},
            {"LOWER": "land"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "chattels"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cheating"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cheating"}, {"LOWER": "public"}, {"LOWER": "revenue"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cheating"},
            {"LOWER": "the"},
            {"LOWER": "public"},
            {"LOWER": "revenue"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "chemical"}, {"LOWER": "solvent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cheque"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "chevron"}, {"LOWER": "north"}, {"LOWER": "america"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "chicago"}, {"LOWER": "convention"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "chief"},
            {"LOWER": "magistrate"},
            {"LOWER": "sending"},
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "case"},
            {"LOWER": "to"},
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "for"},
            {"LOWER": "decision"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "extradition"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}, {"LOWER": "abduction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "child"},
            {"LOWER": "abduction"},
            {"LOWER": "and"},
            {"LOWER": "custody"},
            {"LOWER": "act"},
            {"LOWER": "1985"},
            {"ORTH": "("},
            {"LOWER": "c"},
            {"LOWER": "60"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "sch"},
            {"LOWER": "1"},
            {"ORTH": ","},
            {"LOWER": "arts"},
            {"LOWER": "3"},
            {"ORTH": ","},
            {"LOWER": "13"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}, {"LOWER": "abuse"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "abuse"}, {"LOWER": "register"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "child"},
            {"LOWER": "aged"},
            {"LOWER": "14"},
            {"LOWER": "refusing"},
            {"LOWER": "to"},
            {"LOWER": "return"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}, {"LOWER": "allowance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "arrangement"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "arrangements"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "arrangements"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "child"},
            {"LOWER": "born"},
            {"LOWER": "to"},
            {"LOWER": "mother"},
            {"LOWER": "aged"},
            {"LOWER": "under"},
            {"LOWER": "16"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "in"}, {"LOWER": "care"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "child"},
            {"LOWER": "in"},
            {"LOWER": "care"},
            {"LOWER": "of"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}, {"LOWER": "maintenance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}, {"LOWER": "maintenance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "maintenance"}, {"LOWER": "orders"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}, {"LOWER": "minder"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "child"},
            {"LOWER": "of"},
            {"LOWER": "hungarian"},
            {"LOWER": "parents"},
            {"LOWER": "abandoned"},
            {"LOWER": "in"},
            {"LOWER": "england"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}, {"LOWER": "parent"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "child"},
            {"LOWER": "received"},
            {"LOWER": "into"},
            {"LOWER": "care"},
            {"LOWER": "by"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "sexual"}, {"LOWER": "exploitation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "support"}, {"LOWER": "agency"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "child"}, {"LOWER": "support"}, {"LOWER": "maintenance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indecent"},
            {"LOWER": "photograph"},
            {"LOWER": "of"},
            {"LOWER": "child"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "child"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unborn"}, {"LOWER": "child"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "childminder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "children"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "and"},
            {"LOWER": "family"},
            {"LOWER": "reporter"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "and"},
            {"LOWER": "young"},
            {"LOWER": "persons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "and"},
            {"LOWER": "young"},
            {"LOWER": "persons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "at"},
            {"LOWER": "high"},
            {"LOWER": "risk"},
            {"LOWER": "of"},
            {"LOWER": "emotional"},
            {"LOWER": "harm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "between"},
            {"LOWER": "ages"},
            {"LOWER": "of"},
            {"LOWER": "10"},
            {"LOWER": "and"},
            {"LOWER": "14"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "family"},
            {"LOWER": "proceedings"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "children"}, {"LOWER": "in"}, {"LOWER": "need"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "children"}, {"LOWER": "of"}, {"LOWER": "immigrants"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "orphaned"},
            {"LOWER": "when"},
            {"LOWER": "parents"},
            {"LOWER": "murdered"},
            {"LOWER": "in"},
            {"LOWER": "france"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "sent"},
            {"LOWER": "to"},
            {"LOWER": "pakistan"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "children"},
            {"LOWER": "unlawfully"},
            {"LOWER": "removed"},
            {"LOWER": "and"},
            {"LOWER": "retained"},
            {"LOWER": "in"},
            {"LOWER": "sudan"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offences"}, {"LOWER": "against"}, {"LOWER": "children"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offences"}, {"LOWER": "by"}, {"LOWER": "children"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "childrens"}, {"LOWER": "guardian"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "children"}, {"LOWER": "’s"}, {"LOWER": "hearings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "childs"}, {"LOWER": "maintenance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "childs"}, {"LOWER": "views"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "chirpopractor"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "choice"}, {"LOWER": "of"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "choice"}, {"LOWER": "of"}, {"LOWER": "representative"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "choked"}, {"LOWER": "watercourse"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "choko"}, {"LOWER": "star"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "chose"}, {"LOWER": "in"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "chose"}, {"LOWER": "in"}, {"LOWER": "action"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "church"}, {"LOWER": "interior"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "church"}, {"LOWER": "minister"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "churchyard"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "circuit"}, {"LOWER": "judge"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "citation"}, {"LOWER": "of"}, {"LOWER": "authorities"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "citation"}, {"LOWER": "of"}, {"LOWER": "authority"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "citizenship"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "citizenship"}, {"LOWER": "of"}, {"LOWER": "union"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "action"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "civil"},
            {"LOWER": "and"},
            {"LOWER": "family"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "appeals"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "appeals"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "aviation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "claims"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "contempt"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "courts"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "civil"}, {"LOWER": "legal"}, {"LOWER": "services"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "matter"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "recovery"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "civil"}, {"LOWER": "restraint"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "right"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "civil"}, {"LOWER": "servant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "civil"},
            {"LOWER": "service"},
            {"LOWER": "appeal"},
            {"LOWER": "board"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "claim"}, {"LOWER": "for"}, {"LOWER": "asylum"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "asylum"},
            {"LOWER": "refused"},
            {"LOWER": "on"},
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "immigration"},
            {"LOWER": "judge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "procurement"},
            {"LOWER": "process"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "clinical"},
            {"LOWER": "negligence"},
            {"LOWER": "in"},
            {"LOWER": "respect"},
            {"LOWER": "of"},
            {"LOWER": "treatment"},
            {"LOWER": "received"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "claim"}, {"LOWER": "for"}, {"LOWER": "damages"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "claim"}, {"LOWER": "for"}, {"LOWER": "declaration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "infringement"},
            {"LOWER": "of"},
            {"LOWER": "registered"},
            {"LOWER": "design"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "claim"}, {"LOWER": "for"}, {"LOWER": "possession"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "possession"},
            {"LOWER": "of"},
            {"LOWER": "farmhouse"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "claim"}, {"LOWER": "form"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "of"},
            {"LOWER": "race"},
            {"LOWER": "and"},
            {"LOWER": "religious"},
            {"LOWER": "belief"},
            {"LOWER": "discrimination"},
            {"LOWER": "following"},
            {"LOWER": "withdrawal"},
            {"LOWER": "of"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "security"},
            {"LOWER": "clearance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "on"},
            {"LOWER": "behalf"},
            {"LOWER": "of"},
            {"LOWER": "intestate"},
            {"LOWER": "’s"},
            {"LOWER": "estate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "to"},
            {"LOWER": "beneficial"},
            {"LOWER": "interest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "to"},
            {"LOWER": "estate"},
            {"LOWER": "of"},
            {"LOWER": "deceased"},
            {"LOWER": "person"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claim"},
            {"LOWER": "to"},
            {"LOWER": "proprietary"},
            {"LOWER": "interest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "accepting"},
            {"LOWER": "defendant"},
            {"LOWER": "’s"},
            {"LOWER": "conduct"},
            {"LOWER": "as"},
            {"LOWER": "repudiating"},
            {"LOWER": "partnership"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "appealing"},
            {"LOWER": "against"},
            {"LOWER": "judge"},
            {"LOWER": "’s"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "lawful"},
            {"LOWER": "use"},
            {"LOWER": "and"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "as"},
            {"LOWER": "successful"},
            {"LOWER": "party"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "bringing"},
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "defects"},
            {"LOWER": "in"},
            {"LOWER": "building"},
            {"LOWER": "and"},
            {"LOWER": "negligence"},
            {"LOWER": "against"},
            {"LOWER": "defendants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "bringing"},
            {"LOWER": "negligence"},
            {"LOWER": "claim"},
            {"LOWER": "against"},
            {"LOWER": "defendant"},
            {"LOWER": "doctor"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "bringing"},
            {"LOWER": "personal"},
            {"LOWER": "injury"},
            {"LOWER": "action"},
            {"LOWER": "against"},
            {"LOWER": "two"},
            {"LOWER": "defendants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "claiming"},
            {"LOWER": "defendant"},
            {"LOWER": "misappropriating"},
            {"LOWER": "funds"},
            {"LOWER": "from"},
            {"LOWER": "trust"},
            {"LOWER": "account"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "claiming"},
            {"LOWER": "dyslexia"},
            {"LOWER": "and"},
            {"LOWER": "requesting"},
            {"LOWER": "reasonable"},
            {"LOWER": "adjustments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "company"},
            {"LOWER": "trustee"},
            {"LOWER": "under"},
            {"LOWER": "issues"},
            {"LOWER": "of"},
            {"LOWER": "commercial"},
            {"LOWER": "mortgage"},
            {"LOWER": "backed"},
            {"LOWER": "floating"},
            {"LOWER": "rate"},
            {"LOWER": "notes"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "detained"},
            {"LOWER": "in"},
            {"LOWER": "hospital"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "entering"},
            {"LOWER": "into"},
            {"LOWER": "building"},
            {"LOWER": "design"},
            {"LOWER": "contract"},
            {"LOWER": "with"},
            {"LOWER": "defendants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "failing"},
            {"LOWER": "to"},
            {"LOWER": "finish"},
            {"LOWER": "medical"},
            {"LOWER": "degree"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "having"},
            {"LOWER": "no"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "enter"},
            {"LOWER": "and"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "instructing"},
            {"LOWER": "london"},
            {"LOWER": "solicitor"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "lance"},
            {"LOWER": "corporal"},
            {"LOWER": "suffering"},
            {"LOWER": "suffered"},
            {"LOWER": "a"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "freezing"},
            {"LOWER": "cold"},
            {"LOWER": "injury"},
            {"LOWER": "when"},
            {"LOWER": "engaging"},
            {"LOWER": "in"},
            {"LOWER": "exercises"},
            {"LOWER": "in"},
            {"LOWER": "cold"},
            {"LOWER": "weather"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "purportedly"},
            {"LOWER": "gifting"},
            {"LOWER": "car"},
            {"LOWER": "to"},
            {"LOWER": "first"},
            {"LOWER": "defendant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "receiving"},
            {"LOWER": "grant"},
            {"LOWER": "payments"},
            {"LOWER": "from"},
            {"LOWER": "european"},
            {"LOWER": "regional"},
            {"LOWER": "development"},
            {"LOWER": "fund"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "claimant"}, {"LOWER": "refused"}, {"LOWER": "asylum"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "refused"},
            {"LOWER": "asylum"},
            {"LOWER": "and"},
            {"LOWER": "exhausting"},
            {"LOWER": "appeal"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "refusing"},
            {"LOWER": "to"},
            {"LOWER": "complete"},
            {"LOWER": "application"},
            {"LOWER": "form"},
            {"LOWER": "twice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "seeking"},
            {"LOWER": "judicial"},
            {"LOWER": "review"},
            {"LOWER": "of"},
            {"LOWER": "restricted"},
            {"LOWER": "discretionary"},
            {"LOWER": "leave"},
            {"LOWER": "policy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "seeking"},
            {"LOWER": "to"},
            {"LOWER": "restrain"},
            {"LOWER": "former"},
            {"LOWER": "employee"},
            {"LOWER": "from"},
            {"LOWER": "employing"},
            {"LOWER": "others"},
            {"LOWER": "in"},
            {"LOWER": "competing"},
            {"LOWER": "business"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "serving"},
            {"LOWER": "five"},
            {"LOWER": "year"},
            {"LOWER": "sentence"},
            {"LOWER": "for"},
            {"LOWER": "web"},
            {"ORTH": "-"},
            {"LOWER": "posting"},
            {"LOWER": "action"},
            {"LOWER": "against"},
            {"LOWER": "british"},
            {"LOWER": "and"},
            {"LOWER": "other"},
            {"LOWER": "forces"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "studying"},
            {"LOWER": "at"},
            {"LOWER": "university"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "studying"},
            {"LOWER": "for"},
            {"LOWER": "master"},
            {"LOWER": "’s"},
            {"LOWER": "degree"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "succeeding"},
            {"LOWER": "on"},
            {"LOWER": "liability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "suffering"},
            {"LOWER": "injury"},
            {"LOWER": "at"},
            {"LOWER": "birth"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "suffering"},
            {"LOWER": "severe"},
            {"LOWER": "obstetric"},
            {"LOWER": "injury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "sustaining"},
            {"LOWER": "injuries"},
            {"LOWER": "when"},
            {"LOWER": "coming"},
            {"LOWER": "into"},
            {"LOWER": "contact"},
            {"LOWER": "with"},
            {"LOWER": "defendant"},
            {"LOWER": "’s"},
            {"LOWER": "car"},
            {"LOWER": "in"},
            {"LOWER": "car"},
            {"LOWER": "park"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "sustaining"},
            {"LOWER": "injuries"},
            {"LOWER": "while"},
            {"LOWER": "detained"},
            {"LOWER": "at"},
            {"LOWER": "hospital"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "sustaining"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "freezing"},
            {"LOWER": "cold"},
            {"LOWER": "injury"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "defendant"},
            {"LOWER": "’s"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "duty"},
            {"LOWER": "of"},
            {"LOWER": "care"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "claimant"}, {"LOWER": "under"}, {"LOWER": "disability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimants"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimants"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "order"},
            {"LOWER": "requiring"},
            {"LOWER": "defendant"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "information"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "its"},
            {"LOWER": "insurance"},
            {"LOWER": "cover"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimants"},
            {"LOWER": "challenging"},
            {"LOWER": "transposition"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "directive"},
            {"LOWER": "into"},
            {"LOWER": "uk"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "fridge"},
            {"LOWER": "freezer"},
            {"LOWER": "catching"},
            {"LOWER": "fire"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claims"},
            {"LOWER": "against"},
            {"LOWER": "company"},
            {"LOWER": "assigned"},
            {"LOWER": "to"},
            {"LOWER": "director"},
            {"LOWER": "and"},
            {"LOWER": "again"},
            {"LOWER": "assigned"},
            {"LOWER": "by"},
            {"LOWER": "him"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claims"},
            {"LOWER": "against"},
            {"LOWER": "former"},
            {"LOWER": "trustee"},
            {"LOWER": "for"},
            {"LOWER": "deceit"},
            {"ORTH": ","},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "trust"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "claims"},
            {"LOWER": "for"},
            {"LOWER": "damages"},
            {"LOWER": "and"},
            {"LOWER": "costs"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "clandestine"}, {"LOWER": "entrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "class"}, {"LOWER": "a"}, {"LOWER": "drugs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "class"}, {"LOWER": "a"}, {"LOWER": "drug"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "class"}, {"LOWER": "gift"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "class"}, {"LOWER": "meeting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "class"}, {"LOWER": "meetings"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "classes"}, {"LOWER": "of"}, {"LOWER": "shares"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "classification"}, {"LOWER": "society"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "cleaner"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "clearly"}, {"LOWER": "unfounded"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "clergyman"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "clerical"}, {"LOWER": "error"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "client"}, {"LOWER": "’s"}, {"LOWER": "instructions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "clifford"}, {"LOWER": "maersk"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "clinical"}, {"LOWER": "case"}, {"LOWER": "manager"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "clinical"}, {"LOWER": "negligence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "close"}, {"LOWER": "company"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "close"}, {"LOWER": "of"}, {"LOWER": "bankruptcy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "closed"}, {"LOWER": "material"}, {"LOWER": "procedure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "closely"}, {"LOWER": "connected"}, {"LOWER": "claims"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "closure"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "closure"},
            {"LOWER": "of"},
            {"LOWER": "independent"},
            {"LOWER": "living"},
            {"LOWER": "fund"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "closure"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
            {"LOWER": "between"},
            {"LOWER": "local"},
            {"LOWER": "authorities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
            {"LOWER": "in"},
            {"LOWER": "civil"},
            {"LOWER": "matters"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "coal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "coalmining"}, {"LOWER": "industry"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "code"}, {"LOWER": "of"}, {"LOWER": "conduct"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "cogeneration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cohabitee"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cohesion"}, {"LOWER": "fund"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "coinage"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cold"}, {"ORTH": "-"}, {"LOWER": "calling"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "collateral"}, {"LOWER": "bond"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "collateral"}, {"LOWER": "contract"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "collateral"},
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "evidence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "collecting"},
            {"LOWER": "information"},
            {"LOWER": "likely"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "useful"},
            {"LOWER": "to"},
            {"LOWER": "terrorist"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "collection"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "collection"},
            {"LOWER": "of"},
            {"LOWER": "information"},
            {"LOWER": "for"},
            {"LOWER": "terrorist"},
            {"LOWER": "purposes"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "collection"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "collective"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "collective"},
            {"LOWER": "agreement"},
            {"LOWER": "incorporated"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "collective"}, {"LOWER": "bargaining"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "collective"},
            {"LOWER": "conditional"},
            {"LOWER": "fee"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "collective"}, {"LOWER": "grievance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "collective"},
            {"LOWER": "investment"},
            {"LOWER": "schemes"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "collective"},
            {"LOWER": "labour"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "college"},
            {"LOWER": "’s"},
            {"LOWER": "student"},
            {"LOWER": "sponsor"},
            {"LOWER": "licence"},
            {"LOWER": "revoked"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "colliery"}, {"LOWER": "closure"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "collision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "colony"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "combating"},
            {"LOWER": "late"},
            {"LOWER": "payments"},
            {"LOWER": "in"},
            {"LOWER": "commercial"},
            {"LOWER": "transactions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "combined"}, {"LOWER": "nomenclature"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commencement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commencement"},
            {"LOWER": "of"},
            {"LOWER": "employment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commencement"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "commercial"}, {"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commercial"}, {"LOWER": "agents"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commercial"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commercial"}, {"LOWER": "debt"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commercial"}, {"LOWER": "policy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commercially"},
            {"LOWER": "sensitive"},
            {"LOWER": "documents"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "commission"}, {"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commission"}, {"LOWER": "decision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commission"}, {"LOWER": "decision"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commission"},
            {"LOWER": "for"},
            {"LOWER": "racial"},
            {"LOWER": "equality"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commission"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "communities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "commission"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commission"},
            {"LOWER": "on"},
            {"LOWER": "money"},
            {"LOWER": "received"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "commission"}, {"LOWER": "regulation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "commission"}, {"LOWER": "tender"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commission"},
            {"LOWER": "to"},
            {"LOWER": "take"},
            {"LOWER": "evidence"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "application"}, {"LOWER": "by"}, {"LOWER": "commission"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "commissioner"},
            {"LOWER": "of"},
            {"LOWER": "deposit"},
            {"ORTH": "-"},
            {"LOWER": "taking"},
            {"LOWER": "companies"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "commissioners"}, {"LOWER": "consent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "committal"}, {"LOWER": "application"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "committal"}, {"LOWER": "for"}, {"LOWER": "contempt"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "committal"}, {"LOWER": "for"}, {"LOWER": "sentence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "committal"},
            {"LOWER": "for"},
            {"LOWER": "sentence"},
            {"LOWER": "to"},
            {"LOWER": "crown"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "committal"}, {"LOWER": "for"}, {"LOWER": "trial"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "committal"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "committal"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "committal"},
            {"LOWER": "proceedings"},
            {"LOWER": "brought"},
            {"LOWER": "by"},
            {"LOWER": "father"},
            {"LOWER": "seeking"},
            {"LOWER": "contact"},
            {"LOWER": "with"},
            {"LOWER": "child"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "committal"},
            {"LOWER": "proceedings"},
            {"ORTH": ","},
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "bring"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "committal"},
            {"LOWER": "to"},
            {"LOWER": "crown"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "committal"}, {"LOWER": "to"}, {"LOWER": "prison"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "committal"}, {"LOWER": "warrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "common"},
            {"LOWER": "agricultural"},
            {"LOWER": "policy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "common"}, {"LOWER": "commercial"}, {"LOWER": "policy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "common"}, {"LOWER": "customs"}, {"LOWER": "tariff"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "common"}, {"LOWER": "fisheries"}, {"LOWER": "policy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "common"},
            {"LOWER": "foreign"},
            {"LOWER": "and"},
            {"LOWER": "security"},
            {"LOWER": "policy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "common"},
            {"LOWER": "foreign"},
            {"LOWER": "and"},
            {"LOWER": "security"},
            {"LOWER": "policy"},
            {"ORTH": "("},
            {"LOWER": "cfsp"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "common"}, {"LOWER": "land"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "common"}, {"LOWER": "law"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "common"}, {"LOWER": "law"}, {"LOWER": "offence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "common"}, {"LOWER": "mistake"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "common"},
            {"LOWER": "organisation"},
            {"LOWER": "of"},
            {"LOWER": "markets"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "common"}, {"LOWER": "pur"}, {"LOWER": "vicinage"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "common"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "common"}, {"LOWER": "sting"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "common"},
            {"LOWER": "system"},
            {"LOWER": "of"},
            {"LOWER": "financial"},
            {"LOWER": "transaction"},
            {"LOWER": "tax"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "common"},
            {"LOWER": "system"},
            {"LOWER": "of"},
            {"LOWER": "value"},
            {"LOWER": "added"},
            {"LOWER": "tax"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "commons"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "commonwealth"}, {"LOWER": "immigrant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "communication"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "communication"}, {"LOWER": "to"}, {"LOWER": "public"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "community"}, {"LOWER": "care"}, {"LOWER": "grant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "community"}, {"LOWER": "care"}, {"LOWER": "services"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "community"}, {"LOWER": "care"}, {"LOWER": "services"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "community"}, {"LOWER": "care"}, {"LOWER": "services"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "community"}, {"LOWER": "charge"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "community"}, {"LOWER": "customs"}, {"LOWER": "code"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "community"}, {"LOWER": "design"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "community"},
            {"LOWER": "plant"},
            {"LOWER": "variety"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "community"},
            {"LOWER": "registered"},
            {"LOWER": "design"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "community"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "challenge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "community"}, {"LOWER": "service"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "community"}, {"LOWER": "trade"}, {"LOWER": "mark"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "companies"},
            {"LOWER": "act"},
            {"ORTH": ","},
            {"LOWER": "1862"},
            {"ORTH": ","},
            {"LOWER": "s."},
            {"LOWER": "165"},
            {"ORTH": "["},
            {"LOWER": "revised"},
            {"LOWER": "ed"},
            {"ORTH": "."},
            {"LOWER": "statutes"},
            {"ORTH": ","},
            {"LOWER": "vol"},
            {"ORTH": "."},
            {"LOWER": "xiv"},
            {"ORTH": "."},
            {"LOWER": "p."},
            {"LOWER": "238"},
            {"ORTH": "]"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "companies"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "companies"},
            {"LOWER": "house"},
            {"LOWER": "and"},
            {"LOWER": "registrar"},
            {"LOWER": "of"},
            {"LOWER": "companies"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "company"},
            {"LOWER": "appearing"},
            {"LOWER": "by"},
            {"LOWER": "director"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "company"}, {"LOWER": "director"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "company"}, {"LOWER": "in"}, {"LOWER": "liquidation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "company"}, {"LOWER": "in"}, {"LOWER": "liquidation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "company"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "company"}, {"LOWER": "prospectus"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "company"},
            {"LOWER": "registered"},
            {"LOWER": "in"},
            {"LOWER": "italy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "company"},
            {"LOWER": "reimbursement"},
            {"LOWER": "policy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "company"},
            {"LOWER": "seeking"},
            {"LOWER": "certification"},
            {"LOWER": "of"},
            {"LOWER": "solar"},
            {"LOWER": "panel"},
            {"LOWER": "technology"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "company"}, {"LOWER": "structure"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "companys"}, {"LOWER": "assets"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "company"}, {"LOWER": "’s"}, {"LOWER": "debts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "comparative"},
            {"LOWER": "and"},
            {"LOWER": "misleading"},
            {"LOWER": "advertising"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "comparative"},
            {"LOWER": "or"},
            {"LOWER": "misleading"},
            {"LOWER": "advertising"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "compatibility"},
            {"LOWER": "with"},
            {"LOWER": "convention"},
            {"LOWER": "rights"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "compensation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "compensation"},
            {"LOWER": "and"},
            {"LOWER": "assistance"},
            {"LOWER": "to"},
            {"LOWER": "passengers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "compensation"},
            {"LOWER": "for"},
            {"LOWER": "ill"},
            {"ORTH": "-"},
            {"LOWER": "treated"},
            {"LOWER": "workers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "compensation"}, {"LOWER": "for"}, {"LOWER": "libel"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "compensation"},
            {"LOWER": "whether"},
            {"LOWER": "payable"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "assessment"},
            {"LOWER": "of"},
            {"LOWER": "compensation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "competence"}, {"LOWER": "of"}, {"LOWER": "witness"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "competing"}, {"LOWER": "proposals"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "competition"},
            {"LOWER": "appeal"},
            {"LOWER": "tribunal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "competitive"}, {"LOWER": "tendering"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "complainants"},
            {"LOWER": "sexual"},
            {"LOWER": "history"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "complainant"},
            {"LOWER": "’s"},
            {"LOWER": "sexual"},
            {"LOWER": "history"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "complainant"},
            {"LOWER": "’s"},
            {"LOWER": "sexual"},
            {"LOWER": "history"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "complaints"}, {"LOWER": "procedure"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "completion"}, {"LOWER": "notice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "completion"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "compliance"},
            {"LOWER": "with"},
            {"LOWER": "order"},
            {"LOWER": "later"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "compromise"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "compromise"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "compromise"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "compulsory"}, {"LOWER": "acquisition"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "compulsory"}, {"LOWER": "admission"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "compulsory"}, {"LOWER": "detention"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "compulsory"},
            {"LOWER": "indemnity"},
            {"LOWER": "insurance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "compulsory"},
            {"LOWER": "insurance"},
            {"LOWER": "against"},
            {"LOWER": "civil"},
            {"LOWER": "liability"},
            {"LOWER": "in"},
            {"LOWER": "respect"},
            {"LOWER": "of"},
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "motor"},
            {"LOWER": "vehicles"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "compulsory"}, {"LOWER": "powers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "compulsory"}, {"LOWER": "purchase"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "compulsory"}, {"LOWER": "purchase"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "compulsory"}, {"LOWER": "purchase"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "compulsory"}, {"LOWER": "purchase"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "compulsory"}, {"LOWER": "treatment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "computation"}, {"LOWER": "of"}, {"LOWER": "gain"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "computer"}, {"LOWER": "misuse"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "computer"}, {"LOWER": "program"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "computer"}, {"LOWER": "software"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concealing"},
            {"LOWER": "debts"},
            {"LOWER": "of"},
            {"LOWER": "company"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concealing"},
            {"LOWER": "proceeds"},
            {"LOWER": "of"},
            {"LOWER": "crime"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concealing"},
            {"ORTH": ","},
            {"LOWER": "disguising"},
            {"LOWER": "or"},
            {"LOWER": "converting"},
            {"LOWER": "criminal"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "concealment"}, {"LOWER": "of"}, {"LOWER": "facts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "concealment"}, {"LOWER": "of"}, {"LOWER": "property"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concentrations"},
            {"LOWER": "of"},
            {"LOWER": "pm10"},
            {"LOWER": "in"},
            {"LOWER": "ambient"},
            {"LOWER": "air"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concept"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "intervention"},
            {"LOWER": "by"},
            {"LOWER": "the"},
            {"LOWER": "state"},
            {"LOWER": "or"},
            {"LOWER": "through"},
            {"LOWER": "state"},
            {"LOWER": "resources"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concept"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "technological"},
            {"LOWER": "measures"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concept"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "working"},
            {"LOWER": "time"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conciliation"}, {"LOWER": "agreement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conciliation"}, {"LOWER": "procedure"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "concurrent"}, {"LOWER": "causes"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "concurrent"},
            {"LOWER": "proceedings"},
            {"LOWER": "before"},
            {"LOWER": "european"},
            {"LOWER": "commission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "concurrent"}, {"LOWER": "tortfeasors"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "condition"}, {"LOWER": "precedent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "condition"}, {"LOWER": "precedent"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conditional"}, {"LOWER": "fee"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conditional"}, {"LOWER": "gift"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conditional"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "defend"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conditional"},
            {"LOWER": "sale"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conditional"}, {"LOWER": "will"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conditions"}, {"LOWER": "for"}, {"LOWER": "execution"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conditions"}, {"LOWER": "for"}, {"LOWER": "obtaining"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conditions"},
            {"LOWER": "for"},
            {"LOWER": "obtaining"},
            {"LOWER": "certificate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conditions"}, {"LOWER": "of"}, {"LOWER": "bail"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conditions"}, {"LOWER": "of"}, {"LOWER": "employment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conditions"}, {"LOWER": "of"}, {"LOWER": "service"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conditions"},
            {"LOWER": "on"},
            {"LOWER": "public"},
            {"LOWER": "processions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conducive"},
            {"LOWER": "to"},
            {"LOWER": "public"},
            {"LOWER": "good"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conduct"},
            {"LOWER": "claimed"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "rise"},
            {"LOWER": "to"},
            {"LOWER": "cause"},
            {"LOWER": "of"},
            {"LOWER": "action"},
            {"LOWER": "occurring"},
            {"LOWER": "before"},
            {"LOWER": "statute"},
            {"LOWER": "in"},
            {"LOWER": "force"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conduct"},
            {"LOWER": "occurring"},
            {"LOWER": "before"},
            {"LOWER": "statute"},
            {"LOWER": "in"},
            {"LOWER": "force"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conduct"}, {"LOWER": "of"}, {"LOWER": "affairs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conduct"}, {"LOWER": "of"}, {"LOWER": "case"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conduct"}, {"LOWER": "of"}, {"LOWER": "company"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conduct"}, {"LOWER": "of"}, {"LOWER": "parties"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conduct"}, {"LOWER": "of"}, {"LOWER": "undertaking"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "confession"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "confidential"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "confidential"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "confidential"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "confidential"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "confidential"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "confidential"},
            {"LOWER": "information"},
            {"LOWER": "and"},
            {"LOWER": "documents"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "confidentiality"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "confinement"}, {"LOWER": "in"}, {"LOWER": "cell"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "confirmatory"}, {"LOWER": "faculty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "confiscation"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "confiscation"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conflict"}, {"LOWER": "of"}, {"LOWER": "interest"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conflict"}, {"LOWER": "of"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conflict"}, {"LOWER": "of"}, {"LOWER": "laws"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "connection"},
            {"LOWER": "with"},
            {"LOWER": "exercise"},
            {"LOWER": "of"},
            {"LOWER": "official"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conscientious"}, {"LOWER": "objection"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "consecutive"}, {"LOWER": "sentence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consecutive"}, {"LOWER": "sentences"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "consent"},
            {"LOWER": "of"},
            {"LOWER": "director"},
            {"LOWER": "general"},
            {"LOWER": "of"},
            {"LOWER": "fair"},
            {"LOWER": "trading"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "consent"}, {"LOWER": "of"}, {"LOWER": "parent"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "consent"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consent"}, {"LOWER": "payments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consent"}, {"LOWER": "to"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "consent"}, {"LOWER": "to"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "consent"},
            {"LOWER": "to"},
            {"LOWER": "sexual"},
            {"LOWER": "relations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "consequences"},
            {"LOWER": "of"},
            {"LOWER": "notification"},
            {"LOWER": "of"},
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "withdraw"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "consequential"}, {"LOWER": "orders"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "conservation"}, {"LOWER": "area"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conservation"}, {"LOWER": "area"}, {"LOWER": "control"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conservation"},
            {"LOWER": "of"},
            {"LOWER": "natural"},
            {"LOWER": "habitats"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conservators"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "considerations"},
            {"LOWER": "affecting"},
            {"LOWER": "leave"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "consistory"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "consolidated"},
            {"LOWER": "practice"},
            {"LOWER": "directions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conspiracy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conspiracy"},
            {"LOWER": "falsely"},
            {"LOWER": "to"},
            {"LOWER": "account"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conspiracy"},
            {"LOWER": "to"},
            {"LOWER": "cheat"},
            {"LOWER": "public"},
            {"LOWER": "revenue"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conspiracy"},
            {"LOWER": "to"},
            {"LOWER": "contravene"},
            {"LOWER": "theft"},
            {"LOWER": "act"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conspiracy"}, {"LOWER": "to"}, {"LOWER": "corrupt"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conspiracy"}, {"LOWER": "to"}, {"LOWER": "defraud"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conspiracy"},
            {"LOWER": "to"},
            {"LOWER": "injure"},
            {"LOWER": "by"},
            {"LOWER": "unlawful"},
            {"LOWER": "means"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conspiracy"},
            {"LOWER": "to"},
            {"LOWER": "injure"},
            {"LOWER": "by"},
            {"LOWER": "unlawful"},
            {"LOWER": "means"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conspiracy"}, {"LOWER": "to"}, {"LOWER": "steal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conspiracy"},
            {"LOWER": "to"},
            {"LOWER": "supply"},
            {"LOWER": "drugs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conspiracy"}, {"LOWER": "with"}, {"LOWER": "spouse"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conspracy"}, {"LOWER": "to"}, {"LOWER": "supply"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "constituencies"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "constitution"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "constitution"}, {"LOWER": "of"}, {"LOWER": "tribunal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "constitutional"}, {"LOWER": "instrument"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "construction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "construction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "construction"}, {"LOWER": "industry"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "construction"},
            {"LOWER": "industry"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "construction"}, {"LOWER": "of"}, {"LOWER": "terms"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "constructive"}, {"LOWER": "delivery"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "constructive"}, {"LOWER": "dismissal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "constructive"}, {"LOWER": "notice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "constructive"}, {"LOWER": "trust"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "consultation"},
            {"LOWER": "and"},
            {"LOWER": "information"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "consumer"}, {"LOWER": "contract"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consumer"}, {"LOWER": "contracts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consumer"}, {"LOWER": "credit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consumer"}, {"LOWER": "credit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consumer"}, {"LOWER": "protection"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consumer"}, {"LOWER": "protection"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "consumer"}, {"LOWER": "protection"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "contact"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "contact"}, {"LOWER": "orders"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contained"},
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "genetically"},
            {"LOWER": "modified"},
            {"LOWER": "micro"},
            {"ORTH": "-"},
            {"LOWER": "organisms"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contaminated"}, {"LOWER": "land"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contemnor"}, {"LOWER": "under"}, {"LOWER": "21"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contempt"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "contempt"}, {"LOWER": "found"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contempt"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contempt"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contempt"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contention"},
            {"LOWER": "that"},
            {"LOWER": "proper"},
            {"LOWER": "process"},
            {"LOWER": "for"},
            {"LOWER": "grading"},
            {"LOWER": "results"},
            {"LOWER": "not"},
            {"LOWER": "followed"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contentious"},
            {"LOWER": "business"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contingency"}, {"LOWER": "fee"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contingent"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "continuing"}, {"LOWER": "obligation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "continuity"}, {"LOWER": "of"}, {"LOWER": "employment"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "continuous"}, {"LOWER": "employment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contra"}, {"LOWER": "bonos"}, {"LOWER": "mores"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contract"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "award"}, {"LOWER": "procedure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "award"}, {"LOWER": "procedure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "awards"}, {"LOWER": "procedure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contract"},
            {"LOWER": "between"},
            {"LOWER": "mobile"},
            {"LOWER": "phone"},
            {"LOWER": "operator"},
            {"LOWER": "and"},
            {"LOWER": "individuals"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contract"},
            {"LOWER": "concerning"},
            {"LOWER": "a"},
            {"LOWER": "local"},
            {"LOWER": "development"},
            {"LOWER": "action"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "for"}, {"LOWER": "assignment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "for"}, {"LOWER": "difference"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "for"}, {"LOWER": "differences"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contract"},
            {"LOWER": "for"},
            {"LOWER": "excess"},
            {"LOWER": "of"},
            {"LOWER": "loss"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "for"}, {"LOWER": "sale"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "for"}, {"LOWER": "services"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "in"}, {"LOWER": "writing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contract"},
            {"LOWER": "made"},
            {"LOWER": "in"},
            {"LOWER": "consumer"},
            {"LOWER": "’s"},
            {"LOWER": "home"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "of"}, {"LOWER": "employment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "of"}, {"LOWER": "employment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "of"}, {"LOWER": "employment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "of"}, {"LOWER": "service"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contract"},
            {"LOWER": "of"},
            {"LOWER": "service"},
            {"LOWER": "or"},
            {"LOWER": "for"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "review"}, {"LOWER": "procedure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "to"}, {"LOWER": "supply"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contract"}, {"LOWER": "under"}, {"LOWER": "seal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contract"}, {"LOWER": "worker"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contract"},
            {"ORTH": ","},
            {"LOWER": "breach"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contracted"}, {"LOWER": "works"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "contracting"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "contracting"}, {"LOWER": "out"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contractors"},
            {"LOWER": "all"},
            {"LOWER": "risk"},
            {"LOWER": "policy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contracts"},
            {"LOWER": "for"},
            {"LOWER": "sale"},
            {"LOWER": "or"},
            {"LOWER": "purchase"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contractual"},
            {"LOWER": "liability"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contravention"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contravention"}, {"LOWER": "of"}, {"LOWER": "rules"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "contribution"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contributory"}, {"LOWER": "infringement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "contributory"}, {"LOWER": "negligence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "contributory"},
            {"LOWER": "retirement"},
            {"LOWER": "pension"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "control"}, {"LOWER": "of"}, {"LOWER": "advertisements"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "control"}, {"LOWER": "of"}, {"LOWER": "concentrations"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "control"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "control"}, {"LOWER": "orders"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "controlled"},
            {"LOWER": "foreign"},
            {"LOWER": "company"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "controlled"}, {"LOWER": "parking"}, {"LOWER": "zone"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "controlling"},
            {"LOWER": "prostitution"},
            {"LOWER": "for"},
            {"LOWER": "gain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "controls"},
            {"LOWER": "of"},
            {"LOWER": "cash"},
            {"LOWER": "entering"},
            {"LOWER": "or"},
            {"LOWER": "leaving"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "convening"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "convention"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "convention"}, {"ORTH": ","}, {"LOWER": "by"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conversion"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conversion"},
            {"LOWER": "from"},
            {"LOWER": "administration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "converting"},
            {"LOWER": "criminal"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "converting"},
            {"LOWER": "proceeds"},
            {"LOWER": "of"},
            {"LOWER": "crime"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conveyance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conveyance"},
            {"LOWER": "of"},
            {"LOWER": "property"},
            {"LOWER": "contrary"},
            {"LOWER": "to"},
            {"LOWER": "instructions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conveyance"},
            {"LOWER": "on"},
            {"LOWER": "transfer"},
            {"LOWER": "or"},
            {"LOWER": "sale"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conveyance"},
            {"LOWER": "or"},
            {"LOWER": "transfer"},
            {"LOWER": "on"},
            {"LOWER": "sale"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conveyancing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "conveyancing"}, {"LOWER": "plan"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "convicted"}, {"LOWER": "fugitive"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "convicted"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "conviction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conviction"},
            {"LOWER": "and"},
            {"LOWER": "sentence"},
            {"LOWER": "in"},
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "absence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conviction"}, {"LOWER": "as"}, {"LOWER": "evidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "conviction"},
            {"LOWER": "as"},
            {"LOWER": "evidence"},
            {"LOWER": "in"},
            {"LOWER": "civil"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conviction"}, {"LOWER": "for"}, {"LOWER": "contumacy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "conviction"}, {"LOWER": "in"}, {"LOWER": "absence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "conviction"}, {"LOWER": "warrant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "copyright"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "copyright"}, {"LOWER": "tribunal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "core"}, {"LOWER": "bundles"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "coroner"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "coroners"},
            {"LOWER": "and"},
            {"LOWER": "justice"},
            {"LOWER": "act"},
            {"LOWER": "2009"},
            {"ORTH": "("},
            {"LOWER": "c"},
            {"LOWER": "25"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "s"},
            {"LOWER": "54"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "coroners"},
            {"LOWER": "and"},
            {"LOWER": "justice"},
            {"LOWER": "act"},
            {"LOWER": "2009"},
            {"ORTH": "("},
            {"LOWER": "c"},
            {"LOWER": "25"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "ss"},
            {"LOWER": "54"},
            {"ORTH": ","},
            {"LOWER": "55"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "coroners"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "coroners"}, {"LOWER": "duties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "coroners"}, {"LOWER": "inquest"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "coroner"}, {"LOWER": "’s"}, {"LOWER": "duties"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "corporation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "corporation"}, {"LOWER": "tax"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "corporation"}, {"LOWER": "tax"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "corporations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "correct"}, {"LOWER": "approach"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "correction"}, {"LOWER": "of"}, {"LOWER": "error"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "correction"}, {"LOWER": "of"}, {"LOWER": "errors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "correlation"},
            {"LOWER": "between"},
            {"LOWER": "duration"},
            {"LOWER": "of"},
            {"LOWER": "sexual"},
            {"LOWER": "harm"},
            {"LOWER": "prevention"},
            {"LOWER": "orders"},
            {"LOWER": "and"},
            {"LOWER": "statutory"},
            {"LOWER": "notification"},
            {"LOWER": "requirements"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "corroboration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "corruption"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cost"}, {"LOWER": "management"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cost"},
            {"LOWER": "of"},
            {"LOWER": "preparation"},
            {"LOWER": "for"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cost"},
            {"LOWER": "of"},
            {"LOWER": "remedying"},
            {"LOWER": "structural"},
            {"LOWER": "defects"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cost"},
            {"ORTH": "-"},
            {"LOWER": "capping"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "costs"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "costs"}, {"LOWER": "against"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"LOWER": "application"},
            {"LOWER": "by"},
            {"LOWER": "successful"},
            {"LOWER": "party"},
            {"LOWER": "resisted"},
            {"LOWER": "by"},
            {"LOWER": "defendants"},
            {"LOWER": "on"},
            {"LOWER": "grounds"},
            {"LOWER": "their"},
            {"LOWER": "objective"},
            {"LOWER": "for"},
            {"LOWER": "retrial"},
            {"LOWER": "achieved"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "costs"}, {"LOWER": "budget"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "costs"}, {"LOWER": "budget"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"LOWER": "consequences"},
            {"LOWER": "following"},
            {"LOWER": "judgment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"LOWER": "follow"},
            {"LOWER": "the"},
            {"LOWER": "event"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "costs"}, {"LOWER": "following"}, {"LOWER": "event"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"LOWER": "following"},
            {"LOWER": "payment"},
            {"LOWER": "in"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "costs"}, {"LOWER": "management"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "costs"}, {"LOWER": "management"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "costs"}, {"LOWER": "of"}, {"LOWER": "administration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "costs"}, {"LOWER": "of"}, {"LOWER": "motion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"LOWER": "of"},
            {"LOWER": "pending"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "costs"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "costs"}, {"LOWER": "of"}, {"LOWER": "taxation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"LOWER": "on"},
            {"LOWER": "indemnity"},
            {"LOWER": "basis"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "costs"}, {"LOWER": "only"}, {"LOWER": "proceedings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "costs"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"LOWER": "order"},
            {"LOWER": "against"},
            {"LOWER": "father"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"ORTH": ","},
            {"LOWER": "interest"},
            {"LOWER": "on"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "costs"},
            {"ORTH": "-"},
            {"LOWER": "only"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "decision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "directive"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "directive"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "85/374"},
            {"LOWER": "/"},
            {"LOWER": "eec"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "4"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "council"},
            {"LOWER": "for"},
            {"LOWER": "the"},
            {"LOWER": "regulation"},
            {"LOWER": "of"},
            {"LOWER": "health"},
            {"LOWER": "care"},
            {"LOWER": "professionals"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "council"},
            {"LOWER": "framework"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "council"},
            {"LOWER": "framework"},
            {"LOWER": "decision2002/584"},
            {"LOWER": "/"},
            {"LOWER": "jha"},
            {"ORTH": "("},
            {"LOWER": "as"},
            {"LOWER": "amended"},
            {"LOWER": "by"},
            {"LOWER": "council"},
            {"LOWER": "framework"},
            {"LOWER": "decision"},
            {"LOWER": "2009/299"},
            {"LOWER": "/"},
            {"LOWER": "jha"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "arts"},
            {"LOWER": "8(1)(c"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "15(2"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "council"},
            {"LOWER": "implementing"},
            {"LOWER": "regulation"},
            {"ORTH": "("},
            {"LOWER": "eu"},
            {"ORTH": ")"},
            {"LOWER": "no"},
            {"LOWER": "610/2010"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "meeting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "meetings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "council"},
            {"LOWER": "providing"},
            {"LOWER": "night"},
            {"ORTH": "-"},
            {"LOWER": "time"},
            {"LOWER": "carer"},
            {"LOWER": "to"},
            {"LOWER": "help"},
            {"LOWER": "claimant"},
            {"LOWER": "access"},
            {"LOWER": "commode"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "regulation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "tax"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "council"}, {"LOWER": "tax"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "council"},
            {"LOWER": "tenant"},
            {"LOWER": "exercising"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "buy"},
            {"LOWER": "council"},
            {"LOWER": "property"},
            {"LOWER": "in"},
            {"LOWER": "her"},
            {"LOWER": "and"},
            {"LOWER": "her"},
            {"LOWER": "partner"},
            {"LOWER": "’s"},
            {"LOWER": "names"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "councils"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "council"}, {"LOWER": "’s"}, {"LOWER": "duty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "counselling"}, {"LOWER": "or"}, {"LOWER": "procuring"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "counsels"}, {"LOWER": "opinion"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "counter"}, {"LOWER": "-"}, {"LOWER": "restitution"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "counterclaim"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "counterfeit"}, {"LOWER": "coins"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "country"}, {"LOWER": "guidance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "counts"},
            {"LOWER": "to"},
            {"LOWER": "lie"},
            {"LOWER": "on"},
            {"LOWER": "file"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "county"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "county"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "county"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "county"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "county"}, {"LOWER": "court"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "county"}, {"LOWER": "court"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "county"}, {"LOWER": "court"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "course"}, {"LOWER": "of"}, {"LOWER": "conduct"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "course"}, {"LOWER": "of"}, {"LOWER": "employment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "allowing"},
            {"LOWER": "claimants"},
            {"LOWER": "’"},
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "admit"},
            {"LOWER": "evidence"},
            {"LOWER": "and"},
            {"LOWER": "for"},
            {"LOWER": "retrial"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "bundles"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "bundles"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "dress"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "fees"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "first"}, {"LOWER": "seised"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "martial"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "martial"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "of"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "of"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "civil"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "civil"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "civil"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "civil"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "civil"},
            {"LOWER": "divsion"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "criminal"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "criminal"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "criminal"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "criminal"},
            {"LOWER": "division"},
            {"ORTH": ")"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "decisions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "of"},
            {"LOWER": "guernsey"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": ","},
            {"LOWER": "criminal"},
            {"LOWER": "division"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "of"}, {"LOWER": "justice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "of"}, {"LOWER": "protection"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "of"}, {"LOWER": "protection"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "of"}, {"LOWER": "session"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "order"},
            {"LOWER": "against"},
            {"LOWER": "father"},
            {"LOWER": "for"},
            {"LOWER": "return"},
            {"LOWER": "of"},
            {"LOWER": "children"},
            {"LOWER": "in"},
            {"LOWER": "child"},
            {"LOWER": "abduction"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "court"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "process"},
            {"ORTH": ","},
            {"LOWER": "abuse"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"ORTH": "-"}, {"LOWER": "martial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"ORTH": "-"}, {"LOWER": "martial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "courts"},
            {"LOWER": "discretion"},
            {"LOWER": "to"},
            {"LOWER": "admit"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "courts"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "courts"},
            {"LOWER": "inherent"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "courts"}, {"LOWER": "jurisdiction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "courts"},
            {"LOWER": "parental"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "courts"}, {"LOWER": "previous"}, {"LOWER": "decisions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "courts"},
            {"LOWER": "supervisory"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "’s"}, {"LOWER": "discretion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "’s"},
            {"LOWER": "inherent"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "’s"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "’s"},
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "court"}, {"LOWER": "’s"}, {"LOWER": "powers"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "court"},
            {"LOWER": "’s"},
            {"LOWER": "supervisory"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "covenant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "covenant"},
            {"LOWER": "against"},
            {"LOWER": "carrying"},
            {"LOWER": "on"},
            {"LOWER": "specified"},
            {"LOWER": "business"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "covenant"},
            {"LOWER": "implied"},
            {"LOWER": "by"},
            {"LOWER": "statute"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "covenant"}, {"ORTH": ","}, {"LOWER": "implied"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "covert"},
            {"LOWER": "human"},
            {"LOWER": "intelligence"},
            {"LOWER": "source"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cpr"}, {"LOWER": "r"}, {"LOWER": "11(2"}, {"ORTH": ")"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cpr"},
            {"LOWER": "r"},
            {"LOWER": "36.14(1)(a"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "credibility"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "credit"}, {"LOWER": "card"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "credit"}, {"LOWER": "hire"}, {"LOWER": "charges"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "credit"}, {"LOWER": "institution"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "credit"}, {"LOWER": "reference"}, {"LOWER": "agency"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "creditor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "creditors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "creditors"}, {"LOWER": "meeting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "creditors"}, {"LOWER": "meeting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "creditors"}, {"LOWER": "petition"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "creditors"},
            {"LOWER": "’"},
            {"LOWER": "votes"},
            {"LOWER": "at"},
            {"LOWER": "meeting"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "creditor"}, {"LOWER": "’s"}, {"LOWER": "obligations"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "creditor"}, {"LOWER": "’s"}, {"LOWER": "petition"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "creditworthiness"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "crime"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "crime"}, {"LOWER": "and"}, {"LOWER": "disorder"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "activity"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "appeal"},
            {"LOWER": "office"},
            {"LOWER": "summaries"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "appeals"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "criminal"}, {"LOWER": "bankruptcy"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "criminal"}, {"LOWER": "behaviour"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "criminal"}, {"LOWER": "behaviour"}, {"LOWER": "orders"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "cases"},
            {"LOWER": "review"},
            {"LOWER": "commission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "cause"},
            {"LOWER": "or"},
            {"LOWER": "matter"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "charge"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "conviction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "damage"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "damage"},
            {"LOWER": "to"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "injuries"},
            {"LOWER": "compensation"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "injuries"},
            {"LOWER": "compensation"},
            {"LOWER": "award"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "injuries"},
            {"LOWER": "compensation"},
            {"LOWER": "board"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "injuries"},
            {"LOWER": "compensation"},
            {"LOWER": "board"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "injuries"},
            {"LOWER": "compensation"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "injury"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "injury"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "criminal"}, {"LOWER": "investigation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "jurisdiction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "jurisdiction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "justice"},
            {"LOWER": "act"},
            {"LOWER": "2003"},
            {"ORTH": "("},
            {"LOWER": "c"},
            {"LOWER": "44"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "s"},
            {"LOWER": "101(1"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "criminal"}, {"LOWER": "legal"}, {"LOWER": "aid"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "criminal"}, {"LOWER": "records"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "criminal"},
            {"LOWER": "standard"},
            {"LOWER": "of"},
            {"LOWER": "proof"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "critically"},
            {"LOWER": "ill"},
            {"LOWER": "patient"},
            {"LOWER": "lacking"},
            {"LOWER": "capacity"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cross"},
            {"LOWER": "-"},
            {"LOWER": "border"},
            {"LOWER": "enforcement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cross"},
            {"LOWER": "-"},
            {"LOWER": "border"},
            {"LOWER": "insolvency"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cross"},
            {"LOWER": "-"},
            {"LOWER": "border"},
            {"LOWER": "merger"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cross"},
            {"LOWER": "-"},
            {"LOWER": "border"},
            {"LOWER": "mergers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "cross"}, {"LOWER": "-"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cross"},
            {"LOWER": "-"},
            {"LOWER": "examination"},
            {"LOWER": "of"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "defendant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cross"},
            {"LOWER": "-"},
            {"LOWER": "examination"},
            {"LOWER": "of"},
            {"LOWER": "translators"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cross"},
            {"LOWER": "-"},
            {"LOWER": "examination"},
            {"LOWER": "on"},
            {"LOWER": "affidavit"},
            {"LOWER": "evidence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "crown"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "crown"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "crown"},
            {"LOWER": "court-"},
            {"LOWER": "appeal"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "crown"}, {"LOWER": "office"}, {"LOWER": "list"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "crown"}, {"LOWER": "office"}, {"LOWER": "list"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "crown"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "crown"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "crown"}, {"LOWER": "prosecution"}, {"LOWER": "service"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "cruelty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "cruise"}, {"LOWER": "ships"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "crypt"}, {"LOWER": "restaurant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "culling"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "currency"}, {"LOWER": "fluctuations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "current"}, {"LOWER": "policies"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "custodial"}, {"LOWER": "sentence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "custodianship"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "custody"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "custody"}, {"LOWER": "appeals"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "custody"}, {"LOWER": "of"}, {"LOWER": "body"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "custody"}, {"LOWER": "officer"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "custody"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "custody"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "customary"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "customary"}, {"LOWER": "right"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "customs"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "customs"}, {"LOWER": "and"}, {"LOWER": "excise"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "customs"}, {"LOWER": "and"}, {"LOWER": "excise"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "customs"},
            {"LOWER": "union"},
            {"LOWER": "and"},
            {"LOWER": "common"},
            {"LOWER": "customs"},
            {"LOWER": "tariff"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cy"},
            {"ORTH": "-"},
            {"LOWER": "pres"},
            {"LOWER": "doctrine"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "cy"},
            {"ORTH": "-"},
            {"LOWER": "près"},
            {"LOWER": "doctrine"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "damage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "damage"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "damage"}, {"LOWER": "or"}, {"LOWER": "distress"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "damage"},
            {"LOWER": "to"},
            {"LOWER": "apparatus"},
            {"LOWER": "in"},
            {"LOWER": "street"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "damage"},
            {"LOWER": "to"},
            {"LOWER": "bailed"},
            {"LOWER": "chattel"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "damage"}, {"LOWER": "to"}, {"LOWER": "cargo"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "damage"}, {"LOWER": "to"}, {"LOWER": "property"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "damages"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "damages"},
            {"LOWER": "for"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "damages"}, {"LOWER": "for"}, {"LOWER": "libel"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "damages"},
            {"LOWER": "for"},
            {"LOWER": "wrongful"},
            {"LOWER": "eviction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "damages"}, {"LOWER": "in"}, {"LOWER": "lieu"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dangerous"}, {"LOWER": "driving"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dangerous"},
            {"LOWER": "machinery"},
            {"ORTH": "("},
            {"LOWER": "fencing"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dangerous"}, {"LOWER": "offenders"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dangerously"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "control"},
            {"LOWER": "in"},
            {"LOWER": "public"},
            {"LOWER": "place"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "data"}, {"LOWER": "on"}, {"LOWER": "computer"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "data"}, {"LOWER": "protection"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "data"}, {"LOWER": "protection"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "data"}, {"LOWER": "protection"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "data"},
            {"LOWER": "protection"},
            {"LOWER": "obligations"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "database"}, {"LOWER": "right"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "date"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "date"}, {"LOWER": "for"}, {"LOWER": "assessment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "date"},
            {"LOWER": "from"},
            {"LOWER": "which"},
            {"LOWER": "payable"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "date"},
            {"LOWER": "from"},
            {"LOWER": "which"},
            {"LOWER": "payable"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "date"}, {"LOWER": "of"}, {"LOWER": "knowledge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "date"},
            {"LOWER": "of"},
            {"LOWER": "knowledge"},
            {"LOWER": "of"},
            {"LOWER": "victim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "death"}, {"LOWER": "in"}, {"LOWER": "service"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "death"},
            {"LOWER": "of"},
            {"LOWER": "british"},
            {"LOWER": "subject"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "death"},
            {"LOWER": "of"},
            {"LOWER": "r."},
            {"LOWER": "a."},
            {"LOWER": "before"},
            {"LOWER": "testator"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "death"}, {"LOWER": "of"}, {"LOWER": "tenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "death"}, {"LOWER": "of"}, {"LOWER": "tenant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "death"}, {"LOWER": "offshore"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "debarring"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "debenture"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "debenture"}, {"LOWER": "holder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "debit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "debt"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "debt"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "employee"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "debt"}, {"LOWER": "security"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "debtors"}, {"LOWER": "summons"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "debts"},
            {"LOWER": "or"},
            {"LOWER": "liabilities"},
            {"LOWER": "incurred"},
            {"LOWER": "by"},
            {"LOWER": "administrator"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deceased"},
            {"LOWER": "making"},
            {"LOWER": "gifts"},
            {"LOWER": "to"},
            {"LOWER": "son"},
            {"LOWER": "and"},
            {"LOWER": "daughter"},
            {"LOWER": "in"},
            {"LOWER": "law"},
            {"LOWER": "of"},
            {"LOWER": "proceeds"},
            {"LOWER": "of"},
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "home"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deceased"},
            {"LOWER": "purportedly"},
            {"LOWER": "leaving"},
            {"LOWER": "entire"},
            {"LOWER": "estate"},
            {"LOWER": "to"},
            {"LOWER": "spouse"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deceit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "deception"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "decision"},
            {"LOWER": "not"},
            {"LOWER": "to"},
            {"LOWER": "prosecute"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "decision"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "decision"},
            {"LOWER": "of"},
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "arches"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "decision"},
            {"LOWER": "of"},
            {"LOWER": "general"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "decision"},
            {"LOWER": "of"},
            {"LOWER": "national"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "decision"}, {"LOWER": "to"}, {"LOWER": "cohabit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "decision"}, {"LOWER": "to"}, {"LOWER": "prosecute"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "decision"}, {"LOWER": "to"}, {"LOWER": "remove"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "decision"},
            {"LOWER": "to"},
            {"LOWER": "terminate"},
            {"LOWER": "investigation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "declaration"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "human"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "declaration"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "status"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "declaration"}, {"LOWER": "of"}, {"LOWER": "age"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "declaration"},
            {"LOWER": "of"},
            {"LOWER": "incompatibility"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "declaration"},
            {"LOWER": "of"},
            {"LOWER": "marital"},
            {"LOWER": "status"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "declaration"}, {"LOWER": "of"}, {"LOWER": "parentage"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "declaration"}, {"LOWER": "of"}, {"LOWER": "trust"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "declaratory"}, {"LOWER": "judgment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "declaratory"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "declaratory"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "declaratory"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "decree"}, {"LOWER": "absolute"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deductability"},
            {"LOWER": "of"},
            {"LOWER": "input"},
            {"LOWER": "tax"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deductibility"},
            {"LOWER": "of"},
            {"LOWER": "input"},
            {"LOWER": "tax"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deductibility"},
            {"LOWER": "of"},
            {"LOWER": "output"},
            {"LOWER": "tax"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deductible"}, {"LOWER": "expenditure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deduction"}, {"LOWER": "from"}, {"LOWER": "salary"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deduction"},
            {"LOWER": "of"},
            {"LOWER": "input"},
            {"LOWER": "tax"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deductions"},
            {"LOWER": "from"},
            {"LOWER": "prisoners"},
            {"LOWER": "’"},
            {"LOWER": "earnings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deed"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deed"}, {"LOWER": "of"}, {"LOWER": "adherence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deed"}, {"LOWER": "of"}, {"LOWER": "assignment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deed"}, {"LOWER": "of"}, {"LOWER": "compromise"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deed"}, {"LOWER": "of"}, {"LOWER": "easement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deed"}, {"LOWER": "of"}, {"LOWER": "variation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deed"}, {"ORTH": ","}, {"LOWER": "by"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deemed"}, {"LOWER": "loan"}, {"LOWER": "transactions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deeming"}, {"LOWER": "provision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defamation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defamatory"}, {"LOWER": "comments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defamatory"}, {"LOWER": "meaning"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "default"}, {"LOWER": "judgment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "default"}, {"LOWER": "summons"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defect"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defective"}, {"LOWER": "cargo"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defective"}, {"LOWER": "design"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defective"}, {"LOWER": "equipment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defective"}, {"LOWER": "premises"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defective"}, {"LOWER": "title"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "defects"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defence"},
            {"LOWER": "of"},
            {"LOWER": "reasonable"},
            {"LOWER": "practicability"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "defence"}, {"LOWER": "statement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "defence"}, {"LOWER": "to"}, {"LOWER": "charge"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "defence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "convicted"},
            {"LOWER": "of"},
            {"LOWER": "causing"},
            {"LOWER": "grievous"},
            {"LOWER": "bodily"},
            {"LOWER": "harm"},
            {"LOWER": "with"},
            {"LOWER": "intent"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "detained"},
            {"LOWER": "and"},
            {"LOWER": "handcuffed"},
            {"LOWER": "by"},
            {"LOWER": "police"},
            {"LOWER": "officers"},
            {"LOWER": "and"},
            {"LOWER": "told"},
            {"LOWER": "to"},
            {"LOWER": "wait"},
            {"LOWER": "for"},
            {"LOWER": "other"},
            {"LOWER": "police"},
            {"LOWER": "officers"},
            {"LOWER": "to"},
            {"LOWER": "come"},
            {"LOWER": "to"},
            {"LOWER": "arrest"},
            {"LOWER": "him"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "employing"},
            {"LOWER": "independent"},
            {"LOWER": "contractor"},
            {"LOWER": "to"},
            {"LOWER": "work"},
            {"LOWER": "at"},
            {"LOWER": "premises"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "engaging"},
            {"LOWER": "in"},
            {"LOWER": "conduct"},
            {"LOWER": "amounting"},
            {"LOWER": "to"},
            {"LOWER": "wilful"},
            {"LOWER": "obstruction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "failing"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "vat"},
            {"LOWER": "payments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "gambling"},
            {"LOWER": "in"},
            {"LOWER": "two"},
            {"LOWER": "years"},
            {"LOWER": "before"},
            {"LOWER": "insolvency"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "in"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "permanent"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "making"},
            {"LOWER": "part"},
            {"LOWER": "36"},
            {"LOWER": "offer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "defendant"}, {"LOWER": "not"}, {"LOWER": "represented"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "pleading"},
            {"LOWER": "by"},
            {"LOWER": "way"},
            {"LOWER": "of"},
            {"LOWER": "defence"},
            {"LOWER": "a"},
            {"LOWER": "set"},
            {"LOWER": "off"},
            {"LOWER": "of"},
            {"LOWER": "damages"},
            {"LOWER": "claimed"},
            {"LOWER": "by"},
            {"LOWER": "counterclaim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "subject"},
            {"LOWER": "of"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "tribunal"},
            {"LOWER": "ordering"},
            {"LOWER": "claimant"},
            {"LOWER": "to"},
            {"LOWER": "pay"},
            {"LOWER": "costs"},
            {"LOWER": "of"},
            {"LOWER": "interested"},
            {"LOWER": "party"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "defendants"}, {"LOWER": "character"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendants"},
            {"LOWER": "charged"},
            {"LOWER": "with"},
            {"LOWER": "laundering"},
            {"LOWER": "proceeds"},
            {"LOWER": "of"},
            {"LOWER": "third"},
            {"LOWER": "party"},
            {"LOWER": "’s"},
            {"LOWER": "conspiracy"},
            {"LOWER": "to"},
            {"LOWER": "traffic"},
            {"LOWER": "drugs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendants"},
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "evidence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendants"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "evidence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendants"},
            {"LOWER": "silence"},
            {"LOWER": "at"},
            {"LOWER": "police"},
            {"LOWER": "interview"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendants"},
            {"LOWER": "suffering"},
            {"LOWER": "from"},
            {"LOWER": "mental"},
            {"LOWER": "disorders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "’s"},
            {"LOWER": "agreement"},
            {"LOWER": "to"},
            {"LOWER": "pay"},
            {"LOWER": "recorded"},
            {"LOWER": "in"},
            {"LOWER": "accountants"},
            {"LOWER": "’"},
            {"LOWER": "letter"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "defendant"},
            {"LOWER": "’s"},
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "evidence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deferred"}, {"LOWER": "sentence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deficiency"}, {"LOWER": "in"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "definition"},
            {"ORTH": ","},
            {"LOWER": "whether"},
            {"LOWER": "within"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "definitive"}, {"LOWER": "map"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "degree"}, {"LOWER": "of"}, {"LOWER": "risk"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delay"},
            {"LOWER": "between"},
            {"LOWER": "charge"},
            {"LOWER": "and"},
            {"LOWER": "trial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delay"},
            {"LOWER": "by"},
            {"LOWER": "parole"},
            {"LOWER": "board"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "delay"}, {"LOWER": "in"}, {"LOWER": "completion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "delay"}, {"LOWER": "in"}, {"LOWER": "delivering"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delay"},
            {"LOWER": "in"},
            {"LOWER": "prosecuting"},
            {"LOWER": "claim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "delay"}, {"LOWER": "in"}, {"LOWER": "prosecution"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delay"},
            {"LOWER": "in"},
            {"LOWER": "prosecution"},
            {"LOWER": "process"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "delegation"}, {"LOWER": "of"}, {"LOWER": "functions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delegation"},
            {"LOWER": "of"},
            {"LOWER": "licensing"},
            {"LOWER": "functions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delegation"},
            {"LOWER": "to"},
            {"LOWER": "independent"},
            {"LOWER": "contractor"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deleted"}, {"LOWER": "words"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deliberate"},
            {"LOWER": "concealment"},
            {"LOWER": "of"},
            {"LOWER": "facts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delivery"},
            {"LOWER": "up"},
            {"LOWER": "of"},
            {"LOWER": "deeds"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "delivery"},
            {"LOWER": "without"},
            {"LOWER": "documents"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "demand"}, {"LOWER": "for"}, {"LOWER": "payment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "demolition"}, {"LOWER": "of"}, {"LOWER": "building"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "demonstration"},
            {"LOWER": "in"},
            {"LOWER": "designated"},
            {"LOWER": "area"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "demoted"}, {"LOWER": "tenancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "demurrage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dental"}, {"LOWER": "practice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dental"}, {"LOWER": "practitioner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dentist"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dentist"},
            {"LOWER": "stuck"},
            {"LOWER": "off"},
            {"LOWER": "as"},
            {"LOWER": "unfit"},
            {"LOWER": "by"},
            {"LOWER": "reason"},
            {"LOWER": "of"},
            {"LOWER": "dishonesty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "department"},
            {"LOWER": "for"},
            {"LOWER": "work"},
            {"LOWER": "and"},
            {"LOWER": "pensions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dependent"},
            {"LOWER": "relative"},
            {"LOWER": "revocation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deportation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deportation"},
            {"LOWER": "of"},
            {"LOWER": "putative"},
            {"LOWER": "refugee"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deportation"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "deportation"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deportation"},
            {"LOWER": "order"},
            {"ORTH": ","},
            {"LOWER": "recommendation"},
            {"LOWER": "for"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deportation"},
            {"ORTH": ","},
            {"LOWER": "recommendation"},
            {"LOWER": "for"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deposit"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deposit"}, {"LOWER": "taking"}, {"LOWER": "business"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deposit"},
            {"ORTH": "-"},
            {"LOWER": "taking"},
            {"LOWER": "business"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "depreciation"},
            {"LOWER": "caused"},
            {"LOWER": "by"},
            {"LOWER": "public"},
            {"LOWER": "works"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "depreciation"},
            {"LOWER": "in"},
            {"LOWER": "value"},
            {"LOWER": "of"},
            {"LOWER": "land"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deprivation"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deprivation"},
            {"LOWER": "of"},
            {"LOWER": "citizenship"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deprivation"}, {"LOWER": "of"}, {"LOWER": "liberty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deprivation"},
            {"LOWER": "of"},
            {"LOWER": "possessions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "deprivation"}, {"LOWER": "of"}, {"LOWER": "property"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "deprivation"}, {"LOWER": "provision"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "deputy"},
            {"LOWER": "director"},
            {"LOWER": "of"},
            {"LOWER": "public"},
            {"LOWER": "prosecutions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "derivative"}, {"LOWER": "action"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "derivative"}, {"LOWER": "agreements"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "derivative"}, {"LOWER": "claim"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "derivative"}, {"LOWER": "contracts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dermatitis"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dermatitis"},
            {"ORTH": ","},
            {"LOWER": "risk"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "design"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "design"}, {"LOWER": "contract"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "designated"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "despina"}, {"LOWER": "g.k"}, {"ORTH": "."}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "destitute"},
            {"LOWER": "domestic"},
            {"LOWER": "violence"},
            {"LOWER": "concession"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "detailed"}, {"LOWER": "assessment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "detention"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "detention"}, {"LOWER": "for"}, {"LOWER": "examination"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "detention"},
            {"LOWER": "for"},
            {"LOWER": "purposes"},
            {"LOWER": "of"},
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "applicant"},
            {"LOWER": "to"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "responsible"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "detention"}, {"LOWER": "in"}, {"LOWER": "afghanistan"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "detention"},
            {"LOWER": "in"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "bill"},
            {"LOWER": "of"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "detention"}, {"LOWER": "in"}, {"LOWER": "hospital"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "detention"},
            {"LOWER": "in"},
            {"LOWER": "police"},
            {"LOWER": "cell"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "detention"},
            {"LOWER": "pending"},
            {"LOWER": "deportation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "detention"},
            {"LOWER": "pending"},
            {"LOWER": "deportation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "detention"},
            {"LOWER": "pending"},
            {"LOWER": "deportation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "detention"}, {"LOWER": "pending"}, {"LOWER": "removal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "detention"},
            {"LOWER": "prior"},
            {"LOWER": "to"},
            {"LOWER": "extradition"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "determination"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"LOWER": "applications"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"LOWER": "applications"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"LOWER": "civil"},
            {"LOWER": "rights"},
            {"LOWER": "and"},
            {"LOWER": "obligations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "determination"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "determination"}, {"LOWER": "of"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "responsible"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "responsible"},
            {"LOWER": "for"},
            {"LOWER": "examining"},
            {"LOWER": "application"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"LOWER": "patient"},
            {"LOWER": "’s"},
            {"LOWER": "best"},
            {"LOWER": "interests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"LOWER": "reference"},
            {"LOWER": "date"},
            {"LOWER": "for"},
            {"LOWER": "purposes"},
            {"LOWER": "of"},
            {"LOWER": "advancement"},
            {"LOWER": "on"},
            {"LOWER": "salary"},
            {"LOWER": "scale"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "civil"},
            {"LOWER": "rights"},
            {"LOWER": "and"},
            {"LOWER": "obligations"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "determination"},
            {"LOWER": "whether"},
            {"LOWER": "prima"},
            {"LOWER": "facie"},
            {"LOWER": "case"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "development"},
            {"LOWER": "affecting"},
            {"LOWER": "listed"},
            {"LOWER": "buildings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "development"},
            {"LOWER": "affecting"},
            {"LOWER": "setting"},
            {"LOWER": "of"},
            {"LOWER": "listed"},
            {"LOWER": "building"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "development"}, {"LOWER": "consent"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "development"}, {"LOWER": "land"}, {"LOWER": "tax"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "development"}, {"LOWER": "of"}, {"LOWER": "railways"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "development"}, {"LOWER": "plan"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "development"}, {"LOWER": "plan"}, {"LOWER": "document"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "devolution"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "devolution"}, {"LOWER": "issue"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "diminished"}, {"LOWER": "responsibility"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "diminished"},
            {"LOWER": "responsibility"},
            {"LOWER": "manslaughter"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "diminution"},
            {"LOWER": "in"},
            {"LOWER": "value"},
            {"LOWER": "of"},
            {"LOWER": "claim"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "diplomatic"}, {"LOWER": "immunity"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "diplomatic"},
            {"LOWER": "immunity-"},
            {"LOWER": "dwelling"},
            {"LOWER": "house"},
            {"LOWER": "leased"},
            {"LOWER": "to"},
            {"LOWER": "foreign"},
            {"LOWER": "state"},
            {"LOWER": "and"},
            {"LOWER": "occupied"},
            {"LOWER": "by"},
            {"LOWER": "diplomatic"},
            {"LOWER": "agent"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "direct"}, {"LOWER": "discrimination"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "direct"}, {"LOWER": "effect"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "direct"},
            {"LOWER": "effect"},
            {"LOWER": "of"},
            {"LOWER": "international"},
            {"LOWER": "treaty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "direct"}, {"LOWER": "life"}, {"LOWER": "assurance"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "direct"}, {"LOWER": "taxation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "direction"},
            {"LOWER": "to"},
            {"LOWER": "investigate"},
            {"LOWER": "child"},
            {"LOWER": "’s"},
            {"LOWER": "circumstances"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "direction"}, {"LOWER": "to"}, {"LOWER": "jury"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "directions"}, {"LOWER": "hearing"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "directions"},
            {"LOWER": "regarding"},
            {"LOWER": "seized"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "directions"}, {"LOWER": "to"}, {"LOWER": "jury"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "director"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "director"},
            {"LOWER": "and"},
            {"LOWER": "senior"},
            {"LOWER": "company"},
            {"LOWER": "employee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "director"},
            {"LOWER": "general"},
            {"LOWER": "of"},
            {"LOWER": "fair"},
            {"LOWER": "trading"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "director"},
            {"LOWER": "general"},
            {"LOWER": "of"},
            {"LOWER": "telecommunications"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "director"},
            {"LOWER": "of"},
            {"LOWER": "defendant"},
            {"LOWER": "company"},
            {"LOWER": "taking"},
            {"LOWER": "over"},
            {"LOWER": "conduct"},
            {"LOWER": "of"},
            {"LOWER": "case"},
            {"LOWER": "after"},
            {"LOWER": "solicitors"},
            {"LOWER": "coming"},
            {"LOWER": "off"},
            {"LOWER": "record"},
            {"LOWER": "18"},
            {"LOWER": "days"},
            {"LOWER": "before"},
            {"LOWER": "trial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "director"},
            {"LOWER": "of"},
            {"LOWER": "public"},
            {"LOWER": "prosecutions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "directors"},
            {"LOWER": "and"},
            {"LOWER": "financial"},
            {"LOWER": "advisers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "directors"}, {"LOWER": "and"}, {"LOWER": "officers"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "directors"},
            {"LOWER": "disqualification"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disability"}, {"LOWER": "discrimination"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disability"},
            {"LOWER": "appeal"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disability"},
            {"LOWER": "living"},
            {"LOWER": "allowance"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disability"}, {"LOWER": "premium"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disability"},
            {"LOWER": "related"},
            {"LOWER": "discrimination"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disability"}, {"ORTH": "-"}, {"LOWER": "related"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disabled"}, {"LOWER": "facilities"}, {"LOWER": "grant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disabled"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disabled"}, {"LOWER": "persons"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disablement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disbursements"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "discharge"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discharge"}, {"LOWER": "from"}, {"LOWER": "liability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discharge"},
            {"LOWER": "of"},
            {"LOWER": "detained"},
            {"LOWER": "patient"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discharge"}, {"LOWER": "of"}, {"LOWER": "patient"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discharge"}, {"LOWER": "or"}, {"LOWER": "modification"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discharged"}, {"LOWER": "mental"}, {"LOWER": "patient"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discharged"},
            {"LOWER": "mental"},
            {"LOWER": "patients"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "discharged"}, {"LOWER": "patient"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disciplinary"}, {"LOWER": "action"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disciplinary"}, {"LOWER": "hearing"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disciplinary"}, {"LOWER": "procedure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disciplinary"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disciplinary"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disciplinary"},
            {"LOWER": "proceedings"},
            {"LOWER": "before"},
            {"LOWER": "nursing"},
            {"LOWER": "and"},
            {"LOWER": "midwifery"},
            {"LOWER": "council"},
            {"ORTH": "("},
            {"ORTH": "“"},
            {"LOWER": "nmc"},
            {"ORTH": "”"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disciplinary"}, {"LOWER": "tribunal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disclaimer"}, {"LOWER": "by"}, {"LOWER": "trustee"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disclaimer"}, {"LOWER": "of"}, {"LOWER": "lease"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclaimer"},
            {"LOWER": "of"},
            {"LOWER": "lease"},
            {"LOWER": "by"},
            {"LOWER": "trustee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclaimer"},
            {"LOWER": "of"},
            {"LOWER": "onerous"},
            {"LOWER": "property"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disclosure"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "and"},
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disclosure"}, {"LOWER": "by"}, {"LOWER": "prosecution"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "by"},
            {"LOWER": "third"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "childs"},
            {"LOWER": "address"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "deliberations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disclosure"}, {"LOWER": "of"}, {"LOWER": "documents"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "draft"},
            {"LOWER": "speeches"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disclosure"}, {"LOWER": "of"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disclosure"}, {"LOWER": "of"}, {"LOWER": "interest"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "material"},
            {"LOWER": "facts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "personal"},
            {"LOWER": "data"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "record"},
            {"LOWER": "of"},
            {"LOWER": "birth"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disclosure"}, {"LOWER": "of"}, {"LOWER": "reports"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "unused"},
            {"LOWER": "material"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disclosure"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disclosures"},
            {"LOWER": "of"},
            {"LOWER": "deliberations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discontinuance"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discontinuance"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discontinuance"},
            {"ORTH": ","},
            {"LOWER": "notice"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discontinuance"}, {"ORTH": ","}, {"LOWER": "whether"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discontinuation"},
            {"LOWER": "of"},
            {"LOWER": "treatment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "discovery"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discovery"},
            {"LOWER": "in"},
            {"LOWER": "aid"},
            {"LOWER": "of"},
            {"LOWER": "execution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discovery"},
            {"LOWER": "of"},
            {"LOWER": "documents"},
            {"LOWER": "retained"},
            {"LOWER": "by"},
            {"LOWER": "former"},
            {"LOWER": "agent"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discretion"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discretion"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discretion"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discretion"}, {"LOWER": "of"}, {"LOWER": "house"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discretion"},
            {"LOWER": "of"},
            {"LOWER": "judge"},
            {"ORTH": ","},
            {"LOWER": "matter"},
            {"LOWER": "within"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discretion"}, {"LOWER": "to"}, {"LOWER": "exclude"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discretion"},
            {"LOWER": "to"},
            {"LOWER": "quash"},
            {"LOWER": "verdict"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discretionary"},
            {"LOWER": "housing"},
            {"LOWER": "payment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discretionary"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discretionary"},
            {"LOWER": "life"},
            {"LOWER": "sentence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "discrimination"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discrimination"},
            {"LOWER": "arising"},
            {"LOWER": "from"},
            {"LOWER": "disability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discrimination"},
            {"LOWER": "by"},
            {"LOWER": "bodies"},
            {"LOWER": "other"},
            {"LOWER": "than"},
            {"LOWER": "employers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discrimination"},
            {"LOWER": "by"},
            {"LOWER": "contracting"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discrimination"},
            {"ORTH": ","},
            {"LOWER": "disability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discrimination"}, {"ORTH": ","}, {"LOWER": "race"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discrimination"}, {"ORTH": ","}, {"LOWER": "sex"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discrimination"}, {"ORTH": ","}, {"LOWER": "sex"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "discrimination"},
            {"ORTH": ","},
            {"LOWER": "sex"},
            {"ORTH": ","},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "discriminaton"}, {"ORTH": ","}, {"LOWER": "sex"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "discriminatory"}, {"LOWER": "act"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dishonest"}, {"LOWER": "appropriation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dishonest"}, {"LOWER": "assistance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dishonest"}, {"LOWER": "evasion"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dishonesty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disinterment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disinterment"},
            {"LOWER": "of"},
            {"LOWER": "remains"},
            {"LOWER": "of"},
            {"LOWER": "yugoslavian"},
            {"LOWER": "royalty"},
            {"LOWER": "for"},
            {"LOWER": "reinterment"},
            {"LOWER": "in"},
            {"LOWER": "yugoslavia"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dismissal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dismissal"}, {"LOWER": "for"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dismissal"},
            {"LOWER": "for"},
            {"LOWER": "reason"},
            {"LOWER": "of"},
            {"LOWER": "pregnancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dismissal"}, {"LOWER": "for"}, {"LOWER": "redundancy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dismissal"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dismissal"},
            {"LOWER": "of"},
            {"LOWER": "action"},
            {"LOWER": "for"},
            {"LOWER": "want"},
            {"LOWER": "of"},
            {"LOWER": "prosecution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dismissal"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "by"},
            {"LOWER": "request"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dismissal"},
            {"LOWER": "of"},
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "want"},
            {"LOWER": "of"},
            {"LOWER": "prosecution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dismissal"}, {"LOWER": "of"}, {"LOWER": "charge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dismissal"},
            {"LOWER": "of"},
            {"LOWER": "charges"},
            {"LOWER": "prior"},
            {"LOWER": "to"},
            {"LOWER": "arraignment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dismissal"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dismissal"}, {"LOWER": "of"}, {"LOWER": "information"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dismissed"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disobeying"}, {"LOWER": "lawful"}, {"LOWER": "command"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dispensing"}, {"LOWER": "with"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dispensing"},
            {"LOWER": "with"},
            {"LOWER": "parents"},
            {"LOWER": "’"},
            {"LOWER": "consent"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dispersal"}, {"LOWER": "of"}, {"LOWER": "groups"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "display"}, {"LOWER": "of"}, {"LOWER": "advertisements"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disposable"}, {"LOWER": "income"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disposal"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disposal"}, {"LOWER": "licence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disposal"}, {"LOWER": "of"}, {"LOWER": "asset"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disposal"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disposal"}, {"LOWER": "of"}, {"LOWER": "chattels"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disposal"}, {"LOWER": "of"}, {"LOWER": "chattels"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disposal"},
            {"LOWER": "of"},
            {"LOWER": "companys"},
            {"LOWER": "assets"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disposal"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disposition"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disposition"},
            {"LOWER": "of"},
            {"LOWER": "equitable"},
            {"LOWER": "interest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disposition"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disposition"}, {"LOWER": "of"}, {"LOWER": "property"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dispute"},
            {"LOWER": "adjudication"},
            {"LOWER": "clause"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dispute"},
            {"LOWER": "between"},
            {"LOWER": "parties"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "meaning"},
            {"LOWER": "of"},
            {"LOWER": "clause"},
            {"LOWER": "in"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dispute"}, {"LOWER": "resolution"}, {"LOWER": "clause"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dispute"},
            {"LOWER": "resolution"},
            {"LOWER": "procedure"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "disputed"}, {"LOWER": "payments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "disqualification"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disqualification"},
            {"LOWER": "from"},
            {"LOWER": "allowance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disqualification"},
            {"LOWER": "from"},
            {"LOWER": "benefit"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disqualification"},
            {"LOWER": "from"},
            {"LOWER": "driving"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "disrupting"},
            {"LOWER": "lawful"},
            {"LOWER": "activity"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dissolution"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "distance"}, {"LOWER": "contract"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "distinction"},
            {"LOWER": "between"},
            {"LOWER": "reasonable"},
            {"LOWER": "suspicion"},
            {"LOWER": "and"},
            {"LOWER": "prima"},
            {"LOWER": "facie"},
            {"LOWER": "proof"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "distinctive"}, {"LOWER": "character"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "distress"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distress"}, {"LOWER": "damage"}, {"LOWER": "feasant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "distress"}, {"LOWER": "warrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distribution"}, {"LOWER": "agreement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distribution"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distribution"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distribution"}, {"LOWER": "of"}, {"LOWER": "business"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distribution"}, {"LOWER": "of"}, {"LOWER": "fund"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distribution"}, {"LOWER": "of"}, {"LOWER": "profits"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "distributorship"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "district"}, {"LOWER": "auditor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "district"}, {"LOWER": "auditors"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disturbance"}, {"ORTH": ","}, {"LOWER": "for"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "disused"}, {"LOWER": "burial"}, {"LOWER": "ground"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dividends"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "divorce"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "divorce"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dna"}, {"LOWER": "evidence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "dna"}, {"LOWER": "profile"}, {"LOWER": "evidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dna"},
            {"LOWER": "profile"},
            {"LOWER": "evidence"},
            {"LOWER": "from"},
            {"LOWER": "moveable"},
            {"LOWER": "object"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dna"}, {"LOWER": "testing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dock"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "documentary"}, {"LOWER": "evidence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "documents"},
            {"LOWER": "disclosed"},
            {"LOWER": "in"},
            {"LOWER": "high"},
            {"LOWER": "court"},
            {"LOWER": "action"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "documents"},
            {"LOWER": "in"},
            {"LOWER": "public"},
            {"LOWER": "procurement"},
            {"LOWER": "process"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "documents"},
            {"ORTH": ","},
            {"LOWER": "admissibility"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dog"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dog"},
            {"LOWER": "breeding"},
            {"LOWER": "establishment"},
            {"LOWER": "licence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "domestic"}, {"LOWER": "remedy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "domestic"}, {"LOWER": "tribunal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "domestic"}, {"LOWER": "violence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "domicile"},
            {"LOWER": "outside"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dominant"}, {"LOWER": "position"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dominant"},
            {"LOWER": "tenement"},
            {"ORTH": ","},
            {"LOWER": "identification"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "donatio"}, {"LOWER": "mortis"}, {"LOWER": "causa"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "donation"},
            {"LOWER": "to"},
            {"LOWER": "registered"},
            {"LOWER": "political"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "donor"}, {"LOWER": "’s"}, {"LOWER": "intention"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "door"}, {"LOWER": "supervisors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "double"}, {"LOWER": "criminality"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "double"}, {"LOWER": "jeopardy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "double"},
            {"LOWER": "murderer"},
            {"LOWER": "held"},
            {"LOWER": "under"},
            {"LOWER": "category"},
            {"LOWER": "a"},
            {"LOWER": "security"},
            {"LOWER": "conditions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "double"}, {"LOWER": "portion"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "double"}, {"LOWER": "taxation"}, {"LOWER": "relief"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "draft"},
            {"LOWER": "order"},
            {"LOWER": "in"},
            {"LOWER": "council"},
            {"LOWER": "laid"},
            {"LOWER": "before"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "drain"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "drainage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dramatic"}, {"LOWER": "work"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "identity"}, {"LOWER": "of"}, {"LOWER": "driver"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "drivers"}, {"LOWER": "hours"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "drivers"}, {"LOWER": "’"}, {"LOWER": "hours"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "driving"}, {"LOWER": "licence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "driving"}, {"LOWER": "other"}, {"LOWER": "cars"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "driving"}, {"LOWER": "uninsured"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "driving"},
            {"LOWER": "with"},
            {"LOWER": "excess"},
            {"LOWER": "alcohol"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "drug"}, {"LOWER": "addiction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "drug"}, {"LOWER": "precursors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "drugs"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "drugs"},
            {"LOWER": "importation"},
            {"ORTH": ","},
            {"LOWER": "prohibition"},
            {"LOWER": "on"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "provision"}, {"LOWER": "of"}, {"LOWER": "drugs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "drunk"},
            {"LOWER": "in"},
            {"LOWER": "charge"},
            {"LOWER": "of"},
            {"LOWER": "motor"},
            {"LOWER": "vehicle"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dual"},
            {"ORTH": "-"},
            {"LOWER": "use"},
            {"LOWER": "items"},
            {"LOWER": "for"},
            {"LOWER": "which"},
            {"LOWER": "export"},
            {"LOWER": "licence"},
            {"LOWER": "might"},
            {"LOWER": "be"},
            {"LOWER": "required"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duke"}, {"LOWER": "of"}, {"LOWER": "yare"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "dumped"},
            {"LOWER": "imports"},
            {"LOWER": "from"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "members"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dumping"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duration"}, {"LOWER": "of"}, {"LOWER": "marriage"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "duress"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duress"}, {"LOWER": "as"}, {"LOWER": "defence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "as"}, {"LOWER": "trustee"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "duty"}, {"LOWER": "chargeable"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "of"}, {"LOWER": "bank"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "of"}, {"LOWER": "candour"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "of"}, {"LOWER": "care"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "of"},
            {"LOWER": "care"},
            {"LOWER": "to"},
            {"LOWER": "whom"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "of"},
            {"LOWER": "care"},
            {"LOWER": "to"},
            {"LOWER": "whom"},
            {"ORTH": "?"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "of"}, {"LOWER": "consultation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "of"}, {"LOWER": "disclosure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "of"},
            {"LOWER": "disclosure"},
            {"LOWER": "to"},
            {"LOWER": "surety"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "of"},
            {"LOWER": "planning"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "of"}, {"LOWER": "trustee"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "on"},
            {"LOWER": "letters"},
            {"LOWER": "of"},
            {"LOWER": "administration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "account"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "advise"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "fair"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "consult"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "customer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "deduct"},
            {"LOWER": "income"},
            {"LOWER": "tax"},
            {"LOWER": "from"},
            {"LOWER": "payment"},
            {"LOWER": "of"},
            {"LOWER": "yearly"},
            {"LOWER": "interest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "reasons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "hold"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "inform"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "maintain"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "maintain"},
            {"LOWER": "highway"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "adjustments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "special"},
            {"LOWER": "educational"},
            {"LOWER": "needs"},
            {"LOWER": "provision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "special"},
            {"LOWER": "educational"},
            {"LOWER": "provision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "prevent"},
            {"LOWER": "obstruction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "duty"}, {"LOWER": "to"}, {"LOWER": "provide"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "accommodation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "accommodation"},
            {"LOWER": "and"},
            {"LOWER": "maintenance"},
            {"LOWER": "for"},
            {"LOWER": "children"},
            {"LOWER": "being"},
            {"LOWER": "looked"},
            {"LOWER": "after"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "requested"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "rate"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "dwelling"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "dwelling"}, {"LOWER": "house"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "early"}, {"LOWER": "conciliation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "early"},
            {"LOWER": "pension"},
            {"LOWER": "without"},
            {"LOWER": "reduction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "early"}, {"LOWER": "retirement"}, {"LOWER": "support"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "early"}, {"LOWER": "stages"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "earnings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "earnings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "earnings"},
            {"LOWER": "from"},
            {"LOWER": "work"},
            {"LOWER": "done"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "earnings"},
            {"ORTH": ","},
            {"LOWER": "loss"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "easement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ecclesiastical"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ecclesiastical"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ecclesiastical"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ecclesiastical"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ecclesiastical"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ecclesiastical"}, {"LOWER": "law"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ecclesiastical"}, {"LOWER": "property"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ecclesiatical"}, {"LOWER": "law"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "economic"},
            {"LOWER": "and"},
            {"LOWER": "monetary"},
            {"LOWER": "policy"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "economic"}, {"LOWER": "loss"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "economic"},
            {"ORTH": ","},
            {"LOWER": "social"},
            {"LOWER": "and"},
            {"LOWER": "territorial"},
            {"LOWER": "cohesion"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "education"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "education"},
            {"LOWER": "authority"},
            {"ORTH": ","},
            {"LOWER": "statutory"},
            {"LOWER": "duties"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "educational"}, {"LOWER": "establishment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "educational"},
            {"LOWER": "institution"},
            {"LOWER": "refused"},
            {"LOWER": "hts"},
            {"LOWER": "status"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "educational"},
            {"LOWER": "provision"},
            {"LOWER": "for"},
            {"LOWER": "improving"},
            {"LOWER": "behaviour"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "effect"}, {"LOWER": "of"}, {"LOWER": "assignment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "effect"},
            {"LOWER": "of"},
            {"LOWER": "country"},
            {"LOWER": "guidance"},
            {"LOWER": "determination"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "effect"},
            {"LOWER": "of"},
            {"LOWER": "earlier"},
            {"LOWER": "faculty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "effect"},
            {"LOWER": "of"},
            {"LOWER": "multiple"},
            {"LOWER": "notices"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "effect"},
            {"LOWER": "of"},
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "nuptial"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "effect"}, {"LOWER": "of"}, {"LOWER": "term"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "effect"}, {"LOWER": "of"}, {"LOWER": "transfer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "effect"},
            {"LOWER": "of"},
            {"LOWER": "withdrawal"},
            {"LOWER": "from"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "election"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "election"},
            {"LOWER": "for"},
            {"LOWER": "trial"},
            {"LOWER": "on"},
            {"LOWER": "indictment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "election"}, {"ORTH": "."}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "elections"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "electricity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "electronic"}, {"LOWER": "commerce"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "electronic"}, {"LOWER": "communications"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "electronic"},
            {"LOWER": "communications"},
            {"LOWER": "networks"},
            {"LOWER": "and"},
            {"LOWER": "services"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "electronic"}, {"LOWER": "data"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "electronically"},
            {"LOWER": "stored"},
            {"LOWER": "information"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "eleftherios"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "elements"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "eligibility"}, {"LOWER": "assessment"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "eligibility"}, {"LOWER": "criteria"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "embracery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "embryo"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "emergency"},
            {"LOWER": "marketing"},
            {"LOWER": "authorisation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "emergency"},
            {"LOWER": "protection"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "emergency"}, {"LOWER": "surgery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employee"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employee"}, {"LOWER": "bonus"}, {"LOWER": "payments"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employee"},
            {"LOWER": "denied"},
            {"LOWER": "maternity"},
            {"LOWER": "leave"},
            {"LOWER": "/"},
            {"LOWER": "pay"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "employee"}, {"LOWER": "inventor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employee"},
            {"LOWER": "liability"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employees"}, {"LOWER": "benefit"}, {"LOWER": "fund"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employees"}, {"LOWER": "compensation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employees"}, {"LOWER": "provident"}, {"LOWER": "fund"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "employer"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employer"}, {"LOWER": "and"}, {"LOWER": "employee"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "employer"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employers"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employers"},
            {"LOWER": "indemnity"},
            {"LOWER": "insurance"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "employers"}, {"LOWER": "insolvency"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employers"}, {"LOWER": "insolvency"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employers"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employers"}, {"LOWER": "liability"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employers"}, {"LOWER": "’"}, {"LOWER": "insolvency"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employer"}, {"LOWER": "’s"}, {"LOWER": "contributions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employer"}, {"LOWER": "’s"}, {"LOWER": "duty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employer"}, {"LOWER": "’s"}, {"LOWER": "insolvency"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employer"}, {"LOWER": "’s"}, {"LOWER": "liability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "employer"}, {"LOWER": "’s"}, {"LOWER": "liabilty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employer"},
            {"LOWER": "’s"},
            {"LOWER": "penalty"},
            {"LOWER": "notice"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "agency"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employment"},
            {"LOWER": "and"},
            {"LOWER": "support"},
            {"LOWER": "allowance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employment"},
            {"LOWER": "appeal"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employment"},
            {"LOWER": "appeal"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employment"},
            {"LOWER": "continuous"},
            {"LOWER": "employment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "contract"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "dismissal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "income"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employment"},
            {"LOWER": "service"},
            {"LOWER": "provider"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "tribunal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "tribunal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "tribunals"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "tribunals"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "employment"}, {"LOWER": "tribunals"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employment"},
            {"ORTH": ","},
            {"LOWER": "skills"},
            {"LOWER": "and"},
            {"LOWER": "enterprise"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "employment"},
            {"ORTH": "-"},
            {"LOWER": "related"},
            {"LOWER": "securities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "encouraging"},
            {"LOWER": "or"},
            {"LOWER": "assisting"},
            {"LOWER": "crime"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "encouraging"},
            {"LOWER": "or"},
            {"LOWER": "assisting"},
            {"LOWER": "offence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "encouraging"},
            {"LOWER": "or"},
            {"LOWER": "assisting"},
            {"LOWER": "offences"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "encouraging"},
            {"LOWER": "or"},
            {"LOWER": "assisting"},
            {"LOWER": "offences"},
            {"LOWER": "believing"},
            {"LOWER": "one"},
            {"LOWER": "or"},
            {"LOWER": "more"},
            {"LOWER": "will"},
            {"LOWER": "be"},
            {"LOWER": "committed"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "encouraging"},
            {"LOWER": "or"},
            {"LOWER": "assisting"},
            {"LOWER": "offences"},
            {"LOWER": "believing"},
            {"LOWER": "one"},
            {"LOWER": "or"},
            {"LOWER": "more"},
            {"LOWER": "would"},
            {"LOWER": "be"},
            {"LOWER": "committed"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "encroaching"}, {"LOWER": "tree"}, {"LOWER": "roots"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "encroachment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "endangered"}, {"LOWER": "species"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enduring"},
            {"LOWER": "power"},
            {"LOWER": "of"},
            {"LOWER": "attorney"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enduring"},
            {"LOWER": "power"},
            {"LOWER": "of"},
            {"LOWER": "attorney"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "energy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "energy"},
            {"LOWER": "labelling"},
            {"LOWER": "of"},
            {"LOWER": "televisions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "enforceability"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforceability"}, {"LOWER": "of"}, {"LOWER": "orders"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "enforcement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enforcement"},
            {"LOWER": "against"},
            {"LOWER": "third"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforcement"}, {"LOWER": "in"}, {"LOWER": "england"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "enforcement"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
            {"LOWER": "adjudicator"},
            {"LOWER": "’s"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforcement"}, {"LOWER": "of"}, {"LOWER": "award"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
            {"LOWER": "california"},
            {"LOWER": "award"},
            {"LOWER": "in"},
            {"LOWER": "england"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforcement"}, {"LOWER": "of"}, {"LOWER": "charge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforcement"}, {"LOWER": "of"}, {"LOWER": "covenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforcement"}, {"LOWER": "of"}, {"LOWER": "fine"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforcement"}, {"LOWER": "of"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
            {"LOWER": "maintenance"},
            {"LOWER": "order"},
            {"LOWER": "registered"},
            {"LOWER": "in"},
            {"LOWER": "magistrates"},
            {"LOWER": "’"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "enforcement"}, {"LOWER": "of"}, {"LOWER": "sentence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "engaged"}, {"LOWER": "couple"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "engineers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "english"},
            {"LOWER": "nationals"},
            {"LOWER": "habitually"},
            {"LOWER": "resident"},
            {"LOWER": "in"},
            {"LOWER": "netherlands"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "enhanced"}, {"LOWER": "interest"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "enterprise"},
            {"LOWER": "investment"},
            {"LOWER": "scheme"},
            {"LOWER": "relief"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "entertainment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "entertainments"}, {"LOWER": "licence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "entitlement"}, {"LOWER": "reference"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "entitlement"},
            {"LOWER": "to"},
            {"LOWER": "bring"},
            {"LOWER": "claim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "entitlement"}, {"LOWER": "to"}, {"LOWER": "interest"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "entitlement"}, {"LOWER": "to"}, {"LOWER": "pension"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "entitlement"},
            {"LOWER": "to"},
            {"LOWER": "pension"},
            {"LOWER": "in"},
            {"LOWER": "several"},
            {"LOWER": "member"},
            {"LOWER": "states"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "entitlement"}, {"LOWER": "to"}, {"LOWER": "surplus"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "entitlement"},
            {"ORTH": ","},
            {"LOWER": "exclusion"},
            {"LOWER": "from"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "entry"},
            {"LOWER": "on"},
            {"LOWER": "land"},
            {"LOWER": "under"},
            {"LOWER": "s."},
            {"LOWER": "85"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "lands"},
            {"LOWER": "clauses"},
            {"LOWER": "consolidation"},
            {"LOWER": "act"},
            {"LOWER": "immediately"},
            {"LOWER": "before"},
            {"LOWER": "expiration"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "period"},
            {"LOWER": "allowed"},
            {"LOWER": "for"},
            {"LOWER": "completion"},
            {"LOWER": "of"},
            {"LOWER": "works"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "environment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "environment"}, {"LOWER": "agency"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "environmental"}, {"LOWER": "assessment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "environmental"}, {"LOWER": "considerations"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "environmental"},
            {"LOWER": "impact"},
            {"LOWER": "assessment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "environmental"},
            {"LOWER": "impact"},
            {"LOWER": "assessment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "environmental"},
            {"LOWER": "impact"},
            {"LOWER": "assessment"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "environmental"}, {"LOWER": "liability"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "environmental"}, {"LOWER": "permit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equal"}, {"LOWER": "pay"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equal"}, {"LOWER": "pay"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equal"}, {"LOWER": "pay"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equal"}, {"LOWER": "treatment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "equal"},
            {"LOWER": "treatment"},
            {"LOWER": "in"},
            {"LOWER": "relation"},
            {"LOWER": "to"},
            {"LOWER": "financial"},
            {"LOWER": "transactions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "equal"},
            {"LOWER": "treatment"},
            {"LOWER": "provisions"},
            {"LOWER": "for"},
            {"LOWER": "men"},
            {"LOWER": "and"},
            {"LOWER": "women"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "equality"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equitable"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equitable"}, {"LOWER": "assignment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "equitable"},
            {"LOWER": "assignment"},
            {"LOWER": "of"},
            {"LOWER": "future"},
            {"LOWER": "receipts"},
            {"LOWER": "of"},
            {"LOWER": "business"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "equitable"}, {"LOWER": "execution"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equitable"}, {"LOWER": "interest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "equity"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "equity"}, {"LOWER": "of"}, {"LOWER": "exoneration"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "equivocal"}, {"LOWER": "plea"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "erection"}, {"LOWER": "of"}, {"LOWER": "fencing"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "erroneous"}, {"LOWER": "advice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "error"}, {"LOWER": "or"}, {"LOWER": "omission"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "errors"}, {"LOWER": "of"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "escape"},
            {"LOWER": "from"},
            {"LOWER": "lawful"},
            {"LOWER": "custody"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "escrow"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "established"}, {"LOWER": "use"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "establishment"},
            {"LOWER": "in"},
            {"LOWER": "great"},
            {"LOWER": "britain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "establishment"},
            {"LOWER": "of"},
            {"LOWER": "new"},
            {"LOWER": "school"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "estate"}, {"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "estate"}, {"LOWER": "contract"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "estate"}, {"LOWER": "management"}, {"LOWER": "scheme"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "estate"},
            {"LOWER": "to"},
            {"LOWER": "wife"},
            {"LOWER": "absolutely"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "estimate"}, {"LOWER": "of"}, {"LOWER": "costs"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "estoppel"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "estoppel"},
            {"LOWER": "raised"},
            {"LOWER": "by"},
            {"LOWER": "assurances"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "estreatment"},
            {"LOWER": "of"},
            {"LOWER": "recognisance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "estreatment"},
            {"LOWER": "of"},
            {"LOWER": "recognizance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "estreatment"},
            {"LOWER": "of"},
            {"LOWER": "recognizances"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ethnic"}, {"LOWER": "origin"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "eu"}, {"LOWER": "grant"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "eu"}, {"LOWER": "registered"}, {"LOWER": "design"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "eu"}, {"LOWER": "trade"}, {"LOWER": "mark"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "agricultural"},
            {"LOWER": "guidance"},
            {"LOWER": "and"},
            {"LOWER": "guarantee"},
            {"LOWER": "fund"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "agricultural"},
            {"LOWER": "guidance"},
            {"LOWER": "and"},
            {"LOWER": "guarantee"},
            {"LOWER": "fund"},
            {"ORTH": "("},
            {"LOWER": "eaggf"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "european"}, {"LOWER": "arrest"}, {"LOWER": "warrant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "chemicals"},
            {"LOWER": "agency"},
            {"ORTH": "("},
            {"LOWER": "echa"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "commission"},
            {"LOWER": "’s"},
            {"LOWER": "margin"},
            {"LOWER": "of"},
            {"LOWER": "assessment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "european"}, {"LOWER": "community"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "european"}, {"LOWER": "community"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "european"}, {"LOWER": "council"}, {"LOWER": "decision"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "economic"},
            {"LOWER": "community"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "enforcement"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "enforcement"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "framework"},
            {"LOWER": "list"},
            {"LOWER": "offence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "payment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "european"}, {"LOWER": "patent"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "regional"},
            {"LOWER": "development"},
            {"LOWER": "fund"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "regional"},
            {"LOWER": "development"},
            {"LOWER": "fund"},
            {"ORTH": "("},
            {"LOWER": "erdf"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "european"}, {"LOWER": "schools"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "european"}, {"LOWER": "union"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "european"}, {"LOWER": "union"}, {"LOWER": "citizen"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "greenhouse"},
            {"LOWER": "gas"},
            {"LOWER": "emission"},
            {"LOWER": "allowance"},
            {"LOWER": "trading"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "european"}, {"LOWER": "union"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "trade"},
            {"LOWER": "mark"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "evading"},
            {"LOWER": "liability"},
            {"LOWER": "by"},
            {"LOWER": "deception"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "evaluation"},
            {"LOWER": "of"},
            {"LOWER": "seismic"},
            {"LOWER": "vulnerability"},
            {"LOWER": "of"},
            {"LOWER": "hospital"},
            {"LOWER": "structure"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hearsay"}, {"LOWER": "evidence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "evidence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "evidence"},
            {"LOWER": "in"},
            {"LOWER": "civil"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "evidence"},
            {"LOWER": "in"},
            {"LOWER": "criminal"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "evidence"}, {"LOWER": "of"}, {"LOWER": "alibi"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "evidence"}, {"LOWER": "of"}, {"LOWER": "character"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "evidence"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "evidence"},
            {"LOWER": "that"},
            {"LOWER": "detention"},
            {"LOWER": "conditions"},
            {"LOWER": "in"},
            {"LOWER": "issuing"},
            {"LOWER": "member"},
            {"LOWER": "states"},
            {"LOWER": "not"},
            {"LOWER": "satisfying"},
            {"LOWER": "fundamental"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "evidence"},
            {"ORTH": ","},
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "evpo"}, {"LOWER": "agnic"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ex"}, {"LOWER": "gratia"}, {"LOWER": "compensation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "ex"},
            {"LOWER": "gratia"},
            {"LOWER": "compensation"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ex"}, {"LOWER": "gratia"}, {"LOWER": "payment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ex"}, {"LOWER": "parte"}, {"LOWER": "application"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ex"}, {"LOWER": "parte"}, {"LOWER": "leave"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "ex"},
            {"LOWER": "parte"},
            {"LOWER": "wardship"},
            {"LOWER": "application"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "examination"}, {"LOWER": "in"}, {"LOWER": "private"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "examination"},
            {"LOWER": "of"},
            {"LOWER": "judgment"},
            {"LOWER": "debtor"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "examination"}, {"LOWER": "of"}, {"LOWER": "witness"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "examination"}, {"LOWER": "of"}, {"LOWER": "witnesses"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "exception"}, {"LOWER": "clause"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exception"}, {"LOWER": "of"}, {"LOWER": "barratry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exceptional"}, {"LOWER": "circumstances"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exceptions"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "the"},
            {"LOWER": "protection"},
            {"LOWER": "of"},
            {"LOWER": "investigations"},
            {"ORTH": ","},
            {"LOWER": "commercial"},
            {"LOWER": "interests"},
            {"LOWER": "and"},
            {"LOWER": "decision"},
            {"ORTH": "-"},
            {"LOWER": "making"},
            {"LOWER": "process"},
            {"LOWER": "of"},
            {"LOWER": "institutions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "excess"}, {"LOWER": "of"}, {"LOWER": "loss"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "excessive"}, {"LOWER": "dust"}, {"LOWER": "deposits"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exchange"}, {"LOWER": "before"}, {"LOWER": "trial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exchange"}, {"LOWER": "of"}, {"LOWER": "contracts"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "excise"}, {"LOWER": "duty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "excise"}, {"LOWER": "licence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "excluded"}, {"LOWER": "classes"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "excluded"}, {"LOWER": "classes"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "excluded"}, {"LOWER": "sectors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exclusion"}, {"LOWER": "clause"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusion"},
            {"LOWER": "from"},
            {"LOWER": "matrimonial"},
            {"LOWER": "home"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusion"},
            {"LOWER": "from"},
            {"LOWER": "patentability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusion"},
            {"LOWER": "from"},
            {"LOWER": "united"},
            {"LOWER": "kingdom"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exclusion"}, {"LOWER": "of"}, {"LOWER": "parent"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusion"},
            {"LOWER": "of"},
            {"LOWER": "possibility"},
            {"LOWER": "for"},
            {"LOWER": "community"},
            {"LOWER": "vessels"},
            {"LOWER": "to"},
            {"LOWER": "carry"},
            {"LOWER": "out"},
            {"LOWER": "fishing"},
            {"LOWER": "activities"},
            {"LOWER": "in"},
            {"LOWER": "moroccan"},
            {"LOWER": "fishing"},
            {"LOWER": "zones"},
            {"LOWER": "on"},
            {"LOWER": "basis"},
            {"LOWER": "of"},
            {"LOWER": "licence"},
            {"LOWER": "issued"},
            {"LOWER": "by"},
            {"LOWER": "moroccan"},
            {"LOWER": "authorities"},
            {"LOWER": "without"},
            {"LOWER": "intervention"},
            {"LOWER": "of"},
            {"LOWER": "competent"},
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "authorities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusion"},
            {"LOWER": "of"},
            {"LOWER": "reproductive"},
            {"LOWER": "cells"},
            {"ORTH": ","},
            {"LOWER": "foetal"},
            {"LOWER": "tissues"},
            {"LOWER": "and"},
            {"LOWER": "embryonic"},
            {"LOWER": "tissues"},
            {"LOWER": "from"},
            {"LOWER": "scope"},
            {"LOWER": "of"},
            {"LOWER": "national"},
            {"LOWER": "legislation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusion"},
            {"LOWER": "or"},
            {"LOWER": "restriction"},
            {"LOWER": "of"},
            {"LOWER": "beneficiaries"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "occupy"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "exclusion"}, {"LOWER": "requirement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exclusive"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exclusive"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusive"},
            {"LOWER": "jurisdiction"},
            {"LOWER": "clause"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "exclusive"}, {"LOWER": "licence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exclusive"}, {"LOWER": "occupation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exclusive"}, {"LOWER": "occupation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exclusive"},
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "burial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "excusing"}, {"LOWER": "from"}, {"LOWER": "service"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "execution"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "execution"}, {"LOWER": "of"}, {"LOWER": "judgment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "execution"}, {"LOWER": "of"}, {"LOWER": "mortgage"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "execution"},
            {"LOWER": "of"},
            {"LOWER": "payment"},
            {"LOWER": "instructions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "execution"},
            {"LOWER": "of"},
            {"LOWER": "scottish"},
            {"LOWER": "warrant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "execution"},
            {"LOWER": "of"},
            {"LOWER": "search"},
            {"LOWER": "warrant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "execution"},
            {"LOWER": "of"},
            {"LOWER": "statutory"},
            {"LOWER": "will"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "executioni"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "executor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "executor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "executor"},
            {"LOWER": "and"},
            {"LOWER": "administrator"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "executor"},
            {"LOWER": "and"},
            {"LOWER": "administrator"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "executor"},
            {"LOWER": "and"},
            {"LOWER": "administrator"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "executor"},
            {"LOWER": "de"},
            {"LOWER": "son"},
            {"LOWER": "tort"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "executory"}, {"LOWER": "trust"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exemplary"}, {"LOWER": "damages"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exempt"}, {"LOWER": "dwelling"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exempt"}, {"LOWER": "dwellings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exempt"}, {"LOWER": "information"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exempt"}, {"LOWER": "information"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exempt"},
            {"LOWER": "information"},
            {"LOWER": "in"},
            {"LOWER": "relation"},
            {"LOWER": "to"},
            {"LOWER": "inquiries"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exemption"},
            {"LOWER": "from"},
            {"LOWER": "obtaining"},
            {"LOWER": "licence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "exemptions"}, {"LOWER": "clause"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exercise"}, {"LOWER": "machine"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exercise"}, {"LOWER": "of"}, {"LOWER": "discretion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "exercise"}, {"LOWER": "of"}, {"LOWER": "power"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exercise"},
            {"LOWER": "of"},
            {"LOWER": "prerogative"},
            {"LOWER": "power"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exercise"},
            {"LOWER": "of"},
            {"LOWER": "prerogative"},
            {"LOWER": "power"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exercise"},
            {"LOWER": "of"},
            {"LOWER": "sovereign"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "exercise"},
            {"LOWER": "of"},
            {"LOWER": "statutory"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "exhaustion"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exhumation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "existence"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "existence"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "expenses"},
            {"LOWER": "of"},
            {"LOWER": "administration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "expenses"},
            {"LOWER": "of"},
            {"LOWER": "administration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "expenses"}, {"LOWER": "of"}, {"LOWER": "liquidation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "expenses"}, {"LOWER": "of"}, {"LOWER": "trade"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "expenses"},
            {"LOWER": "of"},
            {"LOWER": "trade"},
            {"LOWER": "or"},
            {"LOWER": "profession"},
            {"ORTH": "("},
            {"LOWER": "schedule"},
            {"LOWER": "d"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "expenses"},
            {"LOWER": "of"},
            {"LOWER": "trade"},
            {"LOWER": "or"},
            {"LOWER": "profession"},
            {"ORTH": "("},
            {"LOWER": "schedule"},
            {"LOWER": "d"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "experimental"},
            {"LOWER": "traffic"},
            {"LOWER": "regulation"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "experimental"}, {"LOWER": "treatment"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "expert"}, {"LOWER": "witness"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "expert"}, {"LOWER": "determination"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "expert"},
            {"LOWER": "determination"},
            {"LOWER": "clause"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "expert"}, {"LOWER": "evidence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "expert"}, {"LOWER": "’s"}, {"LOWER": "opinion"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "expiring"}, {"LOWER": "licence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "exploratory"}, {"LOWER": "drilling"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "explosive"}, {"LOWER": "substances"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "export"}, {"LOWER": "refunds"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "express"}, {"LOWER": "term"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "expression"}, {"LOWER": "of"}, {"LOWER": "interest"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "extempore"}, {"LOWER": "judgment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "extended"}, {"LOWER": "family"}, {"LOWER": "member"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "extended"}, {"LOWER": "sentence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "extended"}, {"LOWER": "sentences"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "extension"}, {"LOWER": "of"}, {"LOWER": "time"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extension"},
            {"LOWER": "of"},
            {"LOWER": "time"},
            {"LOWER": "for"},
            {"LOWER": "service"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extension"},
            {"LOWER": "of"},
            {"LOWER": "time"},
            {"LOWER": "limit"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "extension"}, {"LOWER": "to"}, {"LOWER": "church"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extent"},
            {"LOWER": "of"},
            {"LOWER": "guarantor"},
            {"LOWER": "’s"},
            {"LOWER": "liability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extent"},
            {"LOWER": "of"},
            {"LOWER": "guarantor"},
            {"LOWER": "’s"},
            {"LOWER": "obligation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "external"},
            {"LOWER": "confiscation"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "external"}, {"LOWER": "relations"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extra"},
            {"ORTH": "-"},
            {"LOWER": "hazardous"},
            {"LOWER": "acts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extra"},
            {"ORTH": "-"},
            {"LOWER": "territorial"},
            {"LOWER": "investigations"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "extraditable"}, {"LOWER": "crime"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "extradition"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "extradition"}, {"LOWER": "crime"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "extradition"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "extradition"}, {"LOWER": "offence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "extradition"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "extradition"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extradition"},
            {"LOWER": "request"},
            {"LOWER": "by"},
            {"LOWER": "us"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "extradition"}, {"LOWER": "to"}, {"LOWER": "lithuania"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extradition"},
            {"LOWER": "to"},
            {"LOWER": "third"},
            {"LOWER": "state"},
            {"LOWER": "of"},
            {"LOWER": "national"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "who"},
            {"LOWER": "exercised"},
            {"LOWER": "his"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "movement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "extradition"},
            {"LOWER": "with"},
            {"LOWER": "view"},
            {"LOWER": "to"},
            {"LOWER": "executing"},
            {"LOWER": "custodial"},
            {"LOWER": "sentences"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "extraterritorial"}, {"LOWER": "effect"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "extrinsic"}, {"LOWER": "evidence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "protection"}, {"LOWER": "of"}, {"LOWER": "eyes"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fac"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "facilitating"},
            {"LOWER": "retention"},
            {"LOWER": "of"},
            {"LOWER": "criminal"},
            {"LOWER": "proceeds"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "facility"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fact"}, {"LOWER": "and"}, {"LOWER": "opinion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fact"}, {"LOWER": "finding"}, {"LOWER": "hearing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fact"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
            {"LOWER": "to"},
            {"LOWER": "establish"},
            {"LOWER": "parentage"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fact"}, {"LOWER": "or"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
            {"LOWER": "in"},
            {"LOWER": "respect"},
            {"LOWER": "of"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "accidental"},
            {"LOWER": "injury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
            {"LOWER": "to"},
            {"LOWER": "establish"},
            {"LOWER": "whether"},
            {"LOWER": "injury"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "accidental"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "factors"}, {"LOWER": "for"}, {"LOWER": "consideration"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "factory"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "failed"}, {"LOWER": "asylum"}, {"LOWER": "seeker"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "failed"}, {"LOWER": "asylum"}, {"LOWER": "seekers"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failing"},
            {"LOWER": "to"},
            {"LOWER": "disclose"},
            {"LOWER": "change"},
            {"LOWER": "of"},
            {"LOWER": "circumstances"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failing"},
            {"LOWER": "to"},
            {"LOWER": "surrender"},
            {"LOWER": "to"},
            {"LOWER": "bail"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "to"},
            {"LOWER": "fulfil"},
            {"LOWER": "obligations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "to"},
            {"LOWER": "fulfil"},
            {"LOWER": "obligation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "to"},
            {"LOWER": "fulfil"},
            {"LOWER": "obligations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "of"},
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "to"},
            {"LOWER": "fulfil"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "trace"},
            {"LOWER": "family"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "account"},
            {"LOWER": "for"},
            {"LOWER": "loss"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "failure"}, {"LOWER": "to"}, {"LOWER": "comply"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "comply"},
            {"LOWER": "with"},
            {"LOWER": "order"},
            {"LOWER": "for"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "failure"}, {"LOWER": "to"}, {"LOWER": "consult"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "diagnose"},
            {"LOWER": "injury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "exercise"},
            {"LOWER": "reasonable"},
            {"LOWER": "care"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "fulfil"},
            {"LOWER": "obligations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "failure"}, {"LOWER": "to"}, {"LOWER": "indicate"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "immigration"},
            {"LOWER": "document"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "specimen"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "failure"}, {"LOWER": "to"}, {"LOWER": "surrender"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "failure"},
            {"LOWER": "to"},
            {"LOWER": "surrender"},
            {"LOWER": "to"},
            {"LOWER": "custody"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "comment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "dealing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "trading"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "trading"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "trading"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "trial"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fair"}, {"LOWER": "trial"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "faith"}, {"LOWER": "healing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "accounting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "declaration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "description"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "document"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "endorsement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "imprisonment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "imprisonment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "imprisonment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "instrument"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "false"}, {"LOWER": "price"}, {"LOWER": "indication"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "statement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "false"}, {"LOWER": "statements"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "benefit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "benefits"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "family"},
            {"LOWER": "benefits"},
            {"LOWER": "for"},
            {"LOWER": "orphans"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "division"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "home"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "home"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "homes"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "family"}, {"LOWER": "income"}, {"LOWER": "supplement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "life"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "member"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "members"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "family"},
            {"LOWER": "members"},
            {"LOWER": "disclaiming"},
            {"LOWER": "all"},
            {"LOWER": "knowledge"},
            {"LOWER": "of"},
            {"LOWER": "mother"},
            {"LOWER": "and"},
            {"LOWER": "child"},
            {"LOWER": "’s"},
            {"LOWER": "whereabouts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "family"}, {"LOWER": "of"}, {"LOWER": "refugee"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "family"},
            {"LOWER": "practitioner"},
            {"LOWER": "committee"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "family"}, {"LOWER": "proceedings"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "family"}, {"LOWER": "proceedings"}, {"LOWER": "court"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "provision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "provision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "family"}, {"LOWER": "provision"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fatal"}, {"LOWER": "accidents"}, {"LOWER": "act"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fatal"}, {"LOWER": "accidents"}, {"LOWER": "acts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fatal"}, {"LOWER": "accidents"}, {"LOWER": "acts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "appealing"},
            {"LOWER": "against"},
            {"LOWER": "order"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "contact"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "reconsideration"},
            {"LOWER": "of"},
            {"LOWER": "order"},
            {"LOWER": "to"},
            {"LOWER": "permit"},
            {"LOWER": "direct"},
            {"LOWER": "contact"},
            {"LOWER": "with"},
            {"LOWER": "child"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "residence"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "applying"},
            {"LOWER": "to"},
            {"LOWER": "discharge"},
            {"LOWER": "care"},
            {"LOWER": "orders"},
            {"LOWER": "and"},
            {"LOWER": "for"},
            {"LOWER": "contact"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "applying"},
            {"LOWER": "to"},
            {"LOWER": "take"},
            {"LOWER": "child"},
            {"LOWER": "abroad"},
            {"LOWER": "on"},
            {"LOWER": "holiday"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "consenting"},
            {"LOWER": "to"},
            {"LOWER": "temporary"},
            {"LOWER": "removal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "persistently"},
            {"LOWER": "failing"},
            {"LOWER": "to"},
            {"LOWER": "procure"},
            {"LOWER": "return"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "or"},
            {"LOWER": "disclose"},
            {"LOWER": "her"},
            {"LOWER": "whereabouts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "persistently"},
            {"LOWER": "seeking"},
            {"LOWER": "to"},
            {"LOWER": "challenge"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "seeking"},
            {"LOWER": "contact"},
            {"LOWER": "with"},
            {"LOWER": "daughter"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "seeking"},
            {"LOWER": "residence"},
            {"LOWER": "and"},
            {"LOWER": "contact"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "seeking"},
            {"LOWER": "stay"},
            {"LOWER": "of"},
            {"LOWER": "english"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "seeking"},
            {"LOWER": "summary"},
            {"LOWER": "return"},
            {"LOWER": "of"},
            {"LOWER": "children"},
            {"LOWER": "to"},
            {"LOWER": "hungary"},
            {"LOWER": "under"},
            {"LOWER": "1980"},
            {"LOWER": "hague"},
            {"LOWER": "convention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "seeking"},
            {"LOWER": "to"},
            {"LOWER": "enforce"},
            {"LOWER": "contact"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "father"},
            {"LOWER": "seeking"},
            {"LOWER": "to"},
            {"LOWER": "retain"},
            {"LOWER": "child"},
            {"LOWER": "in"},
            {"LOWER": "england"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fear"}, {"LOWER": "of"}, {"LOWER": "persecution"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "federal"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fee"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fees"}, {"LOWER": "and"}, {"LOWER": "remuneration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fees"},
            {"LOWER": "for"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "administrative"},
            {"LOWER": "proceedings"},
            {"LOWER": "in"},
            {"LOWER": "field"},
            {"LOWER": "of"},
            {"LOWER": "public"},
            {"LOWER": "procurement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "felling"},
            {"LOWER": "of"},
            {"LOWER": "beech"},
            {"LOWER": "trees"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "female"},
            {"LOWER": "genital"},
            {"LOWER": "mutilation"},
            {"LOWER": "protection"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "female"},
            {"LOWER": "genital"},
            {"LOWER": "mutilation"},
            {"LOWER": "protection"},
            {"LOWER": "orders"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fiduciary"}, {"LOWER": "duties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fiduciary"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fiduciary"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "company"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fiduciary"}, {"LOWER": "relationship"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fiduciary"},
            {"LOWER": "relationship"},
            {"LOWER": "with"},
            {"LOWER": "client"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "final"}, {"LOWER": "charging"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "final"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "final"}, {"LOWER": "salary"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "final"},
            {"LOWER": "warning"},
            {"LOWER": "following"},
            {"LOWER": "disciplinary"},
            {"LOWER": "hearing"},
            {"LOWER": "against"},
            {"LOWER": "employee"},
            {"LOWER": "teacher"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "finance"}, {"LOWER": "leasing"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "advantage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "assistance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "financial"},
            {"LOWER": "compensation"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "financial"},
            {"LOWER": "offences"},
            {"LOWER": "in"},
            {"LOWER": "italy"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "position"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "financial"},
            {"LOWER": "protection"},
            {"LOWER": "for"},
            {"LOWER": "respondent"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "provision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "provision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "provision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "provision"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "financial"},
            {"LOWER": "provision"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "financial"},
            {"LOWER": "provision"},
            {"LOWER": "agreements"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "provisions"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "relief"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "financial"},
            {"LOWER": "relief"},
            {"LOWER": "following"},
            {"LOWER": "divorce"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "financial"},
            {"LOWER": "relief"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "services"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "services"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "financial"}, {"LOWER": "services"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "financial"}, {"LOWER": "transactions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "finding"}, {"LOWER": "of"}, {"LOWER": "fact"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "finding"}, {"LOWER": "of"}, {"LOWER": "identity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "findings"}, {"LOWER": "of"}, {"LOWER": "fact"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fined"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fine"},
            {"LOWER": "for"},
            {"LOWER": "contempt"},
            {"LOWER": "of"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fines"},
            {"LOWER": "imposed"},
            {"LOWER": "by"},
            {"LOWER": "ec"},
            {"LOWER": "commission"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fingerprints"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fire"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fire"},
            {"LOWER": "at"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "premises"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fire"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fire"}, {"LOWER": "brigade"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fire"}, {"LOWER": "brigade"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fire"},
            {"LOWER": "certificate"},
            {"ORTH": ","},
            {"LOWER": "conditions"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fire"}, {"LOWER": "damage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fire"}, {"LOWER": "escape"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fire"}, {"LOWER": "precautions"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fire"}, {"LOWER": "safety"}, {"LOWER": "offences"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "firearms"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "firefighters"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fireman"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"LOWER": "defendant"},
            {"LOWER": "allegedly"},
            {"LOWER": "perpetrating"},
            {"LOWER": "fraud"},
            {"LOWER": "in"},
            {"LOWER": "concert"},
            {"LOWER": "with"},
            {"LOWER": "other"},
            {"LOWER": "defendants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"LOWER": "defendant"},
            {"LOWER": "manager"},
            {"LOWER": "employed"},
            {"LOWER": "by"},
            {"LOWER": "claimant"},
            {"LOWER": "company"},
            {"LOWER": "resigning"},
            {"LOWER": "and"},
            {"LOWER": "employed"},
            {"LOWER": "by"},
            {"LOWER": "second"},
            {"LOWER": "defendant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"LOWER": "trial"},
            {"LOWER": "jury"},
            {"LOWER": "discharged"},
            {"LOWER": "at"},
            {"LOWER": "end"},
            {"LOWER": "of"},
            {"LOWER": "prosecution"},
            {"LOWER": "case"},
            {"LOWER": "owing"},
            {"LOWER": "to"},
            {"LOWER": "work"},
            {"LOWER": "commitments"},
            {"LOWER": "of"},
            {"LOWER": "three"},
            {"LOWER": "jurors"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"ORTH": "-"},
            {"LOWER": "tier"},
            {"LOWER": "and"},
            {"LOWER": "upper"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"ORTH": "-"},
            {"LOWER": "tier"},
            {"LOWER": "and"},
            {"LOWER": "upper"},
            {"LOWER": "tribunals"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"ORTH": "-"},
            {"LOWER": "tier"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"ORTH": "-"},
            {"LOWER": "tier"},
            {"LOWER": "tribunal"},
            {"ORTH": "("},
            {"LOWER": "immigration"},
            {"LOWER": "and"},
            {"LOWER": "asylum"},
            {"LOWER": "chamber"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "first"},
            {"ORTH": "-"},
            {"LOWER": "tier"},
            {"LOWER": "tribunal"},
            {"LOWER": "and"},
            {"LOWER": "upper"},
            {"LOWER": "tribunals"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fisheries"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fisheries"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fishery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fishery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fishing"}, {"LOWER": "quota"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fishing"}, {"LOWER": "quotas"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fishing"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fitness"}, {"LOWER": "for"}, {"LOWER": "purpose"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fitness"}, {"LOWER": "to"}, {"LOWER": "plead"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fitness"}, {"LOWER": "to"}, {"LOWER": "practice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fitness"}, {"LOWER": "to"}, {"LOWER": "practise"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fitness"},
            {"LOWER": "to"},
            {"LOWER": "practise"},
            {"LOWER": "committee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fitness"},
            {"LOWER": "to"},
            {"LOWER": "practise"},
            {"LOWER": "panel"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fitness"},
            {"LOWER": "to"},
            {"LOWER": "practise"},
            {"LOWER": "panel"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fixed"}, {"LOWER": "costs"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fixed"}, {"LOWER": "dates"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fixed"},
            {"LOWER": "or"},
            {"LOWER": "floating"},
            {"LOWER": "charge"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fixed"}, {"LOWER": "penalty"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fixed"}, {"LOWER": "penalty"}, {"LOWER": "notice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fixed"}, {"LOWER": "recoverable"}, {"LOWER": "costs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "fixed"}, {"LOWER": "term"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fixed"},
            {"ORTH": "-"},
            {"LOWER": "term"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fixed"},
            {"ORTH": "-"},
            {"LOWER": "term"},
            {"LOWER": "contracts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fixed"},
            {"ORTH": "-"},
            {"LOWER": "term"},
            {"LOWER": "employee"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fixtures"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "flawed"}, {"LOWER": "finding"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "floating"}, {"LOWER": "charge"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "floating"},
            {"LOWER": "charge"},
            {"LOWER": "on"},
            {"LOWER": "assets"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "floating"}, {"LOWER": "vessel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "flood"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "flying"}, {"LOWER": "freehold"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "follow"}, {"LOWER": "settlement"}, {"LOWER": "clause"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "food"}, {"LOWER": "and"}, {"LOWER": "drugs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "food"}, {"LOWER": "and"}, {"LOWER": "drugs"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "food"}, {"LOWER": "safety"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "football"}, {"LOWER": "banning"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "force"}, {"LOWER": "majeure"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "force"}, {"LOWER": "majeure"}, {"LOWER": "clause"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "forced"},
            {"LOWER": "marriage"},
            {"LOWER": "protection"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "forced"},
            {"LOWER": "marriage"},
            {"LOWER": "protection"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "forced"},
            {"LOWER": "or"},
            {"LOWER": "compulsory"},
            {"LOWER": "labour"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "forcible"}, {"LOWER": "entry"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreclosure"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "bank"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "claimant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "corporation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "criminal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "currency"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "foreign"}, {"LOWER": "currency"}, {"LOWER": "debt"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "debt"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "decree"},
            {"ORTH": ","},
            {"LOWER": "recognition"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "foreign"}, {"LOWER": "domestic"}, {"LOWER": "helpers"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "exchange"},
            {"LOWER": "transaction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "father"},
            {"LOWER": "requiring"},
            {"LOWER": "translation"},
            {"LOWER": "of"},
            {"LOWER": "documents"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "judgment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "jurisdiction"},
            {"LOWER": "clause"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "jurisdiction"},
            {"LOWER": "clause"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "plaintiff"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "proceedings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "spouse"},
            {"LOWER": "or"},
            {"LOWER": "family"},
            {"LOWER": "member"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "spouse"},
            {"LOWER": "or"},
            {"LOWER": "partner"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreign"}, {"LOWER": "taxes"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreign"},
            {"LOWER": "tribunal"},
            {"ORTH": ","},
            {"LOWER": "for"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "foreseeability"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "foreseeability"},
            {"LOWER": "of"},
            {"LOWER": "consequential"},
            {"LOWER": "injury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "foreseeability"}, {"LOWER": "of"}, {"LOWER": "injury"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "foreseeability"}, {"LOWER": "of"}, {"LOWER": "risk"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "forfeiture"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "forfeiture"},
            {"LOWER": "act"},
            {"LOWER": "1982"},
            {"ORTH": "("},
            {"LOWER": "c"},
            {"LOWER": "34"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "s"},
            {"LOWER": "2"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "forfeiture"},
            {"LOWER": "and"},
            {"LOWER": "destruction"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "forfeiture"},
            {"LOWER": "of"},
            {"LOWER": "impermissible"},
            {"LOWER": "donation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "forfeiture"}, {"LOWER": "of"}, {"LOWER": "lease"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "forfeiture"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "forged"}, {"LOWER": "cheque"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "forgery"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "form"}, {"LOWER": "of"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "formal"}, {"LOWER": "investigation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "formation"}, {"LOWER": "of"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "formation"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "contract"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "former"}, {"LOWER": "employee"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "former"},
            {"LOWER": "head"},
            {"LOWER": "of"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "former"}, {"LOWER": "relevant"}, {"LOWER": "child"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "forseeability"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "forseeability"}, {"LOWER": "of"}, {"LOWER": "risk"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "forum"}, {"LOWER": "bar"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "forum"}, {"LOWER": "conveniens"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "forum"}, {"LOWER": "non"}, {"LOWER": "conveniens"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "forum"}, {"LOWER": "non"}, {"LOWER": "conveniens"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "forward"},
            {"LOWER": "freight"},
            {"LOWER": "agreements"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "foster"}, {"LOWER": "parent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "foul"}, {"LOWER": "water"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fraud"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fraud"},
            {"LOWER": "by"},
            {"LOWER": "abuse"},
            {"LOWER": "of"},
            {"LOWER": "position"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fraudulent"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fraudulent"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "trust"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fraudulent"}, {"LOWER": "conveyance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fraudulent"}, {"LOWER": "evasion"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fraudulent"},
            {"LOWER": "evasion"},
            {"LOWER": "of"},
            {"LOWER": "prohibition"},
            {"LOWER": "on"},
            {"LOWER": "importation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fraudulent"},
            {"LOWER": "evasion"},
            {"LOWER": "of"},
            {"LOWER": "restriction"},
            {"LOWER": "on"},
            {"LOWER": "importation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fraudulent"}, {"LOWER": "trading"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "free"}, {"LOWER": "competition"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "free"}, {"LOWER": "competition"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "free"}, {"LOWER": "dompetition"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "free"}, {"LOWER": "elections"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "free"}, {"LOWER": "goddess"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "free"}, {"LOWER": "movement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "free"},
            {"LOWER": "movement"},
            {"LOWER": "of"},
            {"LOWER": "capital"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "free"},
            {"LOWER": "movement"},
            {"LOWER": "of"},
            {"LOWER": "workers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freedom"},
            {"LOWER": "from"},
            {"LOWER": "discrimination"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "assembly"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "establishment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "expression"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "expression"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "information"},
            {"LOWER": "request"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "information"},
            {"LOWER": "requests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "movement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "movement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "movement"},
            {"LOWER": "for"},
            {"LOWER": "persons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "movement"},
            {"LOWER": "for"},
            {"LOWER": "workers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "press"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freedom"}, {"LOWER": "of"}, {"LOWER": "speech"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freedom"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freedom"},
            {"ORTH": ","},
            {"LOWER": "security"},
            {"LOWER": "and"},
            {"LOWER": "justice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freeing"},
            {"LOWER": "child"},
            {"LOWER": "for"},
            {"LOWER": "adoption"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "freezing"},
            {"LOWER": "and"},
            {"LOWER": "search"},
            {"LOWER": "orders"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "freezing"}, {"LOWER": "injunction"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "freezing"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "freezing"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fresh"}, {"LOWER": "claim"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fresh"}, {"LOWER": "evidence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fresh"}, {"LOWER": "summons"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "frozen"},
            {"LOWER": "semen"},
            {"LOWER": "in"},
            {"LOWER": "storage"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "frustration"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fugitive"},
            {"LOWER": "form"},
            {"LOWER": "us"},
            {"LOWER": "accused"},
            {"LOWER": "of"},
            {"LOWER": "parental"},
            {"LOWER": "kidnapping"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fugitive"}, {"LOWER": "offender"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "functi"}, {"LOWER": "officio"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "functions"},
            {"LOWER": "of"},
            {"LOWER": "administrator"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "functions"}, {"LOWER": "of"}, {"LOWER": "adminstrator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "functions"},
            {"LOWER": "of"},
            {"LOWER": "public"},
            {"LOWER": "nature"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "fund"}, {"LOWER": "management"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fundamental"}, {"LOWER": "freedoms"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "fundamental"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "fundamental"},
            {"LOWER": "rights"},
            {"LOWER": "and"},
            {"LOWER": "freedoms"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "furnished"}, {"LOWER": "accommodation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "further"}, {"LOWER": "costs"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "further"}, {"LOWER": "evidence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "further"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "future"}, {"LOWER": "care"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "future"}, {"LOWER": "earnings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "future"},
            {"LOWER": "loss"},
            {"LOWER": "and"},
            {"LOWER": "expenses"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "gambia"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gambling"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "game"}, {"LOWER": "of"}, {"LOWER": "chance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "games"}, {"LOWER": "of"}, {"LOWER": "chance"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "gaming"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gaming"}, {"LOWER": "contract"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gaming"}, {"LOWER": "duty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gaming"}, {"LOWER": "machines"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gaming"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "gang"},
            {"ORTH": "-"},
            {"LOWER": "related"},
            {"LOWER": "violence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "gangmaster"},
            {"LOWER": "without"},
            {"LOWER": "licence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "garage"},
            {"LOWER": "converted"},
            {"LOWER": "into"},
            {"LOWER": "annex"},
            {"LOWER": "to"},
            {"LOWER": "main"},
            {"LOWER": "house"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "garnishee"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gender"}, {"LOWER": "recognition"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "gender"},
            {"LOWER": "recognition"},
            {"LOWER": "certificate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "general"},
            {"LOWER": "and"},
            {"LOWER": "specific"},
            {"LOWER": "powers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "general"},
            {"LOWER": "arrangements"},
            {"LOWER": "for"},
            {"LOWER": "products"},
            {"LOWER": "subject"},
            {"LOWER": "to"},
            {"LOWER": "excise"},
            {"LOWER": "duty"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "general"}, {"LOWER": "average"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "general"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "general"},
            {"LOWER": "court"},
            {"LOWER": "decision"},
            {"LOWER": "annulling"},
            {"LOWER": "commission"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "general"},
            {"LOWER": "permitted"},
            {"LOWER": "development"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "general"}, {"LOWER": "secretary"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "general"}, {"LOWER": "service"}, {"LOWER": "charge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "general"},
            {"LOWER": "words"},
            {"LOWER": "in"},
            {"LOWER": "preamble"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "genetic"},
            {"LOWER": "father"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "contact"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "genuine"},
            {"LOWER": "occupational"},
            {"LOWER": "qualification"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "genuine"}, {"LOWER": "use"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "geographical"}, {"LOWER": "eligibility"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "geographical"}, {"LOWER": "names"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "get"}, {"ORTH": "-"}, {"LOWER": "up"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gibraltar"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gift"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "gift"}, {"LOWER": "inter"}, {"LOWER": "vivos"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "gift"},
            {"LOWER": "of"},
            {"LOWER": "residue"},
            {"LOWER": "unto"},
            {"LOWER": "and"},
            {"LOWER": "equally"},
            {"LOWER": "amongst"},
            {"LOWER": "all"},
            {"LOWER": "the"},
            {"LOWER": "children"},
            {"LOWER": "of"},
            {"LOWER": "j."},
            {"LOWER": "d."},
            {"LOWER": "and"},
            {"LOWER": "r."},
            {"LOWER": "a."},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "vested"},
            {"LOWER": "legacies"},
            {"LOWER": "at"},
            {"LOWER": "testators"},
            {"LOWER": "death"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "gifts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gipsies"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "gladys"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "good"}, {"LOWER": "character"}, {"LOWER": "direction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "good"}, {"LOWER": "character"}, {"LOWER": "directions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "goods"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "goods"}, {"LOWER": "vehicle"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "goring"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "governance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "governing"}, {"LOWER": "body"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "government"}, {"LOWER": "minister"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "grant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "grant"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grant"},
            {"LOWER": "ad"},
            {"LOWER": "colligenda"},
            {"LOWER": "bona"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "grant"}, {"LOWER": "of"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grant"},
            {"LOWER": "of"},
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "building"},
            {"LOWER": "associated"},
            {"LOWER": "with"},
            {"LOWER": "animal"},
            {"LOWER": "husbandry"},
            {"LOWER": "in"},
            {"LOWER": "rural"},
            {"LOWER": "domestic"},
            {"LOWER": "setting"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grant"},
            {"LOWER": "of"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "wind"},
            {"LOWER": "farm"},
            {"LOWER": "in"},
            {"LOWER": "cornwall"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grant"},
            {"LOWER": "of"},
            {"LOWER": "recreational"},
            {"LOWER": "boating"},
            {"LOWER": "licence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grantor"},
            {"LOWER": "without"},
            {"LOWER": "proprietary"},
            {"LOWER": "interest"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "grants"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "gratuity"}, {"LOWER": "on"}, {"LOWER": "discharge"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "grazing"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "great"}, {"LOWER": "peace"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "green"}, {"LOWER": "belt"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "green"}, {"LOWER": "belt"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "green"}, {"LOWER": "belt"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "greenhouse"},
            {"LOWER": "gas"},
            {"LOWER": "emission"},
            {"LOWER": "allowance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "greenhouse"}, {"LOWER": "gas"}, {"LOWER": "emissions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "gregos"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "grenada"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "grievance"}, {"LOWER": "procedure"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "grievance"}, {"LOWER": "procedures"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "grievous"}, {"LOWER": "bodily"}, {"LOWER": "harm"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "gross"},
            {"LOWER": "negligence"},
            {"LOWER": "manslaughter"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "grossly"}, {"LOWER": "offensive"}, {"LOWER": "message"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ground"}, {"LOWER": "for"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "groundless"},
            {"LOWER": "threats"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "grounds"}, {"LOWER": "for"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "grounds"}, {"LOWER": "for"}, {"LOWER": "refusal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grounds"},
            {"LOWER": "for"},
            {"LOWER": "refusal"},
            {"LOWER": "and"},
            {"LOWER": "for"},
            {"LOWER": "declaration"},
            {"LOWER": "of"},
            {"LOWER": "invalidity"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grounds"},
            {"LOWER": "for"},
            {"LOWER": "refusal"},
            {"LOWER": "or"},
            {"LOWER": "invalidation"},
            {"LOWER": "of"},
            {"LOWER": "registration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "grounds"},
            {"LOWER": "for"},
            {"LOWER": "refusing"},
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
            {"LOWER": "judgment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "group"}, {"LOWER": "litigation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "group"}, {"LOWER": "of"}, {"LOWER": "companies"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "group"}, {"LOWER": "relief"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "guarantee"},
            {"LOWER": "in"},
            {"LOWER": "lease"},
            {"LOWER": "under"},
            {"LOWER": "seal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "guarantee"}, {"LOWER": "payment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "guardian"}, {"LOWER": "ad"}, {"LOWER": "litem"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "guardian"},
            {"LOWER": "ad"},
            {"LOWER": "litems"},
            {"LOWER": "report"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "guardians"}, {"LOWER": "allowance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "guernsey"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "guidance"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "levels"},
            {"LOWER": "of"},
            {"LOWER": "fines"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "guidance"},
            {"LOWER": "on"},
            {"LOWER": "proper"},
            {"LOWER": "approach"},
            {"LOWER": "to"},
            {"LOWER": "sentencing"},
            {"LOWER": "where"},
            {"LOWER": "no"},
            {"LOWER": "applicable"},
            {"LOWER": "sentencing"},
            {"LOWER": "guidelines"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "guide"},
            {"LOWER": "to"},
            {"LOWER": "commercial"},
            {"LOWER": "court"},
            {"LOWER": "practice"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "habeas"}, {"LOWER": "corpus"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "habeas"}, {"LOWER": "corpus"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "habeas"}, {"LOWER": "corpus"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "habeas"}, {"LOWER": "corpus"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "habitual"}, {"LOWER": "residence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hackney"}, {"LOWER": "carriage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "halki"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "handling"}, {"LOWER": "stolen"}, {"LOWER": "goods"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "handwriting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hansard"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "harassment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "harassment"}, {"LOWER": "of"}, {"LOWER": "employee"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "harbour"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "having"},
            {"LOWER": "firearm"},
            {"LOWER": "with"},
            {"LOWER": "him"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hawkers"}, {"LOWER": "and"}, {"LOWER": "pedlars"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hazardous"}, {"LOWER": "substances"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "head"}, {"LOWER": "of"}, {"LOWER": "state"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "headstone"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "health"}, {"LOWER": "and"}, {"LOWER": "safety"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "health"}, {"LOWER": "and"}, {"LOWER": "safety"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "health"}, {"LOWER": "and"}, {"LOWER": "safety"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "health"},
            {"LOWER": "and"},
            {"LOWER": "safety"},
            {"LOWER": "offences"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "health"},
            {"LOWER": "and"},
            {"LOWER": "social"},
            {"LOWER": "care"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "health"},
            {"LOWER": "and"},
            {"LOWER": "social"},
            {"LOWER": "care"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "health"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "health"}, {"LOWER": "insurance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "health"},
            {"LOWER": "service"},
            {"LOWER": "commissioner"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "healthcare"}, {"LOWER": "services"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hearing"}, {"LOWER": "adjourned"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "hearing"},
            {"LOWER": "conducted"},
            {"LOWER": "in"},
            {"LOWER": "absence"},
            {"LOWER": "of"},
            {"LOWER": "accused"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hearing"}, {"LOWER": "dates"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hearing"}, {"LOWER": "in"}, {"LOWER": "absentia"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hearing"}, {"LOWER": "in"}, {"LOWER": "camera"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hearing"}, {"LOWER": "in"}, {"LOWER": "private"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hearing"}, {"LOWER": "outside"}, {"LOWER": "london"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hearings"}, {"LOWER": "in"}, {"LOWER": "london"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hearings"}, {"LOWER": "outside"}, {"LOWER": "london"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hearsay"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hearsay"}, {"LOWER": "evidence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "heating"}, {"LOWER": "fuels"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "heavy"}, {"LOWER": "load"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hedging"}, {"LOWER": "products"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hereditary"}, {"LOWER": "peers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "high"},
            {"LOWER": "court"},
            {"LOWER": "decision"},
            {"LOWER": "“"},
            {"LOWER": "final"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "high"}, {"LOWER": "court"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "high"}, {"LOWER": "court"}, {"LOWER": "jurisdiction"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "higher"}, {"LOWER": "education"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "higher"},
            {"LOWER": "education"},
            {"LOWER": "corporation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "higher"},
            {"LOWER": "education"},
            {"LOWER": "funding"},
            {"LOWER": "council"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "highly"},
            {"LOWER": "trusted"},
            {"LOWER": "sponsor"},
            {"ORTH": "("},
            {"ORTH": "“"},
            {"LOWER": "hts"},
            {"ORTH": "”"},
            {"ORTH": ")"},
            {"LOWER": "status"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "highly"},
            {"LOWER": "trusted"},
            {"LOWER": "sponsor"},
            {"LOWER": "status"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "highway"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "highway"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "highway"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "himalaya"}, {"LOWER": "clause"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hire"}, {"LOWER": "charges"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hire"}, {"LOWER": "of"}, {"LOWER": "equipment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "hire"}, {"ORTH": "-"}, {"LOWER": "purchase"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "hire"},
            {"ORTH": "-"},
            {"LOWER": "purchase"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hoegh"}, {"LOWER": "apapa"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "holding"}, {"LOWER": "out"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "holding"},
            {"LOWER": "person"},
            {"LOWER": "in"},
            {"LOWER": "servitude"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "holdover"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hollandia"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "home"},
            {"LOWER": "secretary"},
            {"LOWER": "ordering"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "removal"},
            {"LOWER": "to"},
            {"LOWER": "kosovo"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "home"},
            {"LOWER": "secretary"},
            {"LOWER": "ordering"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "removal"},
            {"LOWER": "to"},
            {"LOWER": "pakistan"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "home"},
            {"LOWER": "secretary"},
            {"LOWER": "refusing"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "homeless"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "homeless"}, {"LOWER": "persons"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "homeless"}, {"LOWER": "persons"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "homelessness"}, {"LOWER": "appeals"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "homicide"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "homicide"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "homosexuality"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "honest"}, {"LOWER": "concurrent"}, {"LOWER": "use"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hong"}, {"LOWER": "kong"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "horse"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "horseplay"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "horserace"}, {"LOWER": "betting"}, {"LOWER": "levy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "horserace"}, {"LOWER": "betting"}, {"LOWER": "levy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hospital"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "hospital"},
            {"LOWER": "and"},
            {"LOWER": "medical"},
            {"LOWER": "care"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "hospital"}, {"LOWER": "direction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hospital"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "hotchpot"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "house"}, {"LOWER": "of"}, {"LOWER": "commons"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "house"}, {"LOWER": "of"}, {"LOWER": "lords"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "house"}, {"LOWER": "of"}, {"LOWER": "lords"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "house"},
            {"LOWER": "of"},
            {"LOWER": "lords"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "house"},
            {"LOWER": "of"},
            {"LOWER": "lords"},
            {"LOWER": "decisions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "house"}, {"LOWER": "purchase"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "house"},
            {"LOWER": "reasonably"},
            {"LOWER": "so"},
            {"LOWER": "called"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "house"},
            {"LOWER": "to"},
            {"LOWER": "house"},
            {"LOWER": "collections"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "householders"}, {"LOWER": "policy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "housing"}, {"LOWER": "act"}, {"LOWER": "1980"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "housing"},
            {"LOWER": "aid"},
            {"LOWER": "granted"},
            {"LOWER": "prior"},
            {"LOWER": "to"},
            {"LOWER": "accession"},
            {"LOWER": "of"},
            {"LOWER": "hungary"},
            {"LOWER": "to"},
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "to"},
            {"LOWER": "certain"},
            {"LOWER": "categories"},
            {"LOWER": "of"},
            {"LOWER": "household"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "housing"}, {"LOWER": "allocation"}, {"LOWER": "policy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "housing"}, {"LOWER": "allocation"}, {"LOWER": "scheme"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}, {"LOWER": "assistance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}, {"LOWER": "benefit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}, {"LOWER": "benefit"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "housing"},
            {"LOWER": "benefit"},
            {"LOWER": "regulations"},
            {"LOWER": "2006"},
            {"ORTH": "("},
            {"LOWER": "si"},
            {"LOWER": "2006/213"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "regs"},
            {"LOWER": "3(1"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"ORTH": "7(1)(2"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "74(7)(a"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}, {"LOWER": "element"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}, {"LOWER": "functions"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "housing"}, {"LOWER": "management"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "how"}, {"LOWER": "far"}, {"LOWER": "binding"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "how"}, {"LOWER": "far"}, {"LOWER": "binding"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "body"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "fertilisation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "human"},
            {"LOWER": "rights"},
            {"LOWER": "and"},
            {"LOWER": "fundamentad"},
            {"LOWER": "freedoms"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "human"},
            {"LOWER": "rights"},
            {"LOWER": "and"},
            {"LOWER": "fundamental"},
            {"LOWER": "freedoms"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "human"}, {"LOWER": "trafficking"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "humanitarian"}, {"LOWER": "protection"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "humber"}, {"LOWER": "bridge"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "husband"}, {"LOWER": "and"}, {"LOWER": "wife"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "husband"}, {"LOWER": "and"}, {"LOWER": "wife"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "husband"}, {"LOWER": "and"}, {"LOWER": "wife"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "husband"}, {"LOWER": "and"}, {"LOWER": "wife"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "husband"},
            {"LOWER": "appealing"},
            {"LOWER": "against"},
            {"LOWER": "strike"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "vary"},
            {"LOWER": "ancillary"},
            {"LOWER": "relief"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "husband"},
            {"LOWER": "persistently"},
            {"LOWER": "refusing"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "required"},
            {"LOWER": "disclosure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "husband"},
            {"LOWER": "seeking"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "incapacitated"},
            {"LOWER": "wife"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ibaraki"}, {"LOWER": "maru"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "icy"}, {"LOWER": "road"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "identification"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "identification"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "identification"},
            {"LOWER": "of"},
            {"LOWER": "list"},
            {"LOWER": "by"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "identification"}, {"LOWER": "parade"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "identity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "identity"}, {"LOWER": "document"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "identity"}, {"LOWER": "documents"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "identity"},
            {"LOWER": "passport"},
            {"LOWER": "service"},
            {"LOWER": "retaining"},
            {"LOWER": "passports"},
            {"LOWER": "of"},
            {"LOWER": "first"},
            {"ORTH": ","},
            {"LOWER": "second"},
            {"LOWER": "and"},
            {"LOWER": "third"},
            {"LOWER": "claimant"},
            {"LOWER": "and"},
            {"LOWER": "refusing"},
            {"LOWER": "to"},
            {"LOWER": "issue"},
            {"LOWER": "fourth"},
            {"LOWER": "with"},
            {"LOWER": "british"},
            {"LOWER": "passport"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ikarian"}, {"LOWER": "reefer"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "ill"},
            {"ORTH": "-"},
            {"LOWER": "health"},
            {"LOWER": "pension"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "illegal"}, {"LOWER": "contract"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "illegal"}, {"LOWER": "display"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "illegal"}, {"LOWER": "entrant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "illegal"}, {"LOWER": "immigrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "illegal"},
            {"LOWER": "immigrant"},
            {"LOWER": "convicted"},
            {"LOWER": "of"},
            {"LOWER": "crimes"},
            {"LOWER": "of"},
            {"LOWER": "dishonesty"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "illegal"}, {"LOWER": "practice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "illegal"}, {"LOWER": "practices"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "illegal"},
            {"ORTH": ","},
            {"LOWER": "unreported"},
            {"LOWER": "and"},
            {"LOWER": "unregulated"},
            {"LOWER": "fishing"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "illegality"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "illegitimate"},
            {"LOWER": "child"},
            {"LOWER": "born"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "illicit"},
            {"LOWER": "radio"},
            {"LOWER": "transmissions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "image"},
            {"LOWER": "of"},
            {"LOWER": "famous"},
            {"LOWER": "person"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "immigrants"},
            {"LOWER": "detained"},
            {"LOWER": "not"},
            {"LOWER": "withstanding"},
            {"LOWER": "their"},
            {"LOWER": "assertion"},
            {"LOWER": "of"},
            {"LOWER": "having"},
            {"LOWER": "been"},
            {"LOWER": "tortured"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "immigration"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "immigration"},
            {"LOWER": "appeal"},
            {"LOWER": "tribunal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "immigration"}, {"LOWER": "cases"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "immigration"}, {"LOWER": "rules"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "immigration"}, {"LOWER": "rules"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "immigration"},
            {"LOWER": "services"},
            {"LOWER": "commissioner"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "immovable"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "immoveable"}, {"LOWER": "property"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "immunity"}, {"LOWER": "from"}, {"LOWER": "suit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "impact"},
            {"LOWER": "of"},
            {"LOWER": "death"},
            {"LOWER": "on"},
            {"LOWER": "level"},
            {"LOWER": "of"},
            {"LOWER": "sentence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "impact"},
            {"LOWER": "of"},
            {"LOWER": "relatively"},
            {"LOWER": "poor"},
            {"LOWER": "profitability"},
            {"LOWER": "where"},
            {"LOWER": "substantial"},
            {"LOWER": "turnover"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "impairment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "impartiality"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "impecuniosity"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "impeding"},
            {"LOWER": "course"},
            {"LOWER": "of"},
            {"LOWER": "justice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "imperfect"}, {"LOWER": "trust"}, {"LOWER": "provision"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "impermissible"}, {"LOWER": "donations"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "implied"}, {"LOWER": "term"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "implied"}, {"LOWER": "terms"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "implied"}, {"LOWER": "trust"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "implied"}, {"LOWER": "trust"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "import"}, {"LOWER": "duties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "import"}, {"LOWER": "licence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "importation"},
            {"LOWER": "of"},
            {"LOWER": "dutiable"},
            {"LOWER": "goods"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "importation"},
            {"LOWER": "of"},
            {"LOWER": "footwear"},
            {"LOWER": "originating"},
            {"LOWER": "outside"},
            {"LOWER": "eu"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "imports"},
            {"LOWER": "of"},
            {"LOWER": "integrated"},
            {"LOWER": "electronic"},
            {"LOWER": "compact"},
            {"LOWER": "fluorescent"},
            {"LOWER": "lamps"},
            {"ORTH": "("},
            {"LOWER": "cfl"},
            {"ORTH": "-"},
            {"LOWER": "i"},
            {"ORTH": ")"},
            {"LOWER": "originating"},
            {"LOWER": "in"},
            {"LOWER": "china"},
            {"ORTH": ","},
            {"LOWER": "vietnam"},
            {"ORTH": ","},
            {"LOWER": "pakistan"},
            {"LOWER": "and"},
            {"LOWER": "the"},
            {"LOWER": "philippines"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "imposition"}, {"LOWER": "of"}, {"LOWER": "limit"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "imprisonment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "imprisonment"},
            {"LOWER": "for"},
            {"LOWER": "public"},
            {"LOWER": "protection"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "imprisonment"},
            {"LOWER": "or"},
            {"LOWER": "detention"},
            {"LOWER": "for"},
            {"LOWER": "public"},
            {"LOWER": "protection"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "improper"}, {"LOWER": "performance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "improper"},
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "personal"},
            {"LOWER": "data"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "improvement"}, {"LOWER": "grant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "improvement"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "in"}, {"LOWER": "transit"}, {"LOWER": "loss"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "in"},
            {"ORTH": "-"},
            {"LOWER": "transit"},
            {"LOWER": "loss"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inadvertently"},
            {"LOWER": "disclosed"},
            {"LOWER": "documents"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "incapable"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "incapable"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "incapacitated"}, {"LOWER": "adult"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "incapacitated"}, {"LOWER": "adults"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "incapacitated"}, {"LOWER": "child"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "incapacitated"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "incapacity"}, {"LOWER": "pension"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "incapacity"},
            {"LOWER": "to"},
            {"LOWER": "consent"},
            {"LOWER": "to"},
            {"LOWER": "proposed"},
            {"LOWER": "treatment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "incidental"}, {"LOWER": "benefit"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "incidents"}, {"LOWER": "of"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inciting"},
            {"LOWER": "child"},
            {"LOWER": "to"},
            {"LOWER": "engage"},
            {"LOWER": "in"},
            {"LOWER": "sexual"},
            {"LOWER": "activity"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "income"}, {"LOWER": "fax"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "income"}, {"LOWER": "from"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "income"}, {"LOWER": "payments"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "income"}, {"LOWER": "support"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "income"}, {"LOWER": "support"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "income"}, {"LOWER": "tax"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "income"}, {"LOWER": "tax"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "income"}, {"LOWER": "tax"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "income"},
            {"LOWER": "tax"},
            {"LOWER": "on"},
            {"LOWER": "self"},
            {"ORTH": "-"},
            {"LOWER": "employed"},
            {"LOWER": "persons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "entitlement"}, {"LOWER": "to"}, {"LOWER": "income"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "inconsistent"}, {"LOWER": "verdicts"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "incorporation"},
            {"LOWER": "of"},
            {"LOWER": "enabling"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "incorporation"}, {"LOWER": "of"}, {"LOWER": "terms"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "incorrect"},
            {"LOWER": "advice"},
            {"LOWER": "or"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "increase"}, {"LOWER": "of"}, {"LOWER": "remuneration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "increasing"},
            {"LOWER": "extent"},
            {"LOWER": "of"},
            {"LOWER": "insolvency"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "increasing"},
            {"LOWER": "extent"},
            {"LOWER": "of"},
            {"LOWER": "insolvency"},
            {"LOWER": "by"},
            {"LOWER": "gambling"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "indecent"}, {"LOWER": "assault"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "indecent"}, {"LOWER": "photograph"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indefinite"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "enter"},
            {"LOWER": "or"},
            {"LOWER": "remain"},
            {"LOWER": "in"},
            {"LOWER": "united"},
            {"LOWER": "kingdom"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indefinite"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "indemnity"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indemnity"},
            {"LOWER": "against"},
            {"LOWER": "future"},
            {"LOWER": "liability"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "indemnity"}, {"LOWER": "basis"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "indemnity"}, {"LOWER": "clause"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "independence"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "independent"}, {"LOWER": "contractor"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "independent"}, {"LOWER": "examination"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "independent"}, {"LOWER": "inquiry"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "independent"}, {"LOWER": "living"}, {"LOWER": "fund"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "independent"},
            {"LOWER": "safeguarding"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "indeterminate"}, {"LOWER": "sentence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indeterminate"},
            {"LOWER": "sentence"},
            {"LOWER": "prisoners"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indian"},
            {"LOWER": "citizen"},
            {"LOWER": "in"},
            {"LOWER": "united"},
            {"LOWER": "kingdom"},
            {"LOWER": "as"},
            {"LOWER": "student"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "further"},
            {"LOWER": "leave"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "indian"}, {"LOWER": "endurance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "indian"}, {"LOWER": "grace"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indication"},
            {"LOWER": "of"},
            {"LOWER": "lower"},
            {"LOWER": "price"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "indication"}, {"LOWER": "of"}, {"LOWER": "plea"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "indictment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indictment"},
            {"LOWER": "charging"},
            {"LOWER": "rape"},
            {"LOWER": "contrary"},
            {"LOWER": "to"},
            {"LOWER": "sexual"},
            {"LOWER": "offences"},
            {"LOWER": "act"},
            {"LOWER": "1956"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "indirect"}, {"LOWER": "copying"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "indirect"}, {"LOWER": "discrimination"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "indirect"}, {"LOWER": "discrimination"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "individual"},
            {"LOWER": "contracts"},
            {"LOWER": "of"},
            {"LOWER": "employment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "individuals"},
            {"LOWER": "right"},
            {"LOWER": "not"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "deprived"},
            {"LOWER": "of"},
            {"LOWER": "life"},
            {"LOWER": "except"},
            {"LOWER": "by"},
            {"LOWER": "due"},
            {"LOWER": "process"},
            {"LOWER": "of"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "indorsement"},
            {"LOWER": "by"},
            {"LOWER": "way"},
            {"LOWER": "of"},
            {"LOWER": "guarantee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inducement"},
            {"LOWER": "of"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contract"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "induction"}, {"LOWER": "period"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "action"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "industrial"}, {"LOWER": "heriditament"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "industrial"},
            {"LOWER": "injuries"},
            {"LOWER": "benefit"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "relations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "relations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "relations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "relations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "training"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "industrial"}, {"LOWER": "training"}, {"LOWER": "levy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "tribunal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "industrial"}, {"LOWER": "tribunals"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "infant"}, {"LOWER": "dependants"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "infant"}, {"LOWER": "plaintiff"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "infanticide"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "infants"}, {"LOWER": "dependency"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "inferior"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "inferior"}, {"LOWER": "tribunal"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inflation"},
            {"LOWER": "up"},
            {"ORTH": "-"},
            {"LOWER": "rating"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "inflation"}, {"LOWER": "uprating"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inflicting"},
            {"LOWER": "grievous"},
            {"LOWER": "bodily"},
            {"LOWER": "harm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "informal"},
            {"LOWER": "surrogacy"},
            {"LOWER": "arrangement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "informal"},
            {"LOWER": "visits"},
            {"LOWER": "by"},
            {"LOWER": "police"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "information"}, {"LOWER": "commissioner"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "information"}, {"LOWER": "notice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "information"}, {"LOWER": "society"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "information"},
            {"LOWER": "supplied"},
            {"LOWER": "under"},
            {"LOWER": "statutory"},
            {"LOWER": "procedure"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "informed"}, {"LOWER": "consent"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "infringement"},
            {"LOWER": "by"},
            {"LOWER": "communication"},
            {"LOWER": "to"},
            {"LOWER": "public"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "infringement"},
            {"LOWER": "of"},
            {"LOWER": "privacy"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "infringement"}, {"LOWER": "of"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ingredients"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ingredients"}, {"LOWER": "of"}, {"LOWER": "tort"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "inherent"}, {"LOWER": "jurisdiction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "inheritance"}, {"LOWER": "tax"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inhuman"},
            {"LOWER": "and"},
            {"LOWER": "degrading"},
            {"LOWER": "treatment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inhuman"},
            {"LOWER": "or"},
            {"LOWER": "degrading"},
            {"LOWER": "punishment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inhuman"},
            {"LOWER": "or"},
            {"LOWER": "degrading"},
            {"LOWER": "treatment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inhuman"},
            {"LOWER": "or"},
            {"LOWER": "degrading"},
            {"LOWER": "treatment"},
            {"LOWER": "or"},
            {"LOWER": "punishment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "injunction"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "injunction"}, {"LOWER": "to"}, {"LOWER": "enforce"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "injunction"}, {"LOWER": "to"}, {"LOWER": "restrain"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "injunction"},
            {"LOWER": "to"},
            {"LOWER": "restrain"},
            {"LOWER": "bank"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "injunction"},
            {"LOWER": "to"},
            {"LOWER": "restrain"},
            {"LOWER": "beneficiary"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "injunction"},
            {"LOWER": "to"},
            {"LOWER": "restrain"},
            {"LOWER": "publication"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "injunctive"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "injunctive"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "injury"}, {"LOWER": "award"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "injury"}, {"LOWER": "by"}, {"LOWER": "lifting"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "injury"}, {"LOWER": "to"}, {"LOWER": "feelings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "injury"}, {"LOWER": "to"}, {"LOWER": "hip"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "injury"}, {"LOWER": "to"}, {"LOWER": "passenger"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "inland"}, {"LOWER": "revenue"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "inland"}, {"LOWER": "waterways"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "innuendo"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "innuendo"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "inquest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "inquiry"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inquiry"},
            {"LOWER": "into"},
            {"LOWER": "companys"},
            {"LOWER": "dealings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inquiry"},
            {"LOWER": "into"},
            {"LOWER": "company"},
            {"LOWER": "’s"},
            {"LOWER": "dealings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "insanity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "inset"}, {"LOWER": "appointment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insider"}, {"LOWER": "dealing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insolvency"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "inspection"}, {"LOWER": "and"}, {"LOWER": "copying"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "inspection"}, {"LOWER": "of"}, {"LOWER": "documents"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inspection"},
            {"LOWER": "of"},
            {"LOWER": "documents"},
            {"LOWER": "retained"},
            {"LOWER": "by"},
            {"LOWER": "former"},
            {"LOWER": "agent"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inspection"},
            {"ORTH": ","},
            {"LOWER": "right"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "inspector"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "installation"},
            {"LOWER": "of"},
            {"LOWER": "transmission"},
            {"LOWER": "system"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "institution"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}, {"LOWER": "benefit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}, {"LOWER": "broker"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}, {"LOWER": "payments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "insurance"}, {"LOWER": "policy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "insurers"},
            {"LOWER": "liability"},
            {"LOWER": "to"},
            {"LOWER": "indemnify"},
            {"LOWER": "insured"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "insurer"}, {"LOWER": "’s"}, {"LOWER": "liability"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "intellectual"}, {"LOWER": "property"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intended"},
            {"LOWER": "beneficiaries"},
            {"LOWER": "under"},
            {"LOWER": "will"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intending"},
            {"LOWER": "administrator"},
            {"LOWER": "’s"},
            {"LOWER": "rights"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "intention"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "objective"}, {"LOWER": "intention"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "subject"}, {"LOWER": "intention"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intent"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "proved"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "intention"}, {"LOWER": "of"}, {"LOWER": "testator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "intention"}, {"LOWER": "of"}, {"LOWER": "testatrix"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "create"},
            {"LOWER": "legal"},
            {"LOWER": "relations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "create"},
            {"LOWER": "legal"},
            {"LOWER": "relationship"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "stay"},
            {"LOWER": "as"},
            {"LOWER": "refugee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intentional"},
            {"LOWER": "infliction"},
            {"LOWER": "of"},
            {"LOWER": "harm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intentionally"},
            {"LOWER": "causing"},
            {"LOWER": "harm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "inter"}, {"LOWER": "vivos"}, {"LOWER": "transaction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interest"}, {"LOWER": "in"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interest"},
            {"LOWER": "of"},
            {"LOWER": "third"},
            {"LOWER": "party"},
            {"LOWER": "in"},
            {"LOWER": "policy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interest"}, {"LOWER": "on"}, {"LOWER": "award"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interest"}, {"LOWER": "on"}, {"LOWER": "costs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interest"}, {"LOWER": "on"}, {"LOWER": "debt"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interest"}, {"LOWER": "rate"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interface"},
            {"LOWER": "with"},
            {"LOWER": "immigration"},
            {"LOWER": "issues"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interference"}, {"LOWER": "with"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interference"}, {"LOWER": "with"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interference"}, {"LOWER": "with"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interference"},
            {"LOWER": "with"},
            {"LOWER": "motor"},
            {"LOWER": "vehicle"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interference"},
            {"LOWER": "with"},
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "way"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interim"}, {"LOWER": "care"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interim"}, {"LOWER": "care"}, {"LOWER": "orders"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "declaration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "injunction"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interim"}, {"LOWER": "injunctive"}, {"LOWER": "relief"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "orders"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interim"}, {"LOWER": "orders"}, {"LOWER": "hearing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interim"}, {"LOWER": "orders"}, {"LOWER": "panel"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interim"}, {"LOWER": "orders"}, {"LOWER": "panel"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "payment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "payment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "payments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interim"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "interlocutory"}, {"LOWER": "appeal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interlocutory"}, {"LOWER": "application"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interlocutory"},
            {"LOWER": "judgment"},
            {"LOWER": "with"},
            {"LOWER": "damages"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "assessed"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interlocutory"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interlocutory"}, {"LOWER": "proceedings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interlocutory"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "intermediate"}, {"LOWER": "power"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "internal"}, {"LOWER": "market"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "internal"},
            {"LOWER": "market"},
            {"LOWER": "in"},
            {"LOWER": "electricity"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "internal"},
            {"LOWER": "market"},
            {"LOWER": "in"},
            {"LOWER": "natural"},
            {"LOWER": "gas"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "internal"},
            {"LOWER": "re"},
            {"LOWER": "-"},
            {"LOWER": "ordering"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "internal"}, {"LOWER": "relocation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "internal"}, {"LOWER": "reordering"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "internal"}, {"LOWER": "reordering"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "internal"}, {"LOWER": "taxes"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "international"}, {"LOWER": "agreements"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "international"},
            {"LOWER": "carriage"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
            {"LOWER": "by"},
            {"LOWER": "road"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "international"},
            {"LOWER": "carriage"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
            {"LOWER": "by"},
            {"LOWER": "road"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "international"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "international"}, {"LOWER": "crime"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "international"}, {"LOWER": "crime"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "international"}, {"LOWER": "jurisdiction"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "international"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "international"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "international"}, {"LOWER": "law"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "international"},
            {"LOWER": "maritime"},
            {"LOWER": "organisation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "international"}, {"LOWER": "organisation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "international"},
            {"LOWER": "organisation"},
            {"ORTH": "-"},
            {"LOWER": "organisation"},
            {"LOWER": "created"},
            {"LOWER": "by"},
            {"LOWER": "treaty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "international"},
            {"LOWER": "removal"},
            {"LOWER": "services"},
            {"LOWER": "market"},
            {"LOWER": "in"},
            {"LOWER": "belgium"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "international"}, {"LOWER": "trade"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "international"}, {"LOWER": "treaties"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "international"}, {"LOWER": "treaties"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "internet"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "internet"}, {"LOWER": "service"}, {"LOWER": "provider"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interpleader"},
            {"LOWER": "proceedings"},
            {"LOWER": "writ"},
            {"LOWER": "of"},
            {"LOWER": "fieri"},
            {"LOWER": "facias"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interpretation"},
            {"LOWER": "of"},
            {"LOWER": "constitution"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interpreter"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interveners"},
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "vary"},
            {"LOWER": "mareva"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intervention"},
            {"LOWER": "by"},
            {"LOWER": "law"},
            {"LOWER": "society"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "interventions"}, {"LOWER": "by"}, {"LOWER": "judge"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "interview"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "interview"},
            {"LOWER": "by"},
            {"LOWER": "security"},
            {"LOWER": "service"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "intestacy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "intimidation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intimidation"},
            {"LOWER": "of"},
            {"LOWER": "witness"},
            {"LOWER": "or"},
            {"LOWER": "juror"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "intra"},
            {"ORTH": "-"},
            {"LOWER": "community"},
            {"LOWER": "acquisition"},
            {"LOWER": "and"},
            {"LOWER": "supply"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "introducing"},
            {"LOWER": "broker"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "introduction"},
            {"LOWER": "and"},
            {"LOWER": "outsourcing"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "introductory"}, {"LOWER": "tenancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "invalidity"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "invasion"}, {"LOWER": "of"}, {"LOWER": "privacy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "invention"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "investigating"}, {"LOWER": "committee"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "investigation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investigation"},
            {"LOWER": "by"},
            {"LOWER": "commission"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "investigation"}, {"LOWER": "into"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investigation"},
            {"LOWER": "into"},
            {"LOWER": "money"},
            {"LOWER": "laundering"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investigation"},
            {"LOWER": "into"},
            {"LOWER": "proceeds"},
            {"LOWER": "of"},
            {"LOWER": "criminal"},
            {"LOWER": "conduct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "investigation"}, {"LOWER": "of"}, {"LOWER": "affairs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investigation"},
            {"LOWER": "of"},
            {"LOWER": "complaint"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "investigations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "investigatory"}, {"LOWER": "powers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investigatory"},
            {"LOWER": "powers"},
            {"LOWER": "tribunal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "investment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "investment"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "investment"}, {"LOWER": "fund"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investment"},
            {"LOWER": "in"},
            {"LOWER": "the"},
            {"LOWER": "processing"},
            {"LOWER": "and"},
            {"LOWER": "marketing"},
            {"LOWER": "of"},
            {"LOWER": "agricultural"},
            {"LOWER": "products"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investment"},
            {"LOWER": "of"},
            {"LOWER": "money"},
            {"LOWER": "in"},
            {"LOWER": "court"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "investment"}, {"LOWER": "property"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investment"},
            {"LOWER": "service"},
            {"LOWER": "and"},
            {"LOWER": "advice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "investment"},
            {"LOWER": "services"},
            {"LOWER": "and"},
            {"LOWER": "activities"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "investments"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inviting"},
            {"LOWER": "support"},
            {"LOWER": "for"},
            {"LOWER": "proscribed"},
            {"LOWER": "organisation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "inviting"},
            {"LOWER": "support"},
            {"LOWER": "for"},
            {"LOWER": "proscribed"},
            {"LOWER": "organisation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "invoice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "irenes"}, {"LOWER": "success"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "irish"}, {"LOWER": "rowan"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "irregular"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "irregularity"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "irregularity"},
            {"LOWER": "in"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "irregularity"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "jury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "irreversible"}, {"LOWER": "brain"}, {"LOWER": "damage"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "isle"}, {"LOWER": "of"}, {"LOWER": "man"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "issue"}, {"LOWER": "estoppel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "issue"}, {"LOWER": "estoppel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "issue"}, {"LOWER": "estoppel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "issue"}, {"LOWER": "of"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jacqueline"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jamaica"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jamaica"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jersey"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jobseekers"}, {"LOWER": "allowance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jobseeker"}, {"LOWER": "’s"}, {"LOWER": "allowance"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "joinder"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "joinder"}, {"LOWER": "of"}, {"LOWER": "charges"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "joinder"}, {"LOWER": "of"}, {"LOWER": "respondent"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "joint"}, {"LOWER": "and"}, {"LOWER": "several"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "joint"},
            {"LOWER": "and"},
            {"LOWER": "several"},
            {"LOWER": "liability"},
            {"LOWER": "of"},
            {"LOWER": "parent"},
            {"LOWER": "companies"},
            {"LOWER": "vis"},
            {"ORTH": "-"},
            {"LOWER": "à"},
            {"ORTH": "-"},
            {"LOWER": "vis"},
            {"LOWER": "creditors"},
            {"LOWER": "of"},
            {"LOWER": "subsidiaries"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "debtors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "enterprise"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "interest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "interest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "interest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "tenancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "tenants"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "tortfeasors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "tortfeasors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "trials"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "joint"}, {"LOWER": "venture"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "joint"}, {"LOWER": "venture"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "journalist"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "granting"},
            {"LOWER": "certificate"},
            {"LOWER": "of"},
            {"LOWER": "fitness"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "in"},
            {"LOWER": "own"},
            {"LOWER": "cause"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "ordering"},
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "extradition"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "ordering"},
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "extradition"},
            {"LOWER": "to"},
            {"LOWER": "france"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "ordering"},
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "extradition"},
            {"LOWER": "to"},
            {"LOWER": "poland"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "ordering"},
            {"LOWER": "defendants"},
            {"LOWER": "extradition"},
            {"LOWER": "to"},
            {"LOWER": "france"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "ruling"},
            {"LOWER": "defendants"},
            {"LOWER": "having"},
            {"LOWER": "no"},
            {"LOWER": "case"},
            {"LOWER": "to"},
            {"LOWER": "answer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "striking"},
            {"LOWER": "out"},
            {"LOWER": "defendants"},
            {"LOWER": "’"},
            {"LOWER": "defence"},
            {"LOWER": "and"},
            {"LOWER": "counterclaim"},
            {"LOWER": "and"},
            {"LOWER": "entering"},
            {"LOWER": "judgment"},
            {"LOWER": "against"},
            {"LOWER": "them"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"ORTH": ","},
            {"LOWER": "change"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "judges"}, {"LOWER": "direction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judges"}, {"LOWER": "directions"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judge"}, {"LOWER": "’s"}, {"LOWER": "conduct"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judge"}, {"LOWER": "’s"}, {"LOWER": "direction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judge"}, {"LOWER": "’s"}, {"LOWER": "directions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "’s"},
            {"LOWER": "directions"},
            {"LOWER": "to"},
            {"LOWER": "jury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judge"},
            {"LOWER": "’s"},
            {"LOWER": "summing"},
            {"LOWER": "up"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judgment"}, {"LOWER": "by"}, {"LOWER": "default"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "judgment"}, {"LOWER": "debt"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judgment"}, {"LOWER": "in"}, {"LOWER": "default"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judgment"}, {"LOWER": "in"}, {"LOWER": "default"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judgment"},
            {"LOWER": "in"},
            {"LOWER": "default"},
            {"LOWER": "of"},
            {"LOWER": "appearance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judgment"},
            {"LOWER": "of"},
            {"LOWER": "general"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judgment"},
            {"ORTH": ","},
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judgments"},
            {"ORTH": ","},
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "act"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judicial"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
            {"LOWER": "in"},
            {"LOWER": "civil"},
            {"LOWER": "and"},
            {"LOWER": "commercial"},
            {"LOWER": "matters"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judicial"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
            {"LOWER": "in"},
            {"LOWER": "civil"},
            {"LOWER": "matters"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "judicial"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
            {"LOWER": "in"},
            {"LOWER": "civil"},
            {"LOWER": "or"},
            {"LOWER": "commercial"},
            {"LOWER": "matters"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "immunity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judicial"}, {"LOWER": "office"}, {"LOWER": "holder"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judicial"}, {"LOWER": "pension"}, {"LOWER": "scheme"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "precedent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "precedent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "precedent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "precedent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "precedent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "review"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "review"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "judicial"}, {"LOWER": "review"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "judicial"}, {"LOWER": "review"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "judiciary"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jurisdiction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "clause"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "of"},
            {"LOWER": "costs"},
            {"LOWER": "judge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "of"},
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "of"},
            {"LOWER": "employment"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "of"},
            {"LOWER": "english"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "of"},
            {"LOWER": "english"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "of"},
            {"LOWER": "national"},
            {"LOWER": "courts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "registration"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "judgment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "to"}, {"LOWER": "award"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "to"}, {"LOWER": "enforce"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "to"}, {"LOWER": "grant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "to"}, {"LOWER": "grant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "to"}, {"LOWER": "make"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "jurisdiction"}, {"LOWER": "to"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "convention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "convention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "council"},
            {"LOWER": "regulation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "european"},
            {"LOWER": "convention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "european"},
            {"LOWER": "regulation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "regulation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "regulation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "jurisdiction"},
            {"LOWER": "under"},
            {"LOWER": "regulation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "jurors"}, {"LOWER": "conduct"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jury"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jury"}, {"LOWER": "trial"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "jurys"}, {"LOWER": "verdict"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "just"},
            {"LOWER": "and"},
            {"LOWER": "equitable"},
            {"LOWER": "grounds"},
            {"LOWER": "for"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "just"}, {"LOWER": "satisfaction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "justice"},
            {"LOWER": "and"},
            {"LOWER": "home"},
            {"LOWER": "affairs"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "justices"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "justices"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "justiciability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "justification"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "juvenile"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "juvenile"}, {"LOWER": "offender"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "k.h"}, {"ORTH": "."}, {"LOWER": "enterprise"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "k.h"},
            {"ORTH": "."},
            {"LOWER": "enterprise"},
            {"ORTH": "-"},
            {"LOWER": "pioneer"},
            {"LOWER": "container"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "keeping"}, {"LOWER": "disorderly"}, {"LOWER": "house"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "keyword"},
            {"LOWER": "advert"},
            {"LOWER": "on"},
            {"LOWER": "internet"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "knowing"}, {"LOWER": "receipt"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "knowledge"}, {"LOWER": "of"}, {"LOWER": "company"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "labelling"},
            {"LOWER": "and"},
            {"LOWER": "presentation"},
            {"LOWER": "of"},
            {"LOWER": "foodstuffs"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "laboratory"}, {"LOWER": "test"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "lack"}, {"LOWER": "of"}, {"LOWER": "novelty"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lady"}, {"LOWER": "korcula"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "lady"}, {"LOWER": "racisce"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "lake"},
            {"LOWER": "in"},
            {"LOWER": "country"},
            {"LOWER": "park"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "land"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "land"}, {"LOWER": "charge"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "land"},
            {"LOWER": "registrar"},
            {"LOWER": "’s"},
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "restrict"},
            {"LOWER": "dealing"},
            {"LOWER": "in"},
            {"LOWER": "registered"},
            {"LOWER": "property"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "land"}, {"LOWER": "registration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "land"}, {"LOWER": "registration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "land"}, {"LOWER": "registration"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recovery"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "landfill"}, {"LOWER": "site"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "landfill"}, {"LOWER": "tax"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "landlord"}, {"LOWER": "and"}, {"LOWER": "tenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "landlord"}, {"LOWER": "and"}, {"LOWER": "tenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "landlord"}, {"LOWER": "and"}, {"LOWER": "tenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "landlord"}, {"LOWER": "and"}, {"LOWER": "tenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlords"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "covenant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlords"},
            {"LOWER": "covenant"},
            {"LOWER": "to"},
            {"LOWER": "insure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlords"},
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "demolish"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlords"},
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "occupy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlords"},
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "reconstruct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlord"},
            {"LOWER": "’s"},
            {"LOWER": "application"},
            {"LOWER": "to"},
            {"LOWER": "dispense"},
            {"LOWER": "with"},
            {"LOWER": "statutory"},
            {"LOWER": "requirement"},
            {"LOWER": "to"},
            {"LOWER": "consult"},
            {"LOWER": "with"},
            {"LOWER": "tenants"},
            {"LOWER": "before"},
            {"LOWER": "carrying"},
            {"LOWER": "out"},
            {"LOWER": "works"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlord"},
            {"LOWER": "’s"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "covenant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlord"},
            {"LOWER": "’s"},
            {"LOWER": "covenant"},
            {"LOWER": "to"},
            {"LOWER": "repair"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlord"},
            {"LOWER": "’s"},
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "reconstruct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "landlord"},
            {"LOWER": "’s"},
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "reconstruct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "landowner"}, {"LOWER": "’s"}, {"LOWER": "duty"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lands"}, {"LOWER": "tribunal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "lapse"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "late"}, {"LOWER": "payment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "law"}, {"LOWER": "enforcement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "law"}, {"LOWER": "enforcement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "law"}, {"LOWER": "reform"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "law"}, {"LOWER": "reform"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "law"}, {"LOWER": "reform"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "law"}, {"LOWER": "reports"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "law"}, {"LOWER": "society"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "law"},
            {"LOWER": "societys"},
            {"LOWER": "indemnity"},
            {"LOWER": "insurance"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "law"},
            {"LOWER": "society"},
            {"LOWER": "’s"},
            {"LOWER": "minimum"},
            {"LOWER": "terms"},
            {"LOWER": "and"},
            {"LOWER": "conditions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lawful"}, {"LOWER": "excuse"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "lawfulness"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "lawfulness"}, {"LOWER": "of"}, {"LOWER": "detention"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lay"}, {"LOWER": "rector"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "lay"},
            {"ORTH": "-"},
            {"LOWER": "off"},
            {"LOWER": "and"},
            {"LOWER": "short"},
            {"ORTH": "-"},
            {"LOWER": "time"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lease"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "lease"},
            {"LOWER": "for"},
            {"LOWER": "fixed"},
            {"LOWER": "term"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leasehold"}, {"LOWER": "enfranchisement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "leasehold"},
            {"LOWER": "interest"},
            {"LOWER": "of"},
            {"LOWER": "bankrupt"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "leasehold"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "leasing"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "of"}, {"LOWER": "absence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "to"}, {"LOWER": "act"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "to"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "time"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "to"}, {"LOWER": "apply"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "bring"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "to"}, {"LOWER": "enter"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "enter"},
            {"LOWER": "and"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "enter"},
            {"LOWER": "or"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "to"}, {"LOWER": "oppose"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "to"}, {"LOWER": "prefer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "leave"}, {"LOWER": "to"}, {"LOWER": "remain"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legacy"}, {"LOWER": "case"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legacy"}, {"LOWER": "to"}, {"LOWER": "executor"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "advice"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legal"},
            {"LOWER": "advice"},
            {"LOWER": "at"},
            {"LOWER": "police"},
            {"LOWER": "station"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legal"}, {"LOWER": "advice"}, {"LOWER": "privilege"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "aid"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "aid"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "aid"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "aid"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legal"}, {"LOWER": "aid"}, {"LOWER": "application"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legal"},
            {"LOWER": "aid"},
            {"LOWER": "contribution"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legal"}, {"LOWER": "aid"}, {"LOWER": "factor"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legal"}, {"LOWER": "aid"}, {"LOWER": "funding"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legal"},
            {"LOWER": "basis"},
            {"LOWER": "of"},
            {"LOWER": "decision"},
            {"LOWER": "at"},
            {"LOWER": "issue"},
            {"LOWER": "in"},
            {"LOWER": "main"},
            {"LOWER": "proceedings"},
            {"LOWER": "no"},
            {"LOWER": "longer"},
            {"LOWER": "present"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "charge"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "charges"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "entrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legal"}, {"LOWER": "expenses"}, {"LOWER": "insurance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legal"}, {"LOWER": "expenses"}, {"LOWER": "insurer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legal"},
            {"LOWER": "or"},
            {"LOWER": "equitable"},
            {"LOWER": "defence"},
            {"LOWER": "to"},
            {"LOWER": "action"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "privilege"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "profession"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legal"},
            {"LOWER": "professional"},
            {"LOWER": "privilege"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legal"},
            {"LOWER": "professional"},
            {"LOWER": "privilege"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legal"},
            {"LOWER": "protection"},
            {"LOWER": "of"},
            {"LOWER": "databases"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legal"}, {"LOWER": "representatives"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legal"}, {"LOWER": "services"}, {"LOWER": "commission"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legality"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legally"}, {"LOWER": "aided"}, {"LOWER": "applicant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legally"}, {"LOWER": "aided"}, {"LOWER": "plaintiff"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legally"}, {"LOWER": "assisted"}, {"LOWER": "person"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legislation"}, {"LOWER": "applicable"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "legislation"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "providing"},
            {"LOWER": "for"},
            {"LOWER": "repayment"},
            {"LOWER": "of"},
            {"LOWER": "price"},
            {"LOWER": "paid"},
            {"LOWER": "by"},
            {"LOWER": "purchaser"},
            {"LOWER": "for"},
            {"LOWER": "purchased"},
            {"LOWER": "shares"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "legitimacy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "legitimate"}, {"LOWER": "expectation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "length"}, {"LOWER": "of"}, {"LOWER": "judgment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "less"}, {"LOWER": "favourable"}, {"LOWER": "treatment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "lessees"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "covenant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "letter"}, {"LOWER": "of"}, {"LOWER": "credit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "letters"}, {"LOWER": "of"}, {"LOWER": "request"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "letters"}, {"LOWER": "rogatory"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "level"},
            {"LOWER": "of"},
            {"LOWER": "direct"},
            {"LOWER": "payments"},
            {"LOWER": "applicable"},
            {"LOWER": "in"},
            {"LOWER": "member"},
            {"LOWER": "states"},
            {"LOWER": "as"},
            {"LOWER": "constituted"},
            {"LOWER": "on"},
            {"LOWER": "30"},
            {"LOWER": "april"},
            {"LOWER": "2004"},
            {"LOWER": "and"},
            {"LOWER": "in"},
            {"LOWER": "member"},
            {"LOWER": "states"},
            {"LOWER": "that"},
            {"LOWER": "joined"},
            {"LOWER": "on"},
            {"LOWER": "1"},
            {"LOWER": "may"},
            {"LOWER": "2004"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "levy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "levy"}, {"LOWER": "on"}, {"LOWER": "goods"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "liability"}, {"LOWER": "for"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "for"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "for"}, {"LOWER": "claims"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "for"},
            {"LOWER": "defective"},
            {"LOWER": "products"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "for"},
            {"LOWER": "import"},
            {"LOWER": "duties"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "for"}, {"LOWER": "injury"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "liability"}, {"LOWER": "insurance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "of"}, {"LOWER": "employee"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "of"},
            {"LOWER": "employers"},
            {"LOWER": "and"},
            {"LOWER": "principals"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "of"}, {"LOWER": "firm"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "of"},
            {"LOWER": "guarantee"},
            {"LOWER": "institution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "of"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "employer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "of"},
            {"LOWER": "original"},
            {"LOWER": "lessee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "of"},
            {"LOWER": "parent"},
            {"LOWER": "company"},
            {"LOWER": "for"},
            {"LOWER": "infringements"},
            {"LOWER": "of"},
            {"LOWER": "law"},
            {"LOWER": "on"},
            {"LOWER": "cartels"},
            {"LOWER": "committed"},
            {"LOWER": "by"},
            {"LOWER": "subsidiary"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "of"}, {"LOWER": "principals"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "of"}, {"LOWER": "signatory"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "of"}, {"LOWER": "trustees"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "to"},
            {"LOWER": "repay"},
            {"LOWER": "loan"},
            {"LOWER": "to"},
            {"LOWER": "employer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "to"}, {"LOWER": "suit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liability"},
            {"LOWER": "to"},
            {"LOWER": "third"},
            {"LOWER": "parties"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "to"}, {"LOWER": "undertakers"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "libel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "libel"}, {"LOWER": "action"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "libel"}, {"LOWER": "and"}, {"LOWER": "slander"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "liberty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "library"}, {"LOWER": "services"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "licence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "licence"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "licence"}, {"LOWER": "conditions"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "licence"}, {"LOWER": "fee"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "licence"}, {"LOWER": "or"}, {"LOWER": "tenancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "licence"}, {"LOWER": "or"}, {"LOWER": "tenancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "licence"}, {"LOWER": "or"}, {"LOWER": "tenancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "licence"},
            {"LOWER": "to"},
            {"LOWER": "occupy"},
            {"LOWER": "for"},
            {"LOWER": "agricultural"},
            {"LOWER": "use"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "licensable"}, {"LOWER": "activity"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "licensed"}, {"LOWER": "betting"}, {"LOWER": "office"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "licensed"}, {"LOWER": "greyhound"}, {"LOWER": "track"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "licensed"}, {"LOWER": "premises"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "licensing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "licensing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "licensing"}, {"LOWER": "authority"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "licensing"},
            {"LOWER": "authority"},
            {"LOWER": "for"},
            {"LOWER": "exhumation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "licensing"}, {"LOWER": "of"}, {"LOWER": "houses"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "licensor"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "licensor"}, {"LOWER": "’s"}, {"LOWER": "liability"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lien"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "lien"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "lien"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "life"}, {"LOWER": "assurance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "life"}, {"LOWER": "imprisonment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "life"}, {"LOWER": "insurance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "life"}, {"LOWER": "interest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "life"}, {"LOWER": "sentence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "lift"}, {"LOWER": "of"}, {"LOWER": "suspension"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "light"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "likelihood"}, {"LOWER": "of"}, {"LOWER": "driving"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "limit"}, {"LOWER": "of"}, {"LOWER": "time"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limit"},
            {"LOWER": "on"},
            {"LOWER": "recoverable"},
            {"LOWER": "costs"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "limitation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "limitation"}, {"LOWER": "period"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "limitation"}, {"LOWER": "action"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "limitation"}, {"LOWER": "clause"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "limitation"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "limitation"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "limitation"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "limitation"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "limitation"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "limitation"}, {"LOWER": "of"}, {"LOWER": "liability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limitation"},
            {"LOWER": "of"},
            {"LOWER": "volatile"},
            {"LOWER": "organic"},
            {"LOWER": "compounds"},
            {"LOWER": "emissions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "limited"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "limited"}, {"LOWER": "company"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limited"},
            {"LOWER": "grant"},
            {"LOWER": "of"},
            {"LOWER": "letters"},
            {"LOWER": "of"},
            {"LOWER": "administration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limited"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "enter"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limited"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limited"},
            {"LOWER": "liability"},
            {"LOWER": "partnership"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "limited"}, {"LOWER": "partnership"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limiting"},
            {"LOWER": "oral"},
            {"LOWER": "argument"},
            {"LOWER": "in"},
            {"LOWER": "certain"},
            {"LOWER": "applications"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "limiting"},
            {"LOWER": "witness"},
            {"LOWER": "statements"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "linked"}, {"LOWER": "contracts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "linking"}, {"LOWER": "factor"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "lion"}, {"LOWER": "intoximeter"}, {"LOWER": "3000"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "liquidated"}, {"LOWER": "damages"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "liquidated"},
            {"LOWER": "pecuniary"},
            {"LOWER": "claim"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "liquidation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "liquidation"}, {"LOWER": "petition"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "liquidator"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "list"}, {"LOWER": "of"}, {"LOWER": "streets"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "listed"}, {"LOWER": "building"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "listed"}, {"LOWER": "building"}, {"LOWER": "control"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "listed"}, {"LOWER": "buildings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "listing"}, {"LOWER": "decision"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "literary"}, {"LOWER": "work"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "litigant"}, {"LOWER": "in"}, {"LOWER": "person"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "litigation"},
            {"LOWER": "expense"},
            {"LOWER": "insurance"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "litigation"}, {"LOWER": "friend"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "litigation"}, {"LOWER": "funding"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "litigation"}, {"LOWER": "privilege"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "live"}, {"LOWER": "video"}, {"LOWER": "link"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "living"},
            {"LOWER": "on"},
            {"LOWER": "earnings"},
            {"LOWER": "of"},
            {"LOWER": "prostitution"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lloyds"}, {"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "loan"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "loan"}, {"LOWER": "deed"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "loan"}, {"LOWER": "facility"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "authority"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "alleging"},
            {"LOWER": "abuse"},
            {"LOWER": "of"},
            {"LOWER": "children"},
            {"LOWER": "perpetrated"},
            {"LOWER": "by"},
            {"LOWER": "parents"},
            {"LOWER": "and"},
            {"LOWER": "family"},
            {"LOWER": "members"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "care"},
            {"LOWER": "and"},
            {"LOWER": "placement"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "committal"},
            {"LOWER": "for"},
            {"LOWER": "contempt"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "forced"},
            {"LOWER": "marriage"},
            {"LOWER": "protection"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "placement"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "placement"},
            {"LOWER": "order"},
            {"LOWER": "with"},
            {"LOWER": "view"},
            {"LOWER": "to"},
            {"LOWER": "adoption"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "reporting"},
            {"LOWER": "restriction"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "withdrawal"},
            {"LOWER": "of"},
            {"LOWER": "treatment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "applying"},
            {"LOWER": "to"},
            {"LOWER": "withdraw"},
            {"LOWER": "findings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "bringing"},
            {"LOWER": "best"},
            {"LOWER": "interests"},
            {"LOWER": "proceedings"},
            {"LOWER": "in"},
            {"LOWER": "relation"},
            {"LOWER": "to"},
            {"LOWER": "patient"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "changing"},
            {"LOWER": "care"},
            {"LOWER": "plan"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "local"}, {"LOWER": "authority"}, {"LOWER": "employee"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "local"}, {"LOWER": "authority"}, {"LOWER": "notice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "promoting"},
            {"LOWER": "plan"},
            {"LOWER": "restricting"},
            {"LOWER": "housing"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "seeking"},
            {"LOWER": "care"},
            {"LOWER": "and"},
            {"LOWER": "placement"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "seeking"},
            {"LOWER": "care"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "seeking"},
            {"LOWER": "removal"},
            {"LOWER": "of"},
            {"LOWER": "radicalised"},
            {"LOWER": "child"},
            {"LOWER": "from"},
            {"LOWER": "home"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "seeking"},
            {"LOWER": "to"},
            {"LOWER": "remove"},
            {"LOWER": "child"},
            {"LOWER": "at"},
            {"LOWER": "birth"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authoritys"},
            {"LOWER": "development"},
            {"LOWER": "plan"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "’s"},
            {"LOWER": "allocation"},
            {"LOWER": "criteria"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "’s"},
            {"LOWER": "development"},
            {"LOWER": "plan"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "’s"},
            {"LOWER": "education"},
            {"LOWER": "functions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "board"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "commissioner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "commissioner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "connection"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "local"}, {"LOWER": "connection"}, {"LOWER": "referral"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "local"}, {"LOWER": "education"}, {"LOWER": "authority"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "government"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "government"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "government"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "government"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "government"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "government"},
            {"LOWER": "pension"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "government"},
            {"LOWER": "reorganisation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "local"}, {"LOWER": "government"}, {"LOWER": "scheme"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "local"}, {"LOWER": "government"}, {"ORTH": "—"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "knowledge"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "plan"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "plan"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "planning"},
            {"LOWER": "authority"},
            {"LOWER": "granting"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "new"},
            {"LOWER": "food"},
            {"LOWER": "store"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "local"},
            {"LOWER": "planning"},
            {"LOWER": "authority"},
            {"LOWER": "issuing"},
            {"LOWER": "discontinuance"},
            {"LOWER": "notice"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "local"}, {"LOWER": "practices"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "local"}, {"LOWER": "valuation"}, {"LOWER": "court"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "locus"}, {"LOWER": "standi"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "london"}, {"LOWER": "plan"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "long"}, {"LOWER": "delay"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "long"}, {"LOWER": "vacation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "long"},
            {"ORTH": "-"},
            {"LOWER": "term"},
            {"LOWER": "disability"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "lord"},
            {"LOWER": "chancellor"},
            {"LOWER": "introducing"},
            {"LOWER": "changes"},
            {"LOWER": "to"},
            {"LOWER": "criminal"},
            {"LOWER": "legal"},
            {"LOWER": "aid"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "lordlord"}, {"LOWER": "and"}, {"LOWER": "tenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "chance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"ORTH": "“"},
            {"LOWER": "wilful"},
            {"LOWER": "misconduct"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "loss"}, {"LOWER": "of"}, {"LOWER": "chance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "loss"}, {"LOWER": "of"}, {"LOWER": "citizenship"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "loss"}, {"LOWER": "of"}, {"LOWER": "earnings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "expectation"},
            {"LOWER": "of"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "opportunity"},
            {"LOWER": "to"},
            {"LOWER": "have"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "or"},
            {"LOWER": "damage"},
            {"LOWER": "to"},
            {"LOWER": "ship"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "pension"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "personal"},
            {"LOWER": "injuries"},
            {"LOWER": "claim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "self"},
            {"ORTH": "-"},
            {"LOWER": "control"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "loss"}, {"LOWER": "of"}, {"LOWER": "support"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "loss"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "lottery"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "low"},
            {"LOWER": "value"},
            {"LOWER": "personal"},
            {"LOWER": "injury"},
            {"LOWER": "claim"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "luka"}, {"LOWER": "botic"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "lump"}, {"LOWER": "sum"}, {"LOWER": "award"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "lusitania"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maersk"}, {"LOWER": "nimrod"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "magistrates"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "magistrates"}, {"LOWER": "’"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "magistrates"},
            {"LOWER": "’"},
            {"LOWER": "court"},
            {"ORTH": ","},
            {"LOWER": "from"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mahkutai"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "maintained"}, {"LOWER": "by"}, {"LOWER": "deceased"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "maintenance"}, {"LOWER": "charge"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maintenance"}, {"LOWER": "charges"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maintenance"}, {"LOWER": "clause"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "maintenance"}, {"LOWER": "obligation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "maintenance"}, {"LOWER": "of"}, {"LOWER": "equipment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "maintenance"}, {"LOWER": "of"}, {"LOWER": "suit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "maintenance"}, {"LOWER": "of"}, {"LOWER": "suit"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "maintenance"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "maintenance"}, {"LOWER": "pending"}, {"LOWER": "suit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "major"},
            {"LOWER": "internal"},
            {"LOWER": "and"},
            {"LOWER": "external"},
            {"LOWER": "reordering"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "majority"}, {"LOWER": "verdict"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "making"},
            {"LOWER": "misleading"},
            {"LOWER": "statements"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "making"}, {"LOWER": "of"}, {"LOWER": "threat"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "making"},
            {"LOWER": "off"},
            {"LOWER": "without"},
            {"LOWER": "payment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "making"}, {"LOWER": "product"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maladministration"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "maladministration"},
            {"LOWER": "by"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "malaysia"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "male"},
            {"LOWER": "lacking"},
            {"LOWER": "capacity"},
            {"LOWER": "to"},
            {"LOWER": "consent"},
            {"LOWER": "to"},
            {"LOWER": "sterilisation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "male"},
            {"LOWER": "police"},
            {"LOWER": "officers"},
            {"LOWER": "armed"},
            {"LOWER": "as"},
            {"LOWER": "protection"},
            {"LOWER": "against"},
            {"LOWER": "terrorists"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "malicious"},
            {"LOWER": "administration"},
            {"LOWER": "of"},
            {"LOWER": "poison"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "malicious"}, {"LOWER": "damage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "malicious"}, {"LOWER": "falsehood"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "malicious"}, {"LOWER": "prosecution"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "malicious"}, {"LOWER": "wounding"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "maliciously"},
            {"LOWER": "procuring"},
            {"LOWER": "arrest"},
            {"LOWER": "and"},
            {"LOWER": "detention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "management"}, {"LOWER": "and"}, {"LOWER": "control"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "management"},
            {"LOWER": "of"},
            {"LOWER": "complex"},
            {"LOWER": "cases"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "managing"}, {"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "managing"}, {"LOWER": "director"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "managing"}, {"LOWER": "director"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "manchester"}, {"LOWER": "and"}, {"LOWER": "liverpool"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mandate"}, {"LOWER": "refugees"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mandatory"},
            {"LOWER": "contractual"},
            {"LOWER": "retirement"},
            {"LOWER": "age"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mandatory"}, {"LOWER": "death"}, {"LOWER": "penalty"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mandatory"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mandatory"}, {"LOWER": "retirement"}, {"LOWER": "age"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mandatory"}, {"LOWER": "sentence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mandatory"},
            {"LOWER": "whole"},
            {"LOWER": "life"},
            {"LOWER": "sentence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "manslaughter"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "manslaughter"}, {"LOWER": "of"}, {"LOWER": "mother"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "manslaughter"}, {"LOWER": "of"}, {"LOWER": "testator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "manual"},
            {"LOWER": "handling"},
            {"LOWER": "operations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "manufacture"}, {"LOWER": "of"}, {"LOWER": "medicines"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "manufacturer"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "manzanillo"}, {"ORTH": "ii"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "marchioness"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mareva"}, {"LOWER": "injunction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maria"}, {"LOWER": "d"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "marine"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "marine"}, {"LOWER": "insurance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maritime"}, {"LOWER": "transport"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "market"}, {"LOWER": "abuse"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "market"}, {"LOWER": "practice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "marketing"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "marketing"}, {"LOWER": "authorisation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "markets"},
            {"LOWER": "in"},
            {"LOWER": "financial"},
            {"LOWER": "instruments"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "marriage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "marriage"}, {"LOWER": "settlement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "married"}, {"LOWER": "woman"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "master"}, {"LOWER": "and"}, {"LOWER": "servant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "master"}, {"LOWER": "and"}, {"LOWER": "servant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "master"}, {"LOWER": "services"}, {"LOWER": "agreement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "material"},
            {"LOWER": "amendment"},
            {"LOWER": "to"},
            {"LOWER": "contract"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "material"}, {"LOWER": "circumstance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "material"}, {"LOWER": "consideration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "material"}, {"LOWER": "considerations"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "material"}, {"LOWER": "facts"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "material"},
            {"LOWER": "obtained"},
            {"LOWER": "under"},
            {"LOWER": "compulsory"},
            {"LOWER": "powers"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "maternity"}, {"LOWER": "leave"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maternity"}, {"LOWER": "pay"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maternity"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "matrimonial"}, {"LOWER": "home"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "matrimonial"}, {"LOWER": "matters"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "matrimonial"},
            {"LOWER": "proceedings"},
            {"LOWER": "brought"},
            {"LOWER": "by"},
            {"LOWER": "husband"},
            {"LOWER": "in"},
            {"LOWER": "italy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "matters"},
            {"LOWER": "outside"},
            {"LOWER": "framework"},
            {"LOWER": "of"},
            {"LOWER": "community"},
            {"LOWER": "law"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "maule"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mauritius"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mauritius"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "maximum"}, {"LOWER": "term"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "maximum"}, {"LOWER": "weekly"}, {"LOWER": "hours"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "maysora"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mckenzie"}, {"LOWER": "friends"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "meaning"}, {"LOWER": "of"}, {"LOWER": "words"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "measure"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "measure"}, {"LOWER": "of"}, {"LOWER": "damages"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "measure"}, {"LOWER": "of"}, {"LOWER": "damages"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "measure"}, {"LOWER": "of"}, {"LOWER": "indemnity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mechanically"},
            {"LOWER": "propelled"},
            {"LOWER": "invalid"},
            {"LOWER": "carriage"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mechanically"},
            {"LOWER": "propelled"},
            {"LOWER": "vehicles"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "media"}, {"LOWER": "interviews"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "media"}, {"LOWER": "representatives"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "devices"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medical"},
            {"LOWER": "disciplinary"},
            {"LOWER": "panel"},
            {"LOWER": "decision"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "evidence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "practioner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "practitioner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "practitioner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "practitioner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "practitioner"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medical"},
            {"LOWER": "practitioners"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medical"},
            {"LOWER": "practitioners"},
            {"LOWER": "tribunal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "reports"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "treatment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "treatment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "medical"}, {"LOWER": "treatment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medical"},
            {"LOWER": "treatment"},
            {"LOWER": "in"},
            {"LOWER": "another"},
            {"LOWER": "member"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medical"},
            {"LOWER": "treatment"},
            {"ORTH": ","},
            {"LOWER": "consent"},
            {"LOWER": "to"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "medicinal"}, {"LOWER": "products"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medicinal"},
            {"LOWER": "products"},
            {"LOWER": "for"},
            {"LOWER": "human"},
            {"LOWER": "use"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medicine"},
            {"ORTH": ","},
            {"LOWER": "sale"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "medicine"},
            {"ORTH": ","},
            {"LOWER": "supply"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "meeting"},
            {"LOWER": "of"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "meetings"}, {"LOWER": "of"}, {"LOWER": "directors"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "member"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "parliament"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "member"},
            {"LOWER": "of"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "member"}, {"LOWER": "of"}, {"LOWER": "mission"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "withdrawal"},
            {"LOWER": "from"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "members"}, {"LOWER": "club"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "members"},
            {"LOWER": "incapacity"},
            {"LOWER": "to"},
            {"LOWER": "sit"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "members"},
            {"LOWER": "of"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "membership"}, {"LOWER": "of"}, {"LOWER": "profession"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "membership"}, {"LOWER": "of"}, {"LOWER": "tribunal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "membership"}, {"LOWER": "policy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "membership"}, {"LOWER": "registration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "members"}, {"LOWER": "’"}, {"LOWER": "club"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "memorandum"}, {"LOWER": "of"}, {"LOWER": "association"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "memorandum"}, {"LOWER": "of"}, {"LOWER": "association"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "memorial"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "memorial"}, {"LOWER": "grave"}, {"LOWER": "stone"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "memorial"}, {"LOWER": "inscription"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "memorial"}, {"LOWER": "monument"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "memorial"},
            {"LOWER": "not"},
            {"LOWER": "conforming"},
            {"LOWER": "to"},
            {"LOWER": "churchyard"},
            {"LOWER": "rules"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "memorial"}, {"LOWER": "tablet"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mens"}, {"LOWER": "rea"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "capacity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "capacity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "disorder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "disorder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "disorder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "disorder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "element"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "health"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mental"},
            {"LOWER": "health"},
            {"LOWER": "review"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mental"}, {"LOWER": "health"}, {"LOWER": "tribunal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mental"},
            {"LOWER": "health"},
            {"LOWER": "tribunal"},
            {"LOWER": "for"},
            {"LOWER": "scotland"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "illness"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mental"}, {"LOWER": "incapacity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mercantile"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mercantile"}, {"LOWER": "list"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "merchantable"}, {"LOWER": "quality"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "merger"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "merger"},
            {"LOWER": "of"},
            {"LOWER": "parent"},
            {"LOWER": "company"},
            {"LOWER": "established"},
            {"LOWER": "in"},
            {"LOWER": "one"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "with"},
            {"LOWER": "subsidiary"},
            {"LOWER": "established"},
            {"LOWER": "in"},
            {"LOWER": "another"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "merger"}, {"ORTH": "."}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "message"},
            {"LOWER": "of"},
            {"LOWER": "menacing"},
            {"LOWER": "character"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "method"},
            {"LOWER": "of"},
            {"LOWER": "allocating"},
            {"LOWER": "allowances"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "method"}, {"LOWER": "of"}, {"LOWER": "estimation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "migrant"}, {"LOWER": "workers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "migrants"},
            {"LOWER": "from"},
            {"LOWER": "eu"},
            {"LOWER": "member"},
            {"LOWER": "states"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "military"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "military"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "military"}, {"LOWER": "law"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "military"}, {"LOWER": "service"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mind"}, {"LOWER": "sports"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mineral"}, {"LOWER": "transporter"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mines"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mines"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "minister"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "minister"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "minister"},
            {"ORTH": ","},
            {"LOWER": "determination"},
            {"LOWER": "by"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ministerial"}, {"LOWER": "directions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ministerial"}, {"LOWER": "directions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ministers"}, {"LOWER": "powers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "minor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "minors"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "miscarriage"}, {"LOWER": "of"}, {"LOWER": "justice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "miscarriage"}, {"LOWER": "of"}, {"LOWER": "justice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "misconduct"},
            {"LOWER": "in"},
            {"LOWER": "public"},
            {"LOWER": "office"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "misdescription"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "misfeasance"},
            {"LOWER": "in"},
            {"LOWER": "public"},
            {"LOWER": "office"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "misleading"},
            {"LOWER": "commercial"},
            {"LOWER": "practice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "misleading"},
            {"LOWER": "price"},
            {"LOWER": "indication"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "misrepresentation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "misrepresentations"},
            {"LOWER": "during"},
            {"LOWER": "naturalisation"},
            {"LOWER": "process"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "missing"},
            {"LOWER": "trader"},
            {"LOWER": "intra"},
            {"ORTH": "-"},
            {"LOWER": "community"},
            {"LOWER": "fraud"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mistake"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "defence"}, {"LOWER": "of"}, {"LOWER": "mistake"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "misuse"}, {"LOWER": "of"}, {"LOWER": "drugs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "misuse"}, {"LOWER": "of"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "misuse"},
            {"LOWER": "of"},
            {"LOWER": "private"},
            {"LOWER": "information"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mitigation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mitigation"}, {"LOWER": "of"}, {"LOWER": "harm"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mnaghten"}, {"LOWER": "rules"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mobile"},
            {"LOWER": "communication"},
            {"LOWER": "services"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mobile"}, {"LOWER": "homes"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mobility"}, {"LOWER": "allowance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mode"}, {"LOWER": "of"}, {"LOWER": "trial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "model"},
            {"LOWER": "code"},
            {"LOWER": "of"},
            {"LOWER": "conduct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "modification"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "modified"}, {"LOWER": "procedure"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "money"}, {"LOWER": "circulation"}, {"LOWER": "scheme"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "money"},
            {"LOWER": "had"},
            {"LOWER": "and"},
            {"LOWER": "received"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "money"}, {"LOWER": "in"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "money"},
            {"LOWER": "in"},
            {"LOWER": "court"},
            {"ORTH": ","},
            {"LOWER": "investment"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "money"}, {"LOWER": "judgment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "money"}, {"LOWER": "laundering"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "money"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "money"},
            {"LOWER": "paid"},
            {"LOWER": "under"},
            {"LOWER": "mistake"},
            {"LOWER": "of"},
            {"LOWER": "fact"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "money"},
            {"LOWER": "paid"},
            {"LOWER": "under"},
            {"LOWER": "mistake"},
            {"LOWER": "of"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "moneys"},
            {"LOWER": "had"},
            {"LOWER": "and"},
            {"LOWER": "received"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "monopolies"}, {"LOWER": "and"}, {"LOWER": "mergers"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mortgage"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mortgage"},
            {"LOWER": "backed"},
            {"LOWER": "securities"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mortgage"}, {"LOWER": "debt"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mortgage"}, {"LOWER": "debt"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mortgage"}, {"LOWER": "fraud"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mortgage"},
            {"LOWER": "to"},
            {"LOWER": "building"},
            {"LOWER": "society"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mortgaged"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mortgagees"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mortgagees"},
            {"LOWER": "power"},
            {"LOWER": "of"},
            {"LOWER": "sale"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mortgagees"}, {"LOWER": "right"}, {"LOWER": "to"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mortgagee"}, {"LOWER": "’s"}, {"LOWER": "duty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mortgagee"},
            {"LOWER": "’s"},
            {"LOWER": "right"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "admitted"},
            {"LOWER": "to"},
            {"LOWER": "hospital"},
            {"LOWER": "for"},
            {"LOWER": "induction"},
            {"LOWER": "of"},
            {"LOWER": "labour"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "alleging"},
            {"LOWER": "abusive"},
            {"LOWER": "or"},
            {"LOWER": "inappropriate"},
            {"LOWER": "behaviour"},
            {"LOWER": "by"},
            {"LOWER": "father"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "alleging"},
            {"LOWER": "domestic"},
            {"LOWER": "violence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mother"}, {"LOWER": "and"}, {"LOWER": "daughter"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "mother"}, {"LOWER": "and"}, {"LOWER": "son"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "female"},
            {"LOWER": "genital"},
            {"LOWER": "mutilation"},
            {"LOWER": "protection"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "residence"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "summary"},
            {"LOWER": "return"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "applying"},
            {"LOWER": "to"},
            {"LOWER": "discharge"},
            {"LOWER": "special"},
            {"LOWER": "guardianship"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "applying"},
            {"LOWER": "to"},
            {"LOWER": "remove"},
            {"LOWER": "child"},
            {"LOWER": "permanently"},
            {"LOWER": "from"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "disclaiming"},
            {"LOWER": "all"},
            {"LOWER": "knowledge"},
            {"LOWER": "of"},
            {"LOWER": "whereabouts"},
            {"LOWER": "of"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "making"},
            {"LOWER": "allegations"},
            {"LOWER": "against"},
            {"LOWER": "father"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "seeking"},
            {"LOWER": "injunctions"},
            {"LOWER": "and"},
            {"LOWER": "no"},
            {"LOWER": "contact"},
            {"LOWER": "orders"},
            {"LOWER": "against"},
            {"LOWER": "father"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "seeking"},
            {"LOWER": "orders"},
            {"LOWER": "for"},
            {"LOWER": "residence"},
            {"ORTH": ","},
            {"LOWER": "change"},
            {"LOWER": "of"},
            {"LOWER": "name"},
            {"LOWER": "and"},
            {"LOWER": "termination"},
            {"LOWER": "of"},
            {"LOWER": "parental"},
            {"LOWER": "responsibility"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "seeking"},
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "remove"},
            {"LOWER": "child"},
            {"LOWER": "from"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "seeking"},
            {"LOWER": "residence"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "seeking"},
            {"LOWER": "to"},
            {"LOWER": "enforce"},
            {"LOWER": "russian"},
            {"LOWER": "order"},
            {"LOWER": "in"},
            {"LOWER": "english"},
            {"LOWER": "courts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "under"},
            {"ORTH": "-"},
            {"LOWER": "age"},
            {"LOWER": "at"},
            {"LOWER": "time"},
            {"LOWER": "of"},
            {"LOWER": "forced"},
            {"LOWER": "marriage"},
            {"LOWER": "in"},
            {"LOWER": "pakistan"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "wishing"},
            {"LOWER": "to"},
            {"LOWER": "relinquish"},
            {"LOWER": "child"},
            {"LOWER": "for"},
            {"LOWER": "adoption"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mother"},
            {"LOWER": "’s"},
            {"LOWER": "subsequent"},
            {"LOWER": "high"},
            {"LOWER": "court"},
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "return"},
            {"LOWER": "order"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
            {"LOWER": "granted"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "motion"}, {"LOWER": "for"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "motion"},
            {"LOWER": "for"},
            {"LOWER": "judgment"},
            {"LOWER": "judicature"},
            {"LOWER": "act"},
            {"ORTH": ","},
            {"LOWER": "1873"},
            {"ORTH": ","},
            {"LOWER": "ss"},
            {"ORTH": "."},
            {"LOWER": "56"},
            {"ORTH": ","},
            {"LOWER": "57"},
            {"ORTH": ","},
            {"LOWER": "58"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "motion"},
            {"LOWER": "supported"},
            {"LOWER": "by"},
            {"LOWER": "affidavit"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "motion"}, {"LOWER": "to"}, {"LOWER": "remit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "motor"}, {"LOWER": "insurers"}, {"LOWER": "bureau"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "motor"},
            {"LOWER": "insurers"},
            {"LOWER": "’"},
            {"LOWER": "bureau"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "motor"}, {"LOWER": "vehicle"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "motor"}, {"LOWER": "vehicle"}, {"LOWER": "insurance"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "movables"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "move"},
            {"LOWER": "to"},
            {"LOWER": "creditors"},
            {"LOWER": "voluntary"},
            {"LOWER": "liquidation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "move"},
            {"LOWER": "to"},
            {"LOWER": "creditors"},
            {"LOWER": "’"},
            {"LOWER": "voluntary"},
            {"LOWER": "liquidation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "move"},
            {"LOWER": "to"},
            {"LOWER": "creditor"},
            {"LOWER": "’s"},
            {"LOWER": "voluntary"},
            {"LOWER": "liquidation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "multi"},
            {"LOWER": "-"},
            {"LOWER": "functional"},
            {"LOWER": "media"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "multiple"},
            {"LOWER": "actions"},
            {"LOWER": "raising"},
            {"LOWER": "similar"},
            {"LOWER": "issues"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "multiple"},
            {"LOWER": "entry"},
            {"LOWER": "visit"},
            {"LOWER": "visa"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "multiple"},
            {"LOWER": "judgments"},
            {"LOWER": "from"},
            {"LOWER": "preliminary"},
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
            {"LOWER": "ex"},
            {"LOWER": "tempore"},
            {"LOWER": "judgment"},
            {"LOWER": "in"},
            {"LOWER": "preliminary"},
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
            {"LOWER": "in"},
            {"LOWER": "care"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "multiple"}, {"LOWER": "occupation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "municipal"}, {"LOWER": "cemetery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "murder"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "murder"}, {"LOWER": "/"}, {"LOWER": "manslaughter"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mussel"}, {"LOWER": "fishing"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mutual"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
            {"LOWER": "and"},
            {"LOWER": "assistance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mutual"},
            {"LOWER": "mistake"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "essential"},
            {"LOWER": "matter"},
            {"LOWER": "of"},
            {"LOWER": "fact"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "mutual"},
            {"LOWER": "recognition"},
            {"LOWER": "of"},
            {"LOWER": "qualifications"},
            {"LOWER": "in"},
            {"LOWER": "the"},
            {"LOWER": "field"},
            {"LOWER": "of"},
            {"LOWER": "architecture"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "mutual"}, {"LOWER": "wills"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mv"}, {"LOWER": "alkyon"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "mv"}, {"LOWER": "renos"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "myrto"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nagasaki"}, {"LOWER": "spirit"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "name"}, {"LOWER": "of"}, {"LOWER": "deceased"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "naming"}, {"LOWER": "of"}, {"LOWER": "streets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "national"}, {"LOWER": "coal"}, {"LOWER": "board"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "national"}, {"LOWER": "health"}, {"LOWER": "service"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "national"}, {"LOWER": "health"}, {"LOWER": "service"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "national"}, {"LOWER": "health"}, {"LOWER": "service"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "national"},
            {"LOWER": "health"},
            {"LOWER": "service"},
            {"LOWER": "trust"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "national"}, {"LOWER": "heritage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "national"}, {"LOWER": "insurance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "national"},
            {"LOWER": "insurance"},
            {"LOWER": "contributions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "national"},
            {"LOWER": "legislation"},
            {"LOWER": "concerning"},
            {"LOWER": "corporate"},
            {"LOWER": "income"},
            {"LOWER": "tax"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "national"},
            {"LOWER": "legislation"},
            {"LOWER": "recognising"},
            {"LOWER": "equivalence"},
            {"LOWER": "of"},
            {"LOWER": "qualifications"},
            {"LOWER": "in"},
            {"LOWER": "architecture"},
            {"LOWER": "and"},
            {"LOWER": "civil"},
            {"LOWER": "engineering"},
            {"ORTH": ","},
            {"LOWER": "but"},
            {"LOWER": "reserving"},
            {"LOWER": "work"},
            {"LOWER": "on"},
            {"LOWER": "classified"},
            {"LOWER": "heritage"},
            {"LOWER": "buildings"},
            {"LOWER": "to"},
            {"LOWER": "architects"},
            {"LOWER": "with"},
            {"LOWER": "special"},
            {"LOWER": "qualifications"},
            {"LOWER": "in"},
            {"LOWER": "field"},
            {"LOWER": "of"},
            {"LOWER": "cultural"},
            {"LOWER": "assets"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "national"}, {"LOWER": "minimum"}, {"LOWER": "wage"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "national"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "economic"},
            {"LOWER": "community"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "national"}, {"LOWER": "park"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "national"},
            {"LOWER": "planning"},
            {"LOWER": "policy"},
            {"LOWER": "statement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "national"}, {"LOWER": "policy"}, {"LOWER": "statement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "national"}, {"LOWER": "security"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nationality"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nationality"},
            {"LOWER": "as"},
            {"LOWER": "basis"},
            {"LOWER": "of"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nationally"},
            {"LOWER": "significant"},
            {"LOWER": "infrastructure"},
            {"LOWER": "project"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "native"}, {"LOWER": "land"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "natural"}, {"LOWER": "gas"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "natural"}, {"LOWER": "habitats"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "natural"}, {"LOWER": "justice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "natural"}, {"LOWER": "justice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "natural"}, {"LOWER": "justice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "natural"}, {"LOWER": "justice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "natural"}, {"LOWER": "parent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "naturalisation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "nature"}, {"LOWER": "of"}, {"LOWER": "duty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "nature"}, {"LOWER": "of"}, {"LOWER": "right"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nautical"},
            {"LOWER": "and"},
            {"LOWER": "other"},
            {"LOWER": "assessors"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "nautical"}, {"LOWER": "assessors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "navigation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "necessary"},
            {"LOWER": "goods"},
            {"LOWER": "or"},
            {"LOWER": "services"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "necessary"}, {"LOWER": "obligation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "necessity"}, {"LOWER": "as"}, {"LOWER": "defence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "negative"}, {"LOWER": "covenant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "negative"}, {"LOWER": "declaration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "neglect"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "negligence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "negligent"}, {"LOWER": "advice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "negligent"}, {"LOWER": "misstatement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "negligent"},
            {"LOWER": "performance"},
            {"LOWER": "of"},
            {"LOWER": "duty"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "negligent"}, {"LOWER": "valuation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "negotiations"},
            {"LOWER": "to"},
            {"LOWER": "reach"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "neighbourhood"},
            {"LOWER": "area"},
            {"LOWER": "designation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "neighbourhood"},
            {"LOWER": "area"},
            {"LOWER": "designation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "neighbourhood"},
            {"LOWER": "development"},
            {"LOWER": "plan"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "neighbourhood"}, {"LOWER": "plan"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nema"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nemo"},
            {"LOWER": "debet"},
            {"LOWER": "bis"},
            {"LOWER": "vexari"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "nervous"}, {"LOWER": "shock"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "new"},
            {"LOWER": "claim"},
            {"LOWER": "in"},
            {"LOWER": "pending"},
            {"LOWER": "action"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "new"},
            {"LOWER": "claims"},
            {"LOWER": "in"},
            {"LOWER": "pending"},
            {"LOWER": "action"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "new"},
            {"LOWER": "point"},
            {"LOWER": "raised"},
            {"LOWER": "on"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "new"},
            {"LOWER": "points"},
            {"LOWER": "on"},
            {"LOWER": "appeal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "new"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "new"}, {"LOWER": "south"}, {"LOWER": "wales"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "new"},
            {"LOWER": "york"},
            {"LOWER": "convention"},
            {"LOWER": "award"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "new"}, {"LOWER": "zealand"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "newspaper"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "newspaper"}, {"LOWER": "article"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nicholas"}, {"LOWER": "h"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "niedersachsen"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nightclub"}, {"LOWER": "owner"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ninemia"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nmc"},
            {"LOWER": "making"},
            {"LOWER": "findings"},
            {"LOWER": "of"},
            {"LOWER": "misconduct"},
            {"LOWER": "against"},
            {"LOWER": "registered"},
            {"LOWER": "nurses"},
            {"LOWER": "managing"},
            {"LOWER": "care"},
            {"LOWER": "home"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "no"}, {"LOWER": "relief"}, {"LOWER": "granted"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "no"},
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "review"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "noise"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "noise"}, {"LOWER": "policy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "accidental"}, {"LOWER": "injury"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "est"}, {"LOWER": "factum"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "eu"}, {"LOWER": "national"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "transfer"}, {"LOWER": "taxes"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "accidental"},
            {"LOWER": "injury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "commonwealth"},
            {"LOWER": "nationals"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "-"}, {"LOWER": "compliance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "contractual"},
            {"LOWER": "obligation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "-"}, {"LOWER": "disclosure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "evidence"},
            {"LOWER": "of"},
            {"LOWER": "police"},
            {"LOWER": "informer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "relevant"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "-"}, {"LOWER": "discrimination"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "domestic"},
            {"LOWER": "premises"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "domestic"},
            {"LOWER": "rates"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "eu"},
            {"LOWER": "national"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "national"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "exclusive"},
            {"LOWER": "trade"},
            {"LOWER": "description"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "-"}, {"LOWER": "justiciability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "justiciable"},
            {"LOWER": "issue"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "-"}, {"LOWER": "party"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "party"},
            {"LOWER": "costs"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "-"}, {"LOWER": "patrial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "payment"},
            {"LOWER": "of"},
            {"LOWER": "confiscation"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "payment"},
            {"LOWER": "of"},
            {"LOWER": "fines"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "payment"},
            {"LOWER": "of"},
            {"LOWER": "maintenance"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "proliferation"},
            {"LOWER": "of"},
            {"LOWER": "nuclear"},
            {"LOWER": "weapons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "non"}, {"LOWER": "-"}, {"LOWER": "repair"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "resident"},
            {"LOWER": "status"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "secure"},
            {"LOWER": "tenancies"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "tidal"},
            {"LOWER": "waters"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "witness"},
            {"LOWER": "list"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "nordglimt"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nordkap"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "northern"}, {"LOWER": "ireland"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "notice"}, {"LOWER": "of"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "lodged"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "time"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "time"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "notice"}, {"LOWER": "of"}, {"LOWER": "appearance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "of"},
            {"LOWER": "creditors"},
            {"LOWER": "meeting"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "notice"}, {"LOWER": "of"}, {"LOWER": "discontinuance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "of"},
            {"LOWER": "open"},
            {"LOWER": "competition"},
            {"LOWER": "for"},
            {"LOWER": "recruitment"},
            {"LOWER": "of"},
            {"LOWER": "administrators"},
            {"LOWER": "and"},
            {"LOWER": "assistants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "notice"}, {"LOWER": "of"}, {"LOWER": "termination"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "notice"}, {"LOWER": "pay"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "terminating"},
            {"LOWER": "tenancy"},
            {"LOWER": "given"},
            {"LOWER": "by"},
            {"LOWER": "one"},
            {"LOWER": "joint"},
            {"LOWER": "tenant"},
            {"LOWER": "without"},
            {"LOWER": "authority"},
            {"LOWER": "of"},
            {"LOWER": "other"},
            {"LOWER": "joint"},
            {"LOWER": "tenant"},
            {"LOWER": "effective"},
            {"LOWER": "to"},
            {"LOWER": "terminate"},
            {"LOWER": "tenancy"},
            {"LOWER": "entirely"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "to"},
            {"LOWER": "admit"},
            {"LOWER": "facts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "to"},
            {"LOWER": "break"},
            {"LOWER": "tenancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "notice"}, {"LOWER": "to"}, {"LOWER": "determine"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notice"},
            {"LOWER": "to"},
            {"LOWER": "produce"},
            {"LOWER": "evidence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "notice"}, {"LOWER": "to"}, {"LOWER": "quit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "notices"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"LOWER": "and"},
            {"LOWER": "applications"},
            {"LOWER": "for"},
            {"LOWER": "leave"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "notification"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "notification"}, {"LOWER": "requirements"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "novus"},
            {"LOWER": "actus"},
            {"LOWER": "interveniens"},
            {"LOWER": "/break"},
            {"LOWER": "in"},
            {"LOWER": "chain"},
            {"LOWER": "of"},
            {"LOWER": "causation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "nuclear"}, {"LOWER": "installations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nuisance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nullity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nurse"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nurse"},
            {"LOWER": "struck"},
            {"LOWER": "off"},
            {"LOWER": "register"},
            {"LOWER": "for"},
            {"LOWER": "incompetence"},
            {"LOWER": "and"},
            {"LOWER": "dishonesty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nursing"},
            {"LOWER": "and"},
            {"LOWER": "midwifery"},
            {"LOWER": "council"},
            {"LOWER": "conduct"},
            {"LOWER": "panel"},
            {"LOWER": "hearing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "nursing"},
            {"LOWER": "and"},
            {"LOWER": "midwifery"},
            {"LOWER": "council"},
            {"LOWER": "conduct"},
            {"LOWER": "panel"},
            {"LOWER": "hearing"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "nursing"}, {"LOWER": "care"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "nursing"}, {"LOWER": "care"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "oath"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "administration"}, {"LOWER": "of"}, {"LOWER": "oath"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "obiter"}, {"LOWER": "dicta"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "object"}, {"LOWER": "of"}, {"LOWER": "act"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "objection"}, {"LOWER": "at"}, {"LOWER": "audit"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "objective"}, {"LOWER": "test"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obligation"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "railway"},
            {"LOWER": "infrastructure"},
            {"LOWER": "manager"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "railway"},
            {"LOWER": "undertakings"},
            {"LOWER": "in"},
            {"LOWER": "real"},
            {"LOWER": "time"},
            {"LOWER": "with"},
            {"LOWER": "all"},
            {"LOWER": "information"},
            {"LOWER": "concerning"},
            {"LOWER": "train"},
            {"LOWER": "movements"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "obligation"}, {"LOWER": "on"}, {"LOWER": "claimant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "include"},
            {"LOWER": "additional"},
            {"LOWER": "costs"},
            {"LOWER": "necessarily"},
            {"LOWER": "incurred"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "install"},
            {"LOWER": "recording"},
            {"LOWER": "equipment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "obligation"}, {"LOWER": "to"}, {"LOWER": "refer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "state"},
            {"LOWER": "reasons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "obligations"}, {"LOWER": "to"}, {"LOWER": "borrower"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "obscene"}, {"LOWER": "libel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "obscene"}, {"LOWER": "publication"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "obscenity"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obtaining"},
            {"LOWER": "a"},
            {"LOWER": "pecuniary"},
            {"LOWER": "advantage"},
            {"LOWER": "by"},
            {"LOWER": "deception"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obtaining"},
            {"LOWER": "further"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obtaining"},
            {"LOWER": "money"},
            {"LOWER": "transfer"},
            {"LOWER": "by"},
            {"LOWER": "deception"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obtaining"},
            {"LOWER": "pecuniary"},
            {"LOWER": "advantage"},
            {"LOWER": "by"},
            {"LOWER": "deception"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "obtaining"},
            {"LOWER": "property"},
            {"LOWER": "by"},
            {"LOWER": "deception"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "obvious"}, {"LOWER": "drafting"}, {"LOWER": "error"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "occasional"}, {"LOWER": "permission"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "occupier"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "occupation"},
            {"LOWER": "by"},
            {"LOWER": "government"},
            {"LOWER": "department"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "occupation"},
            {"LOWER": "for"},
            {"LOWER": "purpose"},
            {"LOWER": "of"},
            {"LOWER": "tenants"},
            {"LOWER": "business"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "occupation"},
            {"ORTH": ","},
            {"LOWER": "right"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "occupier"}, {"LOWER": "of"}, {"LOWER": "premises"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "occupiers"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "occupiers"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "occupiers"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "occupiers"}, {"LOWER": "liability"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "occupiers"},
            {"LOWER": "of"},
            {"LOWER": "adjoining"},
            {"LOWER": "premises"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "occupiers"}, {"LOWER": "’"}, {"LOWER": "liability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "occupiers"}, {"LOWER": "’"}, {"LOWER": "liability"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ocean"}, {"LOWER": "blessing"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offence"}, {"LOWER": "committed"}, {"LOWER": "abroad"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offence"},
            {"LOWER": "triable"},
            {"LOWER": "either"},
            {"LOWER": "way"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "offences"}, {"LOWER": "against"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offences"}, {"LOWER": "against"}, {"LOWER": "byelaws"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offences"},
            {"LOWER": "against"},
            {"LOWER": "the"},
            {"LOWER": "person"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offences"},
            {"LOWER": "against"},
            {"LOWER": "the"},
            {"LOWER": "person"},
            {"LOWER": "act"},
            {"LOWER": "1861"},
            {"ORTH": ","},
            {"LOWER": "s"},
            {"LOWER": "18"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offences"},
            {"LOWER": "committed"},
            {"LOWER": "many"},
            {"LOWER": "years"},
            {"LOWER": "previously"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offences"},
            {"LOWER": "created"},
            {"LOWER": "by"},
            {"LOWER": "statute"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offences"},
            {"LOWER": "triable"},
            {"LOWER": "either"},
            {"LOWER": "way"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "offensive"}, {"LOWER": "weapon"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offer"}, {"LOWER": "and"}, {"LOWER": "acceptance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offer"}, {"LOWER": "of"}, {"LOWER": "amends"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offer"}, {"LOWER": "of"}, {"LOWER": "settlement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offer"},
            {"LOWER": "of"},
            {"LOWER": "suitable"},
            {"LOWER": "employment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "offer"},
            {"LOWER": "to"},
            {"LOWER": "grant"},
            {"LOWER": "tenancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "offer"}, {"LOWER": "to"}, {"LOWER": "settle"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "office"}, {"LOWER": "for"}, {"LOWER": "harmonisation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "office"}, {"LOWER": "holder"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "office"}, {"LOWER": "of"}, {"LOWER": "communications"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "office"},
            {"LOWER": "of"},
            {"LOWER": "president"},
            {"LOWER": "and"},
            {"LOWER": "vice"},
            {"LOWER": "-"},
            {"LOWER": "president"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "office"}, {"ORTH": "-"}, {"LOWER": "holder"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "officer"},
            {"LOWER": "of"},
            {"LOWER": "corporate"},
            {"LOWER": "body"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "officer"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "officer"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "official"}, {"LOWER": "receiver"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "official"}, {"LOWER": "secrets"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "official"},
            {"LOWER": "solicitor"},
            {"LOWER": "as"},
            {"LOWER": "guardian"},
            {"LOWER": "ad"},
            {"LOWER": "litem"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "officials"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "oil"}, {"LOWER": "refinery"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "old"},
            {"LOWER": "age"},
            {"LOWER": "and"},
            {"LOWER": "survivors"},
            {"LOWER": "’"},
            {"LOWER": "insurance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "old"},
            {"ORTH": "-"},
            {"LOWER": "age"},
            {"LOWER": "benefit"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ombudsman"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ombudsman"}, {"LOWER": "scheme"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "omission"},
            {"LOWER": "of"},
            {"LOWER": "material"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "one"},
            {"LOWER": "of"},
            {"LOWER": "several"},
            {"LOWER": "executors"},
            {"LOWER": "a"},
            {"LOWER": "lunatic"},
            {"ORTH": "."},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "online"}, {"LOWER": "marketplace"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "open"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "open"}, {"LOWER": "justice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "operating"}, {"LOWER": "licences"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "operator"}, {"LOWER": "’s"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "opportunity"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "heard"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "opportunity"},
            {"LOWER": "to"},
            {"LOWER": "meet"},
            {"LOWER": "charge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "opportunity"},
            {"LOWER": "to"},
            {"LOWER": "meet"},
            {"LOWER": "charge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "opportunity"},
            {"LOWER": "to"},
            {"LOWER": "meet"},
            {"LOWER": "charge"},
            {"LOWER": "licensing"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "opposition"}, {"LOWER": "proceedings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "oppression"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "optician"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "opticians"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "option"}, {"LOWER": "to"}, {"LOWER": "purchase"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "oral"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "oral"}, {"LOWER": "communication"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "oral"},
            {"LOWER": "examination"},
            {"LOWER": "of"},
            {"LOWER": "witnesses"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "oral"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "order"}, {"LOWER": "absolute"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "disqualifying"},
            {"LOWER": "defendant"},
            {"LOWER": "from"},
            {"LOWER": "working"},
            {"LOWER": "with"},
            {"LOWER": "children"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "order"}, {"LOWER": "for"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "adoption"},
            {"LOWER": "placement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "costs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "costs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "discovery"},
            {"LOWER": "or"},
            {"LOWER": "inspection"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "examination"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "expenses"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "first"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "removal"},
            {"LOWER": "to"},
            {"LOWER": "congo"},
            {"LOWER": "following"},
            {"LOWER": "failed"},
            {"LOWER": "asylum"},
            {"LOWER": "claim"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "no"},
            {"LOWER": "further"},
            {"LOWER": "hearing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "obtaining"},
            {"LOWER": "evidence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "payment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "possession"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "retrial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "retrial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "for"}, {"LOWER": "sale"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "in"}, {"LOWER": "council"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "made"},
            {"LOWER": "forbidding"},
            {"LOWER": "disclosure"},
            {"LOWER": "of"},
            {"LOWER": "information"},
            {"LOWER": "in"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "of"},
            {"LOWER": "application"},
            {"LOWER": "of"},
            {"LOWER": "assets"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "child"},
            {"LOWER": "subject"},
            {"LOWER": "to"},
            {"LOWER": "care"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "order"},
            {"LOWER": "restoring"},
            {"LOWER": "parties"},
            {"LOWER": "position"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "to"}, {"LOWER": "pay"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "order"}, {"LOWER": "to"}, {"LOWER": "proceed"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "in"},
            {"LOWER": "best"},
            {"LOWER": "interests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "in"},
            {"LOWER": "respect"},
            {"LOWER": "of"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "with"},
            {"LOWER": "regard"},
            {"LOWER": "to"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "with"},
            {"LOWER": "respect"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "with"},
            {"LOWER": "respect"},
            {"LOWER": "to"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "with"},
            {"LOWER": "respect"},
            {"LOWER": "to"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "orders"},
            {"LOWER": "with"},
            {"LOWER": "respect"},
            {"LOWER": "to"},
            {"LOWER": "children"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ordinary"}, {"LOWER": "residence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "organisation"},
            {"LOWER": "of"},
            {"LOWER": "working"},
            {"LOWER": "time"},
            {"LOWER": "of"},
            {"LOWER": "doctors"},
            {"LOWER": "in"},
            {"LOWER": "training"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "original"}, {"LOWER": "idea"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "original"},
            {"LOWER": "scientific"},
            {"LOWER": "research"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "original"},
            {"LOWER": "set"},
            {"ORTH": "-"},
            {"LOWER": "off"},
            {"LOWER": "or"},
            {"LOWER": "counterclaim"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "originating"}, {"LOWER": "process"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ornament"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ostensible"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "osteopath"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "osteopath"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "out"}, {"LOWER": "of"}, {"LOWER": "control"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "outline"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "outraging"}, {"LOWER": "public"}, {"LOWER": "decency"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "overcharging"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "overlapping"}, {"LOWER": "claims"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "overpayment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "overpayment"}, {"LOWER": "of"}, {"LOWER": "benefit"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "overriding"}, {"LOWER": "interest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "oversea"}, {"LOWER": "companies"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "overseas"}, {"LOWER": "company"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "overseas"}, {"LOWER": "freezing"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "overseas"}, {"LOWER": "visitors"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "overstayer"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "overstayer"},
            {"LOWER": "intending"},
            {"LOWER": "to"},
            {"LOWER": "marry"},
            {"LOWER": "uk"},
            {"LOWER": "citizen"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "overstayers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "overtime"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "overview"},
            {"LOWER": "report"},
            {"LOWER": "produced"},
            {"LOWER": "by"},
            {"LOWER": "local"},
            {"LOWER": "safeguarding"},
            {"LOWER": "children"},
            {"LOWER": "board"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "own"}, {"LOWER": "resources"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "owner"},
            {"LOWER": "of"},
            {"LOWER": "adjoining"},
            {"LOWER": "terraced"},
            {"LOWER": "property"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "owners"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ownership"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "p.a.y.e"}, {"ORTH": "."}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "package"}, {"LOWER": "tour"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "paid"}, {"LOWER": "annual"}, {"LOWER": "leave"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "paid"}, {"LOWER": "annual"}, {"LOWER": "leave"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pain"},
            {"ORTH": ","},
            {"LOWER": "suffering"},
            {"LOWER": "and"},
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "amenity"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "panel"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "panel"},
            {"LOWER": "finding"},
            {"LOWER": "nurse"},
            {"LOWER": "’s"},
            {"LOWER": "fitness"},
            {"LOWER": "to"},
            {"LOWER": "practise"},
            {"LOWER": "impaired"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "panel"},
            {"LOWER": "on"},
            {"LOWER": "take"},
            {"ORTH": "-"},
            {"LOWER": "overs"},
            {"LOWER": "and"},
            {"LOWER": "mergers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "panel"},
            {"LOWER": "striking"},
            {"LOWER": "claimant"},
            {"LOWER": "off"},
            {"LOWER": "register"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "parallel"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pardon"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "parens"}, {"LOWER": "patriae"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "parent"}, {"LOWER": "company"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parent"},
            {"LOWER": "pleading"},
            {"LOWER": "guilty"},
            {"LOWER": "to"},
            {"LOWER": "murder"},
            {"LOWER": "of"},
            {"LOWER": "child"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "parentage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "parental"}, {"LOWER": "contact"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "parental"}, {"LOWER": "leave"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "parental"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "parental"}, {"LOWER": "responsibility"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "parental"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parents"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "oppose"},
            {"LOWER": "adoption"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parents"},
            {"LOWER": "claiming"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "human"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parents"},
            {"LOWER": "of"},
            {"LOWER": "deceased"},
            {"LOWER": "child"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parents"},
            {"LOWER": "suspected"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "abuse"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parents"},
            {"LOWER": "unable"},
            {"LOWER": "to"},
            {"LOWER": "consent"},
            {"LOWER": "to"},
            {"LOWER": "child"},
            {"LOWER": "’s"},
            {"LOWER": "blood"},
            {"LOWER": "transfusion"},
            {"LOWER": "for"},
            {"LOWER": "religious"},
            {"LOWER": "reasons"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "parking"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "parliament"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "parliament"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "2002/21"},
            {"LOWER": "/"},
            {"LOWER": "ec"},
            {"ORTH": ","},
            {"LOWER": "arts"},
            {"LOWER": "7"},
            {"ORTH": ","},
            {"LOWER": "20"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "2002/22"},
            {"LOWER": "/"},
            {"LOWER": "ec"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "28"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "2003/88"},
            {"LOWER": "/"},
            {"LOWER": "ec"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "7"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "2003/88"},
            {"LOWER": "/"},
            {"LOWER": "ec"},
            {"ORTH": ","},
            {"LOWER": "arts"},
            {"LOWER": "2"},
            {"ORTH": ","},
            {"LOWER": "15"},
            {"ORTH": ","},
            {"LOWER": "17(3)(iii"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "2008/94"},
            {"LOWER": "/"},
            {"LOWER": "ec"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "8"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "2011/36"},
            {"LOWER": "/"},
            {"LOWER": "eu"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "11"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliament"},
            {"LOWER": "and"},
            {"LOWER": "council"},
            {"LOWER": "regulation"},
            {"ORTH": "("},
            {"LOWER": "eu"},
            {"ORTH": ")"},
            {"LOWER": "no"},
            {"LOWER": "604/2013"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "27"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliamentary"},
            {"LOWER": "advisory"},
            {"LOWER": "body"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parliamentary"},
            {"LOWER": "commissioner"},
            {"LOWER": "for"},
            {"LOWER": "standards"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "parole"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "part"}, {"LOWER": "20"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "part"},
            {"LOWER": "time"},
            {"LOWER": "fee"},
            {"ORTH": "-"},
            {"LOWER": "paid"},
            {"LOWER": "judges"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "part"},
            {"ORTH": "-"},
            {"LOWER": "time"},
            {"LOWER": "fee"},
            {"ORTH": "-"},
            {"LOWER": "paid"},
            {"LOWER": "judge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "part"},
            {"ORTH": "-"},
            {"LOWER": "time"},
            {"LOWER": "worker"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "part"},
            {"ORTH": "-"},
            {"LOWER": "time"},
            {"LOWER": "worker"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "partial"},
            {"LOWER": "defence"},
            {"LOWER": "to"},
            {"LOWER": "murder"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "partial"}, {"LOWER": "incapacity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "partial"}, {"LOWER": "success"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "participants"},
            {"LOWER": "in"},
            {"LOWER": "criminal"},
            {"LOWER": "offences"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "participants"},
            {"LOWER": "in"},
            {"LOWER": "criminal"},
            {"LOWER": "offences"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "particular"}, {"LOWER": "disadvantage"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "particular"}, {"LOWER": "social"}, {"LOWER": "group"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "particulars"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "particulars"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parties"},
            {"LOWER": "entering"},
            {"LOWER": "into"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parties"},
            {"LOWER": "entering"},
            {"LOWER": "into"},
            {"LOWER": "compromise"},
            {"LOWER": "agreement"},
            {"LOWER": "by"},
            {"LOWER": "which"},
            {"LOWER": "first"},
            {"LOWER": "defendant"},
            {"LOWER": "agreeing"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "residential"},
            {"LOWER": "accommodation"},
            {"LOWER": "to"},
            {"LOWER": "claimant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parties"},
            {"LOWER": "entering"},
            {"LOWER": "into"},
            {"LOWER": "licence"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "parties"}, {"LOWER": "to"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "partner"},
            {"LOWER": "of"},
            {"LOWER": "limited"},
            {"LOWER": "liability"},
            {"LOWER": "partnership"},
            {"LOWER": "expelled"},
            {"LOWER": "from"},
            {"LOWER": "partnership"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "partnership"},
            {"LOWER": "acting"},
            {"LOWER": "on"},
            {"LOWER": "own"},
            {"LOWER": "account"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "partnership"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "parts"},
            {"LOWER": "a"},
            {"ORTH": ","},
            {"LOWER": "b"},
            {"ORTH": ","},
            {"LOWER": "c"},
            {"ORTH": ","},
            {"LOWER": "d."},
            {"LOWER": "and"},
            {"LOWER": "e"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "party"}, {"LOWER": "to"}, {"LOWER": "proceedings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "party"}, {"LOWER": "wall"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "party"}, {"LOWER": "wall"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "partys"}, {"LOWER": "documents"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "passage"}, {"LOWER": "of"}, {"LOWER": "time"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "passing"}, {"LOWER": "of"}, {"LOWER": "property"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "passing"}, {"LOWER": "off"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "passing"}, {"LOWER": "off"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "passing"}, {"LOWER": "off"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "passport"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "passport"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pastoral"}, {"LOWER": "scheme"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "patent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "patent"}, {"LOWER": "application"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "patent"},
            {"LOWER": "attorney"},
            {"LOWER": "litigators"},
            {"LOWER": "having"},
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "audience"},
            {"LOWER": "to"},
            {"LOWER": "conduct"},
            {"LOWER": "intellectual"},
            {"LOWER": "property"},
            {"LOWER": "litigation"},
            {"LOWER": "in"},
            {"LOWER": "high"},
            {"LOWER": "court"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "patentability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "patents"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "patents"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "paternity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "patient"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "patient"},
            {"LOWER": "in"},
            {"LOWER": "minimally"},
            {"LOWER": "conscious"},
            {"LOWER": "state"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "patient"}, {"LOWER": "information"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "patient"}, {"LOWER": "information"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "patient"}, {"LOWER": "lacking"}, {"LOWER": "capacity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "patient"},
            {"LOWER": "lacking"},
            {"LOWER": "capacity"},
            {"LOWER": "to"},
            {"LOWER": "take"},
            {"LOWER": "decisions"},
            {"LOWER": "on"},
            {"LOWER": "medical"},
            {"LOWER": "treatment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "patients"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "patrials"}, {"LOWER": "wife"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pay"},
            {"ORTH": ","},
            {"LOWER": "hours"},
            {"LOWER": "and"},
            {"LOWER": "holidays"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "payment"}, {"LOWER": "by"}, {"LOWER": "insurers"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "payment"},
            {"LOWER": "for"},
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "dental"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "payment"},
            {"LOWER": "in"},
            {"LOWER": "lieu"},
            {"LOWER": "of"},
            {"LOWER": "remuneration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "payment"}, {"LOWER": "into"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "payment"}, {"LOWER": "of"}, {"LOWER": "fees"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "payment"}, {"LOWER": "of"}, {"LOWER": "grant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "payment"}, {"LOWER": "or"}, {"LOWER": "reward"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "payment"}, {"LOWER": "out"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "payment"}, {"LOWER": "out"}, {"LOWER": "application"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "payment"}, {"LOWER": "services"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "payment"}, {"LOWER": "without"}, {"LOWER": "authority"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "payment"},
            {"ORTH": ","},
            {"LOWER": "calculation"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "payments"},
            {"LOWER": "in"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "fiduciary"},
            {"LOWER": "duty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "peaceful"},
            {"LOWER": "enjoyment"},
            {"LOWER": "of"},
            {"LOWER": "possessions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "peaceful"},
            {"LOWER": "enjoyment"},
            {"LOWER": "of"},
            {"LOWER": "possessions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pedlar"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "peerage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "penal"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "penalties"}, {"LOWER": "and"}, {"LOWER": "injunction"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "penalty"}, {"LOWER": "clause"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "pending"}, {"LOWER": "actions"}, {"LOWER": "register"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "pending"}, {"LOWER": "land"}, {"LOWER": "action"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pending"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pension"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "pension"}, {"LOWER": "protection"}, {"LOWER": "fund"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pension"}, {"LOWER": "scheme"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pension"}, {"LOWER": "scheme"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "pension"}, {"LOWER": "sharing"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pensions"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pensions"}, {"LOWER": "ombudsman"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pensions"}, {"LOWER": "regulator"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "per"}, {"LOWER": "incuriam"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "per"}, {"LOWER": "rem"}, {"LOWER": "judicatam"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "per"}, {"LOWER": "rem"}, {"LOWER": "judicatem"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "peremptory"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "performance"}, {"LOWER": "bond"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "performance"}, {"LOWER": "bonds"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "performance"}, {"LOWER": "guarantee"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "performance"},
            {"LOWER": "of"},
            {"LOWER": "existing"},
            {"LOWER": "duty"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "performing"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "period"}, {"LOWER": "of"}, {"LOWER": "letting"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "period"}, {"LOWER": "of"}, {"LOWER": "limitation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "period"}, {"LOWER": "of"}, {"LOWER": "tenancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "periodical"}, {"LOWER": "payments"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "periodical"},
            {"LOWER": "payments"},
            {"LOWER": "order"},
            {"LOWER": "and"},
            {"LOWER": "lump"},
            {"LOWER": "sum"},
            {"LOWER": "award"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "perjury"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "permanent"}, {"LOWER": "injury"}, {"LOWER": "benefit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permanent"},
            {"LOWER": "injury"},
            {"LOWER": "benefit"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "permanent"}, {"LOWER": "removal"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "extension"},
            {"LOWER": "of"},
            {"LOWER": "subsisting"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "moving"},
            {"LOWER": "brow"},
            {"LOWER": "bridging"},
            {"LOWER": "at"},
            {"LOWER": "barge"},
            {"LOWER": "moorings"},
            {"LOWER": "in"},
            {"LOWER": "river"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "granted"},
            {"LOWER": "for"},
            {"LOWER": "wind"},
            {"LOWER": "turbine"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "permission"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "permission"}, {"LOWER": "to"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "permission"}, {"LOWER": "to"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "apply"},
            {"LOWER": "for"},
            {"LOWER": "adoption"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "bring"},
            {"LOWER": "claim"},
            {"LOWER": "in"},
            {"LOWER": "upper"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "bring"},
            {"LOWER": "committal"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "bring"},
            {"LOWER": "or"},
            {"LOWER": "continue"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "bring"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "invoke"},
            {"LOWER": "inherent"},
            {"LOWER": "jurisdiction"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "order"},
            {"LOWER": "aside"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "permitted"}, {"LOWER": "development"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "permitted"}, {"LOWER": "display"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "person"},
            {"LOWER": "deemed"},
            {"LOWER": "to"},
            {"LOWER": "belong"},
            {"LOWER": "to"},
            {"LOWER": "bermuda"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "person"}, {"LOWER": "from"}, {"LOWER": "abroad"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "person"},
            {"LOWER": "held"},
            {"LOWER": "out"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "partner"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "person"}, {"LOWER": "lacking"}, {"LOWER": "capacity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "person"},
            {"LOWER": "lacking"},
            {"LOWER": "capacity"},
            {"LOWER": "to"},
            {"LOWER": "litigate"},
            {"LOWER": "and"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "decisions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "person"},
            {"LOWER": "surrendered"},
            {"LOWER": "by"},
            {"LOWER": "foreign"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "person"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "deported"},
            {"LOWER": "on"},
            {"LOWER": "national"},
            {"LOWER": "security"},
            {"LOWER": "grounds"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "person"}, {"LOWER": "under"}, {"LOWER": "disability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "person"}, {"LOWER": "under"}, {"LOWER": "disability"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "person"},
            {"LOWER": "“"},
            {"LOWER": "aggrieved"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "personal"}, {"LOWER": "correspondence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "personal"}, {"LOWER": "data"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "personal"}, {"LOWER": "injuries"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "personal"}, {"LOWER": "injuries"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "personal"}, {"LOWER": "injury"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "personal"}, {"LOWER": "injury"}, {"LOWER": "cases"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "personal"}, {"LOWER": "injury"}, {"LOWER": "claims"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "personal"},
            {"LOWER": "injury"},
            {"LOWER": "damages"},
            {"LOWER": "award"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "personal"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "personal"}, {"LOWER": "occupation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "personal"},
            {"LOWER": "protective"},
            {"LOWER": "equipment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "personal"}, {"LOWER": "relief"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "personal"}, {"LOWER": "representative"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "personal"}, {"LOWER": "representative"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "personal"},
            {"LOWER": "representatives"},
            {"LOWER": "liability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "personal"},
            {"LOWER": "service"},
            {"LOWER": "companies"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "persons"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "persons"},
            {"LOWER": "carrying"},
            {"LOWER": "on"},
            {"LOWER": "business"},
            {"LOWER": "together"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "persons"}, {"LOWER": "in"}, {"LOWER": "custody"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "persons"},
            {"LOWER": "in"},
            {"LOWER": "need"},
            {"LOWER": "of"},
            {"LOWER": "care"},
            {"LOWER": "and"},
            {"LOWER": "attention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "persons"},
            {"LOWER": "in"},
            {"LOWER": "need"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "care"},
            {"LOWER": "and"},
            {"LOWER": "support"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "persons"}, {"LOWER": "lacking"}, {"LOWER": "capacity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "persons"},
            {"LOWER": "needing"},
            {"LOWER": "international"},
            {"LOWER": "protection"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "persons"},
            {"LOWER": "occupying"},
            {"LOWER": "building"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "persons"},
            {"LOWER": "suspected"},
            {"LOWER": "of"},
            {"LOWER": "association"},
            {"LOWER": "with"},
            {"LOWER": "terrorism"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "perverting"},
            {"LOWER": "course"},
            {"LOWER": "of"},
            {"LOWER": "justice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "perverting"},
            {"LOWER": "course"},
            {"LOWER": "of"},
            {"LOWER": "public"},
            {"LOWER": "justice"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pesticides"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pet"}, {"LOWER": "shop"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "brought"},
            {"LOWER": "on"},
            {"LOWER": "grounds"},
            {"LOWER": "of"},
            {"LOWER": "mistake"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "petition"}, {"LOWER": "for"}, {"LOWER": "exhumation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "extension"},
            {"LOWER": "to"},
            {"LOWER": "church"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "faculty"},
            {"LOWER": "for"},
            {"LOWER": "removal"},
            {"LOWER": "and"},
            {"LOWER": "replacement"},
            {"LOWER": "of"},
            {"LOWER": "various"},
            {"LOWER": "items"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "faculty"},
            {"LOWER": "for"},
            {"LOWER": "restoration"},
            {"LOWER": "of"},
            {"LOWER": "church"},
            {"LOWER": "bells"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "faculty"},
            {"LOWER": "to"},
            {"LOWER": "install"},
            {"LOWER": "broadband"},
            {"LOWER": "equipment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "faculty"},
            {"LOWER": "to"},
            {"LOWER": "install"},
            {"LOWER": "telecommunications"},
            {"LOWER": "equipment"},
            {"LOWER": "and"},
            {"LOWER": "grant"},
            {"LOWER": "licence"},
            {"LOWER": "permitting"},
            {"LOWER": "retention"},
            {"LOWER": "and"},
            {"LOWER": "use"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "faculty"},
            {"LOWER": "to"},
            {"LOWER": "replace"},
            {"LOWER": "organ"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "handrail"},
            {"LOWER": "in"},
            {"LOWER": "bell"},
            {"LOWER": "tower"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "improvements"},
            {"LOWER": "to"},
            {"LOWER": "memorial"},
            {"LOWER": "gardens"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "installation"},
            {"LOWER": "of"},
            {"LOWER": "memorial"},
            {"LOWER": "plaque"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "installation"},
            {"LOWER": "of"},
            {"LOWER": "oak"},
            {"LOWER": "table"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "used"},
            {"LOWER": "as"},
            {"LOWER": "an"},
            {"LOWER": "altar"},
            {"LOWER": "and"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "refreshments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "internal"},
            {"LOWER": "re"},
            {"LOWER": "-"},
            {"LOWER": "ordering"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "internal"},
            {"LOWER": "re"},
            {"LOWER": "-"},
            {"LOWER": "ordering"},
            {"LOWER": "works"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "internal"},
            {"LOWER": "reordering"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "vestments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "works"},
            {"LOWER": "of"},
            {"LOWER": "internal"},
            {"LOWER": "re"},
            {"LOWER": "-"},
            {"LOWER": "ordering"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "for"},
            {"LOWER": "works"},
            {"LOWER": "to"},
            {"LOWER": "church"},
            {"LOWER": "hall"},
            {"LOWER": "and"},
            {"LOWER": "churchyard"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "to"},
            {"LOWER": "add"},
            {"LOWER": "bar"},
            {"LOWER": "over"},
            {"LOWER": "finials"},
            {"LOWER": "to"},
            {"LOWER": "railings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "to"},
            {"LOWER": "convert"},
            {"LOWER": "permanent"},
            {"LOWER": "loan"},
            {"LOWER": "into"},
            {"LOWER": "gift"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "to"},
            {"LOWER": "erect"},
            {"LOWER": "memorial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "to"},
            {"LOWER": "european"},
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "human"},
            {"LOWER": "rights"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "to"},
            {"LOWER": "install"},
            {"LOWER": "access"},
            {"LOWER": "ramp"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petition"},
            {"LOWER": "to"},
            {"LOWER": "reserve"},
            {"LOWER": "churchyard"},
            {"LOWER": "plot"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petitioner"},
            {"LOWER": "seeking"},
            {"LOWER": "faculty"},
            {"LOWER": "for"},
            {"LOWER": "headstone"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petitioner"},
            {"LOWER": "seeking"},
            {"LOWER": "faculty"},
            {"LOWER": "for"},
            {"LOWER": "proposed"},
            {"LOWER": "memorial"},
            {"LOWER": "with"},
            {"LOWER": "kerbstones"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petitioners"},
            {"LOWER": "seeking"},
            {"LOWER": "confirmatory"},
            {"LOWER": "faculty"},
            {"LOWER": "in"},
            {"LOWER": "respect"},
            {"LOWER": "of"},
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "painting"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petitioners"},
            {"LOWER": "seeking"},
            {"LOWER": "faculty"},
            {"LOWER": "for"},
            {"LOWER": "disposal"},
            {"LOWER": "of"},
            {"LOWER": "church"},
            {"LOWER": "chattels"},
            {"LOWER": "not"},
            {"LOWER": "amounting"},
            {"LOWER": "to"},
            {"LOWER": "church"},
            {"LOWER": "treasures"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petitioning"},
            {"LOWER": "creditors"},
            {"LOWER": "debt"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "petrol"}, {"LOWER": "bomb"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petroleum"},
            {"LOWER": "and"},
            {"LOWER": "natural"},
            {"LOWER": "gas"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "petroleum"},
            {"LOWER": "exploration"},
            {"LOWER": "and"},
            {"LOWER": "development"},
            {"LOWER": "licence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "petty"}, {"LOWER": "sessional"}, {"LOWER": "division"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pharmacist"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pharmacy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "photographs"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "photographs"},
            {"LOWER": "taken"},
            {"LOWER": "in"},
            {"LOWER": "public"},
            {"LOWER": "place"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "physical"},
            {"LOWER": "or"},
            {"LOWER": "mental"},
            {"LOWER": "condition"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "physical"},
            {"LOWER": "training"},
            {"LOWER": "and"},
            {"LOWER": "activity"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "physiotherapist"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pickets"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "picture"},
            {"LOWER": "archive"},
            {"ORTH": ","},
            {"LOWER": "damage"},
            {"LOWER": "to"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pigeon"}, {"LOWER": "droppings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pioneer"}, {"LOWER": "container"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "place"}, {"LOWER": "of"}, {"LOWER": "performance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "place"},
            {"LOWER": "of"},
            {"LOWER": "performance"},
            {"LOWER": "of"},
            {"LOWER": "contractual"},
            {"LOWER": "obligation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "place"},
            {"LOWER": "of"},
            {"LOWER": "safety"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "place"},
            {"LOWER": "of"},
            {"LOWER": "safety"},
            {"ORTH": ","},
            {"LOWER": "removal"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "place"}, {"LOWER": "of"}, {"LOWER": "trial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "place"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "public"},
            {"LOWER": "religious"},
            {"LOWER": "worship"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "placement"}, {"LOWER": "for"}, {"LOWER": "adoption"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "placement"}, {"LOWER": "for"}, {"LOWER": "adoption"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "placement"}, {"LOWER": "of"}, {"LOWER": "child"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "placement"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "national"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "placement"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "placement"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "placement"},
            {"LOWER": "order"},
            {"LOWER": "opposed"},
            {"LOWER": "by"},
            {"LOWER": "mother"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "places"},
            {"LOWER": "of"},
            {"LOWER": "religious"},
            {"LOWER": "worship"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "places"},
            {"LOWER": "of"},
            {"LOWER": "religious"},
            {"LOWER": "worship"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "plaintiff"},
            {"LOWER": "acting"},
            {"LOWER": "as"},
            {"LOWER": "appropriate"},
            {"LOWER": "adult"},
            {"LOWER": "at"},
            {"LOWER": "police"},
            {"LOWER": "interviews"},
            {"LOWER": "of"},
            {"LOWER": "murder"},
            {"LOWER": "suspect"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "plaintiff"},
            {"LOWER": "severely"},
            {"LOWER": "injured"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "defendants"},
            {"LOWER": "negligence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "plan"},
            {"LOWER": "altered"},
            {"LOWER": "so"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "preclude"},
            {"LOWER": "borough"},
            {"ORTH": "-"},
            {"LOWER": "wide"},
            {"LOWER": "cap"},
            {"LOWER": "on"},
            {"LOWER": "affordable"},
            {"LOWER": "rentals"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "planning"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "planning"}, {"LOWER": "control"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "planning"},
            {"LOWER": "development"},
            {"LOWER": "under"},
            {"LOWER": "private"},
            {"LOWER": "parliamentary"},
            {"LOWER": "act"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "planning"},
            {"LOWER": "inspector"},
            {"LOWER": "dismissing"},
            {"LOWER": "appeal"},
            {"LOWER": "against"},
            {"LOWER": "enforcement"},
            {"LOWER": "notice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "planning"},
            {"LOWER": "inspector"},
            {"LOWER": "’s"},
            {"LOWER": "decision"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "planning"}, {"LOWER": "obligation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "planning"}, {"LOWER": "application"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "planning"}, {"LOWER": "permission"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "planning"}, {"LOWER": "permission"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "planning"},
            {"LOWER": "permission"},
            {"LOWER": "for"},
            {"LOWER": "domestic"},
            {"LOWER": "housing"},
            {"LOWER": "granted"},
            {"LOWER": "on"},
            {"LOWER": "basis"},
            {"LOWER": "proposal"},
            {"LOWER": "not"},
            {"LOWER": "environmental"},
            {"LOWER": "impact"},
            {"LOWER": "assessment"},
            {"ORTH": "("},
            {"ORTH": "“"},
            {"LOWER": "eia"},
            {"LOWER": "”"},
            {"ORTH": ")"},
            {"LOWER": "development"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "planning"},
            {"LOWER": "permission"},
            {"ORTH": ","},
            {"LOWER": "grant"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "plant"}, {"LOWER": "protection"}, {"LOWER": "product"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "playing"}, {"LOWER": "fields"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "plea"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "plea"}, {"LOWER": "bargain"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "plea"}, {"LOWER": "of"}, {"LOWER": "guilty"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pleading"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pleadings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pleadings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pleadings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "point"},
            {"LOWER": "of"},
            {"LOWER": "law"},
            {"LOWER": "not"},
            {"LOWER": "taken"},
            {"LOWER": "below"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "points"}, {"LOWER": "based"}, {"LOWER": "system"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "points"}, {"LOWER": "based"}, {"LOWER": "system"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "police"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "police"},
            {"ORTH": "("},
            {"LOWER": "property"},
            {"ORTH": ")"},
            {"LOWER": "act"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "police"},
            {"LOWER": "and"},
            {"LOWER": "crime"},
            {"LOWER": "commissioner"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "police"},
            {"LOWER": "and"},
            {"LOWER": "judicial"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "operation"},
            {"LOWER": "in"},
            {"LOWER": "criminal"},
            {"LOWER": "matters"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "police"}, {"LOWER": "authority"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "police"},
            {"LOWER": "complaints"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "police"}, {"LOWER": "complaints"}, {"LOWER": "board"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "police"}, {"LOWER": "conduct"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "police"}, {"LOWER": "investigation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "police"}, {"LOWER": "officer"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "police"},
            {"LOWER": "officers"},
            {"LOWER": "undergoing"},
            {"LOWER": "promotion"},
            {"LOWER": "selection"},
            {"LOWER": "process"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "police"},
            {"LOWER": "officer"},
            {"LOWER": "’s"},
            {"LOWER": "injury"},
            {"LOWER": "award"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "police"},
            {"LOWER": "public"},
            {"LOWER": "procession"},
            {"LOWER": "and"},
            {"LOWER": "assembly"},
            {"LOWER": "control"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "police"}, {"LOWER": "service"}, {"LOWER": "commission"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "police"}, {"LOWER": "statements"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "police"}, {"LOWER": "surgeon"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "policemans"}, {"LOWER": "injury"}, {"LOWER": "award"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "political"}, {"LOWER": "advertising"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "political"}, {"LOWER": "opinion"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "political"}, {"LOWER": "party"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "political"}, {"LOWER": "party"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "political"},
            {"LOWER": "portrayal"},
            {"LOWER": "empowering"},
            {"LOWER": "“"},
            {"LOWER": "disaffiliation"},
            {"LOWER": "of"},
            {"LOWER": "an"},
            {"LOWER": "organization"},
            {"LOWER": "or"},
            {"LOWER": "expulsion"},
            {"LOWER": "of"},
            {"LOWER": "an"},
            {"LOWER": "individual"},
            {"ORTH": "”"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "pollution"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pollution"}, {"LOWER": "control"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pollution"},
            {"LOWER": "of"},
            {"LOWER": "underground"},
            {"LOWER": "water"},
            {"LOWER": "supply"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pollution"},
            {"ORTH": ","},
            {"LOWER": "control"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "polygamous"}, {"LOWER": "system"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "port"}, {"LOWER": "development"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "position"},
            {"LOWER": "where"},
            {"LOWER": "finding"},
            {"LOWER": "of"},
            {"LOWER": "no"},
            {"LOWER": "marriage"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "positive"}, {"LOWER": "covenant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "possessing"},
            {"LOWER": "article"},
            {"LOWER": "for"},
            {"LOWER": "terrorist"},
            {"LOWER": "purposes"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "possessing"}, {"LOWER": "false"}, {"LOWER": "document"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "possessing"},
            {"LOWER": "false"},
            {"LOWER": "identity"},
            {"LOWER": "document"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "possessing"},
            {"LOWER": "prohibited"},
            {"LOWER": "weapon"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "possession"}, {"LOWER": "application"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "possession"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "possession"}, {"LOWER": "of"}, {"LOWER": "drugs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "possession"},
            {"LOWER": "of"},
            {"LOWER": "article"},
            {"LOWER": "for"},
            {"LOWER": "use"},
            {"LOWER": "in"},
            {"LOWER": "fraud"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "possession"},
            {"LOWER": "of"},
            {"LOWER": "firearm"},
            {"LOWER": "with"},
            {"LOWER": "intent"},
            {"LOWER": "to"},
            {"LOWER": "endanger"},
            {"LOWER": "life"},
            {"LOWER": "or"},
            {"LOWER": "to"},
            {"LOWER": "enable"},
            {"LOWER": "another"},
            {"LOWER": "to"},
            {"LOWER": "endanger"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "possession"},
            {"LOWER": "of"},
            {"LOWER": "imitation"},
            {"LOWER": "firearm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "possession"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "possession"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "possession"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "possession"},
            {"LOWER": "with"},
            {"LOWER": "intent"},
            {"LOWER": "to"},
            {"LOWER": "supply"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "not"},
            {"LOWER": "to"},
            {"LOWER": "part"},
            {"LOWER": "with"},
            {"LOWER": "possession"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "possession"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "possessory"}, {"LOWER": "lien"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "possessory"}, {"LOWER": "title"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "post"}, {"LOWER": "mortem"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "post"},
            {"LOWER": "-"},
            {"LOWER": "incident"},
            {"LOWER": "procedure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "post"},
            {"LOWER": "-"},
            {"LOWER": "nuptial"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "post"}, {"LOWER": "-"}, {"LOWER": "settlement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "postal"}, {"LOWER": "ballot"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "postal"}, {"LOWER": "services"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "potential"},
            {"LOWER": "expert"},
            {"LOWER": "legal"},
            {"LOWER": "assistance"},
            {"LOWER": "to"},
            {"LOWER": "court"},
            {"LOWER": "by"},
            {"LOWER": "additional"},
            {"LOWER": "solicitor"},
            {"LOWER": "on"},
            {"LOWER": "behalf"},
            {"LOWER": "of"},
            {"LOWER": "litigant"},
            {"LOWER": "in"},
            {"LOWER": "family"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "poultry"},
            {"LOWER": "processing"},
            {"LOWER": "factory"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "advancement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "advancement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "appointment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "arrest"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "attorney"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "attorney"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "attorney"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "of"},
            {"LOWER": "court"},
            {"LOWER": "to"},
            {"LOWER": "act"},
            {"LOWER": "of"},
            {"LOWER": "own"},
            {"LOWER": "motion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "exclusion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "of"}, {"LOWER": "leasing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "or"}, {"LOWER": "trust"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "adjudicate"},
            {"LOWER": "debtor"},
            {"LOWER": "a"},
            {"LOWER": "bankrupt"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "to"}, {"LOWER": "appoint"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "to"}, {"LOWER": "appoint"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "appoint"},
            {"ORTH": ","},
            {"LOWER": "suspend"},
            {"LOWER": "and"},
            {"LOWER": "remove"},
            {"LOWER": "chief"},
            {"LOWER": "constable"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "to"}, {"LOWER": "award"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "award"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "control"},
            {"LOWER": "evidence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "modify"},
            {"LOWER": "statute"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "order"},
            {"LOWER": "retrial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "overrule"},
            {"LOWER": "previous"},
            {"LOWER": "decisions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "power"}, {"LOWER": "to"}, {"LOWER": "quash"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "question"},
            {"LOWER": "at"},
            {"LOWER": "port"},
            {"LOWER": "or"},
            {"LOWER": "border"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "rectify"},
            {"LOWER": "matters"},
            {"LOWER": "where"},
            {"LOWER": "there"},
            {"LOWER": "has"},
            {"LOWER": "been"},
            {"LOWER": "procedural"},
            {"LOWER": "error"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "remand"},
            {"LOWER": "to"},
            {"LOWER": "hospital"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "strike"},
            {"LOWER": "out"},
            {"LOWER": "statement"},
            {"LOWER": "of"},
            {"LOWER": "case"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "vary"},
            {"LOWER": "or"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "powers"}, {"LOWER": "and"}, {"LOWER": "duties"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "powers"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "powers"},
            {"LOWER": "of"},
            {"LOWER": "law"},
            {"LOWER": "society"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "powers"}, {"LOWER": "of"}, {"LOWER": "regulator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "powers"},
            {"LOWER": "of"},
            {"LOWER": "revenue"},
            {"LOWER": "officers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "powers"}, {"LOWER": "of"}, {"LOWER": "seizure"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "practical"}, {"LOWER": "benefit"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "practice"}, {"LOWER": "and"}, {"LOWER": "procedure"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "practice"}, {"LOWER": "and"}, {"LOWER": "procedure"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "practice"}, {"LOWER": "directions"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "practice"}, {"LOWER": "jurisdiction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "practices"},
            {"LOWER": "which"},
            {"LOWER": "might"},
            {"LOWER": "affect"},
            {"LOWER": "trade"},
            {"LOWER": "between"},
            {"LOWER": "eu"},
            {"LOWER": "member"},
            {"LOWER": "states"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "action"},
            {"LOWER": "disclosure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "contract"},
            {"LOWER": "inquiries"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "contract"},
            {"LOWER": "negotiations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "contractual"},
            {"LOWER": "information"},
            {"LOWER": "duties"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "emptive"},
            {"LOWER": "costs"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "hearing"},
            {"LOWER": "review"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "incorporation"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "inquest"},
            {"LOWER": "review"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "inquest"},
            {"LOWER": "review"},
            {"LOWER": "hearing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "judgment"},
            {"LOWER": "interest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "permission"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "sentence"},
            {"LOWER": "report"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "trial"},
            {"LOWER": "pardon"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "trial"},
            {"LOWER": "review"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "precedent"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "preclusion"}, {"LOWER": "of"}, {"LOWER": "torture"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "preconditions"}, {"LOWER": "for"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prefabricated"}, {"LOWER": "structure"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "preferential"}, {"LOWER": "debt"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "pregnancy"}, {"LOWER": "and"}, {"LOWER": "maternity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prejudicing"}, {"LOWER": "fair"}, {"LOWER": "trial"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "preliminary"},
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "hearing"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "preliminary"}, {"LOWER": "issue"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "preliminary"}, {"LOWER": "issues"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "preliminary"}, {"LOWER": "point"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "preliminary"},
            {"LOWER": "point"},
            {"LOWER": "of"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "preliminary"},
            {"LOWER": "ruling"},
            {"LOWER": "on"},
            {"LOWER": "brussels"},
            {"LOWER": "convention"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "premises"}, {"LOWER": "licence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "premises"}, {"LOWER": "within"}, {"LOWER": "acts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "premises"},
            {"LOWER": "within"},
            {"LOWER": "the"},
            {"LOWER": "acts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "preparation"}, {"LOWER": "and"}, {"LOWER": "conduct"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "preparation"}, {"LOWER": "for"}, {"LOWER": "hearings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "preparation"}, {"LOWER": "of"}, {"LOWER": "case"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "preparatory"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "preparatory"}, {"LOWER": "hearing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prerogative"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prerogative"}, {"LOWER": "of"}, {"LOWER": "mercy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prerogative"}, {"LOWER": "power"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "preservation"},
            {"LOWER": "and"},
            {"LOWER": "enhancement"},
            {"LOWER": "policy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "preservation"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "preservation"},
            {"LOWER": "of"},
            {"LOWER": "maori"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "preservation"}, {"LOWER": "of"}, {"LOWER": "property"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "presumption"},
            {"LOWER": "against"},
            {"LOWER": "taking"},
            {"LOWER": "advantage"},
            {"LOWER": "of"},
            {"LOWER": "own"},
            {"LOWER": "wrong"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "presumption"}, {"LOWER": "of"}, {"LOWER": "continuity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "presumption"}, {"LOWER": "of"}, {"LOWER": "damage"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "presumption"}, {"LOWER": "of"}, {"LOWER": "innocence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "presumption"}, {"LOWER": "of"}, {"LOWER": "validity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prevalence"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prevention"}, {"LOWER": "of"}, {"LOWER": "use"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prevention"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "use"},
            {"ORTH": "”"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prevention"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prevention"}, {"LOWER": "orders"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "previous"},
            {"LOWER": "convictions"},
            {"LOWER": "or"},
            {"LOWER": "antecedents"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "previously"},
            {"LOWER": "expressed"},
            {"LOWER": "wish"},
            {"LOWER": "to"},
            {"LOWER": "have"},
            {"LOWER": "no"},
            {"LOWER": "contact"},
            {"LOWER": "with"},
            {"LOWER": "family"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "price"}, {"LOWER": "controls"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "price"}, {"LOWER": "transparency"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "price"},
            {"ORTH": "-"},
            {"LOWER": "fixing"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "price"}, {"LOWER": "tariff"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "principal"},
            {"LOWER": "in"},
            {"LOWER": "first"},
            {"LOWER": "degree"},
            {"LOWER": "to"},
            {"LOWER": "murder"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "principal"}, {"LOWER": "registry"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "principle"},
            {"LOWER": "of"},
            {"LOWER": "equal"},
            {"LOWER": "treatment"},
            {"LOWER": "between"},
            {"LOWER": "persons"},
            {"LOWER": "irrespective"},
            {"LOWER": "of"},
            {"LOWER": "racial"},
            {"LOWER": "or"},
            {"LOWER": "ethnic"},
            {"LOWER": "origin"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "principles"},
            {"LOWER": "of"},
            {"LOWER": "effectiveness"},
            {"LOWER": "and"},
            {"LOWER": "legal"},
            {"LOWER": "certainty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "principles"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "applied"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prior"}, {"LOWER": "clearance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prior"}, {"LOWER": "publication"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "priority"}, {"LOWER": "date"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "priority"}, {"LOWER": "need"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "priority"}, {"LOWER": "need"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "priority"}, {"LOWER": "of"}, {"LOWER": "entries"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "priority"},
            {"LOWER": "of"},
            {"LOWER": "mortgages"},
            {"LOWER": "and"},
            {"LOWER": "charges"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prison"}, {"LOWER": "authorities"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prison"},
            {"LOWER": "conditions"},
            {"LOWER": "in"},
            {"LOWER": "issuing"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prison"},
            {"LOWER": "conditions"},
            {"LOWER": "in"},
            {"LOWER": "requesting"},
            {"LOWER": "state"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prison"}, {"LOWER": "governor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prison"}, {"LOWER": "mutiny"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prison"}, {"LOWER": "officer"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prison"}, {"LOWER": "officers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prison"},
            {"LOWER": "rules"},
            {"LOWER": "1964"},
            {"ORTH": "("},
            {"LOWER": "s.i"},
            {"ORTH": "."},
            {"LOWER": "1964"},
            {"LOWER": "no"},
            {"LOWER": "."},
            {"LOWER": "388"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "rr"},
            {"ORTH": "."},
            {"LOWER": "33"},
            {"ORTH": "("},
            {"LOWER": "3"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"ORTH": "37a"},
            {"ORTH": "("},
            {"LOWER": "1"},
            {"ORTH": ")"},
            {"ORTH": "("},
            {"LOWER": "as"},
            {"LOWER": "amended"},
            {"LOWER": "by"},
            {"LOWER": "prison"},
            {"ORTH": "("},
            {"LOWER": "amendment"},
            {"ORTH": ")"},
            {"LOWER": "rules"},
            {"LOWER": "1972"},
            {"ORTH": "("},
            {"LOWER": "s.i"},
            {"ORTH": "."},
            {"LOWER": "1972"},
            {"LOWER": "no"},
            {"LOWER": "."},
            {"LOWER": "1860"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "r."},
            {"LOWER": "3"},
            {"LOWER": "and"},
            {"LOWER": "prison"},
            {"ORTH": "("},
            {"LOWER": "amendment"},
            {"ORTH": ")"},
            {"LOWER": "rules"},
            {"LOWER": "1974"},
            {"ORTH": "("},
            {"LOWER": "s.i"},
            {"ORTH": "."},
            {"LOWER": "1974"},
            {"LOWER": "no"},
            {"LOWER": "."},
            {"LOWER": "713"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "r."},
            {"LOWER": "3"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prison"}, {"LOWER": "s"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prisoner"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prisoner"}, {"LOWER": "of"}, {"LOWER": "war"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prisoner"},
            {"LOWER": "requesting"},
            {"LOWER": "downgrade"},
            {"LOWER": "to"},
            {"LOWER": "category"},
            {"LOWER": "b"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prisoner"}, {"LOWER": "serving"}, {"LOWER": "sentence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prisoners"},
            {"LOWER": "legal"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prisoners"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prisoners"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prisoners"}, {"LOWER": "rights"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prisoners"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prisoners"}, {"LOWER": "’"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prisoners"}, {"LOWER": "’"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prisoners"}, {"LOWER": "’"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prisoner"}, {"LOWER": "’s"}, {"LOWER": "repatriation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prisoner"}, {"LOWER": "’s"}, {"LOWER": "rights"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prisons"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prisons"},
            {"LOWER": "board"},
            {"LOWER": "of"},
            {"LOWER": "visitor"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prisons"},
            {"LOWER": "board"},
            {"LOWER": "of"},
            {"LOWER": "visitors"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "privacy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "privacy"}, {"LOWER": "claim"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "private"}, {"LOWER": "life"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "private"},
            {"LOWER": "and"},
            {"LOWER": "family"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "private"},
            {"LOWER": "consultation"},
            {"LOWER": "with"},
            {"LOWER": "lawyers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "private"},
            {"LOWER": "detectives"},
            {"LOWER": "acting"},
            {"LOWER": "for"},
            {"LOWER": "supervisory"},
            {"LOWER": "body"},
            {"LOWER": "of"},
            {"LOWER": "regulated"},
            {"LOWER": "profession"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "private"},
            {"LOWER": "divorce"},
            {"LOWER": "obtained"},
            {"LOWER": "before"},
            {"LOWER": "religious"},
            {"LOWER": "court"},
            {"LOWER": "in"},
            {"LOWER": "third"},
            {"LOWER": "country"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "private"}, {"LOWER": "footpath"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "private"}, {"LOWER": "hire"}, {"LOWER": "vehicle"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "private"}, {"LOWER": "law"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "private"}, {"LOWER": "law"}, {"LOWER": "applications"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "private"}, {"LOWER": "law"}, {"LOWER": "matters"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "private"}, {"LOWER": "law"}, {"LOWER": "orders"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "private"}, {"LOWER": "law"}, {"LOWER": "proceedings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "private"}, {"LOWER": "life"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "private"}, {"LOWER": "prosecution"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "private"}, {"LOWER": "rented"}, {"LOWER": "sector"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "private"}, {"LOWER": "residence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "private"}, {"LOWER": "security"}, {"LOWER": "industry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "privatised"}, {"LOWER": "public"}, {"LOWER": "company"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "law"}, {"LOWER": "of"}, {"LOWER": "privilege"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "privileged"}, {"LOWER": "communication"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "privilege"},
            {"LOWER": "against"},
            {"LOWER": "self"},
            {"LOWER": "incrimination"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "privilege"},
            {"LOWER": "against"},
            {"LOWER": "self"},
            {"ORTH": "-"},
            {"LOWER": "incrimination"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "privy"}, {"LOWER": "council"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "privy"}, {"LOWER": "council"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "probate"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "probation"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "probation"},
            {"LOWER": "order"},
            {"ORTH": ","},
            {"LOWER": "breach"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "probationer"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "problem"}, {"LOWER": "gambler"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "procedural"}, {"LOWER": "defect"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "procedural"}, {"LOWER": "fairness"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "procedural"}, {"LOWER": "impropriety"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "procedural"}, {"LOWER": "irregularity"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "procedural"}, {"LOWER": "unfairness"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "procedure"},
            {"LOWER": "for"},
            {"LOWER": "returning"},
            {"LOWER": "illegally"},
            {"LOWER": "staying"},
            {"LOWER": "third"},
            {"ORTH": "-"},
            {"LOWER": "country"},
            {"LOWER": "nationals"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "procedure"}, {"LOWER": "while"}, {"LOWER": "pending"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "against"},
            {"LOWER": "company"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "between"},
            {"LOWER": "subjects"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proceedings"}, {"LOWER": "concerning"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "concerning"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "for"},
            {"LOWER": "approval"},
            {"LOWER": "of"},
            {"LOWER": "settlement"},
            {"LOWER": "or"},
            {"LOWER": "compromise"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "for"},
            {"LOWER": "possession"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "identifying"},
            {"LOWER": "unacceptably"},
            {"LOWER": "poor"},
            {"LOWER": "practice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "in"},
            {"LOWER": "england"},
            {"LOWER": "and"},
            {"LOWER": "italy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proceedings"}, {"LOWER": "in"}, {"LOWER": "parliament"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proceedings"}, {"LOWER": "relating"}, {"LOWER": "to"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "children"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "transferred"},
            {"LOWER": "from"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceedings"},
            {"LOWER": "transferred"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proceeds"}, {"LOWER": "of"}, {"LOWER": "crime"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proceeds"}, {"LOWER": "of"}, {"LOWER": "crime"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proceeds"}, {"LOWER": "of"}, {"LOWER": "crime"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proceeds"}, {"LOWER": "of"}, {"LOWER": "crime"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proceeds"},
            {"LOWER": "of"},
            {"LOWER": "criminal"},
            {"LOWER": "conduct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "processing"}, {"LOWER": "and"}, {"LOWER": "access"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "processing"},
            {"LOWER": "of"},
            {"LOWER": "personal"},
            {"LOWER": "data"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "procuring"}, {"LOWER": "deadly"}, {"LOWER": "weapon"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "procuring"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "procuring"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "trust"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "procuring"},
            {"LOWER": "execution"},
            {"LOWER": "of"},
            {"LOWER": "valuable"},
            {"LOWER": "secutiy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "procuring"},
            {"LOWER": "or"},
            {"LOWER": "inciting"},
            {"LOWER": "murder"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "procuring"},
            {"ORTH": "“"},
            {"LOWER": "thing"},
            {"ORTH": "”"},
            {"LOWER": "with"},
            {"LOWER": "intent"},
            {"LOWER": "to"},
            {"LOWER": "procure"},
            {"LOWER": "miscarriage"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "producer"}, {"LOWER": "organisations"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "product"}, {"LOWER": "licence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "production"},
            {"LOWER": "of"},
            {"LOWER": "controlled"},
            {"LOWER": "drug"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "production"}, {"LOWER": "of"}, {"LOWER": "documents"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "production"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "professional"},
            {"LOWER": "conduct"},
            {"LOWER": "committee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "professional"},
            {"LOWER": "conduct"},
            {"LOWER": "committee"},
            {"LOWER": "of"},
            {"LOWER": "general"},
            {"LOWER": "council"},
            {"LOWER": "of"},
            {"LOWER": "bar"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "professional"},
            {"LOWER": "conduct"},
            {"LOWER": "committee"},
            {"LOWER": "of"},
            {"LOWER": "general"},
            {"LOWER": "medical"},
            {"LOWER": "council"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "professional"}, {"LOWER": "indemnity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "professional"},
            {"LOWER": "indemnity"},
            {"LOWER": "insurance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "professional"}, {"LOWER": "misconduct"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "professional"}, {"LOWER": "negligence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "professional"}, {"LOWER": "negligence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "professional"}, {"LOWER": "standards"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "professional"}, {"LOWER": "training"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "profits"}, {"LOWER": "of"}, {"LOWER": "investment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "profits"}, {"LOWER": "of"}, {"LOWER": "trade"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "profits"}, {"LOWER": "tax"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "computation"}, {"LOWER": "of"}, {"LOWER": "profits"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "programmes"},
            {"LOWER": "and"},
            {"LOWER": "audiovisual"},
            {"LOWER": "media"},
            {"LOWER": "service"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prohibited"}, {"LOWER": "steps"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prohibited"}, {"LOWER": "weapon"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prohibited"}, {"LOWER": "weapons"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prohibition"},
            {"LOWER": "against"},
            {"LOWER": "identification"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prohibition"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prohibition"}, {"LOWER": "of"}, {"LOWER": "harassment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prohibition"},
            {"LOWER": "on"},
            {"LOWER": "advertising"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prohibition"},
            {"LOWER": "on"},
            {"LOWER": "doing"},
            {"LOWER": "estate"},
            {"LOWER": "agency"},
            {"LOWER": "work"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prohibition"},
            {"LOWER": "on"},
            {"LOWER": "import"},
            {"LOWER": "and"},
            {"LOWER": "export"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prohibition"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prohibitory"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "project"}, {"LOWER": "monitor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "projects"},
            {"LOWER": "of"},
            {"LOWER": "common"},
            {"LOWER": "interest"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "territory"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "promotion"}, {"LOWER": "money"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "pronouncement"}, {"LOWER": "of"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "proof"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proof"}, {"LOWER": "in"}, {"LOWER": "bankruptcy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proof"}, {"LOWER": "in"}, {"LOWER": "liquidation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "proof"}, {"LOWER": "of"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proof"}, {"LOWER": "of"}, {"LOWER": "debt"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "proper"}, {"LOWER": "law"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proper"}, {"LOWER": "law"}, {"LOWER": "clause"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proper"},
            {"LOWER": "level"},
            {"LOWER": "of"},
            {"LOWER": "sentence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"LOWER": "and"},
            {"LOWER": "affairs"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"LOWER": "belonging"},
            {"LOWER": "to"},
            {"LOWER": "another"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"LOWER": "held"},
            {"LOWER": "by"},
            {"LOWER": "police"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"LOWER": "obtained"},
            {"LOWER": "through"},
            {"LOWER": "unlawful"},
            {"LOWER": "conduct"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "property"}, {"LOWER": "of"}, {"LOWER": "deceased"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "property"}, {"LOWER": "ownership"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"LOWER": "passing"},
            {"LOWER": "to"},
            {"LOWER": "trustee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"LOWER": "passing"},
            {"LOWER": "to"},
            {"LOWER": "trustee"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "property"}, {"LOWER": "purchase"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "property"}, {"LOWER": "tax"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "property"}, {"LOWER": "transactions"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"ORTH": ","},
            {"LOWER": "when"},
            {"LOWER": "passing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "property"},
            {"ORTH": ","},
            {"LOWER": "whether"},
            {"LOWER": "passing"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "proportionality"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proportionate"}, {"LOWER": "costs"}, {"LOWER": "or"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proposal"}, {"LOWER": "to"}, {"LOWER": "merge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proposed"},
            {"LOWER": "works"},
            {"LOWER": "to"},
            {"LOWER": "listed"},
            {"LOWER": "church"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "proprietary"}, {"LOWER": "estoppel"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proprietor"},
            {"LOWER": "of"},
            {"LOWER": "food"},
            {"LOWER": "business"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "proprietors"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proprietor"}, {"LOWER": "’s"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proprietor"}, {"LOWER": "’s"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "proscribed"}, {"LOWER": "organisation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prosecuting"}, {"LOWER": "counsel"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prosecution"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prosecution"}, {"LOWER": "evidence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "prosecution"}, {"LOWER": "evidence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prosecution"},
            {"LOWER": "for"},
            {"LOWER": "drug"},
            {"LOWER": "related"},
            {"LOWER": "offence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "prosecution"},
            {"LOWER": "for"},
            {"LOWER": "tax"},
            {"LOWER": "evasion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prosecution"}, {"LOWER": "of"}, {"LOWER": "offences"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prosecutors"}, {"LOWER": "costs"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "prosecutor"}, {"LOWER": "’s"}, {"LOWER": "conduct"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "prospectus"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "protected"}, {"LOWER": "disclosure"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "protected"}, {"LOWER": "tenancy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "protecting"},
            {"LOWER": "public"},
            {"LOWER": "from"},
            {"LOWER": "dangerous"},
            {"LOWER": "offenders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "protection"},
            {"LOWER": "against"},
            {"LOWER": "transmissible"},
            {"LOWER": "spongiform"},
            {"LOWER": "encephalopathies"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "protection"},
            {"LOWER": "and"},
            {"LOWER": "indemnity"},
            {"LOWER": "risks"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "protection"}, {"LOWER": "of"}, {"LOWER": "children"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "protection"}, {"LOWER": "of"}, {"LOWER": "employees"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "protection"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "essential"},
            {"LOWER": "security"},
            {"LOWER": "interests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "protection"},
            {"LOWER": "of"},
            {"LOWER": "military"},
            {"LOWER": "remains"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "protection"},
            {"LOWER": "of"},
            {"LOWER": "waters"},
            {"LOWER": "against"},
            {"LOWER": "pollution"},
            {"LOWER": "caused"},
            {"LOWER": "by"},
            {"LOWER": "nitrates"},
            {"LOWER": "from"},
            {"LOWER": "agricultural"},
            {"LOWER": "sources"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "protection"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "protective"}, {"LOWER": "award"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "protective"}, {"LOWER": "costs"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "protective"}, {"LOWER": "costs"}, {"LOWER": "orders"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provender"},
            {"LOWER": "mill"},
            {"LOWER": "and"},
            {"LOWER": "chicken"},
            {"LOWER": "processing"},
            {"LOWER": "factory"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "province"}, {"LOWER": "of"}, {"LOWER": "york"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "from"},
            {"LOWER": "former"},
            {"LOWER": "spouses"},
            {"LOWER": "estate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "accommodation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "basic"},
            {"LOWER": "package"},
            {"LOWER": "of"},
            {"LOWER": "radio"},
            {"LOWER": "and"},
            {"LOWER": "television"},
            {"LOWER": "programmes"},
            {"LOWER": "via"},
            {"LOWER": "cable"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
            {"ORTH": ","},
            {"LOWER": "facilities"},
            {"LOWER": "or"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "nursing"},
            {"LOWER": "care"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "sanitary"},
            {"LOWER": "appliances"},
            {"LOWER": "at"},
            {"ORTH": "“"},
            {"LOWER": "relevant"},
            {"LOWER": "place"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "suitable"},
            {"LOWER": "education"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provision"},
            {"LOWER": "of"},
            {"LOWER": "transport"},
            {"LOWER": "for"},
            {"LOWER": "pupils"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provisional"},
            {"LOWER": "and"},
            {"LOWER": "protective"},
            {"LOWER": "measures"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "provisional"}, {"LOWER": "damages"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "provisional"}, {"LOWER": "liquidator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provisional"},
            {"LOWER": "liquidators"},
            {"LOWER": "costs"},
            {"LOWER": "and"},
            {"LOWER": "expenses"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "provisional"}, {"LOWER": "registration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provisions"},
            {"LOWER": "governing"},
            {"LOWER": "institutions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "provisions"},
            {"LOWER": "governing"},
            {"LOWER": "the"},
            {"LOWER": "institutions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "proviso"},
            {"LOWER": "to"},
            {"LOWER": "section"},
            {"LOWER": "2(1"},
            {"ORTH": ")"},
            {"LOWER": "of"},
            {"LOWER": "act"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "provocation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "pseudo"}, {"ORTH": "-"}, {"LOWER": "photograph"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "psychiatric"}, {"LOWER": "report"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "environmental"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "information"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "assembly"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "authority"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "authority"},
            {"ORTH": ","},
            {"LOWER": "duty"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "contract"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "contracts"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "election"}, {"LOWER": "broadcast"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "examination"},
            {"LOWER": "of"},
            {"LOWER": "bankrupt"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "footpath"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "general"}, {"LOWER": "act"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "good"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "health"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "health"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "health"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "information"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "interest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "interest"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "interest"},
            {"LOWER": "disclosure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "interest"}, {"LOWER": "grounds"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "interest"}, {"LOWER": "immunity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "interest"},
            {"LOWER": "immunity"},
            {"LOWER": "claimed"},
            {"LOWER": "re"},
            {"LOWER": "material"},
            {"LOWER": "relevant"},
            {"LOWER": "to"},
            {"LOWER": "inquest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "law"}, {"LOWER": "category"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "nuisance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "office"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "or"}, {"LOWER": "private"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "order"},
            {"LOWER": "offences"},
            {"LOWER": "by"},
            {"LOWER": "miners"},
            {"LOWER": "picketing"},
            {"LOWER": "local"},
            {"LOWER": "collieries"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "park"},
            {"LOWER": "or"},
            {"LOWER": "pleasure"},
            {"LOWER": "ground"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "participation"},
            {"LOWER": "in"},
            {"LOWER": "decision‑making"},
            {"LOWER": "and"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "justice"},
            {"LOWER": "in"},
            {"LOWER": "environmental"},
            {"LOWER": "matters"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "passenger"},
            {"LOWER": "transport"},
            {"LOWER": "services"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "policy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "policy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "procurement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "procurement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "procurement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "procurement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "procurement"}, {"LOWER": "regime"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "documents"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "navigation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "sector"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "public"},
            {"LOWER": "sector"},
            {"LOWER": "equality"},
            {"LOWER": "duty"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "security"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "service"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "service"}, {"LOWER": "contracts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "service"}, {"LOWER": "vehicle"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "services"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "services"}, {"LOWER": "contracts"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "trust"}, {"LOWER": "office"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "trustee"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "public"}, {"LOWER": "utilities"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "works"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "public"}, {"LOWER": "works"}, {"LOWER": "contracts"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "publication"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "publication"},
            {"LOWER": "by"},
            {"LOWER": "co"},
            {"LOWER": "-"},
            {"LOWER": "owner"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "publication"},
            {"LOWER": "of"},
            {"LOWER": "information"},
            {"LOWER": "on"},
            {"LOWER": "internet"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "publication"}, {"LOWER": "on"}, {"LOWER": "internet"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "publication"}, {"LOWER": "on"}, {"LOWER": "internet"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "publicly"},
            {"LOWER": "funded"},
            {"LOWER": "legal"},
            {"LOWER": "work"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "publicly"}, {"LOWER": "funded"}, {"LOWER": "work"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "pupil"}, {"LOWER": "at"}, {"LOWER": "school"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "purchase"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "purchase"}, {"LOWER": "of"}, {"LOWER": "shares"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "purpose"}, {"LOWER": "of"}, {"LOWER": "transaction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "qualified"},
            {"LOWER": "one"},
            {"ORTH": "-"},
            {"LOWER": "way"},
            {"LOWER": "costs"},
            {"LOWER": "shifting"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "qualifying"}, {"LOWER": "disclosure"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "quality"},
            {"LOWER": "of"},
            {"LOWER": "article"},
            {"LOWER": "demanded"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "quantitative"}, {"LOWER": "restriction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "quantitative"},
            {"LOWER": "restriction"},
            {"LOWER": "on"},
            {"LOWER": "imports"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "quantitative"}, {"LOWER": "restrictions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "quantitative"},
            {"LOWER": "restrictions"},
            {"LOWER": "on"},
            {"LOWER": "exports"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "quantitative"},
            {"LOWER": "restrictions"},
            {"LOWER": "on"},
            {"LOWER": "imports"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "quantum"}, {"LOWER": "meruit"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "quasi"}, {"LOWER": "-"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "queens"}, {"LOWER": "bench"}, {"LOWER": "division"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "queen"}, {"LOWER": "’s"}, {"LOWER": "proctor"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "quia"}, {"LOWER": "timet"}, {"LOWER": "action"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "quiet"}, {"LOWER": "enjoyment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "quorum"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "r.i.b.a"}, {"ORTH": "."}, {"LOWER": "form"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "r.l.b.a"}, {"ORTH": "."}, {"LOWER": "form"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "racial"}, {"LOWER": "group"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "racial"}, {"LOWER": "group"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "racially"},
            {"LOWER": "aggravated"},
            {"LOWER": "offence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "radioactive"},
            {"LOWER": "substances"},
            {"LOWER": "in"},
            {"LOWER": "water"},
            {"LOWER": "intended"},
            {"LOWER": "for"},
            {"LOWER": "human"},
            {"LOWER": "consumption"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rafaela"}, {"LOWER": "s"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "railway"},
            {"LOWER": "administration"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "railway"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "railways"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rape"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rape"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "under"},
            {"LOWER": "13"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rastafarian"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rate"}, {"LOWER": "of"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rateable"}, {"LOWER": "occupation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rateable"}, {"LOWER": "value"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rateable"}, {"LOWER": "value"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ratification"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rating"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rating"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rating"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ratio"}, {"LOWER": "decidendi"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "re"}, {"LOWER": "-"}, {"LOWER": "engagement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "re"}, {"LOWER": "-"}, {"LOWER": "entry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "re"}, {"LOWER": "-"}, {"LOWER": "ordering"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "real"},
            {"LOWER": "prospect"},
            {"LOWER": "of"},
            {"LOWER": "success"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reason"}, {"LOWER": "for"}, {"LOWER": "absence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reason"}, {"LOWER": "for"}, {"LOWER": "dismissal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reason"}, {"LOWER": "for"}, {"LOWER": "dismissal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "reasonable"}, {"LOWER": "adjustment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reasonable"}, {"LOWER": "adjustments"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "reasonable"}, {"LOWER": "belief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reasonable"}, {"LOWER": "excuse"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reasonable"},
            {"LOWER": "financial"},
            {"LOWER": "provision"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "reasonable"}, {"LOWER": "force"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reasonableness"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reasonable"}, {"LOWER": "prospect"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reasonable"}, {"LOWER": "expectation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reasonableness"},
            {"LOWER": "of"},
            {"LOWER": "dismissal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reasonableness"},
            {"LOWER": "of"},
            {"LOWER": "dismissal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reasonably"}, {"LOWER": "practicable"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reasoning"},
            {"LOWER": "of"},
            {"LOWER": "first"},
            {"ORTH": "-"},
            {"LOWER": "tier"},
            {"LOWER": "tribunal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "reasoning"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reasons"}, {"LOWER": "for"}, {"LOWER": "award"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reasons"}, {"LOWER": "for"}, {"LOWER": "decision"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reasons"}, {"LOWER": "for"}, {"LOWER": "dismissal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rebate"}, {"LOWER": "payment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rebranding"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "recall"}, {"LOWER": "warrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "receipt"},
            {"LOWER": "of"},
            {"LOWER": "stolen"},
            {"LOWER": "money"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "receiver"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "receivers"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "receivers"}, {"LOWER": "of"}, {"LOWER": "company"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "receivers"}, {"LOWER": "remuneration"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "receivership"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "receivership"}, {"LOWER": "accounts"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "receiver"}, {"LOWER": "’s"}, {"LOWER": "remuneration"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "recipients"}, {"LOWER": "duty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reckless"}, {"LOWER": "driving"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recognisance"}, {"LOWER": "to"}, {"LOWER": "keep"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognised"},
            {"LOWER": "terms"},
            {"LOWER": "and"},
            {"LOWER": "conditions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "and"},
            {"LOWER": "enforcement"},
            {"LOWER": "in"},
            {"LOWER": "other"},
            {"LOWER": "member"},
            {"LOWER": "states"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "and"},
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "and"},
            {"LOWER": "registration"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "in"},
            {"LOWER": "other"},
            {"LOWER": "contracting"},
            {"LOWER": "states"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "divorce"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "judgment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "proceeding"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recognition"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reconsideration"},
            {"LOWER": "of"},
            {"LOWER": "earlier"},
            {"LOWER": "decision"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "recorded"}, {"LOWER": "information"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "recorder"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recording"}, {"LOWER": "of"}, {"LOWER": "complaint"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recordings"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recoverability"},
            {"LOWER": "against"},
            {"LOWER": "third"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recovery"},
            {"LOWER": "by"},
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recovery"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recovery"}, {"LOWER": "of"}, {"LOWER": "assets"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recovery"},
            {"LOWER": "of"},
            {"LOWER": "overpaid"},
            {"LOWER": "benefits"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recovery"}, {"LOWER": "of"}, {"LOWER": "overpayment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recovery"},
            {"LOWER": "of"},
            {"LOWER": "parking"},
            {"LOWER": "charge"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recovery"}, {"LOWER": "of"}, {"LOWER": "payments"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recovery"}, {"LOWER": "of"}, {"LOWER": "proceeds"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recovery"}, {"LOWER": "of"}, {"LOWER": "taxes"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recovery"},
            {"LOWER": "of"},
            {"LOWER": "trust"},
            {"LOWER": "property"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "recovery"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "recruitment"}, {"LOWER": "consultancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "recruitment"},
            {"LOWER": "of"},
            {"LOWER": "police"},
            {"LOWER": "officers"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rectification"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rectification"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rectification"}, {"LOWER": "of"}, {"LOWER": "register"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "redaction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "redress"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reduced"}, {"LOWER": "funding"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reduced"}, {"LOWER": "rate"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reduction"},
            {"LOWER": "of"},
            {"LOWER": "basic"},
            {"LOWER": "practice"},
            {"LOWER": "allowance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reduction"}, {"LOWER": "of"}, {"LOWER": "capital"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reduction"},
            {"LOWER": "of"},
            {"LOWER": "financial"},
            {"LOWER": "assistance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reduction"},
            {"LOWER": "or"},
            {"LOWER": "remission"},
            {"LOWER": "of"},
            {"LOWER": "rates"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "redundancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "redundant"}, {"LOWER": "font"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "by"},
            {"LOWER": "attorney"},
            {"LOWER": "general"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "by"},
            {"LOWER": "criminal"},
            {"LOWER": "cases"},
            {"LOWER": "review"},
            {"LOWER": "commission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "by"},
            {"LOWER": "enforcing"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "by"},
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "for"},
            {"LOWER": "preliminary"},
            {"LOWER": "ruling"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reference"}, {"LOWER": "to"}, {"LOWER": "adjudicator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reference"}, {"LOWER": "to"}, {"LOWER": "arbitration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "to"},
            {"LOWER": "central"},
            {"LOWER": "arbitration"},
            {"LOWER": "committee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "to"},
            {"LOWER": "city"},
            {"LOWER": "panel"},
            {"LOWER": "on"},
            {"LOWER": "take"},
            {"ORTH": "-"},
            {"LOWER": "overs"},
            {"LOWER": "and"},
            {"LOWER": "mergers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "to"},
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "justice"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reference"},
            {"LOWER": "to"},
            {"LOWER": "monopolies"},
            {"LOWER": "and"},
            {"LOWER": "mergers"},
            {"LOWER": "commission"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "referendum"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "referral"},
            {"LOWER": "by"},
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "for"},
            {"LOWER": "sentencing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "referral"}, {"LOWER": "of"}, {"LOWER": "application"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "refugee"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refugee"},
            {"LOWER": "or"},
            {"LOWER": "subsidiary"},
            {"LOWER": "protection"},
            {"LOWER": "status"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "refugee"}, {"LOWER": "status"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refugee"},
            {"ORTH": ","},
            {"LOWER": "intention"},
            {"LOWER": "to"},
            {"LOWER": "stay"},
            {"LOWER": "as"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refund"},
            {"LOWER": "on"},
            {"LOWER": "bad"},
            {"LOWER": "debts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "by"},
            {"LOWER": "national"},
            {"LOWER": "supervisory"},
            {"LOWER": "authority"},
            {"LOWER": "on"},
            {"LOWER": "grounds"},
            {"LOWER": "of"},
            {"LOWER": "confidentiality"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "accommodation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "amputation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "judicial"},
            {"LOWER": "review"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "asylum"},
            {"LOWER": "support"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "blood"},
            {"LOWER": "transfusion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "caesarean"},
            {"LOWER": "section"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "employment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "entry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "food"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "leave"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "offer"},
            {"LOWER": "of"},
            {"LOWER": "accommodation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "offered"},
            {"LOWER": "accommodation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "permission"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "permission"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "refund"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "relief"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "of"}, {"LOWER": "support"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "of"},
            {"LOWER": "travel"},
            {"LOWER": "document"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "to"}, {"LOWER": "consent"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "to"},
            {"LOWER": "stay"},
            {"LOWER": "or"},
            {"LOWER": "postpone"},
            {"LOWER": "criminal"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "refusal"}, {"LOWER": "to"}, {"LOWER": "transmit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "refusal"},
            {"LOWER": "to"},
            {"LOWER": "vary"},
            {"LOWER": "leave"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "regional"}, {"LOWER": "strategy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "register"}, {"LOWER": "of"}, {"LOWER": "members"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "register"},
            {"ORTH": ","},
            {"LOWER": "vacation"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "registered"}, {"LOWER": "design"}, {"LOWER": "right"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registered"},
            {"LOWER": "geographical"},
            {"LOWER": "indication"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registered"},
            {"LOWER": "sex"},
            {"LOWER": "offender"},
            {"LOWER": "required"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "details"},
            {"LOWER": "of"},
            {"LOWER": "bank"},
            {"ORTH": ","},
            {"LOWER": "debit"},
            {"LOWER": "or"},
            {"LOWER": "credit"},
            {"LOWER": "card"},
            {"LOWER": "accounts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registered"},
            {"LOWER": "social"},
            {"LOWER": "landlord"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registered"},
            {"LOWER": "trade"},
            {"LOWER": "marks"},
            {"LOWER": "with"},
            {"LOWER": "reputation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "registrars"}, {"LOWER": "discretion"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "registration"}, {"LOWER": "appeal"}, {"LOWER": "panel"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "registration"}, {"LOWER": "as"}, {"LOWER": "citizen"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registration"},
            {"LOWER": "as"},
            {"LOWER": "excise"},
            {"LOWER": "dealer"},
            {"LOWER": "and"},
            {"LOWER": "shipper"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registration"},
            {"LOWER": "in"},
            {"LOWER": "official"},
            {"LOWER": "roll"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registration"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "minders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registration"},
            {"LOWER": "of"},
            {"LOWER": "insurance"},
            {"LOWER": "companies"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registration"},
            {"LOWER": "of"},
            {"LOWER": "motor"},
            {"LOWER": "vehicles"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registration"},
            {"LOWER": "of"},
            {"LOWER": "resolutions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "registration"}, {"LOWER": "of"}, {"LOWER": "vessels"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "registration"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "registration"},
            {"ORTH": ","},
            {"LOWER": "evaluation"},
            {"LOWER": "and"},
            {"LOWER": "authorisation"},
            {"LOWER": "of"},
            {"LOWER": "chemical"},
            {"LOWER": "substances"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "regulated"}, {"LOWER": "activities"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "regulated"},
            {"LOWER": "mortgage"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "regulation"},
            {"LOWER": "of"},
            {"LOWER": "court"},
            {"LOWER": "advocacy"},
            {"LOWER": "standards"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "regulation"},
            {"LOWER": "of"},
            {"LOWER": "licensed"},
            {"LOWER": "activities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "regulation"},
            {"LOWER": "of"},
            {"LOWER": "postal"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "regulations"},
            {"LOWER": "1985"},
            {"ORTH": "("},
            {"LOWER": "s.i"},
            {"ORTH": "."},
            {"LOWER": "1985"},
            {"LOWER": "no"},
            {"LOWER": "."},
            {"LOWER": "16"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "sch"},
            {"ORTH": "."},
            {"LOWER": "1"},
            {"ORTH": ","},
            {"LOWER": "r."},
            {"LOWER": "12(2)(e"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "regulator"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "regulatory"}, {"LOWER": "enforcement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "regulatory"}, {"LOWER": "powers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "regulatory"},
            {"LOWER": "reform"},
            {"ORTH": "("},
            {"LOWER": "fire"},
            {"LOWER": "safety"},
            {"ORTH": ")"},
            {"LOWER": "order"},
            {"LOWER": "2005"},
            {"ORTH": "("},
            {"LOWER": "si"},
            {"LOWER": "2005/1541"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "32(1)(a"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rehearing"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reinstatement"},
            {"LOWER": "of"},
            {"LOWER": "applications"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "reinsurance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reinsurance"}, {"LOWER": "contract"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rejection"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "related"}, {"LOWER": "action"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "related"}, {"LOWER": "actions"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "related"},
            {"LOWER": "proceedings"},
            {"LOWER": "before"},
            {"LOWER": "tribunal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "relation"},
            {"LOWER": "back"},
            {"LOWER": "of"},
            {"LOWER": "trustees"},
            {"LOWER": "title"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "relationship"},
            {"LOWER": "akin"},
            {"LOWER": "to"},
            {"LOWER": "employment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "release"}, {"LOWER": "of"}, {"LOWER": "trustee"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "release"},
            {"LOWER": "on"},
            {"LOWER": "home"},
            {"LOWER": "detention"},
            {"LOWER": "curfew"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "release"}, {"LOWER": "on"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "release"}, {"LOWER": "on"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "relevance"}, {"LOWER": "of"}, {"LOWER": "documents"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "relevance"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "decisions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "relevant"}, {"LOWER": "merger"}, {"LOWER": "situation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "relevant"}, {"LOWER": "transfer"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reliance"},
            {"LOWER": "on"},
            {"LOWER": "closed"},
            {"LOWER": "material"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reliance"}, {"LOWER": "on"}, {"LOWER": "statement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "relief"}, {"LOWER": "from"}, {"LOWER": "forfeiture"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "relief"}, {"LOWER": "from"}, {"LOWER": "sanction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "relief"}, {"LOWER": "from"}, {"LOWER": "sanctions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "relief"}, {"LOWER": "of"}, {"LOWER": "poverty"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "religion"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "religion"}, {"LOWER": "or"}, {"LOWER": "belief"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "religion"}, {"LOWER": "or"}, {"LOWER": "belief"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "religious"},
            {"LOWER": "belief"},
            {"LOWER": "or"},
            {"LOWER": "political"},
            {"LOWER": "opinion"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "religious"}, {"LOWER": "upbringing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "remedies"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "remedy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "remoteness"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "removal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "removal"}, {"LOWER": "direction"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "removal"}, {"LOWER": "from"}, {"LOWER": "association"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "removal"}, {"LOWER": "from"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "removal"}, {"LOWER": "from"}, {"LOWER": "jurisdiction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "removal"}, {"LOWER": "of"}, {"LOWER": "arbitrator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "removal"}, {"LOWER": "of"}, {"LOWER": "documents"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "removal"},
            {"LOWER": "of"},
            {"LOWER": "name"},
            {"LOWER": "form"},
            {"LOWER": "record"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "removal"},
            {"LOWER": "of"},
            {"LOWER": "name"},
            {"LOWER": "from"},
            {"LOWER": "record"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "removal"}, {"LOWER": "of"}, {"LOWER": "trustee"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "remuera"}, {"LOWER": "bay"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "remuneration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "renewable"}, {"LOWER": "energy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "renewables"},
            {"LOWER": "obligation"},
            {"LOWER": "certificates"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "renewal"}, {"LOWER": "of"}, {"LOWER": "application"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "renewal"}, {"LOWER": "of"}, {"LOWER": "licence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "renewed"}, {"LOWER": "application"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "renewed"},
            {"LOWER": "applications"},
            {"LOWER": "for"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
            {"LOWER": "on"},
            {"LOWER": "different"},
            {"LOWER": "grounds"},
            {"LOWER": "from"},
            {"LOWER": "those"},
            {"LOWER": "refused"},
            {"LOWER": "by"},
            {"LOWER": "single"},
            {"LOWER": "judge"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "renewed"}, {"LOWER": "claim"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "renewed"},
            {"LOWER": "ex"},
            {"LOWER": "parte"},
            {"LOWER": "application"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rent"}, {"LOWER": "assessment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rent"}, {"LOWER": "payable"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rent"}, {"LOWER": "restriction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "rented"}, {"LOWER": "accommodation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reordering"},
            {"LOWER": "of"},
            {"LOWER": "church"},
            {"LOWER": "interior"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reordering"},
            {"LOWER": "of"},
            {"LOWER": "lady"},
            {"LOWER": "chapel"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "repair"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "repairs"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "repairs"},
            {"LOWER": "and"},
            {"LOWER": "reglazing"},
            {"LOWER": "of"},
            {"LOWER": "west"},
            {"LOWER": "window"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "repairs"}, {"LOWER": "to"}, {"LOWER": "chancel"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "repairs"},
            {"LOWER": "to"},
            {"LOWER": "church"},
            {"LOWER": "tower"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "repayment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "repayment"},
            {"LOWER": "of"},
            {"LOWER": "financial"},
            {"LOWER": "assistance"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "repeal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "replacement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "replacement"},
            {"LOWER": "of"},
            {"LOWER": "boiler"},
            {"LOWER": "and"},
            {"LOWER": "ancillary"},
            {"LOWER": "heating"},
            {"LOWER": "works"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "report"},
            {"LOWER": "of"},
            {"LOWER": "official"},
            {"LOWER": "referee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "reporting"},
            {"LOWER": "of"},
            {"LOWER": "judicial"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "reporting"}, {"LOWER": "restriction"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reporting"}, {"LOWER": "restrictions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "representation"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "representation"},
            {"ORTH": ","},
            {"LOWER": "grant"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "representation"},
            {"ORTH": ","},
            {"LOWER": "grants"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "representative"}, {"LOWER": "action"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reprocessed"}, {"LOWER": "substance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reproduction"}, {"LOWER": "right"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "repudiation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "repudiatory"}, {"LOWER": "breach"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "repudiatory"}, {"LOWER": "conduct"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "reputation"}, {"LOWER": "and"}, {"LOWER": "goodwill"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "request"}, {"LOWER": "for"}, {"LOWER": "arbitration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "request"},
            {"LOWER": "for"},
            {"LOWER": "new"},
            {"LOWER": "tenancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "request"},
            {"LOWER": "for"},
            {"LOWER": "no"},
            {"LOWER": "newspaper"},
            {"LOWER": "or"},
            {"LOWER": "other"},
            {"LOWER": "media"},
            {"LOWER": "report"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "request"},
            {"LOWER": "for"},
            {"LOWER": "preliminary"},
            {"LOWER": "ruling"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "request"},
            {"LOWER": "for"},
            {"LOWER": "temporary"},
            {"LOWER": "transfer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "requirement"},
            {"LOWER": "for"},
            {"LOWER": "authorisation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "requirement"}, {"LOWER": "or"}, {"LOWER": "condition"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "requirement"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "notice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "requirement"},
            {"LOWER": "to"},
            {"LOWER": "give"},
            {"LOWER": "security"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "requirement"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "directions"},
            {"LOWER": "specifying"},
            {"LOWER": "timely"},
            {"LOWER": "service"},
            {"LOWER": "of"},
            {"LOWER": "documents"},
            {"LOWER": "on"},
            {"LOWER": "litigants"},
            {"LOWER": "in"},
            {"LOWER": "person"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "res"}, {"LOWER": "cogitans"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "res"}, {"LOWER": "ipsa"}, {"LOWER": "loquitur"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rescission"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rescission"}, {"LOWER": "of"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rescuer"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "research"},
            {"LOWER": "and"},
            {"LOWER": "technological"},
            {"LOWER": "development"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "reserved"}, {"LOWER": "judgment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "reserved"}, {"LOWER": "judgments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "residence"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "residence"}, {"LOWER": "test"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "residential"}, {"LOWER": "accommodation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "residential"}, {"LOWER": "care"}, {"LOWER": "homes"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "residential"}, {"LOWER": "occupier"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "residential"}, {"LOWER": "property"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "residential"},
            {"LOWER": "tenancy"},
            {"LOWER": "agreement"},
            {"LOWER": "between"},
            {"LOWER": "landlord"},
            {"LOWER": "and"},
            {"LOWER": "tenant"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "residuary"}, {"LOWER": "legatee"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "resiting"},
            {"LOWER": "of"},
            {"LOWER": "war"},
            {"LOWER": "memorial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "resolution"},
            {"LOWER": "of"},
            {"LOWER": "disputes"},
            {"LOWER": "between"},
            {"LOWER": "undertakings"},
            {"LOWER": "providing"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "family"},
            {"LOWER": "and"},
            {"LOWER": "private"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "family"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "respect"}, {"LOWER": "for"}, {"LOWER": "home"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "private"},
            {"LOWER": "and"},
            {"LOWER": "family"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "private"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "respondents"},
            {"LOWER": "claiming"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "parents"},
            {"LOWER": "of"},
            {"LOWER": "child"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "respondent"}, {"LOWER": "’s"}, {"LOWER": "behaviour"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "responsibility"},
            {"LOWER": "for"},
            {"LOWER": "publication"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rest"}, {"LOWER": "periods"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restitution"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restoration"},
            {"LOWER": "of"},
            {"LOWER": "names"},
            {"LOWER": "to"},
            {"LOWER": "register"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restoration"},
            {"LOWER": "of"},
            {"LOWER": "priority"},
            {"LOWER": "rights"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "restoration"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restoration"}, {"LOWER": "to"}, {"LOWER": "register"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restoration"}, {"LOWER": "to"}, {"LOWER": "register"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restrain"}, {"LOWER": "of"}, {"LOWER": "trade"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "restraining"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restraint"},
            {"LOWER": "of"},
            {"LOWER": "activity"},
            {"LOWER": "pending"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restraint"},
            {"LOWER": "of"},
            {"LOWER": "decision"},
            {"LOWER": "pending"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restraint"},
            {"LOWER": "of"},
            {"LOWER": "domestic"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restraint"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restraint"}, {"LOWER": "of"}, {"LOWER": "trade"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restraint"}, {"LOWER": "of"}, {"LOWER": "trade"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restraint"}, {"LOWER": "of"}, {"LOWER": "trade"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restraint"},
            {"LOWER": "on"},
            {"LOWER": "leaving"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "restraint"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restricted"}, {"LOWER": "patient"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restriction"},
            {"LOWER": "of"},
            {"LOWER": "private"},
            {"LOWER": "access"},
            {"LOWER": "to"},
            {"LOWER": "highway"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restriction"}, {"LOWER": "on"}, {"LOWER": "access"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restriction"},
            {"LOWER": "on"},
            {"LOWER": "publication"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restriction"}, {"LOWER": "on"}, {"LOWER": "return"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restriction"},
            {"LOWER": "on"},
            {"LOWER": "termination"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restriction"},
            {"LOWER": "or"},
            {"LOWER": "distortion"},
            {"LOWER": "of"},
            {"LOWER": "competition"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restriction"},
            {"LOWER": "or"},
            {"LOWER": "distortion"},
            {"LOWER": "of"},
            {"LOWER": "competition"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "restriction"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restriction"},
            {"LOWER": "pending"},
            {"LOWER": "deportation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restrictions"}, {"LOWER": "on"}, {"LOWER": "access"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restrictions"}, {"LOWER": "on"}, {"LOWER": "reporting"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restrictions"}, {"LOWER": "on"}, {"LOWER": "return"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "restrictions"}, {"LOWER": "on"}, {"LOWER": "shares"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restrictions"},
            {"LOWER": "on"},
            {"LOWER": "taking"},
            {"LOWER": "financial"},
            {"LOWER": "advantage"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "restrictive"}, {"LOWER": "covenant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restrictive"}, {"LOWER": "covenant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restrictive"}, {"LOWER": "covenant"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restrictive"}, {"LOWER": "measures"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restrictive"},
            {"LOWER": "measures"},
            {"LOWER": "adopted"},
            {"LOWER": "against"},
            {"LOWER": "persons"},
            {"LOWER": "and"},
            {"LOWER": "entities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "restrictive"},
            {"LOWER": "measures"},
            {"LOWER": "taken"},
            {"LOWER": "against"},
            {"LOWER": "persons"},
            {"LOWER": "and"},
            {"LOWER": "entities"},
            {"LOWER": "associated"},
            {"LOWER": "with"},
            {"LOWER": "usama"},
            {"LOWER": "bin"},
            {"LOWER": "laden"},
            {"ORTH": ","},
            {"LOWER": "al"},
            {"ORTH": "-"},
            {"LOWER": "qaeda"},
            {"LOWER": "and"},
            {"LOWER": "taliban"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "restrictive"}, {"LOWER": "practice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restrictive"}, {"LOWER": "practices"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restrictive"}, {"LOWER": "practices"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restructuring"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "restructuring"}, {"LOWER": "aid"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "resulting"}, {"LOWER": "trust"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retail"},
            {"LOWER": "food"},
            {"LOWER": "products"},
            {"LOWER": "comparison"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "retainer"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retention"},
            {"LOWER": "of"},
            {"LOWER": "conviction"},
            {"LOWER": "data"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retention"},
            {"LOWER": "of"},
            {"LOWER": "conviction"},
            {"LOWER": "data"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "retention"}, {"LOWER": "of"}, {"LOWER": "evidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "retention"}, {"LOWER": "of"}, {"LOWER": "evidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "retention"}, {"LOWER": "of"}, {"LOWER": "money"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retention"},
            {"LOWER": "of"},
            {"LOWER": "personal"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retention"},
            {"LOWER": "of"},
            {"LOWER": "personal"},
            {"LOWER": "information"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retention"},
            {"LOWER": "of"},
            {"LOWER": "seized"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "retention"}, {"LOWER": "of"}, {"LOWER": "title"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "retirement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "retirement"}, {"LOWER": "of"}, {"LOWER": "tenant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "retirement"}, {"LOWER": "pension"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "retirement"}, {"LOWER": "pension"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "retirement"}, {"LOWER": "relief"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retirement"},
            {"LOWER": "to"},
            {"LOWER": "consider"},
            {"LOWER": "verdict"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "retrenchment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "retrial"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retrial"},
            {"LOWER": "following"},
            {"LOWER": "on"},
            {"LOWER": "immediately"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "retroactive"}, {"LOWER": "effect"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "retrospective"}, {"LOWER": "application"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "retrospective"},
            {"LOWER": "criminal"},
            {"LOWER": "law"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "retrospectivity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "return"}, {"LOWER": "decision"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "return"}, {"LOWER": "to"}, {"LOWER": "work"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "return"},
            {"LOWER": "to"},
            {"LOWER": "work"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "revenue"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "revenue"}, {"LOWER": "list"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "revenue"}, {"ORTH": "—"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "review"}, {"LOWER": "of"}, {"LOWER": "decision"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "review"}, {"LOWER": "of"}, {"LOWER": "detention"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "review"},
            {"LOWER": "of"},
            {"LOWER": "local"},
            {"LOWER": "authority"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "review"},
            {"LOWER": "of"},
            {"LOWER": "original"},
            {"LOWER": "judgment"},
            {"LOWER": "granting"},
            {"LOWER": "faculty"},
            {"LOWER": "following"},
            {"LOWER": "parochial"},
            {"LOWER": "church"},
            {"LOWER": "council"},
            {"LOWER": "objections"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "review"}, {"LOWER": "of"}, {"LOWER": "services"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "review"}, {"LOWER": "or"}, {"LOWER": "rehearing"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "revocation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "revocation"},
            {"LOWER": "of"},
            {"LOWER": "debt"},
            {"LOWER": "relief"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "revocation"}, {"LOWER": "of"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "revocation"},
            {"LOWER": "of"},
            {"LOWER": "order"},
            {"LOWER": "dispensing"},
            {"LOWER": "with"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "revocation"},
            {"LOWER": "of"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "rialto"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "in"}, {"LOWER": "equity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "access"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "audience"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "cancellation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "economic"},
            {"LOWER": "operator"},
            {"LOWER": "to"},
            {"LOWER": "rely"},
            {"LOWER": "on"},
            {"LOWER": "capacities"},
            {"LOWER": "of"},
            {"LOWER": "other"},
            {"LOWER": "entities"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "fencing"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "movement"},
            {"LOWER": "and"},
            {"LOWER": "residence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "licensee"},
            {"LOWER": "to"},
            {"LOWER": "sue"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "market"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "occupation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "pre"},
            {"LOWER": "-"},
            {"LOWER": "emption"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "reproduction"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "sole"},
            {"LOWER": "vesture"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "of"}, {"LOWER": "way"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "accompanied"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "heard"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "represented"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "bring"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "buy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "charge"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "election"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "enter"},
            {"LOWER": "and"},
            {"LOWER": "reside"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "fair"},
            {"LOWER": "hearing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "fair"},
            {"LOWER": "trial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "family"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "expression"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "information"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "inherit"},
            {"LOWER": "land"},
            {"LOWER": "in"},
            {"LOWER": "england"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "jury"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "legal"},
            {"LOWER": "advice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "liberty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "liberty"},
            {"LOWER": "and"},
            {"LOWER": "security"},
            {"LOWER": "of"},
            {"LOWER": "person"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "life"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "light"},
            {"LOWER": "by"},
            {"LOWER": "prescription"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "manage"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "marry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "move"},
            {"LOWER": "and"},
            {"LOWER": "reside"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "peaceful"},
            {"LOWER": "enjoyment"},
            {"LOWER": "of"},
            {"LOWER": "possessions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "prevent"},
            {"LOWER": "re"},
            {"LOWER": "-"},
            {"LOWER": "utilisation"},
            {"LOWER": "of"},
            {"LOWER": "contents"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "protection"},
            {"LOWER": "against"},
            {"LOWER": "degrading"},
            {"LOWER": "treatment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "remain"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "reside"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "family"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "private"},
            {"LOWER": "and"},
            {"LOWER": "family"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "private"},
            {"LOWER": "and"},
            {"LOWER": "family"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "respect"},
            {"LOWER": "for"},
            {"LOWER": "private"},
            {"LOWER": "life"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "see"},
            {"LOWER": "documents"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "select"},
            {"LOWER": "legal"},
            {"LOWER": "representative"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "send"},
            {"LOWER": "sewage"},
            {"LOWER": "into"},
            {"LOWER": "sewers"},
            {"LOWER": "of"},
            {"LOWER": "adjoining"},
            {"LOWER": "district"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "trade"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "vote"},
            {"LOWER": "in"},
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "and"},
            {"LOWER": "municipal"},
            {"LOWER": "elections"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "right"}, {"LOWER": "to"}, {"LOWER": "work"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rights"},
            {"LOWER": "against"},
            {"LOWER": "company"},
            {"LOWER": "or"},
            {"LOWER": "directors"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rights"}, {"LOWER": "attached"}, {"LOWER": "to"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rights"}, {"LOWER": "of"}, {"LOWER": "audience"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "entry"},
            {"LOWER": "and"},
            {"LOWER": "residence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "lessor"},
            {"LOWER": "against"},
            {"LOWER": "third"},
            {"LOWER": "parties"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "media"},
            {"LOWER": "to"},
            {"LOWER": "attend"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "original"},
            {"LOWER": "landlowner"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rights"},
            {"LOWER": "of"},
            {"LOWER": "use"},
            {"LOWER": "for"},
            {"LOWER": "radio"},
            {"LOWER": "frequencies"},
            {"LOWER": "and"},
            {"LOWER": "numbers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rights"}, {"LOWER": "of"}, {"LOWER": "way"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rights"}, {"LOWER": "to"}, {"LOWER": "remuneration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rights"},
            {"LOWER": "to"},
            {"LOWER": "use"},
            {"LOWER": "radio"},
            {"LOWER": "frequencies"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "ring"},
            {"ORTH": "-"},
            {"LOWER": "fencing"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "riot"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "risk"}, {"LOWER": "assessment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "risk"}, {"LOWER": "of"}, {"LOWER": "confusion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "risk"},
            {"LOWER": "of"},
            {"LOWER": "domestic"},
            {"LOWER": "violence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "risk"},
            {"LOWER": "of"},
            {"LOWER": "physical"},
            {"LOWER": "violence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "risk"},
            {"LOWER": "of"},
            {"LOWER": "suicide"},
            {"LOWER": "by"},
            {"LOWER": "reason"},
            {"LOWER": "of"},
            {"LOWER": "mental"},
            {"LOWER": "condition"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "risk"},
            {"LOWER": "of"},
            {"LOWER": "violence"},
            {"LOWER": "from"},
            {"LOWER": "neighbour"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "river"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "river"}, {"LOWER": "rima"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "riverine"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "road"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "road"}, {"LOWER": "traffic"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "road"}, {"LOWER": "traffic"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "road"}, {"LOWER": "traffic"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "road"}, {"LOWER": "traffic"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "road"}, {"LOWER": "traffic"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "road"}, {"LOWER": "traffic"}, {"LOWER": "accident"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "road"}, {"LOWER": "transport"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "road"}, {"LOWER": "transport"}, {"LOWER": "industry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "road"}, {"LOWER": "widening"}, {"LOWER": "scheme"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "robbery"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "roll"},
            {"ORTH": "-"},
            {"LOWER": "over"},
            {"LOWER": "relief"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "romanian"},
            {"LOWER": "child"},
            {"LOWER": "privately"},
            {"LOWER": "fostered"},
            {"LOWER": "in"},
            {"LOWER": "the"},
            {"LOWER": "united"},
            {"LOWER": "kingdom"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "romanian"},
            {"LOWER": "judgment"},
            {"LOWER": "giving"},
            {"LOWER": "care"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "to"},
            {"LOWER": "father"},
            {"LOWER": "registered"},
            {"LOWER": "in"},
            {"LOWER": "english"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "royal"}, {"LOWER": "charter"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "royal"},
            {"LOWER": "courts"},
            {"LOWER": "of"},
            {"LOWER": "justice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "royal"},
            {"LOWER": "navy"},
            {"LOWER": "resettlement"},
            {"LOWER": "grant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "royal"},
            {"LOWER": "prerogative"},
            {"LOWER": "of"},
            {"LOWER": "mercy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rule"},
            {"LOWER": "book"},
            {"LOWER": "of"},
            {"LOWER": "political"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rule"}, {"LOWER": "of"}, {"LOWER": "law"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rule"},
            {"LOWER": "restricting"},
            {"LOWER": "enforcement"},
            {"LOWER": "of"},
            {"LOWER": "costs"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rule"},
            {"LOWER": "that"},
            {"LOWER": "leasehold"},
            {"LOWER": "term"},
            {"LOWER": "can"},
            {"LOWER": "not"},
            {"LOWER": "be"},
            {"LOWER": "uncertain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rules"}, {"LOWER": "against"}, {"LOWER": "overlapping"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rules"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "rules"},
            {"LOWER": "of"},
            {"LOWER": "unincorporated"},
            {"LOWER": "association"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rylands"}, {"LOWER": "v"}, {"LOWER": "fletcher"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rylands"}, {"LOWER": "v"}, {"LOWER": "fletcher"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "rylands"}, {"LOWER": "v."}, {"LOWER": "fletcher"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "safe"},
            {"LOWER": "operation"},
            {"LOWER": "of"},
            {"LOWER": "ship"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "safe"},
            {"LOWER": "place"},
            {"LOWER": "of"},
            {"LOWER": "work"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "safe"}, {"LOWER": "port"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "safe"},
            {"LOWER": "system"},
            {"LOWER": "of"},
            {"LOWER": "work"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "safeguarding"}, {"LOWER": "checks"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "safeguarding"},
            {"LOWER": "children"},
            {"LOWER": "measures"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "safety"}, {"LOWER": "regulations"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "saint"},
            {"LOWER": "christopher"},
            {"LOWER": "and"},
            {"LOWER": "nevis"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "saint"},
            {"LOWER": "christopher"},
            {"LOWER": "and"},
            {"LOWER": "nevis"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "saint"}, {"LOWER": "lucia"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "saint"},
            {"LOWER": "vincent"},
            {"LOWER": "and"},
            {"LOWER": "the"},
            {"LOWER": "grenadines"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sale"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sale"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sale"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "and"},
            {"LOWER": "purchase"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "by"}, {"LOWER": "mortgagee"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "alcohol"},
            {"LOWER": "to"},
            {"LOWER": "person"},
            {"LOWER": "under"},
            {"LOWER": "18"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "armet"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "birds"},
            {"LOWER": "at"},
            {"LOWER": "bird"},
            {"LOWER": "fair"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "chattels"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "consumer"},
            {"LOWER": "goods"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "council"},
            {"LOWER": "houses"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "hotel"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "land"},
            {"LOWER": "subject"},
            {"LOWER": "to"},
            {"LOWER": "lease"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "property"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "residential"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "of"}, {"LOWER": "ship"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "of"},
            {"LOWER": "trust"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "or"},
            {"LOWER": "lease"},
            {"LOWER": "of"},
            {"LOWER": "trust"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sale"}, {"LOWER": "to"}, {"LOWER": "developer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sale"},
            {"LOWER": "to"},
            {"LOWER": "developer"},
            {"ORTH": "--"},
            {"LOWER": "council"},
            {"LOWER": "flats"},
            {"LOWER": "in"},
            {"LOWER": "state"},
            {"LOWER": "of"},
            {"LOWER": "disrepair"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "salesman"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "salvage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "same"}, {"LOWER": "employment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "same"},
            {"ORTH": "-"},
            {"LOWER": "sex"},
            {"LOWER": "couple"},
            {"LOWER": "entering"},
            {"LOWER": "into"},
            {"LOWER": "surrogacy"},
            {"LOWER": "agreement"},
            {"LOWER": "in"},
            {"LOWER": "india"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sanction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sanction"}, {"LOWER": "hearing"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sanction"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "satellite"},
            {"LOWER": "broadcasting"},
            {"LOWER": "of"},
            {"LOWER": "football"},
            {"LOWER": "matches"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "schedule"}, {"LOWER": "a"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "schedule"},
            {"LOWER": "d"},
            {"ORTH": ","},
            {"LOWER": "case"},
            {"LOWER": "i"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "schedule"},
            {"LOWER": "d"},
            {"ORTH": ","},
            {"LOWER": "case"},
            {"LOWER": "iii"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scheme"}, {"LOWER": "of"}, {"LOWER": "arrangemenrt"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scheme"}, {"LOWER": "of"}, {"LOWER": "arrangement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "schemes"}, {"LOWER": "and"}, {"LOWER": "reductions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "school"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "school"}, {"LOWER": "site"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "schools"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "scientific"}, {"LOWER": "adviser"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "scientific"}, {"LOWER": "procedures"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "of"}, {"LOWER": "appeal"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "of"}, {"LOWER": "convention"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "of"}, {"LOWER": "duty"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "scope"},
            {"LOWER": "of"},
            {"LOWER": "european"},
            {"LOWER": "union"},
            {"LOWER": "law"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "scope"},
            {"LOWER": "of"},
            {"LOWER": "exemption"},
            {"LOWER": "of"},
            {"LOWER": "management"},
            {"LOWER": "of"},
            {"LOWER": "special"},
            {"LOWER": "investment"},
            {"LOWER": "funds"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "of"}, {"LOWER": "inquest"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "of"}, {"LOWER": "regulation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "of"}, {"LOWER": "retainer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "of"}, {"LOWER": "scheme"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "scope"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "liability"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scope"}, {"LOWER": "ratione"}, {"LOWER": "personae"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scotch"}, {"LOWER": "assets"}, {"ORTH": "."}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "scotland"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "scotland"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "scottish"}, {"LOWER": "peers"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "scottish"}, {"LOWER": "permanence"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "screening"}, {"LOWER": "assessment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sea"}, {"LOWER": "fishing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "seal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "seaman"}, {"LOWER": "’s"}, {"LOWER": "will"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "search"}, {"LOWER": "of"}, {"LOWER": "premises"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "search"}, {"LOWER": "order"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "search"}, {"LOWER": "upon"}, {"LOWER": "arrest"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "search"}, {"LOWER": "warrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "search"},
            {"ORTH": ","},
            {"LOWER": "refusal"},
            {"LOWER": "to"},
            {"LOWER": "permit"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "search"},
            {"ORTH": ","},
            {"LOWER": "right"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "seaspeed"}, {"LOWER": "dora"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "seaworthiness"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "second"}, {"LOWER": "action"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "second"},
            {"LOWER": "defendant"},
            {"LOWER": "bringing"},
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "losses"},
            {"LOWER": "following"},
            {"LOWER": "liquidation"},
            {"LOWER": "of"},
            {"LOWER": "company"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "secondary"}, {"LOWER": "loan"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "secret"}, {"LOWER": "commission"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "ordering"},
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "extradition"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "refusing"},
            {"LOWER": "asylum"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "refusing"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "refusing"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "asylum"},
            {"LOWER": "application"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "rejecting"},
            {"LOWER": "claimant"},
            {"LOWER": "’s"},
            {"LOWER": "application"},
            {"LOWER": "for"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "states"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "states"},
            {"LOWER": "powers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "certification"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secretary"},
            {"LOWER": "of"},
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "functions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "section"}, {"LOWER": "47"}, {"LOWER": "inquiry"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "secure"}, {"LOWER": "accommodation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secure"},
            {"LOWER": "accommodation"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "secure"},
            {"LOWER": "accommodation"},
            {"LOWER": "orders"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "secure"}, {"LOWER": "tenancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "secure"}, {"LOWER": "tenancy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "secure"}, {"LOWER": "training"}, {"LOWER": "centre"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "secured"}, {"LOWER": "creditor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "securities"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "security"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "security"}, {"LOWER": "categorisation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "security"}, {"LOWER": "for"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "security"}, {"LOWER": "for"}, {"LOWER": "costs"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "security"}, {"LOWER": "of"}, {"LOWER": "tenure"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sedition"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "segregation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "seizure"}, {"LOWER": "of"}, {"LOWER": "evidence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "seizure"}, {"LOWER": "of"}, {"LOWER": "vehicle"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "selection"}, {"LOWER": "for"}, {"LOWER": "promotion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "selection"}, {"LOWER": "of"}, {"LOWER": "claimants"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "self"}, {"LOWER": "assessment"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "self"}, {"ORTH": "-"}, {"LOWER": "assessment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "self"}, {"ORTH": "-"}, {"LOWER": "defence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "self"},
            {"ORTH": "-"},
            {"LOWER": "employed"},
            {"LOWER": "commercial"},
            {"LOWER": "agents"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sending"},
            {"LOWER": "by"},
            {"LOWER": "means"},
            {"LOWER": "of"},
            {"LOWER": "public"},
            {"LOWER": "electronic"},
            {"LOWER": "communications"},
            {"LOWER": "network"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sending"},
            {"LOWER": "by"},
            {"LOWER": "means"},
            {"LOWER": "of"},
            {"LOWER": "public"},
            {"LOWER": "electronic"},
            {"LOWER": "communications"},
            {"LOWER": "network"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sending"}, {"LOWER": "for"}, {"LOWER": "trial"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "senior"}, {"LOWER": "courts"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "senior"},
            {"LOWER": "facilities"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sensitive"}, {"LOWER": "material"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sentence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sentence"}, {"LOWER": "for"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sentence"}, {"LOWER": "of"}, {"LOWER": "deprivation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sentence"}, {"LOWER": "of"}, {"LOWER": "flogging"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sentence"},
            {"LOWER": "of"},
            {"LOWER": "imprisonment"},
            {"LOWER": "for"},
            {"LOWER": "public"},
            {"LOWER": "protection"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sentence"},
            {"LOWER": "of"},
            {"LOWER": "imprisonment"},
            {"LOWER": "in"},
            {"LOWER": "default"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sentence"},
            {"ORTH": ","},
            {"LOWER": "appeal"},
            {"LOWER": "against"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sentencing"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sentencing"}, {"LOWER": "guidelines"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sentencing"}, {"LOWER": "powers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "separate"}, {"LOWER": "estate"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sequana"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sequestration"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sequestrator"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "series"},
            {"LOWER": "of"},
            {"LOWER": "trade"},
            {"LOWER": "marks"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "serious"}, {"LOWER": "case"}, {"LOWER": "reviews"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "serious"},
            {"LOWER": "crime"},
            {"LOWER": "prevention"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "serious"}, {"LOWER": "fraud"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "serious"}, {"LOWER": "harm"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "serious"}, {"LOWER": "irregularity"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "serious"},
            {"LOWER": "procedural"},
            {"LOWER": "irregularity"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "seriousness"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "service"}, {"LOWER": "charge"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "confidence"},
            {"LOWER": "procedure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "of"},
            {"LOWER": "answer"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "time"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "service"}, {"LOWER": "of"}, {"LOWER": "documents"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "service"}, {"LOWER": "of"}, {"LOWER": "notice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "of"},
            {"LOWER": "notice"},
            {"LOWER": "of"},
            {"LOWER": "motion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "of"},
            {"LOWER": "statement"},
            {"LOWER": "of"},
            {"LOWER": "claim"},
            {"LOWER": "without"},
            {"LOWER": "leave"},
            {"LOWER": "on"},
            {"LOWER": "defendant"},
            {"LOWER": "bank"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "new"},
            {"LOWER": "zealand"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "of"},
            {"ORTH": ","},
            {"LOWER": "or"},
            {"LOWER": "for"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "out"},
            {"LOWER": "of"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "service"},
            {"LOWER": "provider"},
            {"LOWER": "filling"},
            {"LOWER": "ready"},
            {"LOWER": "prepared"},
            {"LOWER": "tins"},
            {"LOWER": "with"},
            {"LOWER": "fizzy"},
            {"LOWER": "drink"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "service"}, {"LOWER": "provision"}, {"LOWER": "change"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "serviceman"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "services"},
            {"LOWER": "offered"},
            {"LOWER": "in"},
            {"LOWER": "spain"},
            {"LOWER": "by"},
            {"LOWER": "pension"},
            {"LOWER": "funds"},
            {"LOWER": "and"},
            {"LOWER": "insurance"},
            {"LOWER": "companies"},
            {"LOWER": "established"},
            {"LOWER": "in"},
            {"LOWER": "another"},
            {"LOWER": "member"},
            {"LOWER": "state"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "servicing"}, {"LOWER": "agreement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "set"}, {"LOWER": "aside"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "set"}, {"ORTH": "-"}, {"LOWER": "off"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "setting"},
            {"LOWER": "aside"},
            {"LOWER": "previous"},
            {"LOWER": "decision"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "setting"}, {"LOWER": "down"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "setting"}, {"LOWER": "down"}, {"LOWER": "action"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "settled"}, {"LOWER": "estate"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "settled"}, {"LOWER": "land"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "settled"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "settlement"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "settlement"}, {"LOWER": "of"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "settlement"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "settlement"}, {"LOWER": "sum"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "several"}, {"LOWER": "defendants"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "several"}, {"LOWER": "fishery"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "several"}, {"LOWER": "parties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "several"}, {"LOWER": "plaintiffs"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "severance"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "severely"},
            {"LOWER": "disabled"},
            {"LOWER": "man"},
            {"LOWER": "subject"},
            {"LOWER": "to"},
            {"LOWER": "restraints"},
            {"LOWER": "to"},
            {"LOWER": "prevent"},
            {"LOWER": "self"},
            {"ORTH": "-"},
            {"LOWER": "harm"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sewage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sewerage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sex"}, {"LOWER": "discrimination"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sex"}, {"LOWER": "establishment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sex"}, {"LOWER": "establishments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sex"}, {"LOWER": "offender"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sex"},
            {"LOWER": "offender"},
            {"LOWER": "notification"},
            {"LOWER": "requirements"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sex"}, {"LOWER": "offender"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "assault"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "offence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "offences"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "offences"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "offences"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sexual"},
            {"LOWER": "offences"},
            {"LOWER": "prevention"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "orientation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "orientation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sexual"}, {"LOWER": "relations"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sexual"},
            {"LOWER": "relations"},
            {"ORTH": ","},
            {"LOWER": "consent"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sexual"},
            {"LOWER": "touching"},
            {"LOWER": "of"},
            {"LOWER": "person"},
            {"LOWER": "with"},
            {"LOWER": "mental"},
            {"LOWER": "disorder"},
            {"LOWER": "impeding"},
            {"LOWER": "choice"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sham"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "shape"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "share"}, {"LOWER": "option"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "share"}, {"LOWER": "option"}, {"LOWER": "agreement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "share"}, {"LOWER": "purchase"}, {"LOWER": "agreement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "share"},
            {"LOWER": "subscription"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "share"},
            {"ORTH": "-"},
            {"LOWER": "sale"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "shared"}, {"LOWER": "assumption"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "shareholder"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "shareholders"}, {"LOWER": "’"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "shares"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "shares"},
            {"LOWER": "vested"},
            {"LOWER": "in"},
            {"LOWER": "trustees"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sheltered"}, {"LOWER": "accommodation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sheriff"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sheriffs"},
            {"LOWER": "representative"},
            {"LOWER": "to"},
            {"LOWER": "inform"},
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "date"},
            {"LOWER": "of"},
            {"LOWER": "expiration"},
            {"LOWER": "of"},
            {"LOWER": "writ"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "shipbuilding"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "shipowner"}, {"LOWER": "’s"}, {"LOWER": "lien"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "shipping"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "shipping"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "shipping"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ships"}, {"LOWER": "name"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ships"}, {"LOWER": "names"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ships"}, {"LOWER": "names"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ships"}, {"LOWER": "names"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ships"}, {"LOWER": "’"}, {"LOWER": "names"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ship"}, {"LOWER": "’s"}, {"LOWER": "names"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "shop"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "shop"}, {"LOWER": "stewards"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "short"},
            {"LOWER": "term"},
            {"LOWER": "investment"},
            {"LOWER": "account"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "short"}, {"LOWER": "time"}, {"LOWER": "working"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "short"}, {"LOWER": "warned"}, {"LOWER": "list"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "short"},
            {"ORTH": "-"},
            {"LOWER": "time"},
            {"LOWER": "working"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sick"},
            {"LOWER": "and"},
            {"LOWER": "disabled"},
            {"LOWER": "persons"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sick"}, {"LOWER": "pay"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sick"},
            {"LOWER": "pay"},
            {"LOWER": "benefit"},
            {"LOWER": "scheme"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sickness"},
            {"LOWER": "and"},
            {"LOWER": "injury"},
            {"LOWER": "pensions"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sickness"}, {"LOWER": "benefit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "signature"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "significant"}, {"LOWER": "harm"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "signs"},
            {"LOWER": "capable"},
            {"LOWER": "of"},
            {"LOWER": "constituting"},
            {"LOWER": "trade"},
            {"LOWER": "marks"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "similar"}, {"LOWER": "facts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "singapore"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "single"}, {"LOWER": "commissioner"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "single"},
            {"LOWER": "judge"},
            {"ORTH": ","},
            {"LOWER": "appeal"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "single"}, {"LOWER": "meaning"}, {"LOWER": "rule"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "single"}, {"LOWER": "payment"}, {"LOWER": "scheme"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "site"},
            {"LOWER": "conversion"},
            {"LOWER": "and"},
            {"LOWER": "building"},
            {"LOWER": "project"},
            {"LOWER": "not"},
            {"LOWER": "carried"},
            {"LOWER": "out"},
            {"LOWER": "in"},
            {"LOWER": "accordance"},
            {"LOWER": "with"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "site"},
            {"LOWER": "licence"},
            {"ORTH": ","},
            {"LOWER": "exemptions"},
            {"LOWER": "from"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "site"},
            {"LOWER": "of"},
            {"LOWER": "special"},
            {"LOWER": "scientific"},
            {"LOWER": "interest"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "site"},
            {"LOWER": "visit"},
            {"LOWER": "by"},
            {"LOWER": "judge"},
            {"LOWER": "and"},
            {"LOWER": "experts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sixth"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "77/388"},
            {"LOWER": "/"},
            {"LOWER": "eec"},
            {"ORTH": "("},
            {"LOWER": "as"},
            {"LOWER": "amended"},
            {"LOWER": "by"},
            {"LOWER": "council"},
            {"LOWER": "directive"},
            {"LOWER": "2004/7"},
            {"LOWER": "/"},
            {"LOWER": "ec"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "art"},
            {"LOWER": "11a(1)(a"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "27(1"},
            {"ORTH": ")"},
            {"ORTH": "("},
            {"LOWER": "5"},
            {"ORTH": ")"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "skeleton"}, {"LOWER": "arguments"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "skipper"}, {"LOWER": "i"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "slander"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "slippery"}, {"LOWER": "floor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "small"},
            {"LOWER": "amount"},
            {"LOWER": "of"},
            {"LOWER": "assets"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "smell"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "smoking"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "smoking"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "social"}, {"LOWER": "fund"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "social"}, {"LOWER": "security"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "social"}, {"LOWER": "security"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "social"}, {"LOWER": "security"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "social"}, {"LOWER": "security"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "social"},
            {"LOWER": "security"},
            {"LOWER": "adjudication"},
            {"LOWER": "officer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "social"}, {"LOWER": "security"}, {"LOWER": "benefit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "social"}, {"LOWER": "security"}, {"LOWER": "benefits"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "social"},
            {"LOWER": "security"},
            {"LOWER": "for"},
            {"LOWER": "migrant"},
            {"LOWER": "workers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "solar"},
            {"LOWER": "energy"},
            {"LOWER": "certification"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "soldier"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "soldier"},
            {"LOWER": "on"},
            {"LOWER": "active"},
            {"LOWER": "service"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "soldier"},
            {"LOWER": "on"},
            {"LOWER": "active"},
            {"LOWER": "service"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "soldiers"},
            {"LOWER": "on"},
            {"LOWER": "active"},
            {"LOWER": "service"},
            {"LOWER": "abroad"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "soldiers"}, {"LOWER": "serving"}, {"LOWER": "overseas"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sole"}, {"LOWER": "interest"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sole"},
            {"LOWER": "or"},
            {"LOWER": "main"},
            {"LOWER": "residence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "sole"},
            {"LOWER": "trader"},
            {"LOWER": "’s"},
            {"LOWER": "business"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "soliciting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "solicitor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "solicitor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "solicitor"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "solicitor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "solicitor"},
            {"LOWER": "acting"},
            {"LOWER": "on"},
            {"LOWER": "own"},
            {"LOWER": "account"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "solicitor"}, {"LOWER": "and"}, {"LOWER": "client"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "solicitors"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "solicitors"},
            {"LOWER": "and"},
            {"LOWER": "counsels"},
            {"LOWER": "fees"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "solicitors"}, {"LOWER": "client"}, {"LOWER": "account"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "solicitors"},
            {"LOWER": "disciplinary"},
            {"LOWER": "appeals"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "solicitors"}, {"LOWER": "fees"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "solicitors"}, {"LOWER": "negligence"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "solicitors"},
            {"LOWER": "seeking"},
            {"LOWER": "summary"},
            {"LOWER": "judgment"},
            {"LOWER": "against"},
            {"LOWER": "clients"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "unpaid"},
            {"LOWER": "bills"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "solicitors"}, {"LOWER": "’"}, {"LOWER": "fees"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "solicitors"}, {"LOWER": "’"}, {"LOWER": "regulator"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "solicitor"}, {"LOWER": "’s"}, {"LOWER": "fees"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "solicitor"}, {"LOWER": "’s"}, {"LOWER": "negligence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "solicitor"}, {"LOWER": "’s"}, {"LOWER": "undertaking"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "somalian"},
            {"LOWER": "claimant"},
            {"LOWER": "originally"},
            {"LOWER": "enjoying"},
            {"LOWER": "exceptional"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "remain"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "son"},
            {"LOWER": "convicted"},
            {"LOWER": "of"},
            {"LOWER": "mother"},
            {"LOWER": "’s"},
            {"LOWER": "manslaughter"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sonia"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sound"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sound"}, {"LOWER": "recordings"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sources"}, {"LOWER": "of"}, {"LOWER": "information"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "south"}, {"LOWER": "australia"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sovereign"}, {"LOWER": "immunity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sovereign"}, {"LOWER": "immunity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "span"}, {"LOWER": "terza"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "special"}, {"LOWER": "administration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "custodial"},
            {"LOWER": "sentence"},
            {"LOWER": "for"},
            {"LOWER": "certain"},
            {"LOWER": "offenders"},
            {"LOWER": "of"},
            {"LOWER": "particular"},
            {"LOWER": "concern"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "extradition"},
            {"LOWER": "arrangements"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "guardianship"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "guardianship"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "guardianship"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "special"}, {"LOWER": "hours"}, {"LOWER": "certificate"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "immigration"},
            {"LOWER": "appeals"},
            {"LOWER": "commission"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "special"}, {"LOWER": "jurisdiction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "special"}, {"LOWER": "power"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "special"}, {"LOWER": "procedure"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "procedure"},
            {"LOWER": "material"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "special"},
            {"LOWER": "removal"},
            {"LOWER": "of"},
            {"LOWER": "licence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "special"}, {"LOWER": "verdict"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "specific"},
            {"LOWER": "bequests"},
            {"LOWER": "of"},
            {"LOWER": "property"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "specific"}, {"LOWER": "disclosure"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "specific"}, {"LOWER": "issue"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "specific"}, {"LOWER": "performance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "specific"}, {"LOWER": "performance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "specific"}, {"LOWER": "performance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "specimen"}, {"LOWER": "counts"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "spent"}, {"LOWER": "conviction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "spiliada"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "spirit"}, {"LOWER": "drinks"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "spoken"}, {"LOWER": "words"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sponsoring"}, {"LOWER": "parent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sport"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "st"},
            {"LOWER": "vincent"},
            {"LOWER": "and"},
            {"LOWER": "the"},
            {"LOWER": "grenadines"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "staff"}, {"LOWER": "regulations"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "stakeholder"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "stale"}, {"LOWER": "claim"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stalking"},
            {"LOWER": "involving"},
            {"LOWER": "fear"},
            {"LOWER": "of"},
            {"LOWER": "violence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "stamp"}, {"LOWER": "duty"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stamp"},
            {"LOWER": "duty"},
            {"LOWER": "land"},
            {"LOWER": "tax"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "standard"}, {"LOWER": "form"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "standard"}, {"LOWER": "of"}, {"LOWER": "care"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "standard"}, {"LOWER": "of"}, {"LOWER": "proof"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "standard"}, {"LOWER": "procedure"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "standards"},
            {"LOWER": "for"},
            {"LOWER": "the"},
            {"LOWER": "reception"},
            {"LOWER": "of"},
            {"LOWER": "applicants"},
            {"LOWER": "for"},
            {"LOWER": "international"},
            {"LOWER": "protection"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "star"}, {"LOWER": "sea"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "starsin"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "state"}, {"LOWER": "aid"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "state"},
            {"LOWER": "aid"},
            {"LOWER": "incompatible"},
            {"LOWER": "with"},
            {"LOWER": "the"},
            {"LOWER": "common"},
            {"LOWER": "market"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "state"}, {"LOWER": "enterprise"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "state"}, {"LOWER": "immunity"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "state"}, {"LOWER": "immunity"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "state"}, {"LOWER": "pension"}, {"LOWER": "credit"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "state"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "stateless"}, {"LOWER": "persons"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statement"}, {"LOWER": "by"}, {"LOWER": "agent"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statement"},
            {"LOWER": "in"},
            {"LOWER": "open"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statement"}, {"LOWER": "of"}, {"LOWER": "offence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statement"},
            {"LOWER": "of"},
            {"LOWER": "reasons"},
            {"LOWER": "for"},
            {"LOWER": "dismissal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statement"}, {"LOWER": "of"}, {"LOWER": "terms"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "states"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "investigate"},
            {"LOWER": "death"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "investigate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "investigate"},
            {"LOWER": "attempted"},
            {"LOWER": "suicide"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "investigate"},
            {"LOWER": "death"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "investigate"},
            {"LOWER": "death"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "duty"},
            {"LOWER": "to"},
            {"LOWER": "investigate"},
            {"LOWER": "life"},
            {"ORTH": "-"},
            {"LOWER": "threatening"},
            {"LOWER": "injury"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "state"},
            {"LOWER": "’s"},
            {"LOWER": "obligation"},
            {"LOWER": "to"},
            {"LOWER": "mitigate"},
            {"LOWER": "risks"},
            {"LOWER": "to"},
            {"LOWER": "soldiers"},
            {"LOWER": "’"},
            {"LOWER": "lives"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statute"}, {"LOWER": "of"}, {"LOWER": "limitations"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statute"}, {"LOWER": "of"}, {"LOWER": "limitations"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statute"},
            {"ORTH": ","},
            {"LOWER": "action"},
            {"LOWER": "on"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statute"},
            {"ORTH": ","},
            {"LOWER": "action"},
            {"LOWER": "on"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statute"},
            {"ORTH": ","},
            {"LOWER": "application"},
            {"LOWER": "to"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statutory"},
            {"LOWER": "and"},
            {"LOWER": "common"},
            {"LOWER": "law"},
            {"LOWER": "claims"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "appeal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "authority"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "board"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "caution"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statutory"}, {"LOWER": "compensation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statutory"},
            {"LOWER": "compensation"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "conspiracy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "defence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "demand"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "demand"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "duties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "duties"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "duty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "duty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "entitlement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "exception"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statutory"},
            {"LOWER": "financial"},
            {"LOWER": "threshold"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "instrument"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "instrument"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "instrument"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statutory"},
            {"LOWER": "interest"},
            {"LOWER": "restriction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statutory"},
            {"LOWER": "limitation"},
            {"LOWER": "period"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "market"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "notice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "nuisance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "offence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "power"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "powers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "procedures"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "provision"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statutory"}, {"LOWER": "requirements"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "restriction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "review"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "scheme"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "statutory"}, {"LOWER": "sick"}, {"LOWER": "pay"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "suspension"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "statutory"}, {"LOWER": "tenancy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "statutory"},
            {"LOWER": "undertaker"},
            {"ORTH": ","},
            {"LOWER": "liability"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stay"},
            {"LOWER": "of"},
            {"LOWER": "costs"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stay"},
            {"LOWER": "of"},
            {"LOWER": "english"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "stay"}, {"LOWER": "of"}, {"LOWER": "execution"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stay"},
            {"LOWER": "of"},
            {"LOWER": "foreign"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stay"},
            {"LOWER": "of"},
            {"LOWER": "judicial"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stay"},
            {"LOWER": "of"},
            {"LOWER": "payment"},
            {"LOWER": "out"},
            {"LOWER": "pending"},
            {"LOWER": "appeal"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "stay"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "stay"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "stay"}, {"LOWER": "of"}, {"LOWER": "removal"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "staying"}, {"LOWER": "proceeding"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "staying"},
            {"LOWER": "trial"},
            {"LOWER": "of"},
            {"LOWER": "issue"},
            {"LOWER": "of"},
            {"LOWER": "fact"},
            {"LOWER": "pending"},
            {"LOWER": "appeal"},
            {"LOWER": "on"},
            {"LOWER": "question"},
            {"LOWER": "of"},
            {"LOWER": "law"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "stena"}, {"LOWER": "cambria"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "step"}, {"LOWER": "in"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "steps"},
            {"LOWER": "to"},
            {"LOWER": "enforce"},
            {"LOWER": "security"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sterilisation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sterilisation"}, {"LOWER": "operation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "stock"},
            {"LOWER": "standing"},
            {"LOWER": "in"},
            {"LOWER": "the"},
            {"LOWER": "name"},
            {"LOWER": "of"},
            {"LOWER": "testator"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "stolen"}, {"LOWER": "money"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "stolen"}, {"LOWER": "painting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "stolen"}, {"LOWER": "property"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "stolen"}, {"LOWER": "vehicle"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "stop"}, {"LOWER": "and"}, {"LOWER": "search"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "stop"}, {"LOWER": "notice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "storaa"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "streetworks"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "strike"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "strike"}, {"LOWER": "action"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "strike"}, {"LOWER": "out"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "striking"}, {"LOWER": "out"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "striking"},
            {"LOWER": "out"},
            {"LOWER": "and"},
            {"LOWER": "summary"},
            {"LOWER": "judgment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "striking"}, {"LOWER": "out"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "striking"},
            {"LOWER": "out"},
            {"ORTH": "-"},
            {"LOWER": "defence"},
            {"LOWER": "of"},
            {"LOWER": "justification"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "structural"}, {"LOWER": "funds"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "structured"}, {"LOWER": "funds"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "student"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "student"}, {"LOWER": "complaints"}, {"LOWER": "scheme"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "student"}, {"LOWER": "loan"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "student"}, {"LOWER": "loans"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sub"}, {"ORTH": "-"}, {"LOWER": "bailment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sub"}, {"ORTH": "-"}, {"LOWER": "postmaster"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "subcontractor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "subject"},
            {"LOWER": "matter"},
            {"LOWER": "of"},
            {"LOWER": "invention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "subject"}, {"LOWER": "matter"}, {"LOWER": "only"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "submission"},
            {"LOWER": "of"},
            {"LOWER": "no"},
            {"LOWER": "case"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "submission"},
            {"LOWER": "of"},
            {"LOWER": "no"},
            {"LOWER": "case"},
            {"LOWER": "to"},
            {"LOWER": "answer"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "subordinate"}, {"LOWER": "legislation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "subpoena"}, {"LOWER": "duces"}, {"LOWER": "tecum"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "subrogation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "subsequent"}, {"LOWER": "bankruptcy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "subsequent"}, {"LOWER": "bankruptcy"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "subsequent"},
            {"LOWER": "planning"},
            {"LOWER": "permission"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "subsequent"},
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "disclosed"},
            {"LOWER": "documents"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "subsidiary"}, {"LOWER": "protection"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "substance"},
            {"LOWER": "of"},
            {"LOWER": "article"},
            {"LOWER": "demanded"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "substantial"}, {"LOWER": "money"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "substituted"}, {"LOWER": "expenses"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "substituted"}, {"LOWER": "service"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "substitution"}, {"LOWER": "of"}, {"LOWER": "charges"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "substitution"}, {"LOWER": "of"}, {"LOWER": "verdict"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "subtenancy"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "success"}, {"LOWER": "fee"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "success"},
            {"LOWER": "fees"},
            {"LOWER": "and"},
            {"LOWER": "after"},
            {"LOWER": "event"},
            {"LOWER": "insurance"},
            {"LOWER": "premiums"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "success"},
            {"LOWER": "fees"},
            {"LOWER": "and"},
            {"LOWER": "after"},
            {"LOWER": "the"},
            {"LOWER": "event"},
            {"LOWER": "insurance"},
            {"LOWER": "premiums"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "successful"}, {"LOWER": "appellant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "successful"},
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contract"},
            {"LOWER": "and"},
            {"LOWER": "procuring"},
            {"LOWER": "breach"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "successful"},
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "contracts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "successful"},
            {"LOWER": "unassisted"},
            {"LOWER": "party"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "succession"},
            {"LOWER": "duty"},
            {"LOWER": "act"},
            {"ORTH": ","},
            {"LOWER": "1853"},
            {"ORTH": "("},
            {"LOWER": "16"},
            {"LOWER": "&"},
            {"LOWER": "17"},
            {"LOWER": "vict"},
            {"ORTH": "."},
            {"LOWER": "c."},
            {"LOWER": "51"},
            {"ORTH": ")"},
            {"ORTH": ","},
            {"LOWER": "s."},
            {"LOWER": "10"},
            {"ORTH": "["},
            {"LOWER": "revised"},
            {"ORTH": ","},
            {"LOWER": "ed"},
            {"ORTH": "."},
            {"LOWER": "statutes"},
            {"ORTH": ","},
            {"LOWER": "vol"},
            {"ORTH": "."},
            {"LOWER": "xi"},
            {"ORTH": "."},
            {"LOWER": "p."},
            {"LOWER": "597"},
            {"ORTH": "]"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "succession"}, {"LOWER": "to"}, {"LOWER": "tenancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "successive"}, {"LOWER": "applications"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "successive"},
            {"LOWER": "fixed"},
            {"ORTH": "-"},
            {"LOWER": "term"},
            {"LOWER": "employment"},
            {"LOWER": "contracts"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "sugar"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "suicide"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "suitability"},
            {"LOWER": "of"},
            {"LOWER": "accommodation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "suitability"}, {"LOWER": "of"}, {"LOWER": "equipment"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "summary"}, {"LOWER": "assessment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "summary"}, {"LOWER": "judgment"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "summary"}, {"LOWER": "jurisdiction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "summary"},
            {"LOWER": "procedure"},
            {"LOWER": "under"},
            {"LOWER": "order"},
            {"LOWER": "14"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "summary"},
            {"LOWER": "procedure"},
            {"LOWER": "under"},
            {"LOWER": "order"},
            {"LOWER": "86"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "summary"},
            {"LOWER": "procedure"},
            {"LOWER": "under"},
            {"LOWER": "part"},
            {"LOWER": "24"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "summary"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "summary"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "summing"}, {"LOWER": "up"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "summons"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "summons"}, {"LOWER": "for"}, {"LOWER": "protection"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "summonses"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sunday"}, {"LOWER": "trading"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "sunken"}, {"LOWER": "wreck"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "supervision"}, {"LOWER": "of"}, {"LOWER": "charity"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supervision"},
            {"LOWER": "of"},
            {"LOWER": "children"},
            {"LOWER": "playing"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "supervision"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "supplementary"}, {"LOWER": "benefit"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supplementary"},
            {"LOWER": "benefit"},
            {"LOWER": "allowance"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supplementary"},
            {"LOWER": "protection"},
            {"LOWER": "certificate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supply"},
            {"LOWER": "from"},
            {"LOWER": "court"},
            {"LOWER": "records"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "supply"}, {"LOWER": "of"}, {"LOWER": "goods"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supply"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
            {"LOWER": "and"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supply"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
            {"LOWER": "and"},
            {"LOWER": "services"},
            {"LOWER": "to"},
            {"LOWER": "public"},
            {"LOWER": "bodies"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supply"},
            {"LOWER": "of"},
            {"LOWER": "goods"},
            {"LOWER": "or"},
            {"LOWER": "services"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supply"},
            {"LOWER": "of"},
            {"LOWER": "immovable"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supply"},
            {"LOWER": "of"},
            {"LOWER": "temporary"},
            {"LOWER": "workers"},
            {"LOWER": "to"},
            {"LOWER": "user"},
            {"LOWER": "undertaking"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supply"},
            {"LOWER": "other"},
            {"LOWER": "than"},
            {"LOWER": "in"},
            {"LOWER": "licensed"},
            {"LOWER": "sex"},
            {"LOWER": "shop"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "supplying"},
            {"LOWER": "to"},
            {"LOWER": "person"},
            {"LOWER": "under"},
            {"LOWER": "age"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "support"},
            {"LOWER": "for"},
            {"LOWER": "rural"},
            {"LOWER": "development"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "supreme"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "supreme"}, {"LOWER": "court"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "supreme"}, {"LOWER": "court"}, {"LOWER": "decision"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "surcharge"}, {"LOWER": "remission"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "suretys"}, {"LOWER": "liability"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "surety"}, {"LOWER": "’s"}, {"LOWER": "liability"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "surface"}, {"LOWER": "water"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "surgeon"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "surplus"}, {"LOWER": "fund"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "surrender"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "surrender"}, {"LOWER": "of"}, {"LOWER": "lease"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "surrender"},
            {"LOWER": "of"},
            {"LOWER": "members"},
            {"LOWER": "’"},
            {"LOWER": "rights"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "surveyors"}, {"LOWER": "report"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "surviving"}, {"LOWER": "spouse"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "survivor"}, {"LOWER": "’s"}, {"LOWER": "pension"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "suspected"},
            {"LOWER": "international"},
            {"LOWER": "terrorists"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "suspected"}, {"LOWER": "terrorist"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "suspected"}, {"LOWER": "terrorists"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "suspended"},
            {"LOWER": "possession"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "suspended"}, {"LOWER": "sentence"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "suspension"}, {"LOWER": "from"}, {"LOWER": "practice"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "suspension"}, {"LOWER": "of"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "suspension"},
            {"LOWER": "of"},
            {"LOWER": "registration"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "suspicion"}, {"LOWER": "of"}, {"LOWER": "abuse"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "sustainable"}, {"LOWER": "development"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "swap"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "system"},
            {"LOWER": "for"},
            {"LOWER": "identification"},
            {"LOWER": "and"},
            {"LOWER": "registration"},
            {"LOWER": "of"},
            {"LOWER": "ovine"},
            {"LOWER": "and"},
            {"LOWER": "caprine"},
            {"LOWER": "animals"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "t.f.l"}, {"ORTH": "."}, {"LOWER": "prosperity"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "tachograph"}, {"LOWER": "records"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "take"},
            {"ORTH": "-"},
            {"LOWER": "overs"},
            {"LOWER": "and"},
            {"LOWER": "mergers"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "taking"},
            {"LOWER": "child"},
            {"LOWER": "into"},
            {"LOWER": "care"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "taking"},
            {"LOWER": "conveyance"},
            {"LOWER": "without"},
            {"LOWER": "authority"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "taking"}, {"LOWER": "of"}, {"LOWER": "account"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "taking"}, {"LOWER": "of"}, {"LOWER": "photograph"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "tariff"}, {"LOWER": "classification"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tariff"}, {"LOWER": "quotas"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tax"}, {"LOWER": "advantage"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tax"}, {"LOWER": "avoidance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tax"}, {"LOWER": "avoidance"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tax"}, {"LOWER": "avoidance"}, {"LOWER": "scheme"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "tax"}, {"LOWER": "element"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tax"}, {"LOWER": "evasion"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tax"}, {"LOWER": "reduction"}, {"LOWER": "measures"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "tax"}, {"LOWER": "relief"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "taxable"}, {"LOWER": "amount"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "taxable"}, {"LOWER": "disposal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "taxation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "taxation"},
            {"LOWER": "costs"},
            {"LOWER": "on"},
            {"LOWER": "indemnity"},
            {"LOWER": "basis"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "taxation"},
            {"LOWER": "costs"},
            {"LOWER": "prior"},
            {"LOWER": "to"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "taxation"},
            {"LOWER": "of"},
            {"LOWER": "energy"},
            {"LOWER": "products"},
            {"LOWER": "and"},
            {"LOWER": "electricity"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "teacher"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "teacher"},
            {"LOWER": "employed"},
            {"LOWER": "from"},
            {"LOWER": "1"},
            {"LOWER": "may"},
            {"LOWER": "but"},
            {"LOWER": "commencing"},
            {"LOWER": "teaching"},
            {"LOWER": "duties"},
            {"LOWER": "on"},
            {"LOWER": "4"},
            {"LOWER": "may"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "teachers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "teachers"},
            {"LOWER": "’"},
            {"LOWER": "pension"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "technical"}, {"LOWER": "drawings"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "technical"},
            {"LOWER": "know"},
            {"ORTH": "-"},
            {"LOWER": "how"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "technical"},
            {"LOWER": "requirements"},
            {"LOWER": "relating"},
            {"LOWER": "to"},
            {"LOWER": "blood"},
            {"LOWER": "and"},
            {"LOWER": "blood"},
            {"LOWER": "components"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "technology"},
            {"LOWER": "and"},
            {"LOWER": "construction"},
            {"LOWER": "court"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "telecommunication"}, {"LOWER": "services"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "telecommunications"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "telecommunications"}, {"LOWER": "installation"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "telephone"}, {"LOWER": "intercept"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "television"}, {"LOWER": "broadcasting"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "television"}, {"LOWER": "reception"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "temporary"}, {"LOWER": "accommodation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "temporary"},
            {"LOWER": "cessation"},
            {"LOWER": "of"},
            {"LOWER": "work"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "temporary"}, {"LOWER": "release"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "temporary"}, {"LOWER": "removal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tenancy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tenancy"}, {"LOWER": "deposit"}, {"LOWER": "scheme"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tenancy"}, {"LOWER": "within"}, {"LOWER": "act"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tenant"}, {"LOWER": "for"}, {"LOWER": "life"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenants"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "covenant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenants"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "new"},
            {"LOWER": "lease"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenants"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "reversion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenants"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "reversion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenants"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "buy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenants"},
            {"LOWER": "’"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "freehold"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenants"},
            {"LOWER": "’"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "reversion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenant"},
            {"LOWER": "’s"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "covenant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenant"},
            {"LOWER": "’s"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "new"},
            {"LOWER": "extended"},
            {"LOWER": "lease"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenant"},
            {"LOWER": "’s"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "new"},
            {"LOWER": "lease"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tenant"},
            {"LOWER": "’s"},
            {"LOWER": "right"},
            {"LOWER": "to"},
            {"LOWER": "acquire"},
            {"LOWER": "reversion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tender"}, {"LOWER": "before"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "tender"}, {"LOWER": "specification"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tennis"}, {"LOWER": "club"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "termination"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "termination"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "termination"}, {"LOWER": "of"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "termination"}, {"LOWER": "of"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "termination"},
            {"LOWER": "of"},
            {"LOWER": "periodic"},
            {"LOWER": "tenancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "termination"}, {"LOWER": "of"}, {"LOWER": "pregnancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "termination"},
            {"LOWER": "of"},
            {"LOWER": "secure"},
            {"LOWER": "tenancy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "termination"}, {"LOWER": "of"}, {"LOWER": "tenancy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "termination"},
            {"ORTH": ","},
            {"LOWER": "effective"},
            {"LOWER": "date"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "terms"}, {"LOWER": "of"}, {"LOWER": "engagement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "terms"}, {"LOWER": "of"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "terms"}, {"LOWER": "of"}, {"LOWER": "service"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "terms"},
            {"LOWER": "on"},
            {"LOWER": "which"},
            {"LOWER": "concluded"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "terms"},
            {"ORTH": "-"},
            {"LOWER": "shippers"},
            {"LOWER": "sub"},
            {"ORTH": "-"},
            {"LOWER": "bailing"},
            {"LOWER": "goods"},
            {"LOWER": "with"},
            {"LOWER": "owners"},
            {"LOWER": "consent"},
            {"LOWER": "under"},
            {"LOWER": "bill"},
            {"LOWER": "of"},
            {"LOWER": "lading"},
            {"LOWER": "providing"},
            {"LOWER": "for"},
            {"LOWER": "determination"},
            {"LOWER": "in"},
            {"LOWER": "taiwan"},
            {"LOWER": "of"},
            {"LOWER": "all"},
            {"LOWER": "disputes"},
            {"LOWER": "under"},
            {"LOWER": "contract"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "terrorism"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "terrorist"}, {"LOWER": "offences"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "test"}, {"LOWER": "of"}, {"LOWER": "fairness"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "test"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "applied"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "testamentary"}, {"LOWER": "capacity"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "testamentary"}, {"LOWER": "intention"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "testator"},
            {"LOWER": "domiciled"},
            {"LOWER": "in"},
            {"LOWER": "scotland"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "testators"},
            {"LOWER": "knowledge"},
            {"LOWER": "and"},
            {"LOWER": "approval"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "testator"},
            {"LOWER": "’s"},
            {"LOWER": "knowledge"},
            {"LOWER": "and"},
            {"LOWER": "approval"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "testator"}, {"LOWER": "’s"}, {"LOWER": "signature"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "testatrix"}, {"LOWER": "’s"}, {"LOWER": "signature"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "the"}, {"LOWER": "bahamas"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "the"}, {"LOWER": "deichland"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "the"}, {"LOWER": "fanti"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "the"}, {"LOWER": "padre"}, {"LOWER": "island"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "theft"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "theft"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "theft"},
            {"LOWER": "of"},
            {"LOWER": "lead"},
            {"LOWER": "from"},
            {"LOWER": "roof"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "things"},
            {"LOWER": "naturally"},
            {"LOWER": "on"},
            {"LOWER": "land"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "country"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "parties"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "third"},
            {"LOWER": "parties"},
            {"LOWER": "’"},
            {"LOWER": "rights"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "third"}, {"LOWER": "party"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "party"}, {"LOWER": "action"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "party"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "third"},
            {"LOWER": "party"},
            {"LOWER": "costs"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "third"},
            {"LOWER": "party"},
            {"LOWER": "debt"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "third"},
            {"LOWER": "party"},
            {"LOWER": "debt"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "third"},
            {"LOWER": "party"},
            {"LOWER": "debt"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "third"},
            {"LOWER": "party"},
            {"LOWER": "financing"},
            {"LOWER": "litigation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "party"}, {"LOWER": "insurance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "party"}, {"LOWER": "insurance"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "party"}, {"LOWER": "interests"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "party"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "third"}, {"LOWER": "party"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "threat"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "threat"}, {"LOWER": "to"}, {"LOWER": "kill"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "threshold"}, {"LOWER": "conditions"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "threshold"}, {"LOWER": "criteria"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tilting"},
            {"LOWER": "device"},
            {"LOWER": "for"},
            {"LOWER": "seat"},
            {"LOWER": "back"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "time"}, {"LOWER": "bar"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "time"}, {"LOWER": "charter"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "time"}, {"LOWER": "charterer"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "time"}, {"LOWER": "for"}, {"LOWER": "complaint"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "time"}, {"LOWER": "for"}, {"LOWER": "completion"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "time"}, {"LOWER": "for"}, {"LOWER": "extradition"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "time"}, {"LOWER": "for"}, {"LOWER": "presenting"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "time"}, {"LOWER": "limit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "time"}, {"LOWER": "limit"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "time"},
            {"LOWER": "limit"},
            {"LOWER": "for"},
            {"LOWER": "execution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "time"},
            {"LOWER": "limit"},
            {"LOWER": "for"},
            {"LOWER": "review"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "time"},
            {"LOWER": "limit"},
            {"LOWER": "on"},
            {"LOWER": "prosecution"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "time"},
            {"LOWER": "limit"},
            {"ORTH": ","},
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "override"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "time"}, {"LOWER": "limits"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "time"},
            {"LOWER": "off"},
            {"LOWER": "for"},
            {"LOWER": "dependants"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "time"}, {"LOWER": "off"}, {"LOWER": "work"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "time"},
            {"LOWER": "when"},
            {"LOWER": "act"},
            {"LOWER": "committed"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "title"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "title"}, {"LOWER": "to"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "title"}, {"LOWER": "to"}, {"LOWER": "goods"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "tobago"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tomlin"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tort"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "of"}, {"LOWER": "crown"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "liability"}, {"LOWER": "of"}, {"LOWER": "state"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "liable"},
            {"LOWER": "in"},
            {"LOWER": "tort"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "total"}, {"LOWER": "clearance"}, {"LOWER": "sale"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "total"},
            {"LOWER": "loss"},
            {"LOWER": "of"},
            {"LOWER": "vessel"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "totting"}, {"LOWER": "up"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "town"},
            {"LOWER": "or"},
            {"LOWER": "village"},
            {"LOWER": "green"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "town"},
            {"LOWER": "or"},
            {"LOWER": "village"},
            {"LOWER": "green"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "town"}, {"LOWER": "planning"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "town"}, {"LOWER": "planning"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "town"}, {"LOWER": "planning"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tracing"}, {"LOWER": "child"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "description"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "description"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "dispute"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "dispute"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "dispute"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trade"},
            {"LOWER": "in"},
            {"LOWER": "endangered"},
            {"LOWER": "species"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "mark"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trade"},
            {"LOWER": "mark"},
            {"LOWER": "with"},
            {"LOWER": "reputation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "name"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "names"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "organisation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "union"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trade"}, {"LOWER": "union"}, {"LOWER": "activities"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trade"}, {"LOWER": "union"}, {"LOWER": "consultation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trade"}, {"LOWER": "union"}, {"LOWER": "recognition"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trade"}, {"LOWER": "union"}, {"LOWER": "recognition"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trade"},
            {"LOWER": "union"},
            {"ORTH": ","},
            {"LOWER": "discrimination"},
            {"LOWER": "by"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trade"}, {"LOWER": "unions"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trademark"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trading"}, {"LOWER": "standards"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "traffic"}, {"LOWER": "management"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "traffic"}, {"LOWER": "regulation"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "traffic"}, {"LOWER": "route"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "traffic"}, {"LOWER": "sign"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trafficking"},
            {"LOWER": "in"},
            {"LOWER": "trade"},
            {"LOWER": "mark"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trafficking"}, {"LOWER": "victim"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transaction"},
            {"LOWER": "at"},
            {"LOWER": "an"},
            {"LOWER": "undervalue"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transaction"}, {"LOWER": "at"}, {"LOWER": "undervalue"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transaction"},
            {"LOWER": "having"},
            {"LOWER": "preferential"},
            {"LOWER": "effect"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transactions"},
            {"LOWER": "at"},
            {"LOWER": "undervalue"},
            {"LOWER": "or"},
            {"LOWER": "preferences"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transcripts"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "an"},
            {"LOWER": "undertaking"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "assets"},
            {"LOWER": "between"},
            {"LOWER": "companies"},
            {"LOWER": "in"},
            {"LOWER": "same"},
            {"LOWER": "group"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "bookings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "business"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "licence"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "obligation"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "proceedings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "properties"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "property"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "registered"},
            {"LOWER": "land"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "rights"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "sentenced"},
            {"LOWER": "persons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "trade"},
            {"LOWER": "on"},
            {"LOWER": "business"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "trade"},
            {"LOWER": "or"},
            {"LOWER": "business"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "undertaking"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "undertakings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "undertakings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "of"}, {"LOWER": "undertakings"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transfer"}, {"LOWER": "to"}, {"LOWER": "minister"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "to"},
            {"LOWER": "open"},
            {"LOWER": "conditions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfer"},
            {"LOWER": "to"},
            {"LOWER": "uk"},
            {"LOWER": "to"},
            {"LOWER": "serve"},
            {"LOWER": "sentence"},
            {"LOWER": "of"},
            {"LOWER": "convict"},
            {"LOWER": "from"},
            {"LOWER": "republic"},
            {"LOWER": "of"},
            {"LOWER": "ireland"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "transferor"}, {"LOWER": "’s"}, {"LOWER": "insolvency"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transfers"},
            {"LOWER": "of"},
            {"LOWER": "property"},
            {"LOWER": "defrauding"},
            {"LOWER": "creditor"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "transport"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "transportation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transport"},
            {"LOWER": "of"},
            {"LOWER": "consignment"},
            {"LOWER": "which"},
            {"LOWER": "became"},
            {"LOWER": "unintentionally"},
            {"LOWER": "mixed"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transportation"},
            {"ORTH": ","},
            {"LOWER": "processing"},
            {"LOWER": "and"},
            {"LOWER": "operating"},
            {"LOWER": "services"},
            {"LOWER": "agreement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "transposition"},
            {"LOWER": "into"},
            {"LOWER": "national"},
            {"LOWER": "law"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "travellers"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "travelling"},
            {"LOWER": "showpeoples"},
            {"LOWER": "site"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "treasure"}, {"LOWER": "trove"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "treatment"},
            {"LOWER": "of"},
            {"LOWER": "transsexual"},
            {"LOWER": "prisoners"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "treatment"},
            {"ORTH": ","},
            {"LOWER": "participating"},
            {"LOWER": "in"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "treaty"}, {"LOWER": "obligations"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "treaty"},
            {"LOWER": "obligations"},
            {"LOWER": "to"},
            {"LOWER": "indian"},
            {"LOWER": "peoples"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "treaty"},
            {"LOWER": "on"},
            {"LOWER": "european"},
            {"LOWER": "union"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "treaty"}, {"LOWER": "rights"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "treaty"},
            {"ORTH": "-"},
            {"LOWER": "making"},
            {"LOWER": "power"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tree"}, {"LOWER": "preservation"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tree"}, {"LOWER": "replacement"}, {"LOWER": "notice"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "tree"}, {"LOWER": "roots"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trees"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trespass"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trespass"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trespass"},
            {"LOWER": "to"},
            {"LOWER": "the"},
            {"LOWER": "person"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trespasser"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trespassers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trespassory"}, {"LOWER": "assembly"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trial"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trial"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trial"}, {"LOWER": "in"}, {"LOWER": "absentia"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tribunal"}, {"LOWER": "amending"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tribunal"}, {"LOWER": "lay"}, {"LOWER": "member"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "tribunal"}, {"LOWER": "of"}, {"LOWER": "inquiry"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trinidad"}, {"LOWER": "and"}, {"LOWER": "tobago"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trinidad"}, {"LOWER": "and"}, {"LOWER": "tobago"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trinidad"},
            {"LOWER": "and"},
            {"LOWER": "tobago"},
            {"LOWER": "constitution"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trips"}, {"LOWER": "agreement"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "true"},
            {"LOWER": "owner"},
            {"LOWER": "of"},
            {"LOWER": "land"},
            {"LOWER": "deprived"},
            {"LOWER": "of"},
            {"LOWER": "title"},
            {"LOWER": "by"},
            {"LOWER": "forged"},
            {"LOWER": "documents"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trust"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trust"}, {"LOWER": "for"}, {"LOWER": "sale"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trust"}, {"LOWER": "for"}, {"LOWER": "sale"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trust"}, {"LOWER": "fund"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trust"}, {"LOWER": "funds"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trust"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trust"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trust"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trust"}, {"LOWER": "of"}, {"LOWER": "land"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trust"}, {"LOWER": "proceedings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trust"}, {"LOWER": "property"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trust"},
            {"LOWER": "property"},
            {"LOWER": "misappropriated"},
            {"LOWER": "by"},
            {"LOWER": "trustee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trust"},
            {"LOWER": "property"},
            {"LOWER": "not"},
            {"LOWER": "received"},
            {"LOWER": "by"},
            {"LOWER": "defendants"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trust"}, {"LOWER": "settlement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "trustee"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trustee"},
            {"LOWER": "act"},
            {"ORTH": ","},
            {"LOWER": "1850"},
            {"ORTH": ","},
            {"LOWER": "s."},
            {"LOWER": "5"},
            {"ORTH": "["},
            {"LOWER": "revised"},
            {"LOWER": "ed"},
            {"ORTH": "."},
            {"LOWER": "statutes"},
            {"ORTH": ","},
            {"LOWER": "vol"},
            {"ORTH": "."},
            {"LOWER": "x."},
            {"LOWER": "p."},
            {"LOWER": "984"},
            {"ORTH": "]"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trustee"},
            {"LOWER": "and"},
            {"LOWER": "cestui"},
            {"LOWER": "que"},
            {"LOWER": "trust"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trustee"}, {"LOWER": "in"}, {"LOWER": "bankruptcy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trustee"}, {"LOWER": "in"}, {"LOWER": "bankruptcy"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "trustee"}, {"LOWER": "in"}, {"LOWER": "bankuptcy"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trustees"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trustees"},
            {"LOWER": "claim"},
            {"LOWER": "to"},
            {"LOWER": "overseas"},
            {"LOWER": "interest"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trustees"}, {"LOWER": "power"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trustees"},
            {"LOWER": "to"},
            {"LOWER": "pay"},
            {"LOWER": "wife"},
            {"LOWER": "“"},
            {"LOWER": "clear"},
            {"LOWER": "yearly"},
            {"LOWER": "sum"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "trustees"},
            {"LOWER": "undertaking"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "appointment"},
            {"LOWER": "in"},
            {"LOWER": "favour"},
            {"LOWER": "of"},
            {"LOWER": "minors"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "trusts"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "turks"},
            {"LOWER": "and"},
            {"LOWER": "caicos"},
            {"LOWER": "islands"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "turnover"}, {"LOWER": "taxes"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "tuyuti"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "twitter"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "two"}, {"LOWER": "wills"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "typographical"}, {"LOWER": "arrangement"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "tyres"},
            {"LOWER": "stoking"},
            {"LOWER": "fire"},
            {"LOWER": "but"},
            {"LOWER": "not"},
            {"LOWER": "themselves"},
            {"LOWER": "escaping"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "ub"}, {"LOWER": "tiger"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "uk"}, {"LOWER": "border"}, {"LOWER": "agency"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unacceptable"},
            {"LOWER": "professional"},
            {"LOWER": "conduct"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unaccompanied"}, {"LOWER": "minor"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unaccompanied"},
            {"LOWER": "minor"},
            {"LOWER": "claiming"},
            {"LOWER": "asylum"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unapproved"}, {"LOWER": "scheme"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unascertained"}, {"LOWER": "goods"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unauthorised"}, {"LOWER": "act"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unauthorised"}, {"LOWER": "deduction"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unauthorised"}, {"LOWER": "expenses"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unauthorised"}, {"LOWER": "investments"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unauthorised"}, {"LOWER": "possession"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unauthorised"},
            {"LOWER": "possession"},
            {"LOWER": "of"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unauthorised"}, {"LOWER": "reception"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unconditional"},
            {"LOWER": "leave"},
            {"LOWER": "to"},
            {"LOWER": "defend"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "under"},
            {"LOWER": "judgment"},
            {"LOWER": "or"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "undertaking"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "damages"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "undertaking"}, {"LOWER": "to"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "undertaking"},
            {"LOWER": "with"},
            {"LOWER": "significant"},
            {"LOWER": "market"},
            {"LOWER": "power"},
            {"LOWER": "on"},
            {"LOWER": "specific"},
            {"LOWER": "market"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "undertaking"},
            {"ORTH": ","},
            {"LOWER": "transfer"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "underwriter"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "underwriting"}, {"LOWER": "agent"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "undischarged"}, {"LOWER": "bankrupt"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "undistributed"}, {"LOWER": "assets"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "undue"}, {"LOWER": "influence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "undue"}, {"LOWER": "influence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "undue"}, {"LOWER": "influence"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unduly"}, {"LOWER": "lenient"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unemployment"}, {"LOWER": "benefit"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unexplained"}, {"LOWER": "wealth"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unfair"},
            {"LOWER": "commercial"},
            {"LOWER": "practice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unfair"},
            {"LOWER": "commercial"},
            {"LOWER": "practices"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unfair"}, {"LOWER": "dismissal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unfair"}, {"LOWER": "dismissal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unfair"}, {"LOWER": "dismissal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unfair"}, {"LOWER": "dismissal"}, {"LOWER": "claim"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unfair"}, {"LOWER": "prejudice"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unfair"}, {"LOWER": "price"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unfair"}, {"LOWER": "terms"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unformed"}, {"LOWER": "company"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unilateral"}, {"LOWER": "mistake"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unilateral"}, {"LOWER": "notice"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unilateral"},
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "amend"},
            {"LOWER": "terms"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unincorporated"}, {"LOWER": "association"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "uninsured"}, {"LOWER": "driver"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "union"},
            {"LOWER": "membership"},
            {"LOWER": "agreement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unitary"}, {"LOWER": "patent"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "united"},
            {"LOWER": "nations"},
            {"LOWER": "security"},
            {"LOWER": "council"},
            {"LOWER": "sanctions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unity"}, {"LOWER": "of"}, {"LOWER": "ownership"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "universal"}, {"LOWER": "credit"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "universal"},
            {"LOWER": "service"},
            {"LOWER": "and"},
            {"LOWER": "users"},
            {"LOWER": "’"},
            {"LOWER": "rights"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "university"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "university"}, {"LOWER": "education"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unjust"}, {"LOWER": "enrichment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unjustifiable"},
            {"LOWER": "threat"},
            {"LOWER": "of"},
            {"LOWER": "proceedings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unjustified"}, {"LOWER": "threat"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unknown"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "unlawful"}, {"LOWER": "act"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unlawful"}, {"LOWER": "act"}, {"LOWER": "manslaughter"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unlawful"}, {"LOWER": "dismissal"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unlawful"},
            {"LOWER": "harassment"},
            {"LOWER": "of"},
            {"LOWER": "debtor"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unlawful"},
            {"LOWER": "interference"},
            {"LOWER": "with"},
            {"LOWER": "body"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unlawful"},
            {"LOWER": "interference"},
            {"LOWER": "with"},
            {"LOWER": "business"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unlawful"},
            {"LOWER": "interference"},
            {"LOWER": "with"},
            {"LOWER": "property"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unlawful"},
            {"LOWER": "means"},
            {"LOWER": "conspiracy"},
            {"LOWER": "and"},
            {"LOWER": "causing"},
            {"LOWER": "loss"},
            {"LOWER": "by"},
            {"LOWER": "unlawful"},
            {"LOWER": "means"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unlawful"}, {"LOWER": "removal"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unlawful"},
            {"LOWER": "“"},
            {"LOWER": "dealing"},
            {"ORTH": "”"},
            {"LOWER": "with"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unlimited"}, {"LOWER": "company"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unmarried"},
            {"LOWER": "father"},
            {"LOWER": "without"},
            {"LOWER": "parental"},
            {"LOWER": "rights"},
            {"LOWER": "and"},
            {"LOWER": "responsibilities"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unnamed"}, {"LOWER": "principal"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unoccupied"}, {"LOWER": "hereditament"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unoccupied"}, {"LOWER": "property"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unreasonably"},
            {"LOWER": "or"},
            {"LOWER": "improperly"},
            {"LOWER": "incurred"},
            {"LOWER": "costs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unregistered"}, {"LOWER": "agreement"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unregistered"}, {"LOWER": "company"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unregistered"}, {"LOWER": "design"}, {"LOWER": "right"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unreported"},
            {"LOWER": "judgments"},
            {"LOWER": "of"},
            {"LOWER": "court"},
            {"LOWER": "of"},
            {"LOWER": "appeal"},
            {"ORTH": "("},
            {"LOWER": "civil"},
            {"LOWER": "division"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "unrepresented"}, {"LOWER": "defendant"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unrepresented"}, {"LOWER": "party"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unresolved"},
            {"ORTH": "("},
            {"ORTH": "“"},
            {"LOWER": "legacy"},
            {"ORTH": "”"},
            {"ORTH": ")"},
            {"LOWER": "cases"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "unseaworthiness"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "unsolicited"},
            {"LOWER": "transmission"},
            {"LOWER": "of"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "untraced"}, {"LOWER": "driver"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "upper"}, {"LOWER": "tribunal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "upper"}, {"LOWER": "tribunal"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "upper"},
            {"LOWER": "tribunal"},
            {"ORTH": "("},
            {"LOWER": "immigration"},
            {"LOWER": "and"},
            {"LOWER": "asylum"},
            {"LOWER": "chamber"},
            {"ORTH": ")"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "urban"},
            {"LOWER": "waste"},
            {"LOWER": "water"},
            {"LOWER": "treatment"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "urgent"}, {"LOWER": "application"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "usa"},
            {"LOWER": "government"},
            {"LOWER": "seeking"},
            {"LOWER": "appellant"},
            {"LOWER": "’s"},
            {"LOWER": "extradition"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "for"},
            {"LOWER": "ambulance"},
            {"LOWER": "purposes"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "for"},
            {"ORTH": "“"},
            {"LOWER": "ambulance"},
            {"ORTH": "”"},
            {"LOWER": "purposes"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "in"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "of"},
            {"LOWER": "private"},
            {"LOWER": "motor"},
            {"LOWER": "vehicle"},
            {"LOWER": "registered"},
            {"LOWER": "in"},
            {"LOWER": "another"},
            {"LOWER": "member"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "use"}, {"LOWER": "of"}, {"LOWER": "documents"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "inside"},
            {"LOWER": "information"},
            {"LOWER": "by"},
            {"LOWER": "primary"},
            {"LOWER": "insider"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "reasonable"},
            {"LOWER": "force"},
            {"LOWER": "in"},
            {"LOWER": "prevention"},
            {"LOWER": "of"},
            {"LOWER": "crime"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "right"},
            {"LOWER": "of"},
            {"LOWER": "way"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "of"},
            {"LOWER": "street"},
            {"LOWER": "pavement"},
            {"LOWER": "area"},
            {"LOWER": "for"},
            {"LOWER": "al"},
            {"LOWER": "fresco"},
            {"LOWER": "dining"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "use"},
            {"LOWER": "with"},
            {"LOWER": "cheque"},
            {"LOWER": "card"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "utilities"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "utilities"}, {"LOWER": "contract"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "vaccine"}, {"LOWER": "damage"}, {"LOWER": "payment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "validity"}, {"LOWER": "of"}, {"LOWER": "claim"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "validity"}, {"LOWER": "of"}, {"LOWER": "detention"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "validity"},
            {"LOWER": "of"},
            {"LOWER": "landlords"},
            {"LOWER": "counter"},
            {"LOWER": "-"},
            {"LOWER": "notice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "validity"},
            {"LOWER": "of"},
            {"LOWER": "notice"},
            {"LOWER": "of"},
            {"LOWER": "enfranchisement"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "validity"},
            {"LOWER": "of"},
            {"LOWER": "subordinate"},
            {"LOWER": "legislation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "validity"},
            {"LOWER": "of"},
            {"LOWER": "tenants"},
            {"LOWER": "’"},
            {"LOWER": "notice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "validity"},
            {"LOWER": "of"},
            {"LOWER": "uniform"},
            {"LOWER": "visa"},
            {"LOWER": "affixed"},
            {"LOWER": "to"},
            {"LOWER": "cancelled"},
            {"LOWER": "travel"},
            {"LOWER": "document"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "valuable"}, {"LOWER": "security"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "valuation"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "valuation"},
            {"LOWER": "date"},
            {"LOWER": "for"},
            {"LOWER": "shares"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "valuation"}, {"LOWER": "list"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "value"}, {"LOWER": "added"}, {"LOWER": "tax"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "value"}, {"LOWER": "added"}, {"LOWER": "tax"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "value"}, {"LOWER": "added"}, {"LOWER": "tax"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "variable"}, {"LOWER": "punishment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "variation"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "material"},
            {"LOWER": "difference"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "variation"},
            {"LOWER": "due"},
            {"LOWER": "to"},
            {"LOWER": "material"},
            {"LOWER": "difference"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "variation"}, {"LOWER": "of"}, {"LOWER": "contract"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "variation"},
            {"LOWER": "of"},
            {"LOWER": "courts"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "variation"},
            {"LOWER": "of"},
            {"LOWER": "courts"},
            {"LOWER": "decisions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "variation"}, {"LOWER": "of"}, {"LOWER": "decision"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "variation"},
            {"LOWER": "of"},
            {"LOWER": "post"},
            {"LOWER": "-"},
            {"LOWER": "nuptial"},
            {"LOWER": "settlement"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "variation"}, {"LOWER": "order"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "vasectomy"}, {"LOWER": "operation"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "vasso"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "vendor"}, {"LOWER": "and"}, {"LOWER": "purchaser"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "vendor"}, {"LOWER": "and"}, {"LOWER": "purchaser"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "vendor"}, {"LOWER": "and"}, {"LOWER": "purchaser"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "vendors"}, {"LOWER": "covenant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "vendors"},
            {"LOWER": "duty"},
            {"LOWER": "pending"},
            {"LOWER": "completion"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "vendors"},
            {"LOWER": "notice"},
            {"LOWER": "to"},
            {"LOWER": "complete"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "vendor"}, {"LOWER": "’s"}, {"LOWER": "covenant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "vendor"},
            {"LOWER": "’s"},
            {"LOWER": "notice"},
            {"LOWER": "to"},
            {"LOWER": "complete"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "verdict"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "verdict"}, {"LOWER": "of"}, {"LOWER": "suicide"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "verdicts"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "very"}, {"LOWER": "large"}, {"LOWER": "corporations"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "veterinary"}, {"LOWER": "surgeon"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "veterinary"}, {"LOWER": "surgeon"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "vexatious"}, {"LOWER": "litigant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "vexatious"},
            {"LOWER": "or"},
            {"LOWER": "repeated"},
            {"LOWER": "requests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "vibration"}, {"LOWER": "white"}, {"LOWER": "finger"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "vicarious"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "vicarious"}, {"LOWER": "liability"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "vicarious"}, {"LOWER": "liability"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "victim"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "sexual"},
            {"LOWER": "exploitation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "victim"},
            {"LOWER": "of"},
            {"LOWER": "child"},
            {"LOWER": "trafficking"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "victim"},
            {"LOWER": "of"},
            {"LOWER": "domestic"},
            {"LOWER": "violence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "victim"},
            {"LOWER": "personal"},
            {"LOWER": "statements"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "victim"}, {"LOWER": "surcharge"}, {"LOWER": "order"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "victimisation"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "victim"}, {"LOWER": "of"}, {"LOWER": "crime"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "victims"}, {"LOWER": "of"}, {"LOWER": "trafficking"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "video"}, {"LOWER": "games"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "video"}, {"LOWER": "recording"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "video"}, {"LOWER": "recordings"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "video"}, {"LOWER": "work"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ville"}, {"LOWER": "de"}, {"LOWER": "capella"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ville"}, {"LOWER": "de"}, {"LOWER": "libra"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "visit"}, {"LOWER": "by"}, {"LOWER": "journalist"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "visit"},
            {"LOWER": "by"},
            {"LOWER": "legal"},
            {"LOWER": "adviser"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "visit"},
            {"LOWER": "by"},
            {"LOWER": "legal"},
            {"LOWER": "advisers"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "visiting"}, {"LOWER": "forces"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "visitors"}, {"LOWER": "jurisdiction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "visitors"}, {"LOWER": "jurisdiction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "vocational"}, {"LOWER": "training"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "void"}, {"LOWER": "settlement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "voidable"}, {"LOWER": "preference"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "voir"}, {"LOWER": "dire"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "volenti"},
            {"LOWER": "non"},
            {"LOWER": "fit"},
            {"LOWER": "injuria"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "voluntary"}, {"LOWER": "arrangement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "voluntary"}, {"LOWER": "association"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "voluntary"},
            {"LOWER": "redundancy"},
            {"LOWER": "scheme"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "volunteer"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "voting"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "voyeurism"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "vulnerable"}, {"LOWER": "adult"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "vulnerable"},
            {"LOWER": "adult"},
            {"LOWER": "with"},
            {"LOWER": "significant"},
            {"LOWER": "learning"},
            {"LOWER": "disability"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "vulnerable"}, {"LOWER": "witnesses"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wages"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "waiver"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "waiver"}, {"LOWER": "of"}, {"LOWER": "breach"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "wales"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "war"}, {"LOWER": "pensions"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ward"}, {"LOWER": "as"}, {"LOWER": "witness"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ward"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "ward"}, {"LOWER": "of"}, {"LOWER": "court"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "wardship"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wardship"}, {"LOWER": "documents"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wardship"}, {"LOWER": "jurisdiction"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "warning"}, {"LOWER": "notices"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "warning"},
            {"LOWER": "to"},
            {"LOWER": "prevent"},
            {"LOWER": "commission"},
            {"LOWER": "of"},
            {"LOWER": "offence"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "warrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "warrant"}, {"LOWER": "of"}, {"LOWER": "commitment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "warrant"}, {"LOWER": "of"}, {"LOWER": "possession"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "warranty"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "waste"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "waste"}, {"LOWER": "facility"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "waste"}, {"LOWER": "management"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "waste"}, {"LOWER": "management"}, {"LOWER": "permit"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "waste"}, {"LOWER": "oil"}, {"LOWER": "recycling"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "waste"}, {"LOWER": "treatment"}, {"LOWER": "facility"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "wasted"}, {"LOWER": "costs"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "wasted"}, {"LOWER": "costs"}, {"LOWER": "orders"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "water"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "water"}, {"LOWER": "authority"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "water"},
            {"LOWER": "ski"},
            {"ORTH": "-"},
            {"LOWER": "ing"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "water"}, {"LOWER": "supply"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "way"}, {"LOWER": "of"}, {"LOWER": "necessity"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "wayleave"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "website"}, {"LOWER": "controller"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "welfare"}, {"LOWER": "benefits"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "welfare"}, {"LOWER": "benefits"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "welfare"}, {"LOWER": "decision"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "welfare"}, {"LOWER": "disposal"}, {"LOWER": "hearing"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "welfare"}, {"LOWER": "report"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "welfare"}, {"LOWER": "service"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "welfare"}, {"LOWER": "services"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "welsh"}, {"LOWER": "language"}, {"LOWER": "scheme"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "west"}, {"LOWER": "indies"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "when"},
            {"LOWER": "assignment"},
            {"LOWER": "taking"},
            {"LOWER": "effect"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "when"},
            {"LOWER": "prescription"},
            {"LOWER": "period"},
            {"LOWER": "commencing"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "when"},
            {"LOWER": "proceedings"},
            {"ORTH": "“"},
            {"LOWER": "begun"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whereabouts"},
            {"LOWER": "of"},
            {"LOWER": "natural"},
            {"LOWER": "parents"},
            {"LOWER": "unknown"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "adoption"},
            {"LOWER": "order"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "made"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "adoption"},
            {"LOWER": "order"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "recognised"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "appeal"}, {"LOWER": "lying"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "applicant"},
            {"LOWER": "entitled"},
            {"LOWER": "to"},
            {"LOWER": "plead"},
            {"LOWER": "incorrect"},
            {"LOWER": "application"},
            {"LOWER": "of"},
            {"LOWER": "criteria"},
            {"LOWER": "for"},
            {"LOWER": "determining"},
            {"LOWER": "responsibility"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "applications"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "granted"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "appropriate"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "arising"},
            {"LOWER": "from"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "statutory"},
            {"LOWER": "obligation"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "binding"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "binding"}, {"LOWER": "crown"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "duty"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "capacity"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "decisions"},
            {"LOWER": "as"},
            {"LOWER": "to"},
            {"LOWER": "contact"},
            {"LOWER": "with"},
            {"LOWER": "other"},
            {"LOWER": "persons"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "capacity"},
            {"LOWER": "to"},
            {"LOWER": "make"},
            {"LOWER": "or"},
            {"LOWER": "contribute"},
            {"LOWER": "to"},
            {"LOWER": "relevant"},
            {"LOWER": "decisions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "charge"},
            {"LOWER": "duplicitous"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "claim"},
            {"LOWER": "for"},
            {"LOWER": "asylum"},
            {"LOWER": "and"},
            {"LOWER": "humanitarian"},
            {"LOWER": "protection"},
            {"LOWER": "clearly"},
            {"LOWER": "unfounded"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "claim"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "reconsidered"},
            {"LOWER": "in"},
            {"LOWER": "interests"},
            {"LOWER": "of"},
            {"LOWER": "justice"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "claimants"},
            {"LOWER": "entitled"},
            {"LOWER": "to"},
            {"LOWER": "enhanced"},
            {"LOWER": "redundancy"},
            {"LOWER": "payments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "contempt"},
            {"LOWER": "of"},
            {"LOWER": "court"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "continuing"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "contract"},
            {"LOWER": "concluded"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "conviction"}, {"LOWER": "safe"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "damages"},
            {"LOWER": "adequate"},
            {"LOWER": "remedy"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "decision"},
            {"LOWER": "open"},
            {"LOWER": "to"},
            {"LOWER": "judicial"},
            {"LOWER": "review"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "defendant"},
            {"LOWER": "liable"},
            {"LOWER": "to"},
            {"LOWER": "pay"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "discretionary"},
            {"LOWER": "benefit"},
            {"LOWER": "under"},
            {"LOWER": "trust"},
            {"LOWER": "caught"},
            {"LOWER": "by"},
            {"LOWER": "forfeiture"},
            {"LOWER": "rule"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "duty"},
            {"LOWER": "of"},
            {"LOWER": "care"},
            {"LOWER": "existing"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "enforceable"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "evidential"},
            {"LOWER": "burden"},
            {"LOWER": "on"},
            {"LOWER": "occupier"},
            {"LOWER": "to"},
            {"LOWER": "show"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "liability"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "excluded"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "fact"},
            {"ORTH": "-"},
            {"LOWER": "finding"},
            {"LOWER": "judgment"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "published"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "finding"},
            {"LOWER": "of"},
            {"LOWER": "no"},
            {"LOWER": "case"},
            {"LOWER": "to"},
            {"LOWER": "answer"},
            {"LOWER": "perverse"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "first"},
            {"LOWER": "defendant"},
            {"LOWER": "in"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "terms"},
            {"LOWER": "of"},
            {"LOWER": "injunction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "foreign"},
            {"LOWER": "limitation"},
            {"LOWER": "period"},
            {"LOWER": "applying"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "impugned"},
            {"LOWER": "agreement"},
            {"LOWER": "giving"},
            {"LOWER": "rise"},
            {"LOWER": "to"},
            {"LOWER": "enforceable"},
            {"LOWER": "obligations"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "in"},
            {"LOWER": "best"},
            {"LOWER": "interests"},
            {"LOWER": "for"},
            {"LOWER": "artificial"},
            {"LOWER": "nutrition"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "withdrawn"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "in"},
            {"LOWER": "child"},
            {"LOWER": "’s"},
            {"LOWER": "best"},
            {"LOWER": "interests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "in"},
            {"LOWER": "pensionable"},
            {"LOWER": "employment"},
            {"LOWER": "as"},
            {"ORTH": "“"},
            {"LOWER": "nurse"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "in"}, {"LOWER": "private"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "insanity"},
            {"LOWER": "a"},
            {"LOWER": "defence"},
            {"LOWER": "to"},
            {"LOWER": "and"},
            {"LOWER": "and"},
            {"LOWER": "what"},
            {"LOWER": "extent"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "interim"},
            {"LOWER": "or"},
            {"LOWER": "final"},
            {"LOWER": "order"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "made"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "judicial"},
            {"LOWER": "review"},
            {"LOWER": "appropriate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "judicial"},
            {"LOWER": "review"},
            {"LOWER": "appropriate"},
            {"LOWER": "appeals"},
            {"LOWER": "procedure"},
            {"LOWER": "against"},
            {"LOWER": "assessments"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "just"},
            {"LOWER": "and"},
            {"LOWER": "equitable"},
            {"LOWER": "to"},
            {"LOWER": "extend"},
            {"LOWER": "time"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "lacking"}, {"LOWER": "capacity"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "lawful"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "lawful"},
            {"LOWER": "to"},
            {"LOWER": "withhold"},
            {"LOWER": "blood"},
            {"LOWER": "transfusion"},
            {"LOWER": "in"},
            {"LOWER": "accordance"},
            {"LOWER": "with"},
            {"LOWER": "patient"},
            {"LOWER": "’s"},
            {"LOWER": "wishes"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "legislation"},
            {"LOWER": "of"},
            {"LOWER": "member"},
            {"LOWER": "state"},
            {"LOWER": "prohibiting"},
            {"LOWER": "any"},
            {"LOWER": "publication"},
            {"LOWER": "for"},
            {"LOWER": "remuneration"},
            {"LOWER": "not"},
            {"LOWER": "identified"},
            {"LOWER": "by"},
            {"LOWER": "the"},
            {"LOWER": "term"},
            {"LOWER": "“"},
            {"LOWER": "advertisement"},
            {"ORTH": "”"},
            {"LOWER": "prohibited"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "necessary"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "new"},
            {"LOWER": "constitution"},
            {"LOWER": "validly"},
            {"LOWER": "adopted"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "official"},
            {"LOWER": "receiver"},
            {"LOWER": "exercising"},
            {"LOWER": "judicial"},
            {"LOWER": "function"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "order"},
            {"LOWER": "for"},
            {"LOWER": "production"},
            {"LOWER": "of"},
            {"LOWER": "documents"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "granted"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "order"},
            {"LOWER": "refusing"},
            {"LOWER": "permission"},
            {"LOWER": "to"},
            {"LOWER": "appeal"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "order"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "made"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "original"},
            {"LOWER": "judgment"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "published"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "parents"},
            {"LOWER": "“"},
            {"LOWER": "can"},
            {"LOWER": "not"},
            {"LOWER": "be"},
            {"LOWER": "found"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "parties"},
            {"LOWER": "entering"},
            {"LOWER": "into"},
            {"LOWER": "agreement"},
            {"LOWER": "for"},
            {"LOWER": "payment"},
            {"LOWER": "of"},
            {"LOWER": "lump"},
            {"LOWER": "sum"},
            {"LOWER": "and"},
            {"LOWER": "transfer"},
            {"LOWER": "of"},
            {"LOWER": "properties"},
            {"LOWER": "to"},
            {"LOWER": "claimant"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "party"},
            {"LOWER": "estopped"},
            {"LOWER": "from"},
            {"LOWER": "determining"},
            {"LOWER": "contract"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "patent"},
            {"LOWER": "invalid"},
            {"LOWER": "for"},
            {"LOWER": "obviousness"},
            {"LOWER": "or"},
            {"LOWER": "lack"},
            {"LOWER": "of"},
            {"LOWER": "novelty"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "permissible"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "portuguese"},
            {"LOWER": "court"},
            {"LOWER": "be"},
            {"LOWER": "requested"},
            {"LOWER": "to"},
            {"LOWER": "assume"},
            {"LOWER": "jurisdiction"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "reopen"},
            {"LOWER": "case"},
            {"LOWER": "to"},
            {"LOWER": "rectify"},
            {"LOWER": "mistake"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "power"},
            {"LOWER": "to"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
            {"LOWER": "return"},
            {"LOWER": "orders"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "whether"}, {"LOWER": "proportionate"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "proposed"},
            {"LOWER": "treatment"},
            {"LOWER": "in"},
            {"LOWER": "best"},
            {"LOWER": "interests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "protection"},
            {"LOWER": "revived"},
            {"LOWER": "by"},
            {"LOWER": "virtue"},
            {"LOWER": "of"},
            {"LOWER": "protection"},
            {"LOWER": "from"},
            {"LOWER": "international"},
            {"LOWER": "convention"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "rebuttable"},
            {"LOWER": "common"},
            {"LOWER": "law"},
            {"LOWER": "presumption"},
            {"LOWER": "of"},
            {"LOWER": "doli"},
            {"LOWER": "incapax"},
            {"LOWER": "relevant"},
            {"LOWER": "to"},
            {"LOWER": "issue"},
            {"LOWER": "of"},
            {"LOWER": "admissibility"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "relevant"}, {"LOWER": "transfer"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "remedy"}, {"LOWER": "available"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "resident"},
            {"LOWER": "parent"},
            {"LOWER": "to"},
            {"LOWER": "provide"},
            {"LOWER": "maintenance"},
            {"LOWER": "to"},
            {"LOWER": "non"},
            {"LOWER": "-"},
            {"LOWER": "resident"},
            {"LOWER": "parent"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "rule"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "modified"},
            {"LOWER": "to"},
            {"LOWER": "permit"},
            {"LOWER": "son"},
            {"LOWER": "to"},
            {"LOWER": "take"},
            {"LOWER": "under"},
            {"LOWER": "will"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "shares"},
            {"LOWER": "capital"},
            {"LOWER": "or"},
            {"LOWER": "income"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "sister"}, {"LOWER": "ship"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "study"},
            {"LOWER": "subject"},
            {"LOWER": "to"},
            {"LOWER": "data"},
            {"LOWER": "sharing"},
            {"LOWER": "provisions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "subject"},
            {"LOWER": "to"},
            {"LOWER": "review"},
            {"LOWER": "by"},
            {"LOWER": "courts"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "testator"},
            {"LOWER": "having"},
            {"LOWER": "testamentary"},
            {"LOWER": "capacity"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "testing"},
            {"LOWER": "for"},
            {"LOWER": "genetic"},
            {"LOWER": "disorder"},
            {"LOWER": "in"},
            {"LOWER": "children"},
            {"LOWER": "’s"},
            {"LOWER": "best"},
            {"LOWER": "interests"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "threshold"},
            {"LOWER": "criteria"},
            {"LOWER": "made"},
            {"LOWER": "out"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "time"},
            {"LOWER": "of"},
            {"LOWER": "essence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "moved"},
            {"LOWER": "from"},
            {"LOWER": "mother"},
            {"LOWER": "’s"},
            {"LOWER": "care"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "reopened"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "set"},
            {"LOWER": "aside"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "transfer"},
            {"LOWER": "to"},
            {"LOWER": "hungary"},
            {"LOWER": "appropriate"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "treaty"},
            {"LOWER": "provisions"},
            {"LOWER": "on"},
            {"LOWER": "freedom"},
            {"LOWER": "of"},
            {"LOWER": "establishment"},
            {"LOWER": "precluding"},
            {"LOWER": "legislation"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "trust"},
            {"LOWER": "property"},
            {"LOWER": "having"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "in"},
            {"LOWER": "possession"},
            {"LOWER": "or"},
            {"LOWER": "receipt"},
            {"LOWER": "of"},
            {"LOWER": "trustee"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "ultra"}, {"LOWER": "vires"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "whether"}, {"LOWER": "validly"}, {"LOWER": "laid"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "victims"},
            {"LOWER": "consent"},
            {"LOWER": "providing"},
            {"LOWER": "a"},
            {"LOWER": "defence"},
            {"ORTH": "—"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "wrongdoer"},
            {"LOWER": "to"},
            {"LOWER": "be"},
            {"LOWER": "identified"},
            {"LOWER": "in"},
            {"LOWER": "anonymised"},
            {"LOWER": "public"},
            {"LOWER": "judgment"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"ORTH": "“"},
            {"LOWER": "concerned"},
            {"LOWER": "in"},
            {"LOWER": "terrorism"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "“"},
            {"LOWER": "homeless"},
            {"LOWER": "intentionally"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"LOWER": "“"},
            {"LOWER": "working"},
            {"LOWER": "time"},
            {"ORTH": "”"},
            {"LOWER": "or"},
            {"ORTH": "“"},
            {"LOWER": "rest"},
            {"LOWER": "periods"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "whether"},
            {"ORTH": ","},
            {"LOWER": "and"},
            {"LOWER": "extent"},
            {"LOWER": "to"},
            {"LOWER": "which"},
            {"ORTH": ","},
            {"LOWER": "mental"},
            {"LOWER": "disorder"},
            {"LOWER": "relevant"},
            {"LOWER": "to"},
            {"LOWER": "assessment"},
            {"LOWER": "of"},
            {"ORTH": "“"},
            {"LOWER": "the"},
            {"LOWER": "circumstances"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "defendant"},
            {"ORTH": "”"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "widow"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "widows"}, {"LOWER": "allowance"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "widows"}, {"LOWER": "dependency"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "wife"}, {"LOWER": "an"}, {"LOWER": "infant"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "wife"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "consequential"},
            {"LOWER": "orders"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "wife"},
            {"LOWER": "applying"},
            {"LOWER": "for"},
            {"LOWER": "financial"},
            {"LOWER": "provision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "wife"},
            {"LOWER": "in"},
            {"LOWER": "pakistan"},
            {"LOWER": "petitioning"},
            {"LOWER": "for"},
            {"LOWER": "divorce"},
            {"LOWER": "on"},
            {"LOWER": "basis"},
            {"LOWER": "of"},
            {"LOWER": "husband"},
            {"LOWER": "’s"},
            {"LOWER": "residence"},
            {"LOWER": "in"},
            {"LOWER": "england"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "wife"},
            {"LOWER": "seeking"},
            {"LOWER": "compensation"},
            {"LOWER": "or"},
            {"LOWER": "enforcement"},
            {"LOWER": "for"},
            {"LOWER": "breach"},
            {"LOWER": "of"},
            {"LOWER": "husband"},
            {"LOWER": "’s"},
            {"LOWER": "undertakings"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "wilful"}, {"LOWER": "neglect"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wind"}, {"LOWER": "turbine"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wind"}, {"LOWER": "turbines"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "winding"}, {"LOWER": "up"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "winding"}, {"LOWER": "up"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "winding"}, {"LOWER": "up"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "winding"}, {"LOWER": "up"}, {"LOWER": "petition"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "winding"}, {"ORTH": "-"}, {"LOWER": "up"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "wireless"}, {"LOWER": "telegraphy"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "withdrawal"}, {"LOWER": "by"}, {"LOWER": "defendants"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "withdrawal"},
            {"LOWER": "by"},
            {"LOWER": "member"},
            {"LOWER": "state"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "withdrawal"}, {"LOWER": "of"}, {"LOWER": "application"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "withdrawal"}, {"LOWER": "of"}, {"LOWER": "feeding"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "withdrawal"}, {"LOWER": "of"}, {"LOWER": "feeding"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "withdrawal"},
            {"LOWER": "of"},
            {"LOWER": "immigration"},
            {"LOWER": "decision"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "withdrawal"}, {"LOWER": "of"}, {"LOWER": "nutrition"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "withdrawal"}, {"LOWER": "of"}, {"LOWER": "treatment"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "without"},
            {"LOWER": "notice"},
            {"LOWER": "application"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "without"},
            {"LOWER": "notice"},
            {"LOWER": "freezing"},
            {"LOWER": "order"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "without"}, {"LOWER": "notice"}, {"LOWER": "order"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "without"}, {"LOWER": "notice"}, {"LOWER": "orders"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "without"},
            {"LOWER": "prejudice"},
            {"LOWER": "communications"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "without"},
            {"LOWER": "prejudice"},
            {"LOWER": "document"},
            {"LOWER": "and"},
            {"LOWER": "meeting"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "without"}, {"LOWER": "warrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "witness"},
            {"LOWER": "action"},
            {"LOWER": "tried"},
            {"LOWER": "by"},
            {"LOWER": "judge"},
            {"LOWER": "alone"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "witness"}, {"LOWER": "signature"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "witness"}, {"LOWER": "statements"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "witness"}, {"LOWER": "summons"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "witness"}, {"LOWER": "warrant"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "witness"}, {"ORTH": ","}, {"LOWER": "by"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "witnesses"}, {"LOWER": "statements"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "witness"}, {"LOWER": "’s"}, {"LOWER": "expenses"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "work"}, {"LOWER": "equipment"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "work"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "particular"},
            {"LOWER": "kind"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "work"},
            {"LOWER": "of"},
            {"LOWER": "equal"},
            {"LOWER": "value"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "work"},
            {"LOWER": "outside"},
            {"LOWER": "great"},
            {"LOWER": "britain"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "work"}, {"LOWER": "permit"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "worker"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "workers"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "working"}, {"LOWER": "abroad"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "working"}, {"LOWER": "time"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "working"}, {"LOWER": "time"}, {"LOWER": "provisions"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "working"}, {"LOWER": "time"}, {"LOWER": "provisions"}],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "workplace"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "workplace"}, {"LOWER": "obstruction"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "wounding"},
            {"LOWER": "with"},
            {"LOWER": "intent"},
            {"LOWER": "to"},
            {"LOWER": "cause"},
            {"LOWER": "grievous"},
            {"LOWER": "bodily"},
            {"LOWER": "harm"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "wounding"},
            {"LOWER": "with"},
            {"LOWER": "intent"},
            {"LOWER": "to"},
            {"LOWER": "do"},
            {"LOWER": "grievous"},
            {"LOWER": "bodily"},
            {"LOWER": "harm"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "writ"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "writ"},
            {"LOWER": "of"},
            {"LOWER": "fieri"},
            {"LOWER": "facias"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "writ"}, {"LOWER": "of"}, {"LOWER": "possession"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "writ"}, {"LOWER": "of"}, {"LOWER": "restitution"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "writ"}, {"LOWER": "of"}, {"LOWER": "sequestration"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "written"},
            {"LOWER": "acknowledgement"},
            {"LOWER": "of"},
            {"LOWER": "indebtedness"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "written"},
            {"LOWER": "representations"},
            {"LOWER": "procedure"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "written"}, {"LOWER": "statement"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "written"}, {"LOWER": "statements"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wrong"}, {"LOWER": "goods"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "wrongful"},
            {"LOWER": "act"},
            {"LOWER": "or"},
            {"LOWER": "omission"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "wrongful"}, {"LOWER": "arrest"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wrongful"}, {"LOWER": "dismissal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wrongful"}, {"LOWER": "removal"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "wrongful"}, {"LOWER": "retention"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "young"}, {"LOWER": "offender"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "young"},
            {"LOWER": "offender"},
            {"LOWER": "institution"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "young"}, {"LOWER": "offenders"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "young"}, {"LOWER": "person"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "youth"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "youth"}, {"LOWER": "court"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "youth"}, {"LOWER": "custody"}]},
    {"label": "CONCEPT", "pattern": [{"LOWER": "ypatianna"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "zero"}, {"ORTH": "-"}, {"LOWER": "rating"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "—"},
            {"LOWER": "one"},
            {"LOWER": "of"},
            {"LOWER": "jurors"},
            {"LOWER": "at"},
            {"LOWER": "second"},
            {"LOWER": "trial"},
            {"LOWER": "having"},
            {"LOWER": "been"},
            {"LOWER": "juror"},
            {"LOWER": "at"},
            {"LOWER": "first"},
            {"LOWER": "trial"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "accidental"},
            {"LOWER": "means"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "act"},
            {"LOWER": "which"},
            {"LOWER": "intimidates"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "action"},
            {"LOWER": "founded"},
            {"LOWER": "on"},
            {"LOWER": "contract"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "actual"},
            {"LOWER": "occupation"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "agricultural"},
            {"LOWER": "land"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "animal"},
            {"LOWER": "fight"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "any"},
            {"LOWER": "mariner"},
            {"LOWER": "or"},
            {"LOWER": "seaman"},
            {"LOWER": "being"},
            {"LOWER": "at"},
            {"LOWER": "sea"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "any"},
            {"LOWER": "other"},
            {"LOWER": "person"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "associated"},
            {"LOWER": "employer"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "bad"},
            {"LOWER": "faith"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "by"},
            {"LOWER": "consent"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "caravan"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "civil"},
            {"LOWER": "proceedings"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "common"},
            {"LOWER": "duty"},
            {"LOWER": "of"},
            {"LOWER": "care"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "converting"},
            {"LOWER": "criminal"},
            {"LOWER": "property"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "converting"},
            {"LOWER": "criminal"},
            {"LOWER": "property"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "criminal"},
            {"LOWER": "cause"},
            {"LOWER": "or"},
            {"LOWER": "matter"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "criminal"},
            {"LOWER": "property"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "curtilage"},
            {"ORTH": "”"},
            {"LOWER": "of"},
            {"LOWER": "listed"},
            {"LOWER": "building"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "dangerous"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "devolution"},
            {"LOWER": "issue"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "dismissal"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "dismissed"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "document"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "dwelling"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "employed"},
            {"LOWER": "under"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "extradition"},
            {"LOWER": "crime"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "former"},
            {"LOWER": "relevant"},
            {"LOWER": "child"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "free"},
            {"LOWER": "from"},
            {"LOWER": "all"},
            {"LOWER": "deductions"},
            {"LOWER": "and"},
            {"LOWER": "abatements"},
            {"LOWER": "whatsoever"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "genuine"},
            {"LOWER": "occupational"},
            {"LOWER": "qualification"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "genuine"},
            {"LOWER": "use"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "in"},
            {"LOWER": "default"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "in"},
            {"LOWER": "the"},
            {"LOWER": "course"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "business"},
            {"LOWER": "carried"},
            {"LOWER": "on"},
            {"LOWER": "by"},
            {"LOWER": "him"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "information"},
            {"LOWER": "about"},
            {"LOWER": "proceedings"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "inhuman"},
            {"LOWER": "or"},
            {"LOWER": "degrading"},
            {"LOWER": "treatment"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "insolvent"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "irrevocable"},
            {"ORTH": "”"},
            {"LOWER": "undertaking"},
            {"LOWER": "by"},
            {"LOWER": "shareholders"},
            {"LOWER": "of"},
            {"LOWER": "company"},
            {"LOWER": "not"},
            {"LOWER": "to"},
            {"LOWER": "enforce"},
            {"LOWER": "loan"},
            {"LOWER": "notes"},
            {"LOWER": "in"},
            {"LOWER": "consideration"},
            {"LOWER": "of"},
            {"LOWER": "receiving"},
            {"LOWER": "payment"},
            {"LOWER": "from"},
            {"LOWER": "company"},
            {"LOWER": "by"},
            {"LOWER": "stipulated"},
            {"LOWER": "date"},
            {"LOWER": "or"},
            {"LOWER": "“"},
            {"LOWER": "in"},
            {"LOWER": "any"},
            {"LOWER": "event"},
            {"ORTH": "”"},
            {"LOWER": "within"},
            {"LOWER": "12"},
            {"LOWER": "months"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "just"},
            {"LOWER": "and"},
            {"LOWER": "equitable"},
            {"ORTH": "”"},
            {"LOWER": "grounds"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"ORTH": "“"}, {"LOWER": "land"}, {"ORTH": "”"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "lawful"},
            {"LOWER": "object"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "leapfrog"},
            {"ORTH": "”"},
            {"LOWER": "procedure"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "like"}, {"LOWER": "work"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "liquidated"},
            {"LOWER": "sum"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "low"}, {"LOWER": "rent"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "“"}, {"LOWER": "machinery"}, {"ORTH": "”"}],
    },
    {"label": "CONCEPT", "pattern": [{"ORTH": "“"}, {"LOWER": "mark"}, {"ORTH": "”"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "material"},
            {"LOWER": "considerations"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "member"},
            {"LOWER": "of"},
            {"LOWER": "the"},
            {"LOWER": "household"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "nhs"},
            {"LOWER": "contract"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "“"}, {"LOWER": "office"}, {"ORTH": "”"}],
    },
    {"label": "CONCEPT", "pattern": [{"ORTH": "“"}, {"LOWER": "order"}, {"ORTH": "”"}]},
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "paid"},
            {"LOWER": "money"},
            {"LOWER": "for"},
            {"LOWER": "…"},
            {"LOWER": "maintenance"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "particular"},
            {"LOWER": "social"},
            {"LOWER": "group"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "period"},
            {"LOWER": "of"},
            {"LOWER": "entitlement"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "persons"},
            {"LOWER": "who"},
            {"LOWER": "have"},
            {"LOWER": "the"},
            {"LOWER": "same"},
            {"LOWER": "interest"},
            {"LOWER": "in"},
            {"LOWER": "a"},
            {"LOWER": "claim"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "physical"},
            {"LOWER": "training"},
            {"LOWER": "and"},
            {"LOWER": "recreation"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "private"},
            {"LOWER": "purchaser"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "property"},
            {"LOWER": "recovered"},
            {"LOWER": "and"},
            {"LOWER": "preserved"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "reasonable"},
            {"LOWER": "financial"},
            {"LOWER": "provision"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "“"}, {"LOWER": "relatives"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "relevant"},
            {"LOWER": "offence"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "“"}, {"LOWER": "remuneration"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "seller"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "serious"},
            {"LOWER": "irregularity"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "serious"},
            {"LOWER": "professional"},
            {"LOWER": "misconduct"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "statutory"},
            {"LOWER": "next"},
            {"LOWER": "of"},
            {"LOWER": "kin"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "supply"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "supply"},
            {"ORTH": "”"},
            {"LOWER": "of"},
            {"LOWER": "controlled"},
            {"LOWER": "drugs"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "the"},
            {"LOWER": "18"},
            {"LOWER": "months"},
            {"LOWER": "’"},
            {"LOWER": "rule"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "treated"},
            {"LOWER": "…"},
            {"LOWER": "as"},
            {"LOWER": "child"},
            {"LOWER": "of"},
            {"LOWER": "…"},
            {"LOWER": "family"},
            {"LOWER": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "unjust"},
            {"LOWER": "or"},
            {"LOWER": "oppressive"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "unless"},
            {"ORTH": "”"},
            {"LOWER": "order"},
        ],
    },
    {"label": "CONCEPT", "pattern": [{"LOWER": "“"}, {"LOWER": "use"}, {"ORTH": "”"}]},
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "“"}, {"LOWER": "victim"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "“"}, {"LOWER": "victim"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"ORTH": "“"},
            {"LOWER": "waste"},
            {"LOWER": "land"},
            {"LOWER": "of"},
            {"LOWER": "a"},
            {"LOWER": "manor"},
            {"ORTH": "”"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"ORTH": "“"}, {"LOWER": "wilfully"}, {"ORTH": "”"}],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "without"},
            {"LOWER": "prejudice"},
            {"ORTH": "”"},
            {"LOWER": "communications"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "without"},
            {"LOWER": "prejudice"},
            {"ORTH": "”"},
            {"LOWER": "correspondence"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [
            {"LOWER": "“"},
            {"LOWER": "without"},
            {"LOWER": "prejudice"},
            {"ORTH": "”"},
            {"LOWER": "discussions"},
        ],
    },
    {
        "label": "CONCEPT",
        "pattern": [{"LOWER": "“"}, {"LOWER": "worker"}, {"ORTH": "”"}],
    },
]
